/* Generated by Cython 3.2.4 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "EC_VER2",
                ""
            ],
            [
                "USE_SOEM_CONFIG_H",
                ""
            ]
        ],
        "depends": [
            "soem/soem/ethercat.h"
        ],
        "include_dirs": [
            "./pysoem",
            "./soem/oshw/linux",
            "./soem/osal/linux",
            "./soem/oshw",
            "./soem/osal",
            "./soem/soem",
            "./src/soem"
        ],
        "libraries": [
            "pthread",
            "rt"
        ],
        "name": "pysoem.pysoem",
        "sources": [
            "src/pysoem/pysoem.pyx",
            "./soem/osal/linux/osal.c",
            "./soem/oshw/linux/oshw.c",
            "./soem/oshw/linux/nicdrv.c",
            "./soem/soem/ethercatbase.c",
            "./soem/soem/ethercatcoe.c",
            "./soem/soem/ethercatconfig.c",
            "./soem/soem/ethercatdc.c",
            "./soem/soem/ethercatfoe.c",
            "./soem/soem/ethercatmain.c",
            "./soem/soem/ethercatprint.c",
            "./soem/soem/ethercatsoe.c",
            "./src/soem/soem_config.c"
        ]
    },
    "module_name": "pysoem.pysoem"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 0
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pysoem__pysoem
#define __PYX_HAVE_API__pysoem__pysoem
/* Early includes */
#include <stdint.h>
#include "ethercat.h"
#include <string.h>
#include <stdio.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "src/pysoem/pysoem.pyx",
  "<stringsource>",
  "cpython/type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts;
struct __pyx_obj_6pysoem_6pysoem_CdefSettings;
struct __pyx_obj_6pysoem_6pysoem_CdefMaster;
struct __pyx_obj_6pysoem_6pysoem__CallbackData;
struct __pyx_obj_6pysoem_6pysoem_CdefSlave;
struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject;
struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry;
struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open;
struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr;
struct __pyx_t_6pysoem_6pysoem_CdefMasterSettings;

/* "pysoem/pysoem.pyx":123
 * settings = CdefSettings()
 * 
 * cpdef enum ec_datatype:             # <<<<<<<<<<<<<<
 *     ECT_BOOLEAN         = 0x0001,
 *     ECT_INTEGER8        = 0x0002,
*/
enum __pyx_t_6pysoem_6pysoem_ec_datatype {
  __pyx_e_6pysoem_6pysoem_ECT_BOOLEAN = 0x0001,
  __pyx_e_6pysoem_6pysoem_ECT_INTEGER8 = 0x0002,
  __pyx_e_6pysoem_6pysoem_ECT_INTEGER16 = 0x0003,
  __pyx_e_6pysoem_6pysoem_ECT_INTEGER32 = 0x0004,
  __pyx_e_6pysoem_6pysoem_ECT_UNSIGNED8 = 0x0005,
  __pyx_e_6pysoem_6pysoem_ECT_UNSIGNED16 = 0x0006,
  __pyx_e_6pysoem_6pysoem_ECT_UNSIGNED32 = 0x0007,
  __pyx_e_6pysoem_6pysoem_ECT_REAL32 = 0x0008,
  __pyx_e_6pysoem_6pysoem_ECT_VISIBLE_STRING = 0x0009,
  __pyx_e_6pysoem_6pysoem_ECT_OCTET_STRING = 0x000A,
  __pyx_e_6pysoem_6pysoem_ECT_UNICODE_STRING = 0x000B,
  __pyx_e_6pysoem_6pysoem_ECT_TIME_OF_DAY = 0x000C,
  __pyx_e_6pysoem_6pysoem_ECT_TIME_DIFFERENCE = 0x000D,
  __pyx_e_6pysoem_6pysoem_ECT_DOMAIN = 0x000F,
  __pyx_e_6pysoem_6pysoem_ECT_INTEGER24 = 0x0010,
  __pyx_e_6pysoem_6pysoem_ECT_REAL64 = 0x0011,
  __pyx_e_6pysoem_6pysoem_ECT_INTEGER64 = 0x0015,
  __pyx_e_6pysoem_6pysoem_ECT_UNSIGNED24 = 0x0016,
  __pyx_e_6pysoem_6pysoem_ECT_UNSIGNED64 = 0x001B,
  __pyx_e_6pysoem_6pysoem_ECT_BIT1 = 0x0030,
  __pyx_e_6pysoem_6pysoem_ECT_BIT2 = 0x0031,
  __pyx_e_6pysoem_6pysoem_ECT_BIT3 = 0x0032,
  __pyx_e_6pysoem_6pysoem_ECT_BIT4 = 0x0033,
  __pyx_e_6pysoem_6pysoem_ECT_BIT5 = 0x0034,
  __pyx_e_6pysoem_6pysoem_ECT_BIT6 = 0x0035,
  __pyx_e_6pysoem_6pysoem_ECT_BIT7 = 0x0036,
  __pyx_e_6pysoem_6pysoem_ECT_BIT8 = 0x0037
};

/* "pysoem/pysoem.pyx":211
 * 
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     EC_MAXSLAVE = 200
 *     EC_MAXGROUP = 1
*/
enum  {

  /* "pysoem/pysoem.pyx":215
 *     EC_MAXGROUP = 1
 *     EC_MAXEEPBITMAP = 128
 *     EC_MAXEEPBUF = EC_MAXEEPBITMAP * 32             # <<<<<<<<<<<<<<
 *     EC_MAXMAPT = 8
 *     EC_IOMAPSIZE = 4096
*/
  __pyx_e_6pysoem_6pysoem_EC_MAXSLAVE = 0xC8,
  __pyx_e_6pysoem_6pysoem_EC_MAXGROUP = 1,
  __pyx_e_6pysoem_6pysoem_EC_MAXEEPBITMAP = 0x80,
  __pyx_e_6pysoem_6pysoem_EC_MAXEEPBUF = (__pyx_e_6pysoem_6pysoem_EC_MAXEEPBITMAP * 32),
  __pyx_e_6pysoem_6pysoem_EC_MAXMAPT = 8,
  __pyx_e_6pysoem_6pysoem_EC_IOMAPSIZE = 0x1000
};

/* "pysoem/pysoem.pyx":783
 * 
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     EC_TIMEOUTRXM = 700000
 *     STATIC_SDO_READ_BUFFER_SIZE = 256
*/
enum  {
  __pyx_e_6pysoem_6pysoem_EC_TIMEOUTRXM = 0xAAE60,
  __pyx_e_6pysoem_6pysoem_STATIC_SDO_READ_BUFFER_SIZE = 0x100
};

/* "pysoem/pysoem.pyx":152
 *     ECT_BIT8            = 0x0037
 * 
 * cdef struct CdefMasterSettings:             # <<<<<<<<<<<<<<
 *     int* sdo_read_timeout
 *     int* sdo_write_timeout
*/
struct __pyx_t_6pysoem_6pysoem_CdefMasterSettings {
  int *sdo_read_timeout;
  int *sdo_write_timeout;
};

/* "pysoem/pysoem.pyx":57
 * ECT_COEDET_SDOCA     = 0x20
 * 
 * cdef class CdefTimeouts:             # <<<<<<<<<<<<<<
 * 
 *     cdef cpysoem.Ttimeouts* _t
*/
struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts {
  PyObject_HEAD
  Ttimeouts *_t;
};


/* "pysoem/pysoem.pyx":112
 *         self._t.state = value
 * 
 * cdef class CdefSettings:             # <<<<<<<<<<<<<<
 * 
 *     cdef public CdefTimeouts timeouts
*/
struct __pyx_obj_6pysoem_6pysoem_CdefSettings {
  PyObject_HEAD
  struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *timeouts;
  boolean always_release_gil;
};


/* "pysoem/pysoem.pyx":219
 *     EC_IOMAPSIZE = 4096
 * 
 * cdef class CdefMaster:             # <<<<<<<<<<<<<<
 *     """Representing a logical EtherCAT master device.
 * 
*/
struct __pyx_obj_6pysoem_6pysoem_CdefMaster {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *__pyx_vtab;
  ec_slavet _ec_slave[__pyx_e_6pysoem_6pysoem_EC_MAXSLAVE];
  int _ec_slavecount;
  ec_groupt _ec_group[__pyx_e_6pysoem_6pysoem_EC_MAXGROUP];
  uint8 _ec_esibuf[__pyx_e_6pysoem_6pysoem_EC_MAXEEPBUF];
  uint32 _ec_esimap[__pyx_e_6pysoem_6pysoem_EC_MAXEEPBITMAP];
  ec_eringt _ec_elist;
  ec_idxstackT _ec_idxstack;
  ec_SMcommtypet _ec_SMcommtype[__pyx_e_6pysoem_6pysoem_EC_MAXMAPT];
  ec_PDOassignt _ec_PDOassign[__pyx_e_6pysoem_6pysoem_EC_MAXMAPT];
  ec_PDOdesct _ec_PDOdesc[__pyx_e_6pysoem_6pysoem_EC_MAXMAPT];
  ec_eepromSMt _ec_SM;
  ec_eepromFMMUt _ec_FMMU;
  boolean _EcatError;
  int64 _ec_DCtime;
  ecx_portt _ecx_port;
  ecx_redportt _ecx_redport;
  ecx_contextt _ecx_contextt;
  char io_map[__pyx_e_6pysoem_6pysoem_EC_IOMAPSIZE];
  struct __pyx_t_6pysoem_6pysoem_CdefMasterSettings _settings;
  int sdo_read_timeout;
  int sdo_write_timeout;
  boolean always_release_gil;
  boolean context_initialized;
};


/* "pysoem/pysoem.pyx":762
 * 
 * 
 * cdef class _CallbackData:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object slave
*/
struct __pyx_obj_6pysoem_6pysoem__CallbackData {
  PyObject_HEAD
  PyObject *slave;
  PyObject *func;
  PyObject *exc_raised;
  PyObject *exc_info;
};


/* "pysoem/pysoem.pyx":788
 * 
 * 
 * cdef class CdefSlave:             # <<<<<<<<<<<<<<
 *     """Represents a slave device
 * 
*/
struct __pyx_obj_6pysoem_6pysoem_CdefSlave {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *__pyx_vtab;
  struct __pyx_obj_6pysoem_6pysoem_CdefMaster *_master;
  ecx_contextt *_ecx_contextt;
  ec_slavet *_ec_slave;
  struct __pyx_t_6pysoem_6pysoem_CdefMasterSettings *_the_masters_settings;
  int _pos;
  struct __pyx_obj_6pysoem_6pysoem__CallbackData *_cd;
  ec_ODlistt _ex_odlist;
  PyObject *_emcy_callbacks;
};


/* "pysoem/pysoem.pyx":1503
 * 
 * 
 * cdef class CdefCoeObject:             # <<<<<<<<<<<<<<
 *     """Object info for objects in the object dictionary.
 * 
*/
struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject {
  PyObject_HEAD
  ecx_contextt *_ecx_context;
  ec_ODlistt *_ex_odlist;
  int _item;
  boolean _is_description_read;
  boolean _are_entries_read;
  ec_OElistt _ex_oelist;
};


/* "pysoem/pysoem.pyx":1596
 * 
 * 
 * cdef class CdefCoeObjectEntry:             # <<<<<<<<<<<<<<
 *     cdef cpysoem.ec_OElistt* _ex_oelist
 *     cdef int _item
*/
struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry {
  PyObject_HEAD
  ec_OElistt *_ex_oelist;
  int _item;
};


/* "pysoem/pysoem.pyx":172
 * 
 * 
 * @contextlib.contextmanager             # <<<<<<<<<<<<<<
 * def open(ifname):
 *     """Context manager function to create a Master object.
*/
struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open {
  PyObject_HEAD
  PyObject *__pyx_v_ifname;
  PyObject *__pyx_v_master;
};


/* "pysoem/pysoem.pyx":667
 *     def __str__(self):
 *         b1w1w2_bytes = bytes([self.b1]) + self.w1.to_bytes(length=2, byteorder='little') + self.w2.to_bytes(length=2, byteorder='little')
 *         b1w1w2_str = ','.join(format(x, '02x') for x in b1w1w2_bytes)             # <<<<<<<<<<<<<<
 *         return f'Slave {self.slave_pos}:  {self.error_code:04x}, {self.error_reg:02x}, ({b1w1w2_str})'
 * 
*/
struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_x;
};



/* "pysoem/pysoem.pyx":219
 *     EC_IOMAPSIZE = 4096
 * 
 * cdef class CdefMaster:             # <<<<<<<<<<<<<<
 *     """Representing a logical EtherCAT master device.
 * 
*/

struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster {
  int (*_CdefMaster__config_init_nogil)(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *, uint8_t);
  boolean (*check_release_gil)(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *, PyObject *, int __pyx_skip_dispatch);
  int (*_CdefMaster__send_processdata_nogil)(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *);
  int (*_CdefMaster__send_overlap_processdata_nogil)(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *);
  int (*_CdefMaster__receive_processdata_nogil)(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *, int);
};
static struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *__pyx_vtabptr_6pysoem_6pysoem_CdefMaster;


/* "pysoem/pysoem.pyx":788
 * 
 * 
 * cdef class CdefSlave:             # <<<<<<<<<<<<<<
 *     """Represents a slave device
 * 
*/

struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave {
  int (*_CdefSlave__sdo_read_nogil)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, uint16_t, uint8_t, int8_t, int *, unsigned char *);
  int (*_CdefSlave__sdo_write_nogil)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, uint16_t, uint8_t, int8_t, int, PyObject *);
  int (*_CdefSlave__foe_write_nogil)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, PyObject *, uint32_t, int, PyObject *, int);
  int (*_CdefSlave__foe_read_nogil)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, PyObject *, uint32_t, int, unsigned char *, int);
  PyObject *(*_on_emergency)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, ec_errort *);
  PyObject *(*_raise_exception)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, ec_errort *);
};
static struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *__pyx_vtabptr_6pysoem_6pysoem_CdefSlave;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by BytesEquals) */
#include <string.h>

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto (used by PyObjectCallOneArg) */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by RejectKeywords) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* decode_c_string_utf16.proto (used by decode_c_string) */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* PyObjectDelAttr.proto (used by PyObjectSetAttrStr) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyConnectionError_Check.proto */
#define __Pyx_PyExc_ConnectionError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ConnectionError)

/* PyObjectCall2Args.proto (used by PyObjectCallMethod1) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod1) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod1.proto (used by append) */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* PyIndexError_Check.proto */
#define __Pyx_PyExc_IndexError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_IndexError)

/* RaiseUnboundLocalError.proto */
static void __Pyx_RaiseUnboundLocalError(const char *varname);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectVectorCallKwBuilder.proto (used by PyObjectVectorCallMethodKwBuilder) */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectVectorCallMethodKwBuilder.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_VectorcallMethod_CallFromBuilder PyObject_VectorcallMethod
#else
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#endif

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* PyObjectFormat.proto */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* s, PyObject* f);
#else
#define __Pyx_PyObject_Format(s, f) PyObject_Format(s, f)
#endif

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* JoinPyUnicode.export */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyUnboundLocalError_Check.proto */
#define __Pyx_PyExc_UnboundLocalError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_UnboundLocalError)

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_LIMITED_API  ||  PY_VERSION_HEX >= 0x030C0000
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  #if __clang__ || __GNUC__
  __attribute__((no_sanitize("thread")))
  #endif
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* PyAssertionError_Check.proto */
#define __Pyx_PyExc_AssertionError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_AssertionError)

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetTopmostException.proto (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyAttributeError_Check.proto */
#define __Pyx_PyExc_AttributeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_AttributeError)

/* py_dict_keys.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_DivideObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_DivideObjC(op1, op2, floatval, inplace, zerodivision_check)\
    ((inplace ? __Pyx_PyNumber_InPlaceDivide(op1, op2) : __Pyx_PyNumber_Divide(op1, op2)))
    #endif

/* PyUserWarning_Check.proto */
#define __Pyx_PyExc_UserWarning_Check(obj)  __Pyx_TypeCheck(obj, PyExc_UserWarning)

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_MultiplyCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_MultiplyCObj(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceMultiply(op1, op2) : PyNumber_Multiply(op1, op2))
#endif

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyFutureWarning_Check.proto */
#define __Pyx_PyExc_FutureWarning_Check(obj)  __Pyx_TypeCheck(obj, PyExc_FutureWarning)

/* decode_c_bytes.proto (used by decode_bytes) */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    as_c_string = PyBytes_AS_STRING(string);
    size = PyBytes_GET_SIZE(string);
#else
    if (PyBytes_AsStringAndSize(string, &as_c_string, &size) < 0) {
        return NULL;
    }
#endif
    return __Pyx_decode_c_bytes(
        as_c_string, size,
        start, stop, encoding, errors, decode_func);
}

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* AllocateExtensionType.proto */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* CheckTypeForFreelists.proto */
#if CYTHON_USE_FREELISTS
#if CYTHON_USE_TYPE_SPECS
#define __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, expected_tp, expected_size) ((int) ((t) == (expected_tp)))
#define __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS  Py_TPFLAGS_IS_ABSTRACT
#else
#define __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, expected_tp, expected_size) ((int) ((t)->tp_basicsize == (expected_size)))
#define __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS  (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)
#endif
#define __PYX_CHECK_TYPE_FOR_FREELISTS(t, expected_tp, expected_size)\
    (__PYX_CHECK_FINAL_TYPE_FOR_FREELISTS((t), (expected_tp), (expected_size)) &\
     (int) (!__Pyx_PyType_HasFeature((t), __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS)))
#endif

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectCallMethod0.proto (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ListPack.proto */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...);

/* GetAttr.proto (used by Globals) */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* PyObjectLookupSpecial.proto (used by Py3ClassCreate) */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CheckUnpickleChecksum.proto */
static CYTHON_INLINE int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int8_t(int8_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int8_t __Pyx_PyLong_As_int8_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_ec_state(ec_state value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyLong_As_uint8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint16_t __Pyx_PyLong_As_uint16_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint16_t(uint16_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint8_t(uint8_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int32_t(int32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int64_t(int64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int32_t __Pyx_PyLong_As_int32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_ec_err_type(ec_err_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint32_t(uint32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(enum __pyx_t_6pysoem_6pysoem_ec_datatype value);

/* UpdateUnpickledDict.proto */
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* IterNextPlain.proto (used by CoroutineBase) */
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next_Plain(PyObject *iterator);
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *__Pyx_GetBuiltinNext_LimitedAPI(void);
#endif

/* ReturnWithStopIteration.proto (used by CoroutineBase) */
static CYTHON_INLINE void __Pyx_ReturnWithStopIteration(PyObject* value, int async, int iternext);

/* CoroutineBase.proto (used by Generator) */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *gi_weakreflist;
#endif
    PyObject *classobj;
    PyObject *yieldfrom;
    __Pyx_pyiter_sendfunc yieldfrom_am_send;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
#if CYTHON_USE_SYS_MONITORING && (CYTHON_PROFILE || CYTHON_TRACE)
    PyMonitoringState __pyx_pymonitoring_state[__Pyx_MonitoringEventTypes_CyGen_count];
    uint64_t __pyx_pymonitoring_version;
#endif
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static __Pyx_PySendResult __Pyx_Coroutine_AmSend(PyObject *self, PyObject *value, PyObject **retval);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static __Pyx_PySendResult __Pyx_Coroutine_Close(PyObject *self, PyObject **retval);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);
static char __Pyx_Coroutine_test_and_set_is_running(__pyx_CoroutineObject *gen);
static void __Pyx_Coroutine_unset_is_running(__pyx_CoroutineObject *gen);
static char __Pyx_Coroutine_get_is_running(__pyx_CoroutineObject *gen);
static PyObject *__Pyx_Coroutine_get_is_running_getter(PyObject *gen, void *closure);
#if __PYX_HAS_PY_AM_SEND == 2
static void __Pyx_SetBackportTypeAmSend(PyTypeObject *type, __Pyx_PyAsyncMethodsStruct *static_amsend_methods, __Pyx_pyiter_sendfunc am_send);
#endif
static PyObject *__Pyx_Coroutine_fail_reduce_ex(PyObject *self, PyObject *arg);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_mstate_global->__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);
static CYTHON_INLINE PyObject *__Pyx_Generator_GetInlinedResult(PyObject *self);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static int __pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__config_init_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, uint8_t __pyx_v_usetable); /* proto*/
static boolean __pyx_f_6pysoem_6pysoem_10CdefMaster_check_release_gil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_release_gil, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__send_processdata_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto*/
static int __pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__send_overlap_processdata_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto*/
static int __pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__receive_processdata_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, int __pyx_v_timeout); /* proto*/
static int __pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__sdo_read_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, uint16_t __pyx_v_index, uint8_t __pyx_v_subindex, int8_t __pyx_v_ca, int *__pyx_v_size_inout, unsigned char *__pyx_v_pbuf); /* proto*/
static int __pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__sdo_write_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, uint16_t __pyx_v_index, uint8_t __pyx_v_subindex, int8_t __pyx_v_ca, int __pyx_v_size, PyObject *__pyx_v_data); /* proto*/
static int __pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__foe_write_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_filename, uint32_t __pyx_v_password, int __pyx_v_size, PyObject *__pyx_v_data, int __pyx_v_timeout); /* proto*/
static int __pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__foe_read_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_filename, uint32_t __pyx_v_password, int __pyx_v_size_inout, unsigned char *__pyx_v_pbuf, int __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_6pysoem_6pysoem_9CdefSlave__on_emergency(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, ec_errort *__pyx_v_emcy); /* proto*/
static PyObject *__pyx_f_6pysoem_6pysoem_9CdefSlave__raise_exception(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, ec_errort *__pyx_v_err); /* proto*/

/* Module declarations from "libc.stdint" */

/* Module declarations from "pysoem.cpysoem" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "pysoem.pysoem" */
static PyObject *__Pyx_EnumBase = 0;
static PyObject *__Pyx_FlagBase = 0;
static int __pyx_f_6pysoem_6pysoem__xPO2SOconfig(uint16, void *); /*proto*/
static int __pyx_f_6pysoem_6pysoem__xPO2SOconfigEx(uint16, void *); /*proto*/
static PyObject *__pyx_f_6pysoem_6pysoem___pyx_unpickle_CdefSettings__set_state(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pysoem_6pysoem___pyx_unpickle__CallbackData__set_state(struct __pyx_obj_6pysoem_6pysoem__CallbackData *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pysoem.pysoem"
extern int __pyx_module_is_main_pysoem__pysoem;
int __pyx_module_is_main_pysoem__pysoem = 0;

/* Implementation of "pysoem.pysoem" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_property;
/* #### Code section: string_decls ### */
static const char __pyx_k_always_release_gil_timeouts[] = "always_release_gil, timeouts";
static const char __pyx_k_exc_info_exc_raised_func_slave[] = "exc_info, exc_raised, func, slave";
static const char __pyx_k_PySOEM_is_a_Cython_wrapper_for_t[] = "PySOEM is a Cython wrapper for the SOEM library.";
/* #### Code section: decls ### */
static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts___cinit__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_3ret___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_3ret_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_4safe___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_4safe_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_6eeprom___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_6eeprom_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_5state___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_5state_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pysoem_6pysoem_12CdefSettings___init__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefSettings_8timeouts___get__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_12CdefSettings_8timeouts_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6pysoem_6pysoem_12CdefSettings_8timeouts_4__del__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefSettings_18always_release_gil___get__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_12CdefSettings_18always_release_gil_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefSettings_2__reduce_cython__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefSettings_4__setstate_cython__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_find_adapters(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_2open(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ifname); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_5al_status_code_to_string(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code); /* proto */
static int __pyx_pf_6pysoem_6pysoem_10CdefMaster___cinit__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_2open(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_ifname, PyObject *__pyx_v_ifname_red); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_4check_context_is_initialized(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_6check_release_gil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_release_gil); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_8config_init(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_usetable, PyObject *__pyx_v_release_gil); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_10config_map(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_12config_overlap_map(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_14_collect_mailbox_errors(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_16config_dc(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_18close(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_20read_state(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_22write_state(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_24state_check(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, int __pyx_v_expected_state, PyObject *__pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_26send_processdata(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_release_gil); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_28send_overlap_processdata(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_release_gil); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_30receive_processdata(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_timeout, PyObject *__pyx_v_release_gil); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_32_get_slave(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, int __pyx_v_pos); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_34_get_state(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_36_set_state(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_38_get_expected_wkc(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_40_get_dc_time(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_42_set_manual_state_change(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, int __pyx_v_manual_state_change); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_44_get_manual_state_change(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout___get__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout___get__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_18always_release_gil___get__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_10CdefMaster_18always_release_gil_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_19context_initialized___get__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_46__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_48__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_8SdoError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_slave_pos, PyObject *__pyx_v_index, PyObject *__pyx_v_subindex, PyObject *__pyx_v_abort_code, PyObject *__pyx_v_desc); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9Emergency___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_slave_pos, PyObject *__pyx_v_error_code, PyObject *__pyx_v_error_reg, PyObject *__pyx_v_b1, PyObject *__pyx_v_w1, PyObject *__pyx_v_w2); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9Emergency_7__str___genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9Emergency_2__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12SdoInfoError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_message); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_12MailboxError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_slave_pos, PyObject *__pyx_v_error_code, PyObject *__pyx_v_desc); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_11PacketError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_slave_pos, PyObject *__pyx_v_error_code); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_11PacketError_2_get_desc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_14ConfigMapError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_error_list); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_11EepromError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_message); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_8WkcError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_wkc); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13_CallbackData___reduce_cython__(struct __pyx_obj_6pysoem_6pysoem__CallbackData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13_CallbackData_2__setstate_cython__(struct __pyx_obj_6pysoem_6pysoem__CallbackData *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pysoem_6pysoem_9CdefSlave___init__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_pos); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_2dc_sync(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_act, PyObject *__pyx_v_sync0_cycle_time, PyObject *__pyx_v_sync0_shift_time, PyObject *__pyx_v_sync1_cycle_time); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_4sdo_read(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_index, uint8_t __pyx_v_subindex, int __pyx_v_size, PyObject *__pyx_v_ca, PyObject *__pyx_v_release_gil); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_6sdo_write(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_index, uint8_t __pyx_v_subindex, PyObject *__pyx_v_data, PyObject *__pyx_v_ca, PyObject *__pyx_v_release_gil); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_8mbx_receive(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_10write_state(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_12state_check(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, int __pyx_v_expected_state, PyObject *__pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_14reconfig(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_16recover(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_18eeprom_read(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, int __pyx_v_word_address, PyObject *__pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_20eeprom_write(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, int __pyx_v_word_address, PyObject *__pyx_v_data, PyObject *__pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_22foe_write(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_password, PyObject *__pyx_v_data, PyObject *__pyx_v_timeout, PyObject *__pyx_v_release_gil); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_24foe_read(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_password, PyObject *__pyx_v_size, PyObject *__pyx_v_timeout, PyObject *__pyx_v_release_gil); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_26amend_mbx(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_mailbox, PyObject *__pyx_v_start_address, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_28set_watchdog(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_wd_type, PyObject *__pyx_v_wd_time_ms); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_30get_watchdog(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_wd_type); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_32_get_watchdog_divider_ns(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_34get_max_watchdog_time(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_36add_emergency_callback(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_callback); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_38_disable_complete_access(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_40_fprd(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, int __pyx_v_address, int __pyx_v_size, PyObject *__pyx_v_timeout_us); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_42_fpwr(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, int __pyx_v_address, PyObject *__pyx_v_data, PyObject *__pyx_v_timeout_us); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_44_get_name(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_46_get_eep_man(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_48_get_eep_id(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_50_get_eep_rev(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_52_get_PO2SOconfig(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_54_get_PO2SOconfigEx(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_56_set_PO2SOconfig(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_58_set_PO2SOconfigEx(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_60_get_state(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_62_set_state(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_64_get_input(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_66_get_output(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_68_set_output(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_70_get_al_status(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_72_get_is_lost(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_74_set_is_lost(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_76_get_od(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_7_master___get__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_3_cd___get__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_9CdefSlave_3_cd_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6pysoem_6pysoem_9CdefSlave_3_cd_4__del__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks___get__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static int __pyx_pf_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_4__del__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_78__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_80__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pysoem_6pysoem_13CdefCoeObject___init__(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self, int __pyx_v_item); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_2_read_description(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_4_read_entries(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_6_get_index(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_8_get_data_type(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_10_get_object_code(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_12_get_name(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_14_get_entries(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_16_get_bit_length(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_18_get_obj_access(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry___init__(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self, int __pyx_v_item); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_2_get_name(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_4_get_data_type(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_6_get_bit_length(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_8_get_obj_access(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_7__pyx_unpickle_CdefSettings(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pysoem_6pysoem_9__pyx_unpickle__CallbackData(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefTimeouts(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefSettings(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefMaster(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pysoem_6pysoem__CallbackData(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefSlave(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefCoeObject(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefCoeObjectEntry(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pysoem_6pysoem___pyx_scope_struct__open(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pysoem_6pysoem___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyObject *__pyx_type_6pysoem_6pysoem_CdefTimeouts;
  PyObject *__pyx_type_6pysoem_6pysoem_CdefSettings;
  PyObject *__pyx_type_6pysoem_6pysoem_CdefMaster;
  PyObject *__pyx_type_6pysoem_6pysoem__CallbackData;
  PyObject *__pyx_type_6pysoem_6pysoem_CdefSlave;
  PyObject *__pyx_type_6pysoem_6pysoem_CdefCoeObject;
  PyObject *__pyx_type_6pysoem_6pysoem_CdefCoeObjectEntry;
  PyObject *__pyx_type_6pysoem_6pysoem___pyx_scope_struct__open;
  PyObject *__pyx_type_6pysoem_6pysoem___pyx_scope_struct_1_genexpr;
  PyTypeObject *__pyx_ptype_6pysoem_6pysoem_CdefTimeouts;
  PyTypeObject *__pyx_ptype_6pysoem_6pysoem_CdefSettings;
  PyTypeObject *__pyx_ptype_6pysoem_6pysoem_CdefMaster;
  PyTypeObject *__pyx_ptype_6pysoem_6pysoem__CallbackData;
  PyTypeObject *__pyx_ptype_6pysoem_6pysoem_CdefSlave;
  PyTypeObject *__pyx_ptype_6pysoem_6pysoem_CdefCoeObject;
  PyTypeObject *__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry;
  PyTypeObject *__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open;
  PyTypeObject *__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_keys;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  PyObject *__pyx_tuple[11];
  PyObject *__pyx_codeobj_tab[103];
  PyObject *__pyx_string_tab[567];
  PyObject *__pyx_number_tab[29];
/* #### Code section: module_state_contents ### */

#if CYTHON_USE_FREELISTS
struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *__pyx_freelist_6pysoem_6pysoem___pyx_scope_struct__open[8];
int __pyx_freecount_6pysoem_6pysoem___pyx_scope_struct__open;
#endif

#if CYTHON_USE_FREELISTS
struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *__pyx_freelist_6pysoem_6pysoem___pyx_scope_struct_1_genexpr[8];
int __pyx_freecount_6pysoem_6pysoem___pyx_scope_struct_1_genexpr;
#endif
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* IterNextPlain.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
PyObject *__Pyx_GetBuiltinNext_LimitedAPI_cache;
#endif

/* Generator.module_state_decls */
PyTypeObject *__pyx_GeneratorType;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u_02x __pyx_string_tab[1]
#define __pyx_kp_u_04x __pyx_string_tab[2]
#define __pyx_kp_u_Data_container_too_small_for_typ __pyx_string_tab[3]
#define __pyx_kp_u_EEPROM_access_error_Attributes_m __pyx_string_tab[4]
#define __pyx_kp_u_EEPROM_write_error __pyx_string_tab[5]
#define __pyx_kp_u_Emergency_message_Attributes_sla __pyx_string_tab[6]
#define __pyx_kp_u_Error_when_a_master_or_slave_met __pyx_string_tab[7]
#define __pyx_kp_u_Errors_during_Object_directory_i __pyx_string_tab[8]
#define __pyx_kp_u_Errors_during_Object_directory_i_2 __pyx_string_tab[9]
#define __pyx_kp_u_Errors_in_mailbox_communication __pyx_string_tab[10]
#define __pyx_kp_u_Errors_related_to_mailbox_commun __pyx_string_tab[11]
#define __pyx_kp_u_Item_offsets_in_SII_general_sect __pyx_string_tab[12]
#define __pyx_kp_u_None __pyx_string_tab[13]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[14]
#define __pyx_kp_u_Representing_a_logical_EtherCAT __pyx_string_tab[15]
#define __pyx_kp_u_SOEM_Network_interface_is_not_in __pyx_string_tab[16]
#define __pyx_kp_u_Sdo_List_Info_read_failed __pyx_string_tab[17]
#define __pyx_kp_u_Sdo_ObjectEntry_Info_read_failed __pyx_string_tab[18]
#define __pyx_kp_u_Sdo_Object_Info_read_failed __pyx_string_tab[19]
#define __pyx_kp_u_Sdo_read_or_write_abort_Attribut __pyx_string_tab[20]
#define __pyx_kp_u_Slave __pyx_string_tab[21]
#define __pyx_kp_u_The_actual_set_watchdog_time __pyx_string_tab[22]
#define __pyx_kp_u_This_way_of_catching_emergency_m __pyx_string_tab[23]
#define __pyx_kp_u_Unexpected_frame_returned __pyx_string_tab[24]
#define __pyx_kp_u_Working_counter_error_Attributes __pyx_string_tab[25]
#define __pyx_kp_u__2 __pyx_string_tab[26]
#define __pyx_kp_u__3 __pyx_string_tab[27]
#define __pyx_kp_u__4 __pyx_string_tab[28]
#define __pyx_kp_u__5 __pyx_string_tab[29]
#define __pyx_kp_u__6 __pyx_string_tab[30]
#define __pyx_kp_u__7 __pyx_string_tab[31]
#define __pyx_kp_u_add_note __pyx_string_tab[32]
#define __pyx_kp_u_could_not_open_interface __pyx_string_tab[33]
#define __pyx_kp_u_disable __pyx_string_tab[34]
#define __pyx_kp_u_ecx_readODdescription __pyx_string_tab[35]
#define __pyx_kp_u_ecx_readODlist __pyx_string_tab[36]
#define __pyx_kp_u_ecx_readOE __pyx_string_tab[37]
#define __pyx_kp_u_enable __pyx_string_tab[38]
#define __pyx_kp_u_gc __pyx_string_tab[39]
#define __pyx_kp_u_io_map_size __pyx_string_tab[40]
#define __pyx_kp_u_isenabled __pyx_string_tab[41]
#define __pyx_kp_u_ms_differs_from_the_requested_w __pyx_string_tab[42]
#define __pyx_kp_u_ms_due_to_resolution_limits_of __pyx_string_tab[43]
#define __pyx_kp_u_no_default___reduce___due_to_non __pyx_string_tab[44]
#define __pyx_kp_u_pysoem_collections __pyx_string_tab[45]
#define __pyx_kp_u_pysoem_contextlib __pyx_string_tab[46]
#define __pyx_kp_u_pysoem_logging __pyx_string_tab[47]
#define __pyx_kp_u_pysoem_pysoem __pyx_string_tab[48]
#define __pyx_kp_u_pysoem_sys __pyx_string_tab[49]
#define __pyx_kp_u_pysoem_time __pyx_string_tab[50]
#define __pyx_kp_u_pysoem_warnings __pyx_string_tab[51]
#define __pyx_kp_u_requested_slave_device_is_not_av __pyx_string_tab[52]
#define __pyx_kp_u_self__ec_slave_self__ecx_context __pyx_string_tab[53]
#define __pyx_kp_u_self__ecx_context_self__ex_odlis __pyx_string_tab[54]
#define __pyx_kp_u_self__ex_oelist_cannot_be_conver __pyx_string_tab[55]
#define __pyx_kp_u_src_pysoem_pysoem_pyx __pyx_string_tab[56]
#define __pyx_kp_u_stringsource __pyx_string_tab[57]
#define __pyx_kp_u_unexpected_error __pyx_string_tab[58]
#define __pyx_kp_u_unexpected_error_Etype __pyx_string_tab[59]
#define __pyx_kp_u_wd_time_ms_is_limited_to_ms __pyx_string_tab[60]
#define __pyx_n_u_Adapter __pyx_string_tab[61]
#define __pyx_n_u_BOOT_RX_MBX __pyx_string_tab[62]
#define __pyx_n_u_BOOT_STATE __pyx_string_tab[63]
#define __pyx_n_u_BOOT_TX_MBX __pyx_string_tab[64]
#define __pyx_n_u_CallbackData __pyx_string_tab[65]
#define __pyx_n_u_CallbackData___reduce_cython __pyx_string_tab[66]
#define __pyx_n_u_CallbackData___setstate_cython __pyx_string_tab[67]
#define __pyx_n_u_CdefCoeObject __pyx_string_tab[68]
#define __pyx_n_u_CdefCoeObjectEntry __pyx_string_tab[69]
#define __pyx_n_u_CdefCoeObjectEntry___reduce_cyth __pyx_string_tab[70]
#define __pyx_n_u_CdefCoeObjectEntry___setstate_cy __pyx_string_tab[71]
#define __pyx_n_u_CdefCoeObjectEntry__get_bit_leng __pyx_string_tab[72]
#define __pyx_n_u_CdefCoeObjectEntry__get_data_typ __pyx_string_tab[73]
#define __pyx_n_u_CdefCoeObjectEntry__get_name __pyx_string_tab[74]
#define __pyx_n_u_CdefCoeObjectEntry__get_obj_acce __pyx_string_tab[75]
#define __pyx_n_u_CdefCoeObject___reduce_cython __pyx_string_tab[76]
#define __pyx_n_u_CdefCoeObject___setstate_cython __pyx_string_tab[77]
#define __pyx_n_u_CdefCoeObject__get_bit_length __pyx_string_tab[78]
#define __pyx_n_u_CdefCoeObject__get_data_type __pyx_string_tab[79]
#define __pyx_n_u_CdefCoeObject__get_entries __pyx_string_tab[80]
#define __pyx_n_u_CdefCoeObject__get_index __pyx_string_tab[81]
#define __pyx_n_u_CdefCoeObject__get_name __pyx_string_tab[82]
#define __pyx_n_u_CdefCoeObject__get_obj_access __pyx_string_tab[83]
#define __pyx_n_u_CdefCoeObject__get_object_code __pyx_string_tab[84]
#define __pyx_n_u_CdefCoeObject__read_description __pyx_string_tab[85]
#define __pyx_n_u_CdefCoeObject__read_entries __pyx_string_tab[86]
#define __pyx_n_u_CdefMaster __pyx_string_tab[87]
#define __pyx_n_u_CdefMaster___reduce_cython __pyx_string_tab[88]
#define __pyx_n_u_CdefMaster___setstate_cython __pyx_string_tab[89]
#define __pyx_n_u_CdefMaster__collect_mailbox_erro __pyx_string_tab[90]
#define __pyx_n_u_CdefMaster__get_dc_time __pyx_string_tab[91]
#define __pyx_n_u_CdefMaster__get_expected_wkc __pyx_string_tab[92]
#define __pyx_n_u_CdefMaster__get_manual_state_cha __pyx_string_tab[93]
#define __pyx_n_u_CdefMaster__get_slave __pyx_string_tab[94]
#define __pyx_n_u_CdefMaster__get_state __pyx_string_tab[95]
#define __pyx_n_u_CdefMaster__set_manual_state_cha __pyx_string_tab[96]
#define __pyx_n_u_CdefMaster__set_state __pyx_string_tab[97]
#define __pyx_n_u_CdefMaster_check_context_is_init __pyx_string_tab[98]
#define __pyx_n_u_CdefMaster_check_release_gil __pyx_string_tab[99]
#define __pyx_n_u_CdefMaster_close __pyx_string_tab[100]
#define __pyx_n_u_CdefMaster_config_dc __pyx_string_tab[101]
#define __pyx_n_u_CdefMaster_config_init __pyx_string_tab[102]
#define __pyx_n_u_CdefMaster_config_map __pyx_string_tab[103]
#define __pyx_n_u_CdefMaster_config_overlap_map __pyx_string_tab[104]
#define __pyx_n_u_CdefMaster_open __pyx_string_tab[105]
#define __pyx_n_u_CdefMaster_read_state __pyx_string_tab[106]
#define __pyx_n_u_CdefMaster_receive_processdata __pyx_string_tab[107]
#define __pyx_n_u_CdefMaster_send_overlap_processd __pyx_string_tab[108]
#define __pyx_n_u_CdefMaster_send_processdata __pyx_string_tab[109]
#define __pyx_n_u_CdefMaster_state_check __pyx_string_tab[110]
#define __pyx_n_u_CdefMaster_write_state __pyx_string_tab[111]
#define __pyx_n_u_CdefSettings __pyx_string_tab[112]
#define __pyx_n_u_CdefSettings___reduce_cython __pyx_string_tab[113]
#define __pyx_n_u_CdefSettings___setstate_cython __pyx_string_tab[114]
#define __pyx_n_u_CdefSlave __pyx_string_tab[115]
#define __pyx_n_u_CdefSlave___reduce_cython __pyx_string_tab[116]
#define __pyx_n_u_CdefSlave___setstate_cython __pyx_string_tab[117]
#define __pyx_n_u_CdefSlave__disable_complete_acce __pyx_string_tab[118]
#define __pyx_n_u_CdefSlave__fprd __pyx_string_tab[119]
#define __pyx_n_u_CdefSlave__fpwr __pyx_string_tab[120]
#define __pyx_n_u_CdefSlave__get_PO2SOconfig __pyx_string_tab[121]
#define __pyx_n_u_CdefSlave__get_PO2SOconfigEx __pyx_string_tab[122]
#define __pyx_n_u_CdefSlave__get_al_status __pyx_string_tab[123]
#define __pyx_n_u_CdefSlave__get_eep_id __pyx_string_tab[124]
#define __pyx_n_u_CdefSlave__get_eep_man __pyx_string_tab[125]
#define __pyx_n_u_CdefSlave__get_eep_rev __pyx_string_tab[126]
#define __pyx_n_u_CdefSlave__get_input __pyx_string_tab[127]
#define __pyx_n_u_CdefSlave__get_is_lost __pyx_string_tab[128]
#define __pyx_n_u_CdefSlave__get_name __pyx_string_tab[129]
#define __pyx_n_u_CdefSlave__get_od __pyx_string_tab[130]
#define __pyx_n_u_CdefSlave__get_output __pyx_string_tab[131]
#define __pyx_n_u_CdefSlave__get_state __pyx_string_tab[132]
#define __pyx_n_u_CdefSlave__get_watchdog_divider __pyx_string_tab[133]
#define __pyx_n_u_CdefSlave__set_PO2SOconfig __pyx_string_tab[134]
#define __pyx_n_u_CdefSlave__set_PO2SOconfigEx __pyx_string_tab[135]
#define __pyx_n_u_CdefSlave__set_is_lost __pyx_string_tab[136]
#define __pyx_n_u_CdefSlave__set_output __pyx_string_tab[137]
#define __pyx_n_u_CdefSlave__set_state __pyx_string_tab[138]
#define __pyx_n_u_CdefSlave_add_emergency_callback __pyx_string_tab[139]
#define __pyx_n_u_CdefSlave_amend_mbx __pyx_string_tab[140]
#define __pyx_n_u_CdefSlave_dc_sync __pyx_string_tab[141]
#define __pyx_n_u_CdefSlave_eeprom_read __pyx_string_tab[142]
#define __pyx_n_u_CdefSlave_eeprom_write __pyx_string_tab[143]
#define __pyx_n_u_CdefSlave_foe_read __pyx_string_tab[144]
#define __pyx_n_u_CdefSlave_foe_write __pyx_string_tab[145]
#define __pyx_n_u_CdefSlave_get_max_watchdog_time __pyx_string_tab[146]
#define __pyx_n_u_CdefSlave_get_watchdog __pyx_string_tab[147]
#define __pyx_n_u_CdefSlave_mbx_receive __pyx_string_tab[148]
#define __pyx_n_u_CdefSlave_reconfig __pyx_string_tab[149]
#define __pyx_n_u_CdefSlave_recover __pyx_string_tab[150]
#define __pyx_n_u_CdefSlave_sdo_read __pyx_string_tab[151]
#define __pyx_n_u_CdefSlave_sdo_write __pyx_string_tab[152]
#define __pyx_n_u_CdefSlave_set_watchdog __pyx_string_tab[153]
#define __pyx_n_u_CdefSlave_state_check __pyx_string_tab[154]
#define __pyx_n_u_CdefSlave_write_state __pyx_string_tab[155]
#define __pyx_n_u_CdefTimeouts __pyx_string_tab[156]
#define __pyx_n_u_CdefTimeouts___reduce_cython __pyx_string_tab[157]
#define __pyx_n_u_CdefTimeouts___setstate_cython __pyx_string_tab[158]
#define __pyx_n_u_ConfigMapError __pyx_string_tab[159]
#define __pyx_n_u_ConfigMapError___init __pyx_string_tab[160]
#define __pyx_n_u_ECT_BIT1 __pyx_string_tab[161]
#define __pyx_n_u_ECT_BIT2 __pyx_string_tab[162]
#define __pyx_n_u_ECT_BIT3 __pyx_string_tab[163]
#define __pyx_n_u_ECT_BIT4 __pyx_string_tab[164]
#define __pyx_n_u_ECT_BIT5 __pyx_string_tab[165]
#define __pyx_n_u_ECT_BIT6 __pyx_string_tab[166]
#define __pyx_n_u_ECT_BIT7 __pyx_string_tab[167]
#define __pyx_n_u_ECT_BIT8 __pyx_string_tab[168]
#define __pyx_n_u_ECT_BOOLEAN __pyx_string_tab[169]
#define __pyx_n_u_ECT_COEDET_PDOASSIGN __pyx_string_tab[170]
#define __pyx_n_u_ECT_COEDET_PDOCONFIG __pyx_string_tab[171]
#define __pyx_n_u_ECT_COEDET_SDO __pyx_string_tab[172]
#define __pyx_n_u_ECT_COEDET_SDOCA __pyx_string_tab[173]
#define __pyx_n_u_ECT_COEDET_SDOINFO __pyx_string_tab[174]
#define __pyx_n_u_ECT_COEDET_UPLOAD __pyx_string_tab[175]
#define __pyx_n_u_ECT_DOMAIN __pyx_string_tab[176]
#define __pyx_n_u_ECT_INTEGER16 __pyx_string_tab[177]
#define __pyx_n_u_ECT_INTEGER24 __pyx_string_tab[178]
#define __pyx_n_u_ECT_INTEGER32 __pyx_string_tab[179]
#define __pyx_n_u_ECT_INTEGER64 __pyx_string_tab[180]
#define __pyx_n_u_ECT_INTEGER8 __pyx_string_tab[181]
#define __pyx_n_u_ECT_OCTET_STRING __pyx_string_tab[182]
#define __pyx_n_u_ECT_REAL32 __pyx_string_tab[183]
#define __pyx_n_u_ECT_REAL64 __pyx_string_tab[184]
#define __pyx_n_u_ECT_REG_SM0 __pyx_string_tab[185]
#define __pyx_n_u_ECT_REG_SM1 __pyx_string_tab[186]
#define __pyx_n_u_ECT_REG_WD_DIV __pyx_string_tab[187]
#define __pyx_n_u_ECT_REG_WD_TIME_PDI __pyx_string_tab[188]
#define __pyx_n_u_ECT_REG_WD_TIME_PROCESSDATA __pyx_string_tab[189]
#define __pyx_n_u_ECT_TIME_DIFFERENCE __pyx_string_tab[190]
#define __pyx_n_u_ECT_TIME_OF_DAY __pyx_string_tab[191]
#define __pyx_n_u_ECT_UNICODE_STRING __pyx_string_tab[192]
#define __pyx_n_u_ECT_UNSIGNED16 __pyx_string_tab[193]
#define __pyx_n_u_ECT_UNSIGNED24 __pyx_string_tab[194]
#define __pyx_n_u_ECT_UNSIGNED32 __pyx_string_tab[195]
#define __pyx_n_u_ECT_UNSIGNED64 __pyx_string_tab[196]
#define __pyx_n_u_ECT_UNSIGNED8 __pyx_string_tab[197]
#define __pyx_n_u_ECT_VISIBLE_STRING __pyx_string_tab[198]
#define __pyx_n_u_EepromError __pyx_string_tab[199]
#define __pyx_n_u_EepromError___init __pyx_string_tab[200]
#define __pyx_n_u_Emergency __pyx_string_tab[201]
#define __pyx_n_u_Emergency___init __pyx_string_tab[202]
#define __pyx_n_u_Emergency___str __pyx_string_tab[203]
#define __pyx_n_u_Emergency___str___locals_genexpr __pyx_string_tab[204]
#define __pyx_n_u_ID __pyx_string_tab[205]
#define __pyx_n_u_INIT_STATE __pyx_string_tab[206]
#define __pyx_n_u_IntEnum __pyx_string_tab[207]
#define __pyx_n_u_IntFlag __pyx_string_tab[208]
#define __pyx_n_u_MAN __pyx_string_tab[209]
#define __pyx_n_u_MBX_PROTO __pyx_string_tab[210]
#define __pyx_n_u_MailboxError __pyx_string_tab[211]
#define __pyx_n_u_MailboxError___init __pyx_string_tab[212]
#define __pyx_n_u_Master __pyx_string_tab[213]
#define __pyx_n_u_NONE_STATE __pyx_string_tab[214]
#define __pyx_n_u_NetworkInterfaceNotOpenError __pyx_string_tab[215]
#define __pyx_n_u_OP_STATE __pyx_string_tab[216]
#define __pyx_n_u_PREOP_STATE __pyx_string_tab[217]
#define __pyx_n_u_PacketError __pyx_string_tab[218]
#define __pyx_n_u_PacketError___init __pyx_string_tab[219]
#define __pyx_n_u_PacketError__get_desc __pyx_string_tab[220]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[221]
#define __pyx_n_u_REV __pyx_string_tab[222]
#define __pyx_n_u_SAFEOP_STATE __pyx_string_tab[223]
#define __pyx_n_u_STATE_ACK __pyx_string_tab[224]
#define __pyx_n_u_STATE_ERROR __pyx_string_tab[225]
#define __pyx_n_u_STD_RX_MBX __pyx_string_tab[226]
#define __pyx_n_u_STD_TX_MBX __pyx_string_tab[227]
#define __pyx_n_u_SdoError __pyx_string_tab[228]
#define __pyx_n_u_SdoError___init __pyx_string_tab[229]
#define __pyx_n_u_SdoInfoError __pyx_string_tab[230]
#define __pyx_n_u_SdoInfoError___init __pyx_string_tab[231]
#define __pyx_n_u_SiiOffset __pyx_string_tab[232]
#define __pyx_n_u_WkcError __pyx_string_tab[233]
#define __pyx_n_u_WkcError___init __pyx_string_tab[234]
#define __pyx_n_u_abort_code __pyx_string_tab[235]
#define __pyx_n_u_act __pyx_string_tab[236]
#define __pyx_n_u_actual_wd_time_ms __pyx_string_tab[237]
#define __pyx_n_u_adapters __pyx_string_tab[238]
#define __pyx_n_u_add_emergency_callback __pyx_string_tab[239]
#define __pyx_n_u_address __pyx_string_tab[240]
#define __pyx_n_u_al_status __pyx_string_tab[241]
#define __pyx_n_u_al_status_code_to_string __pyx_string_tab[242]
#define __pyx_n_u_always_release_gil __pyx_string_tab[243]
#define __pyx_n_u_amend_mbx __pyx_string_tab[244]
#define __pyx_n_u_append __pyx_string_tab[245]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[246]
#define __pyx_n_u_b1 __pyx_string_tab[247]
#define __pyx_n_u_b1w1w2_bytes __pyx_string_tab[248]
#define __pyx_n_u_b1w1w2_str __pyx_string_tab[249]
#define __pyx_n_u_bit_length __pyx_string_tab[250]
#define __pyx_n_u_buf __pyx_string_tab[251]
#define __pyx_n_u_byteorder __pyx_string_tab[252]
#define __pyx_n_u_ca __pyx_string_tab[253]
#define __pyx_n_u_callback __pyx_string_tab[254]
#define __pyx_n_u_cd __pyx_string_tab[255]
#define __pyx_n_u_cd_2 __pyx_string_tab[256]
#define __pyx_n_u_check_context_is_initialized __pyx_string_tab[257]
#define __pyx_n_u_check_release_gil __pyx_string_tab[258]
#define __pyx_n_u_class_getitem __pyx_string_tab[259]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[260]
#define __pyx_n_u_close __pyx_string_tab[261]
#define __pyx_n_u_code __pyx_string_tab[262]
#define __pyx_n_u_code_desc __pyx_string_tab[263]
#define __pyx_n_u_coe_object __pyx_string_tab[264]
#define __pyx_n_u_coe_objects __pyx_string_tab[265]
#define __pyx_n_u_collect_mailbox_errors __pyx_string_tab[266]
#define __pyx_n_u_collections __pyx_string_tab[267]
#define __pyx_n_u_config_dc __pyx_string_tab[268]
#define __pyx_n_u_config_func __pyx_string_tab[269]
#define __pyx_n_u_config_init __pyx_string_tab[270]
#define __pyx_n_u_config_map __pyx_string_tab[271]
#define __pyx_n_u_config_overlap_map __pyx_string_tab[272]
#define __pyx_n_u_contextlib __pyx_string_tab[273]
#define __pyx_n_u_contextmanager __pyx_string_tab[274]
#define __pyx_n_u_data __pyx_string_tab[275]
#define __pyx_n_u_data_type __pyx_string_tab[276]
#define __pyx_n_u_dc_sync __pyx_string_tab[277]
#define __pyx_n_u_dc_time __pyx_string_tab[278]
#define __pyx_n_u_debug __pyx_string_tab[279]
#define __pyx_n_u_desc __pyx_string_tab[280]
#define __pyx_n_u_dict __pyx_string_tab[281]
#define __pyx_n_u_dict_2 __pyx_string_tab[282]
#define __pyx_n_u_disable_complete_access __pyx_string_tab[283]
#define __pyx_n_u_doc __pyx_string_tab[284]
#define __pyx_n_u_ec_adapter __pyx_string_tab[285]
#define __pyx_n_u_ec_datatype __pyx_string_tab[286]
#define __pyx_n_u_eeprom_read __pyx_string_tab[287]
#define __pyx_n_u_eeprom_write __pyx_string_tab[288]
#define __pyx_n_u_encode __pyx_string_tab[289]
#define __pyx_n_u_entries __pyx_string_tab[290]
#define __pyx_n_u_entry __pyx_string_tab[291]
#define __pyx_n_u_enum __pyx_string_tab[292]
#define __pyx_n_u_err __pyx_string_tab[293]
#define __pyx_n_u_error_code __pyx_string_tab[294]
#define __pyx_n_u_error_list __pyx_string_tab[295]
#define __pyx_n_u_error_reg __pyx_string_tab[296]
#define __pyx_n_u_ethercat_slave __pyx_string_tab[297]
#define __pyx_n_u_exc_info __pyx_string_tab[298]
#define __pyx_n_u_expected_state __pyx_string_tab[299]
#define __pyx_n_u_expected_wkc __pyx_string_tab[300]
#define __pyx_n_u_filename __pyx_string_tab[301]
#define __pyx_n_u_find_adapters __pyx_string_tab[302]
#define __pyx_n_u_foe_read __pyx_string_tab[303]
#define __pyx_n_u_foe_write __pyx_string_tab[304]
#define __pyx_n_u_format __pyx_string_tab[305]
#define __pyx_n_u_fprd __pyx_string_tab[306]
#define __pyx_n_u_fprd_fpwr_timeout_us __pyx_string_tab[307]
#define __pyx_n_u_fpwr __pyx_string_tab[308]
#define __pyx_n_u_fpwr_timeout_us __pyx_string_tab[309]
#define __pyx_n_u_from_bytes __pyx_string_tab[310]
#define __pyx_n_u_func __pyx_string_tab[311]
#define __pyx_n_u_genexpr __pyx_string_tab[312]
#define __pyx_n_u_get __pyx_string_tab[313]
#define __pyx_n_u_getLogger __pyx_string_tab[314]
#define __pyx_n_u_get_PO2SOconfig __pyx_string_tab[315]
#define __pyx_n_u_get_PO2SOconfigEx __pyx_string_tab[316]
#define __pyx_n_u_get_al_status __pyx_string_tab[317]
#define __pyx_n_u_get_bit_length __pyx_string_tab[318]
#define __pyx_n_u_get_data_type __pyx_string_tab[319]
#define __pyx_n_u_get_dc_time __pyx_string_tab[320]
#define __pyx_n_u_get_desc __pyx_string_tab[321]
#define __pyx_n_u_get_eep_id __pyx_string_tab[322]
#define __pyx_n_u_get_eep_man __pyx_string_tab[323]
#define __pyx_n_u_get_eep_rev __pyx_string_tab[324]
#define __pyx_n_u_get_entries __pyx_string_tab[325]
#define __pyx_n_u_get_expected_wkc __pyx_string_tab[326]
#define __pyx_n_u_get_index __pyx_string_tab[327]
#define __pyx_n_u_get_input __pyx_string_tab[328]
#define __pyx_n_u_get_is_lost __pyx_string_tab[329]
#define __pyx_n_u_get_manual_state_change __pyx_string_tab[330]
#define __pyx_n_u_get_max_watchdog_time __pyx_string_tab[331]
#define __pyx_n_u_get_name __pyx_string_tab[332]
#define __pyx_n_u_get_obj_access __pyx_string_tab[333]
#define __pyx_n_u_get_object_code __pyx_string_tab[334]
#define __pyx_n_u_get_od __pyx_string_tab[335]
#define __pyx_n_u_get_output __pyx_string_tab[336]
#define __pyx_n_u_get_slave __pyx_string_tab[337]
#define __pyx_n_u_get_state __pyx_string_tab[338]
#define __pyx_n_u_get_watchdog __pyx_string_tab[339]
#define __pyx_n_u_get_watchdog_divider_ns __pyx_string_tab[340]
#define __pyx_n_u_getstate __pyx_string_tab[341]
#define __pyx_n_u_i __pyx_string_tab[342]
#define __pyx_n_u_id __pyx_string_tab[343]
#define __pyx_n_u_ifname __pyx_string_tab[344]
#define __pyx_n_u_ifname_red __pyx_string_tab[345]
#define __pyx_n_u_in __pyx_string_tab[346]
#define __pyx_n_u_index __pyx_string_tab[347]
#define __pyx_n_u_init __pyx_string_tab[348]
#define __pyx_n_u_input __pyx_string_tab[349]
#define __pyx_n_u_is_coroutine __pyx_string_tab[350]
#define __pyx_n_u_is_lost __pyx_string_tab[351]
#define __pyx_n_u_item __pyx_string_tab[352]
#define __pyx_n_u_items __pyx_string_tab[353]
#define __pyx_n_u_keys __pyx_string_tab[354]
#define __pyx_n_u_length __pyx_string_tab[355]
#define __pyx_n_u_little __pyx_string_tab[356]
#define __pyx_n_u_logger __pyx_string_tab[357]
#define __pyx_n_u_logging __pyx_string_tab[358]
#define __pyx_n_u_mailbox __pyx_string_tab[359]
#define __pyx_n_u_main __pyx_string_tab[360]
#define __pyx_n_u_man __pyx_string_tab[361]
#define __pyx_n_u_manual_state_change __pyx_string_tab[362]
#define __pyx_n_u_master __pyx_string_tab[363]
#define __pyx_n_u_mbx_receive __pyx_string_tab[364]
#define __pyx_n_u_member_names __pyx_string_tab[365]
#define __pyx_n_u_members __pyx_string_tab[366]
#define __pyx_n_u_message __pyx_string_tab[367]
#define __pyx_n_u_metaclass __pyx_string_tab[368]
#define __pyx_n_u_module __pyx_string_tab[369]
#define __pyx_n_u_module_2 __pyx_string_tab[370]
#define __pyx_n_u_mro_entries __pyx_string_tab[371]
#define __pyx_n_u_name __pyx_string_tab[372]
#define __pyx_n_u_name_2 __pyx_string_tab[373]
#define __pyx_n_u_namedtuple __pyx_string_tab[374]
#define __pyx_n_u_new __pyx_string_tab[375]
#define __pyx_n_u_next __pyx_string_tab[376]
#define __pyx_n_u_num_bytes __pyx_string_tab[377]
#define __pyx_n_u_obj_access __pyx_string_tab[378]
#define __pyx_n_u_object_code __pyx_string_tab[379]
#define __pyx_n_u_od __pyx_string_tab[380]
#define __pyx_n_u_open __pyx_string_tab[381]
#define __pyx_n_u_out __pyx_string_tab[382]
#define __pyx_n_u_output __pyx_string_tab[383]
#define __pyx_n_u_password __pyx_string_tab[384]
#define __pyx_n_u_pbuf __pyx_string_tab[385]
#define __pyx_n_u_pdi __pyx_string_tab[386]
#define __pyx_n_u_pop __pyx_string_tab[387]
#define __pyx_n_u_pos __pyx_string_tab[388]
#define __pyx_n_u_prepare __pyx_string_tab[389]
#define __pyx_n_u_processdata __pyx_string_tab[390]
#define __pyx_n_u_property __pyx_string_tab[391]
#define __pyx_n_u_pysoem_pysoem __pyx_string_tab[392]
#define __pyx_n_u_pyx_checksum __pyx_string_tab[393]
#define __pyx_n_u_pyx_result __pyx_string_tab[394]
#define __pyx_n_u_pyx_state __pyx_string_tab[395]
#define __pyx_n_u_pyx_type __pyx_string_tab[396]
#define __pyx_n_u_pyx_unpickle_CdefSettings __pyx_string_tab[397]
#define __pyx_n_u_pyx_unpickle__CallbackData __pyx_string_tab[398]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[399]
#define __pyx_n_u_qualname __pyx_string_tab[400]
#define __pyx_n_u_read_description __pyx_string_tab[401]
#define __pyx_n_u_read_entries __pyx_string_tab[402]
#define __pyx_n_u_read_state __pyx_string_tab[403]
#define __pyx_n_u_receive_processdata __pyx_string_tab[404]
#define __pyx_n_u_reconfig __pyx_string_tab[405]
#define __pyx_n_u_recover __pyx_string_tab[406]
#define __pyx_n_u_reduce __pyx_string_tab[407]
#define __pyx_n_u_reduce_cython __pyx_string_tab[408]
#define __pyx_n_u_reduce_ex __pyx_string_tab[409]
#define __pyx_n_u_release_gil __pyx_string_tab[410]
#define __pyx_n_u_result __pyx_string_tab[411]
#define __pyx_n_u_ret_val __pyx_string_tab[412]
#define __pyx_n_u_rev __pyx_string_tab[413]
#define __pyx_n_u_sdo_read __pyx_string_tab[414]
#define __pyx_n_u_sdo_write __pyx_string_tab[415]
#define __pyx_n_u_self __pyx_string_tab[416]
#define __pyx_n_u_send __pyx_string_tab[417]
#define __pyx_n_u_send_overlap_processdata __pyx_string_tab[418]
#define __pyx_n_u_send_processdata __pyx_string_tab[419]
#define __pyx_n_u_set_PO2SOconfig __pyx_string_tab[420]
#define __pyx_n_u_set_PO2SOconfigEx __pyx_string_tab[421]
#define __pyx_n_u_set_is_lost __pyx_string_tab[422]
#define __pyx_n_u_set_manual_state_change __pyx_string_tab[423]
#define __pyx_n_u_set_name __pyx_string_tab[424]
#define __pyx_n_u_set_output __pyx_string_tab[425]
#define __pyx_n_u_set_state __pyx_string_tab[426]
#define __pyx_n_u_set_watchdog __pyx_string_tab[427]
#define __pyx_n_u_setdefault __pyx_string_tab[428]
#define __pyx_n_u_setstate __pyx_string_tab[429]
#define __pyx_n_u_setstate_cython __pyx_string_tab[430]
#define __pyx_n_u_settings __pyx_string_tab[431]
#define __pyx_n_u_setup_func __pyx_string_tab[432]
#define __pyx_n_u_signed __pyx_string_tab[433]
#define __pyx_n_u_size __pyx_string_tab[434]
#define __pyx_n_u_size_inout __pyx_string_tab[435]
#define __pyx_n_u_slave __pyx_string_tab[436]
#define __pyx_n_u_slave_pos __pyx_string_tab[437]
#define __pyx_n_u_slaves __pyx_string_tab[438]
#define __pyx_n_u_start_address __pyx_string_tab[439]
#define __pyx_n_u_state __pyx_string_tab[440]
#define __pyx_n_u_state_check __pyx_string_tab[441]
#define __pyx_n_u_std_buffer __pyx_string_tab[442]
#define __pyx_n_u_str __pyx_string_tab[443]
#define __pyx_n_u_subindex __pyx_string_tab[444]
#define __pyx_n_u_sum __pyx_string_tab[445]
#define __pyx_n_u_sync0_cycle_time __pyx_string_tab[446]
#define __pyx_n_u_sync0_shift_time __pyx_string_tab[447]
#define __pyx_n_u_sync1_cycle_time __pyx_string_tab[448]
#define __pyx_n_u_sys __pyx_string_tab[449]
#define __pyx_n_u_test __pyx_string_tab[450]
#define __pyx_n_u_throw __pyx_string_tab[451]
#define __pyx_n_u_time __pyx_string_tab[452]
#define __pyx_n_u_timeout __pyx_string_tab[453]
#define __pyx_n_u_timeout_us __pyx_string_tab[454]
#define __pyx_n_u_tmp __pyx_string_tab[455]
#define __pyx_n_u_to_bytes __pyx_string_tab[456]
#define __pyx_n_u_update __pyx_string_tab[457]
#define __pyx_n_u_use_setstate __pyx_string_tab[458]
#define __pyx_n_u_usetable __pyx_string_tab[459]
#define __pyx_n_u_utf8 __pyx_string_tab[460]
#define __pyx_n_u_value __pyx_string_tab[461]
#define __pyx_n_u_values __pyx_string_tab[462]
#define __pyx_n_u_w1 __pyx_string_tab[463]
#define __pyx_n_u_w2 __pyx_string_tab[464]
#define __pyx_n_u_warn __pyx_string_tab[465]
#define __pyx_n_u_warnings __pyx_string_tab[466]
#define __pyx_n_u_wd_div_ns __pyx_string_tab[467]
#define __pyx_n_u_wd_div_reg __pyx_string_tab[468]
#define __pyx_n_u_wd_time_ms __pyx_string_tab[469]
#define __pyx_n_u_wd_time_ms_limit __pyx_string_tab[470]
#define __pyx_n_u_wd_time_reg __pyx_string_tab[471]
#define __pyx_n_u_wd_type __pyx_string_tab[472]
#define __pyx_n_u_wd_type_to_reg_map __pyx_string_tab[473]
#define __pyx_n_u_wkc __pyx_string_tab[474]
#define __pyx_n_u_wkt __pyx_string_tab[475]
#define __pyx_n_u_word_address __pyx_string_tab[476]
#define __pyx_n_u_write_state __pyx_string_tab[477]
#define __pyx_n_u_x __pyx_string_tab[478]
#define __pyx_kp_b_iso88591_1F __pyx_string_tab[479]
#define __pyx_kp_b_iso88591_43DA_k_A_q_l_Q_q_q_5_q_A_k_1 __pyx_string_tab[480]
#define __pyx_kp_b_iso88591_4A_H_1_47Gt7R_a_a __pyx_string_tab[481]
#define __pyx_kp_b_iso88591_6 __pyx_string_tab[482]
#define __pyx_kp_b_iso88591_6a_1A_iq_N_Z_ddjjppq_4s_aq_AWF __pyx_string_tab[483]
#define __pyx_kp_b_iso88591_6a_1_4_8_31AT __pyx_string_tab[484]
#define __pyx_kp_b_iso88591_6a_H_1_oQd_gEUUV __pyx_string_tab[485]
#define __pyx_kp_b_iso88591_6a_oQat_3_Na __pyx_string_tab[486]
#define __pyx_kp_b_iso88591_8_iq_N_Z_ddiillmmt_u_E_E_K_K_L __pyx_string_tab[487]
#define __pyx_kp_b_iso88591_9_UUV_d_4M_4_c_1_H_1_5_1_1_r_aq __pyx_string_tab[488]
#define __pyx_kp_b_iso88591_9_d_AQ_Ja_1_d_aq_Qat3C1_82Q_e5 __pyx_string_tab[489]
#define __pyx_kp_b_iso88591_A __pyx_string_tab[490]
#define __pyx_kp_b_iso88591_A_1_4q __pyx_string_tab[491]
#define __pyx_kp_b_iso88591_A_4r_AQ_4s_a_AQ_3aq_k_q_A_m1D_3a __pyx_string_tab[492]
#define __pyx_kp_b_iso88591_A_4t1_Kq_O4xt_XYY_t7_A_l_1_Q __pyx_string_tab[493]
#define __pyx_kp_b_iso88591_A_4t1_aq __pyx_string_tab[494]
#define __pyx_kp_b_iso88591_A_4t1_aq_0__D_PTTU_T_L_a __pyx_string_tab[495]
#define __pyx_kp_b_iso88591_A_4t_Q_1_xt_Q __pyx_string_tab[496]
#define __pyx_kp_b_iso88591_A_4t_Q_1_xt_Q_2 __pyx_string_tab[497]
#define __pyx_kp_b_iso88591_A_6aq_LATQZZ_IT_a_r_b_Ry_Biq_fA __pyx_string_tab[498]
#define __pyx_kp_b_iso88591_A_D_1_7_Jb_q __pyx_string_tab[499]
#define __pyx_kp_b_iso88591_A_D_D_Rt4z_Jiq __pyx_string_tab[500]
#define __pyx_kp_b_iso88591_A_D_D_Rt4z_Jj __pyx_string_tab[501]
#define __pyx_kp_b_iso88591_A_D_JhgT_6_A __pyx_string_tab[502]
#define __pyx_kp_b_iso88591_A_H_1_1AQ_oQd2B_gQeST_q_4AQ_7_E __pyx_string_tab[503]
#define __pyx_kp_b_iso88591_A_H_1_4s_6_A_awau_F_1_6Fd_Q__dd __pyx_string_tab[504]
#define __pyx_kp_b_iso88591_A_H_1_83a_a_E_j_1_Qb_Q_Qb_A_A_1 __pyx_string_tab[505]
#define __pyx_kp_b_iso88591_A_H_1_G1A __pyx_string_tab[506]
#define __pyx_kp_b_iso88591_A_H_1_oQd_a __pyx_string_tab[507]
#define __pyx_kp_b_iso88591_A_H_1_q_1_1_87_E_4_5Q_b_A_D_1_c __pyx_string_tab[508]
#define __pyx_kp_b_iso88591_A_Jar __pyx_string_tab[509]
#define __pyx_kp_b_iso88591_A_Jar_Rs_D_2Q __pyx_string_tab[510]
#define __pyx_kp_b_iso88591_A_Jiq __pyx_string_tab[511]
#define __pyx_kp_b_iso88591_A_Jj __pyx_string_tab[512]
#define __pyx_kp_b_iso88591_A_JoQa __pyx_string_tab[513]
#define __pyx_kp_b_iso88591_A_Kq __pyx_string_tab[514]
#define __pyx_kp_b_iso88591_A_M_IQ_L_N_HA __pyx_string_tab[515]
#define __pyx_kp_b_iso88591_A_M_N __pyx_string_tab[516]
#define __pyx_kp_b_iso88591_A_M_N_HA __pyx_string_tab[517]
#define __pyx_kp_b_iso88591_A_M_N_M_F_F_F __pyx_string_tab[518]
#define __pyx_kp_b_iso88591_A_N __pyx_string_tab[519]
#define __pyx_kp_b_iso88591_A_N_q __pyx_string_tab[520]
#define __pyx_kp_b_iso88591_A_Q_1D_8_s_1_3a_a_a_AATTUUXXccjj __pyx_string_tab[521]
#define __pyx_kp_b_iso88591_A_a_N_4_hc_1_a_U_4_gQa_1_1D_waq __pyx_string_tab[522]
#define __pyx_kp_b_iso88591_A_a_N_4_hc_4_AQ_1 __pyx_string_tab[523]
#define __pyx_kp_b_iso88591_A_a_t_e1D __pyx_string_tab[524]
#define __pyx_kp_b_iso88591_A_a_t_iq_A __pyx_string_tab[525]
#define __pyx_kp_b_iso88591_A_a_t_k_a __pyx_string_tab[526]
#define __pyx_kp_b_iso88591_A_aq_4DAT_RS_IT_a_r_b_b_b_fA_gQa __pyx_string_tab[527]
#define __pyx_kp_b_iso88591_A_awd_JgWCq __pyx_string_tab[528]
#define __pyx_kp_b_iso88591_A_fAQ_5ET_PQQUUV_4wb_aq_a_E_at_a __pyx_string_tab[529]
#define __pyx_kp_b_iso88591_A_m1AT __pyx_string_tab[530]
#define __pyx_kp_b_iso88591_A_nAQd __pyx_string_tab[531]
#define __pyx_kp_b_iso88591_A_oQat_1 __pyx_string_tab[532]
#define __pyx_kp_b_iso88591_A_q_1_1_87_E_D_1_c_AT_q0B_SPQ_2 __pyx_string_tab[533]
#define __pyx_kp_b_iso88591_A_q_S_1D_a_s_a_1_Cs_Rq_q __pyx_string_tab[534]
#define __pyx_kp_b_iso88591_A_q_a_N_4_hc_4_AQ_U_4_gQa_t_j_1 __pyx_string_tab[535]
#define __pyx_kp_b_iso88591_A_t __pyx_string_tab[536]
#define __pyx_kp_b_iso88591_A_t1 __pyx_string_tab[537]
#define __pyx_kp_b_iso88591_A_t_Q __pyx_string_tab[538]
#define __pyx_kp_b_iso88591_A_t_Q_2 __pyx_string_tab[539]
#define __pyx_kp_b_iso88591_A_t_Q_3 __pyx_string_tab[540]
#define __pyx_kp_b_iso88591_A_t_Qb __pyx_string_tab[541]
#define __pyx_kp_b_iso88591_A_t_V7_1 __pyx_string_tab[542]
#define __pyx_kp_b_iso88591_A_t_at1 __pyx_string_tab[543]
#define __pyx_kp_b_iso88591_A_t_e1D __pyx_string_tab[544]
#define __pyx_kp_b_iso88591_A_t_fAT __pyx_string_tab[545]
#define __pyx_kp_b_iso88591_A_t_iq_A __pyx_string_tab[546]
#define __pyx_kp_b_iso88591_A_t_j_Q __pyx_string_tab[547]
#define __pyx_kp_b_iso88591_A_uAQd_c_7_ZzY___bbkkllssv_w_A_A __pyx_string_tab[548]
#define __pyx_kp_b_iso88591_A_wa_1_t1 __pyx_string_tab[549]
#define __pyx_kp_b_iso88591_BBYYZ_d_4M_4_c_1_H_1_1A_1_T_1Jj __pyx_string_tab[550]
#define __pyx_kp_b_iso88591_Q __pyx_string_tab[551]
#define __pyx_kp_b_iso88591_Q_Kq_G1 __pyx_string_tab[552]
#define __pyx_kp_b_iso88591_RRS_H_1_Cq_q_4D_uDVVW_5T_EWWiij __pyx_string_tab[553]
#define __pyx_kp_b_iso88591_RRS_d_4M_4_c_1_H_1_2U_1_Q_1_T_Z __pyx_string_tab[554]
#define __pyx_kp_b_iso88591_T_D_T_A_G1F_a_vWE_Q_q_t_WE_D_G5 __pyx_string_tab[555]
#define __pyx_kp_b_iso88591_T_d_G1F_a_vWE_Q_q_t_WA_q_T_G1_T __pyx_string_tab[556]
#define __pyx_kp_b_iso88591_VVW_d_4M_H_1_1A_1_T_1G_T_q_M_d __pyx_string_tab[557]
#define __pyx_kp_b_iso88591__8 __pyx_string_tab[558]
#define __pyx_kp_b_iso88591_a_H_1_T __pyx_string_tab[559]
#define __pyx_kp_b_iso88591_a_c_Yaq_F_0_6Fat_Z_gghhqq_C_C_D __pyx_string_tab[560]
#define __pyx_kp_b_iso88591_a_d_AQ_1_4_0_1AT __pyx_string_tab[561]
#define __pyx_kp_b_iso88591_q_0_kQR_7_q0_a_1 __pyx_string_tab[562]
#define __pyx_kp_b_iso88591_q_0_kQR_xq_7_a_nA_1 __pyx_string_tab[563]
#define __pyx_kp_b_iso88591_q_H_1_d __pyx_string_tab[564]
#define __pyx_kp_b_iso88591_q_d_AQ_1_4_31A_aq_4DA __pyx_string_tab[565]
#define __pyx_kp_b_iso88591_waq __pyx_string_tab[566]
#define __pyx_float_1000000_0 __pyx_number_tab[0]
#define __pyx_int_0 __pyx_number_tab[1]
#define __pyx_int_1 __pyx_number_tab[2]
#define __pyx_int_2 __pyx_number_tab[3]
#define __pyx_int_3 __pyx_number_tab[4]
#define __pyx_int_4 __pyx_number_tab[5]
#define __pyx_int_8 __pyx_number_tab[6]
#define __pyx_int_10 __pyx_number_tab[7]
#define __pyx_int_11 __pyx_number_tab[8]
#define __pyx_int_16 __pyx_number_tab[9]
#define __pyx_int_20 __pyx_number_tab[10]
#define __pyx_int_22 __pyx_number_tab[11]
#define __pyx_int_24 __pyx_number_tab[12]
#define __pyx_int_26 __pyx_number_tab[13]
#define __pyx_int_28 __pyx_number_tab[14]
#define __pyx_int_32 __pyx_number_tab[15]
#define __pyx_int_40 __pyx_number_tab[16]
#define __pyx_int_500 __pyx_number_tab[17]
#define __pyx_int_1024 __pyx_number_tab[18]
#define __pyx_int_1040 __pyx_number_tab[19]
#define __pyx_int_1056 __pyx_number_tab[20]
#define __pyx_int_2000 __pyx_number_tab[21]
#define __pyx_int_2048 __pyx_number_tab[22]
#define __pyx_int_20000 __pyx_number_tab[23]
#define __pyx_int_50000 __pyx_number_tab[24]
#define __pyx_int_65535 __pyx_number_tab[25]
#define __pyx_int_200000 __pyx_number_tab[26]
#define __pyx_int_9788398 __pyx_number_tab[27]
#define __pyx_int_241895382 __pyx_number_tab[28]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts);
  Py_CLEAR(clear_module_state->__pyx_type_6pysoem_6pysoem_CdefTimeouts);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pysoem_6pysoem_CdefSettings);
  Py_CLEAR(clear_module_state->__pyx_type_6pysoem_6pysoem_CdefSettings);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pysoem_6pysoem_CdefMaster);
  Py_CLEAR(clear_module_state->__pyx_type_6pysoem_6pysoem_CdefMaster);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pysoem_6pysoem__CallbackData);
  Py_CLEAR(clear_module_state->__pyx_type_6pysoem_6pysoem__CallbackData);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pysoem_6pysoem_CdefSlave);
  Py_CLEAR(clear_module_state->__pyx_type_6pysoem_6pysoem_CdefSlave);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject);
  Py_CLEAR(clear_module_state->__pyx_type_6pysoem_6pysoem_CdefCoeObject);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry);
  Py_CLEAR(clear_module_state->__pyx_type_6pysoem_6pysoem_CdefCoeObjectEntry);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open);
  Py_CLEAR(clear_module_state->__pyx_type_6pysoem_6pysoem___pyx_scope_struct__open);
  Py_CLEAR(clear_module_state->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_6pysoem_6pysoem___pyx_scope_struct_1_genexpr);
  for (int i=0; i<11; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<103; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<567; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<29; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* Generator.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_GeneratorType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts);
  Py_VISIT(traverse_module_state->__pyx_type_6pysoem_6pysoem_CdefTimeouts);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pysoem_6pysoem_CdefSettings);
  Py_VISIT(traverse_module_state->__pyx_type_6pysoem_6pysoem_CdefSettings);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pysoem_6pysoem_CdefMaster);
  Py_VISIT(traverse_module_state->__pyx_type_6pysoem_6pysoem_CdefMaster);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pysoem_6pysoem__CallbackData);
  Py_VISIT(traverse_module_state->__pyx_type_6pysoem_6pysoem__CallbackData);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pysoem_6pysoem_CdefSlave);
  Py_VISIT(traverse_module_state->__pyx_type_6pysoem_6pysoem_CdefSlave);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject);
  Py_VISIT(traverse_module_state->__pyx_type_6pysoem_6pysoem_CdefCoeObject);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry);
  Py_VISIT(traverse_module_state->__pyx_type_6pysoem_6pysoem_CdefCoeObjectEntry);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open);
  Py_VISIT(traverse_module_state->__pyx_type_6pysoem_6pysoem___pyx_scope_struct__open);
  Py_VISIT(traverse_module_state->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_6pysoem_6pysoem___pyx_scope_struct_1_genexpr);
  for (int i=0; i<11; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<103; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<567; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<29; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* Generator.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_GeneratorType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "pysoem/pysoem.pyx":61
 *     cdef cpysoem.Ttimeouts* _t
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._t = &cpysoem.soem_timeouts
 * 
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts___cinit__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts___cinit__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self) {
  int __pyx_r;

  /* "pysoem/pysoem.pyx":62
 * 
 *     def __cinit__(self):
 *         self._t = &cpysoem.soem_timeouts             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_v_self->_t = (&soem_timeouts);

  /* "pysoem/pysoem.pyx":61
 *     cdef cpysoem.Ttimeouts* _t
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._t = &cpysoem.soem_timeouts
 * 
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":64
 *         self._t = &cpysoem.soem_timeouts
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ret(self) -> int:
 *         return self._t.ret
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_3ret_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_3ret_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_3ret___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_3ret___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pysoem/pysoem.pyx":66
 *     @property
 *     def ret(self) -> int:
 *         return self._t.ret             # <<<<<<<<<<<<<<
 * 
 *     @ret.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->_t->ret); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":64
 *         self._t = &cpysoem.soem_timeouts
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ret(self) -> int:
 *         return self._t.ret
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.ret.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":68
 *         return self._t.ret
 * 
 *     @ret.setter             # <<<<<<<<<<<<<<
 *     def ret(self, value: int):
 *         self._t.ret = value
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_3ret_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_3ret_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyLong_Type), 0, "value", 2))) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_3ret_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_3ret_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pysoem/pysoem.pyx":70
 *     @ret.setter
 *     def ret(self, value: int):
 *         self._t.ret = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 70, __pyx_L1_error)
  __pyx_v_self->_t->ret = __pyx_t_1;

  /* "pysoem/pysoem.pyx":68
 *         return self._t.ret
 * 
 *     @ret.setter             # <<<<<<<<<<<<<<
 *     def ret(self, value: int):
 *         self._t.ret = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.ret.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":72
 *         self._t.ret = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def safe(self) -> int:
 *         return self._t.safe
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_4safe_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_4safe_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_4safe___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_4safe___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pysoem/pysoem.pyx":74
 *     @property
 *     def safe(self) -> int:
 *         return self._t.safe             # <<<<<<<<<<<<<<
 * 
 *     @safe.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->_t->safe); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":72
 *         self._t.ret = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def safe(self) -> int:
 *         return self._t.safe
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.safe.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":76
 *         return self._t.safe
 * 
 *     @safe.setter             # <<<<<<<<<<<<<<
 *     def safe(self, value: int):
 *         self._t.safe = value
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_4safe_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_4safe_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyLong_Type), 0, "value", 2))) __PYX_ERR(0, 77, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_4safe_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_4safe_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pysoem/pysoem.pyx":78
 *     @safe.setter
 *     def safe(self, value: int):
 *         self._t.safe = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 78, __pyx_L1_error)
  __pyx_v_self->_t->safe = __pyx_t_1;

  /* "pysoem/pysoem.pyx":76
 *         return self._t.safe
 * 
 *     @safe.setter             # <<<<<<<<<<<<<<
 *     def safe(self, value: int):
 *         self._t.safe = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.safe.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":80
 *         self._t.safe = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def eeprom(self) -> int:
 *         return self._t.eeprom
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_6eeprom_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_6eeprom_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_6eeprom___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_6eeprom___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pysoem/pysoem.pyx":82
 *     @property
 *     def eeprom(self) -> int:
 *         return self._t.eeprom             # <<<<<<<<<<<<<<
 * 
 *     @eeprom.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->_t->eeprom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":80
 *         self._t.safe = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def eeprom(self) -> int:
 *         return self._t.eeprom
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.eeprom.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":84
 *         return self._t.eeprom
 * 
 *     @eeprom.setter             # <<<<<<<<<<<<<<
 *     def eeprom(self, value: int):
 *         self._t.eeprom = value
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_6eeprom_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_6eeprom_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyLong_Type), 0, "value", 2))) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_6eeprom_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_6eeprom_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pysoem/pysoem.pyx":86
 *     @eeprom.setter
 *     def eeprom(self, value: int):
 *         self._t.eeprom = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L1_error)
  __pyx_v_self->_t->eeprom = __pyx_t_1;

  /* "pysoem/pysoem.pyx":84
 *         return self._t.eeprom
 * 
 *     @eeprom.setter             # <<<<<<<<<<<<<<
 *     def eeprom(self, value: int):
 *         self._t.eeprom = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.eeprom.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":88
 *         self._t.eeprom = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tx_mailbox(self) -> int:
 *         return self._t.tx_mailbox
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pysoem/pysoem.pyx":90
 *     @property
 *     def tx_mailbox(self) -> int:
 *         return self._t.tx_mailbox             # <<<<<<<<<<<<<<
 * 
 *     @tx_mailbox.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->_t->tx_mailbox); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":88
 *         self._t.eeprom = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tx_mailbox(self) -> int:
 *         return self._t.tx_mailbox
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.tx_mailbox.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":92
 *         return self._t.tx_mailbox
 * 
 *     @tx_mailbox.setter             # <<<<<<<<<<<<<<
 *     def tx_mailbox(self, value: int):
 *         self._t.tx_mailbox = value
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyLong_Type), 0, "value", 2))) __PYX_ERR(0, 93, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pysoem/pysoem.pyx":94
 *     @tx_mailbox.setter
 *     def tx_mailbox(self, value: int):
 *         self._t.tx_mailbox = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 94, __pyx_L1_error)
  __pyx_v_self->_t->tx_mailbox = __pyx_t_1;

  /* "pysoem/pysoem.pyx":92
 *         return self._t.tx_mailbox
 * 
 *     @tx_mailbox.setter             # <<<<<<<<<<<<<<
 *     def tx_mailbox(self, value: int):
 *         self._t.tx_mailbox = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.tx_mailbox.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":96
 *         self._t.tx_mailbox = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rx_mailbox(self) -> int:
 *         return self._t.rx_mailbox
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pysoem/pysoem.pyx":98
 *     @property
 *     def rx_mailbox(self) -> int:
 *         return self._t.rx_mailbox             # <<<<<<<<<<<<<<
 * 
 *     @rx_mailbox.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->_t->rx_mailbox); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":96
 *         self._t.tx_mailbox = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rx_mailbox(self) -> int:
 *         return self._t.rx_mailbox
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.rx_mailbox.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":100
 *         return self._t.rx_mailbox
 * 
 *     @rx_mailbox.setter             # <<<<<<<<<<<<<<
 *     def rx_mailbox(self, value: int):
 *         self._t.rx_mailbox = value
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyLong_Type), 0, "value", 2))) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pysoem/pysoem.pyx":102
 *     @rx_mailbox.setter
 *     def rx_mailbox(self, value: int):
 *         self._t.rx_mailbox = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 102, __pyx_L1_error)
  __pyx_v_self->_t->rx_mailbox = __pyx_t_1;

  /* "pysoem/pysoem.pyx":100
 *         return self._t.rx_mailbox
 * 
 *     @rx_mailbox.setter             # <<<<<<<<<<<<<<
 *     def rx_mailbox(self, value: int):
 *         self._t.rx_mailbox = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.rx_mailbox.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":104
 *         self._t.rx_mailbox = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def state(self) -> int:
 *         return self._t.state
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_5state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_5state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_5state___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_5state___get__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pysoem/pysoem.pyx":106
 *     @property
 *     def state(self) -> int:
 *         return self._t.state             # <<<<<<<<<<<<<<
 * 
 *     @state.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->_t->state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":104
 *         self._t.rx_mailbox = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def state(self) -> int:
 *         return self._t.state
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":108
 *         return self._t.state
 * 
 *     @state.setter             # <<<<<<<<<<<<<<
 *     def state(self, value: int):
 *         self._t.state = value
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_5state_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_5state_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyLong_Type), 0, "value", 2))) __PYX_ERR(0, 109, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_5state_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_5state_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pysoem/pysoem.pyx":110
 *     @state.setter
 *     def state(self, value: int):
 *         self._t.state = value             # <<<<<<<<<<<<<<
 * 
 * cdef class CdefSettings:
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __pyx_v_self->_t->state = __pyx_t_1;

  /* "pysoem/pysoem.pyx":108
 *         return self._t.state
 * 
 *     @state.setter             # <<<<<<<<<<<<<<
 *     def state(self, value: int):
 *         self._t.state = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.state.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_12CdefTimeouts_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_2__reduce_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_12CdefTimeouts_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefTimeouts_4__setstate_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefTimeouts_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefTimeouts.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":117
 *     cdef public cpysoem.boolean always_release_gil
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.timeouts = CdefTimeouts()
 *         self.always_release_gil = False
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefSettings_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefSettings_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__init__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefSettings___init__(((struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefSettings___init__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pysoem/pysoem.pyx":118
 * 
 *     def __init__(self):
 *         self.timeouts = CdefTimeouts()             # <<<<<<<<<<<<<<
 *         self.always_release_gil = False
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __Pyx_GIVEREF((PyObject *)__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->timeouts);
  __Pyx_DECREF((PyObject *)__pyx_v_self->timeouts);
  __pyx_v_self->timeouts = ((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":119
 *     def __init__(self):
 *         self.timeouts = CdefTimeouts()
 *         self.always_release_gil = False             # <<<<<<<<<<<<<<
 * 
 * settings = CdefSettings()
*/
  __pyx_v_self->always_release_gil = 0;

  /* "pysoem/pysoem.pyx":117
 *     cdef public cpysoem.boolean always_release_gil
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.timeouts = CdefTimeouts()
 *         self.always_release_gil = False
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSettings.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":114
 * cdef class CdefSettings:
 * 
 *     cdef public CdefTimeouts timeouts             # <<<<<<<<<<<<<<
 *     cdef public cpysoem.boolean always_release_gil
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefSettings_8timeouts_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefSettings_8timeouts_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefSettings_8timeouts___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefSettings_8timeouts___get__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->timeouts);
  __pyx_r = ((PyObject *)__pyx_v_self->timeouts);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefSettings_8timeouts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefSettings_8timeouts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefSettings_8timeouts_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefSettings_8timeouts_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts))))) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->timeouts);
  __Pyx_DECREF((PyObject *)__pyx_v_self->timeouts);
  __pyx_v_self->timeouts = ((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSettings.timeouts.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefSettings_8timeouts_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefSettings_8timeouts_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefSettings_8timeouts_4__del__(((struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefSettings_8timeouts_4__del__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->timeouts);
  __Pyx_DECREF((PyObject *)__pyx_v_self->timeouts);
  __pyx_v_self->timeouts = ((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":115
 * 
 *     cdef public CdefTimeouts timeouts
 *     cdef public cpysoem.boolean always_release_gil             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefSettings_18always_release_gil_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefSettings_18always_release_gil_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefSettings_18always_release_gil___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefSettings_18always_release_gil___get__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int8_t(__pyx_v_self->always_release_gil); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSettings.always_release_gil.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_12CdefSettings_18always_release_gil_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_12CdefSettings_18always_release_gil_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefSettings_18always_release_gil_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_12CdefSettings_18always_release_gil_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  boolean __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyLong_As_int8_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(0, 115, __pyx_L1_error)
  __pyx_v_self->always_release_gil = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefSettings.always_release_gil.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefSettings_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_12CdefSettings_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_12CdefSettings_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefSettings_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefSettings_2__reduce_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefSettings_2__reduce_cython__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.always_release_gil, self.timeouts)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = __Pyx_PyLong_From_int8_t(__pyx_v_self->always_release_gil); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->timeouts);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->timeouts);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->timeouts)) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.always_release_gil, self.timeouts)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.always_release_gil, self.timeouts)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.timeouts is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.always_release_gil, self.timeouts)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.timeouts is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, None), state
*/
  /*else*/ {
    __pyx_t_3 = (((PyObject *)__pyx_v_self->timeouts) != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.timeouts is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.timeouts is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_CdefSettings); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_241895382);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_241895382);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_mstate_global->__pyx_int_241895382) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.timeouts is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, None), state
 *     else:
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CdefSettings__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_CdefSettings); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_241895382);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_241895382);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_mstate_global->__pyx_int_241895382) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSettings.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CdefSettings__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefSettings_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_12CdefSettings_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_12CdefSettings_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_12CdefSettings_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSettings.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12CdefSettings_4__setstate_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12CdefSettings_4__setstate_cython__(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CdefSettings__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_6pysoem_6pysoem___pyx_unpickle_CdefSettings__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CdefSettings__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSettings.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":156
 *     int* sdo_write_timeout
 * 
 * def find_adapters():             # <<<<<<<<<<<<<<
 *     """Create a list of available network adapters.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_1find_adapters(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_find_adapters, "Create a list of available network adapters.\n\n    Returns:\n        list[Adapter]: Each element of the list has a name an desc attribute.\n\n    ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_1find_adapters = {"find_adapters", (PyCFunction)__pyx_pw_6pysoem_6pysoem_1find_adapters, METH_NOARGS, __pyx_doc_6pysoem_6pysoem_find_adapters};
static PyObject *__pyx_pw_6pysoem_6pysoem_1find_adapters(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_adapters (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_find_adapters(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_find_adapters(CYTHON_UNUSED PyObject *__pyx_self) {
  ec_adaptert *__pyx_v__ec_adapter;
  PyObject *__pyx_v_Adapter = NULL;
  PyObject *__pyx_v_adapters = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  char *__pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  ec_adaptert *__pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_adapters", 0);

  /* "pysoem/pysoem.pyx":163
 * 
 *     """
 *     cdef cpysoem.ec_adaptert* _ec_adapter = cpysoem.ec_find_adapters()             # <<<<<<<<<<<<<<
 *     Adapter = collections.namedtuple('Adapter', ['name', 'desc'])
 *     adapters = []
*/
  __pyx_v__ec_adapter = ec_find_adapters();

  /* "pysoem/pysoem.pyx":164
 *     """
 *     cdef cpysoem.ec_adaptert* _ec_adapter = cpysoem.ec_find_adapters()
 *     Adapter = collections.namedtuple('Adapter', ['name', 'desc'])             # <<<<<<<<<<<<<<
 *     adapters = []
 *     while not _ec_adapter == NULL:
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_collections); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_namedtuple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_name);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_name);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_mstate_global->__pyx_n_u_name) != (0)) __PYX_ERR(0, 164, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_desc);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_desc);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_mstate_global->__pyx_n_u_desc) != (0)) __PYX_ERR(0, 164, __pyx_L1_error);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Adapter, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_Adapter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":165
 *     cdef cpysoem.ec_adaptert* _ec_adapter = cpysoem.ec_find_adapters()
 *     Adapter = collections.namedtuple('Adapter', ['name', 'desc'])
 *     adapters = []             # <<<<<<<<<<<<<<
 *     while not _ec_adapter == NULL:
 *         adapters.append(Adapter(_ec_adapter.name.decode('utf8'), _ec_adapter.desc))
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":166
 *     Adapter = collections.namedtuple('Adapter', ['name', 'desc'])
 *     adapters = []
 *     while not _ec_adapter == NULL:             # <<<<<<<<<<<<<<
 *         adapters.append(Adapter(_ec_adapter.name.decode('utf8'), _ec_adapter.desc))
 *         _ec_adapter = _ec_adapter.next
*/
  while (1) {
    __pyx_t_6 = (!(__pyx_v__ec_adapter == NULL));
    if (!__pyx_t_6) break;

    /* "pysoem/pysoem.pyx":167
 *     adapters = []
 *     while not _ec_adapter == NULL:
 *         adapters.append(Adapter(_ec_adapter.name.decode('utf8'), _ec_adapter.desc))             # <<<<<<<<<<<<<<
 *         _ec_adapter = _ec_adapter.next
 *     return adapters
*/
    __pyx_t_4 = NULL;
    __Pyx_INCREF(__pyx_v_Adapter);
    __pyx_t_3 = __pyx_v_Adapter; 
    __pyx_t_7 = __pyx_v__ec_adapter->name;
    __pyx_t_8 = __Pyx_ssize_strlen(__pyx_t_7); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 167, __pyx_L1_error)
    __pyx_t_2 = __Pyx_decode_c_string(__pyx_t_7, 0, __pyx_t_8, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyBytes_FromString(__pyx_v__ec_adapter->desc); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_2, __pyx_t_9};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_adapters, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":168
 *     while not _ec_adapter == NULL:
 *         adapters.append(Adapter(_ec_adapter.name.decode('utf8'), _ec_adapter.desc))
 *         _ec_adapter = _ec_adapter.next             # <<<<<<<<<<<<<<
 *     return adapters
 * 
*/
    __pyx_t_11 = __pyx_v__ec_adapter->next;
    __pyx_v__ec_adapter = __pyx_t_11;
  }

  /* "pysoem/pysoem.pyx":169
 *         adapters.append(Adapter(_ec_adapter.name.decode('utf8'), _ec_adapter.desc))
 *         _ec_adapter = _ec_adapter.next
 *     return adapters             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_adapters);
  __pyx_r = __pyx_v_adapters;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":156
 *     int* sdo_write_timeout
 * 
 * def find_adapters():             # <<<<<<<<<<<<<<
 *     """Create a list of available network adapters.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pysoem.pysoem.find_adapters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_Adapter);
  __Pyx_XDECREF(__pyx_v_adapters);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6pysoem_6pysoem_4generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "pysoem/pysoem.pyx":172
 * 
 * 
 * @contextlib.contextmanager             # <<<<<<<<<<<<<<
 * def open(ifname):
 *     """Context manager function to create a Master object.
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_3open(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_2open, "Context manager function to create a Master object.\n\n    .. versionadded:: 1.1.0\n    ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_3open = {"open", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_3open, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_2open};
static PyObject *__pyx_pw_6pysoem_6pysoem_3open(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_ifname = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("open (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_ifname,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 172, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 172, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "open", 0) < (0)) __PYX_ERR(0, 172, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("open", 1, 1, 1, i); __PYX_ERR(0, 172, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 172, __pyx_L3_error)
    }
    __pyx_v_ifname = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("open", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 172, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_2open(__pyx_self, __pyx_v_ifname);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_2open(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ifname) {
  struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("open", 0);
  __pyx_cur_scope = (struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *)__pyx_tp_new_6pysoem_6pysoem___pyx_scope_struct__open(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 172, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_ifname = __pyx_v_ifname;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ifname);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ifname);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6pysoem_6pysoem_4generator, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0]), (PyObject *) __pyx_cur_scope, __pyx_mstate_global->__pyx_n_u_open, __pyx_mstate_global->__pyx_n_u_open, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem); if (unlikely(!gen)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6pysoem_6pysoem_4generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *__pyx_cur_scope = ((struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("open", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L4_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(__pyx_sent_value != Py_None)) {
    if (unlikely(__pyx_sent_value)) PyErr_SetString(PyExc_TypeError, "can't send non-None value to a just-started generator");
    __PYX_ERR(0, 172, __pyx_L1_error)
  }

  /* "pysoem/pysoem.pyx":178
 *     .. versionadded:: 1.1.0
 *     """
 *     master = Master()             # <<<<<<<<<<<<<<
 *     master.open(ifname)
 *     yield master
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Master); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_master = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":179
 *     """
 *     master = Master()
 *     master.open(ifname)             # <<<<<<<<<<<<<<
 *     yield master
 *     master.close()
*/
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_master;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_cur_scope->__pyx_v_ifname};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_open, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":180
 *     master = Master()
 *     master.open(ifname)
 *     yield master             # <<<<<<<<<<<<<<
 *     master.close()
 * 
*/
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_master);
  __pyx_r = __pyx_cur_scope->__pyx_v_master;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  /* return from generator, yielding value */
  __pyx_generator->resume_label = 1;
  return __pyx_r;
  __pyx_L4_resume_from_yield:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 180, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":181
 *     master.open(ifname)
 *     yield master
 *     master.close()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_master;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_close, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "pysoem/pysoem.pyx":172
 * 
 * 
 * @contextlib.contextmanager             # <<<<<<<<<<<<<<
 * def open(ifname):
 *     """Context manager function to create a Master object.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__Pyx_PyErr_Occurred()) {
    __Pyx_AddTraceback("open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":184
 * 
 * 
 * def al_status_code_to_string(code):             # <<<<<<<<<<<<<<
 *     """Look up text string that belongs to AL status code.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_6al_status_code_to_string(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_5al_status_code_to_string, "Look up text string that belongs to AL status code.\n    \n    Args:\n        arg1 (uint16): AL status code as defined in EtherCAT protocol.\n    \n    Returns:\n        str: A verbal description of status code\n    \n    ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_6al_status_code_to_string = {"al_status_code_to_string", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_6al_status_code_to_string, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_5al_status_code_to_string};
static PyObject *__pyx_pw_6pysoem_6pysoem_6al_status_code_to_string(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("al_status_code_to_string (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 184, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 184, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "al_status_code_to_string", 0) < (0)) __PYX_ERR(0, 184, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("al_status_code_to_string", 1, 1, 1, i); __PYX_ERR(0, 184, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 184, __pyx_L3_error)
    }
    __pyx_v_code = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("al_status_code_to_string", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 184, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.al_status_code_to_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_5al_status_code_to_string(__pyx_self, __pyx_v_code);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_5al_status_code_to_string(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16 __pyx_t_1;
  char *__pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("al_status_code_to_string", 0);

  /* "pysoem/pysoem.pyx":194
 * 
 *     """
 *     return cpysoem.ec_ALstatuscode2string(code).decode('utf8');             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_As_uint16_t(__pyx_v_code); if (unlikely((__pyx_t_1 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 194, __pyx_L1_error)
  __pyx_t_2 = ec_ALstatuscode2string(__pyx_t_1);
  __pyx_t_3 = __Pyx_ssize_strlen(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 194, __pyx_L1_error)
  __pyx_t_4 = __Pyx_decode_c_string(__pyx_t_2, 0, __pyx_t_3, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":184
 * 
 * 
 * def al_status_code_to_string(code):             # <<<<<<<<<<<<<<
 *     """Look up text string that belongs to AL status code.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pysoem.pysoem.al_status_code_to_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":257
 *     manual_state_change = property(_get_manual_state_change, _set_manual_state_change)
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.port = &self._ecx_port
 *         self._ecx_contextt.slavelist = &self._ec_slave[0]
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_10CdefMaster_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_10CdefMaster_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster___cinit__(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_10CdefMaster___cinit__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  boolean __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pysoem/pysoem.pyx":258
 * 
 *     def __cinit__(self):
 *         self._ecx_contextt.port = &self._ecx_port             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.slavelist = &self._ec_slave[0]
 *         self._ecx_contextt.slavecount = &self._ec_slavecount
*/
  __pyx_v_self->_ecx_contextt.port = (&__pyx_v_self->_ecx_port);

  /* "pysoem/pysoem.pyx":259
 *     def __cinit__(self):
 *         self._ecx_contextt.port = &self._ecx_port
 *         self._ecx_contextt.slavelist = &self._ec_slave[0]             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.slavecount = &self._ec_slavecount
 *         self._ecx_contextt.maxslave = EC_MAXSLAVE
*/
  __pyx_v_self->_ecx_contextt.slavelist = (&(__pyx_v_self->_ec_slave[0]));

  /* "pysoem/pysoem.pyx":260
 *         self._ecx_contextt.port = &self._ecx_port
 *         self._ecx_contextt.slavelist = &self._ec_slave[0]
 *         self._ecx_contextt.slavecount = &self._ec_slavecount             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.maxslave = EC_MAXSLAVE
 *         self._ecx_contextt.grouplist = &self._ec_group[0]
*/
  __pyx_v_self->_ecx_contextt.slavecount = (&__pyx_v_self->_ec_slavecount);

  /* "pysoem/pysoem.pyx":261
 *         self._ecx_contextt.slavelist = &self._ec_slave[0]
 *         self._ecx_contextt.slavecount = &self._ec_slavecount
 *         self._ecx_contextt.maxslave = EC_MAXSLAVE             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.grouplist = &self._ec_group[0]
 *         self._ecx_contextt.maxgroup = EC_MAXGROUP
*/
  __pyx_v_self->_ecx_contextt.maxslave = __pyx_e_6pysoem_6pysoem_EC_MAXSLAVE;

  /* "pysoem/pysoem.pyx":262
 *         self._ecx_contextt.slavecount = &self._ec_slavecount
 *         self._ecx_contextt.maxslave = EC_MAXSLAVE
 *         self._ecx_contextt.grouplist = &self._ec_group[0]             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.maxgroup = EC_MAXGROUP
 *         self._ecx_contextt.esibuf = &self._ec_esibuf[0]
*/
  __pyx_v_self->_ecx_contextt.grouplist = (&(__pyx_v_self->_ec_group[0]));

  /* "pysoem/pysoem.pyx":263
 *         self._ecx_contextt.maxslave = EC_MAXSLAVE
 *         self._ecx_contextt.grouplist = &self._ec_group[0]
 *         self._ecx_contextt.maxgroup = EC_MAXGROUP             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.esibuf = &self._ec_esibuf[0]
 *         self._ecx_contextt.esimap = &self._ec_esimap[0]
*/
  __pyx_v_self->_ecx_contextt.maxgroup = __pyx_e_6pysoem_6pysoem_EC_MAXGROUP;

  /* "pysoem/pysoem.pyx":264
 *         self._ecx_contextt.grouplist = &self._ec_group[0]
 *         self._ecx_contextt.maxgroup = EC_MAXGROUP
 *         self._ecx_contextt.esibuf = &self._ec_esibuf[0]             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.esimap = &self._ec_esimap[0]
 *         self._ecx_contextt.esislave = 0
*/
  __pyx_v_self->_ecx_contextt.esibuf = (&(__pyx_v_self->_ec_esibuf[0]));

  /* "pysoem/pysoem.pyx":265
 *         self._ecx_contextt.maxgroup = EC_MAXGROUP
 *         self._ecx_contextt.esibuf = &self._ec_esibuf[0]
 *         self._ecx_contextt.esimap = &self._ec_esimap[0]             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.esislave = 0
 *         self._ecx_contextt.elist = &self._ec_elist
*/
  __pyx_v_self->_ecx_contextt.esimap = (&(__pyx_v_self->_ec_esimap[0]));

  /* "pysoem/pysoem.pyx":266
 *         self._ecx_contextt.esibuf = &self._ec_esibuf[0]
 *         self._ecx_contextt.esimap = &self._ec_esimap[0]
 *         self._ecx_contextt.esislave = 0             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.elist = &self._ec_elist
 *         self._ecx_contextt.idxstack = &self._ec_idxstack
*/
  __pyx_v_self->_ecx_contextt.esislave = 0;

  /* "pysoem/pysoem.pyx":267
 *         self._ecx_contextt.esimap = &self._ec_esimap[0]
 *         self._ecx_contextt.esislave = 0
 *         self._ecx_contextt.elist = &self._ec_elist             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.idxstack = &self._ec_idxstack
 *         self._EcatError = 0
*/
  __pyx_v_self->_ecx_contextt.elist = (&__pyx_v_self->_ec_elist);

  /* "pysoem/pysoem.pyx":268
 *         self._ecx_contextt.esislave = 0
 *         self._ecx_contextt.elist = &self._ec_elist
 *         self._ecx_contextt.idxstack = &self._ec_idxstack             # <<<<<<<<<<<<<<
 *         self._EcatError = 0
 *         self._ecx_contextt.ecaterror = &self._EcatError
*/
  __pyx_v_self->_ecx_contextt.idxstack = (&__pyx_v_self->_ec_idxstack);

  /* "pysoem/pysoem.pyx":269
 *         self._ecx_contextt.elist = &self._ec_elist
 *         self._ecx_contextt.idxstack = &self._ec_idxstack
 *         self._EcatError = 0             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.ecaterror = &self._EcatError
 *         self._ecx_contextt.DCtime = &self._ec_DCtime
*/
  __pyx_v_self->_EcatError = 0;

  /* "pysoem/pysoem.pyx":270
 *         self._ecx_contextt.idxstack = &self._ec_idxstack
 *         self._EcatError = 0
 *         self._ecx_contextt.ecaterror = &self._EcatError             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.DCtime = &self._ec_DCtime
 *         self._ecx_contextt.SMcommtype = &self._ec_SMcommtype[0]
*/
  __pyx_v_self->_ecx_contextt.ecaterror = (&__pyx_v_self->_EcatError);

  /* "pysoem/pysoem.pyx":271
 *         self._EcatError = 0
 *         self._ecx_contextt.ecaterror = &self._EcatError
 *         self._ecx_contextt.DCtime = &self._ec_DCtime             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.SMcommtype = &self._ec_SMcommtype[0]
 *         self._ecx_contextt.PDOassign = &self._ec_PDOassign[0]
*/
  __pyx_v_self->_ecx_contextt.DCtime = (&__pyx_v_self->_ec_DCtime);

  /* "pysoem/pysoem.pyx":272
 *         self._ecx_contextt.ecaterror = &self._EcatError
 *         self._ecx_contextt.DCtime = &self._ec_DCtime
 *         self._ecx_contextt.SMcommtype = &self._ec_SMcommtype[0]             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.PDOassign = &self._ec_PDOassign[0]
 *         self._ecx_contextt.PDOdesc = &self._ec_PDOdesc[0]
*/
  __pyx_v_self->_ecx_contextt.SMcommtype = (&(__pyx_v_self->_ec_SMcommtype[0]));

  /* "pysoem/pysoem.pyx":273
 *         self._ecx_contextt.DCtime = &self._ec_DCtime
 *         self._ecx_contextt.SMcommtype = &self._ec_SMcommtype[0]
 *         self._ecx_contextt.PDOassign = &self._ec_PDOassign[0]             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.PDOdesc = &self._ec_PDOdesc[0]
 *         self._ecx_contextt.eepSM = &self._ec_SM
*/
  __pyx_v_self->_ecx_contextt.PDOassign = (&(__pyx_v_self->_ec_PDOassign[0]));

  /* "pysoem/pysoem.pyx":274
 *         self._ecx_contextt.SMcommtype = &self._ec_SMcommtype[0]
 *         self._ecx_contextt.PDOassign = &self._ec_PDOassign[0]
 *         self._ecx_contextt.PDOdesc = &self._ec_PDOdesc[0]             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.eepSM = &self._ec_SM
 *         self._ecx_contextt.eepFMMU = &self._ec_FMMU
*/
  __pyx_v_self->_ecx_contextt.PDOdesc = (&(__pyx_v_self->_ec_PDOdesc[0]));

  /* "pysoem/pysoem.pyx":275
 *         self._ecx_contextt.PDOassign = &self._ec_PDOassign[0]
 *         self._ecx_contextt.PDOdesc = &self._ec_PDOdesc[0]
 *         self._ecx_contextt.eepSM = &self._ec_SM             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.eepFMMU = &self._ec_FMMU
 *         self._ecx_contextt.FOEhook = NULL
*/
  __pyx_v_self->_ecx_contextt.eepSM = (&__pyx_v_self->_ec_SM);

  /* "pysoem/pysoem.pyx":276
 *         self._ecx_contextt.PDOdesc = &self._ec_PDOdesc[0]
 *         self._ecx_contextt.eepSM = &self._ec_SM
 *         self._ecx_contextt.eepFMMU = &self._ec_FMMU             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.FOEhook = NULL
 *         self._ecx_contextt.manualstatechange = 0
*/
  __pyx_v_self->_ecx_contextt.eepFMMU = (&__pyx_v_self->_ec_FMMU);

  /* "pysoem/pysoem.pyx":277
 *         self._ecx_contextt.eepSM = &self._ec_SM
 *         self._ecx_contextt.eepFMMU = &self._ec_FMMU
 *         self._ecx_contextt.FOEhook = NULL             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.manualstatechange = 0
 * 
*/
  __pyx_v_self->_ecx_contextt.FOEhook = NULL;

  /* "pysoem/pysoem.pyx":278
 *         self._ecx_contextt.eepFMMU = &self._ec_FMMU
 *         self._ecx_contextt.FOEhook = NULL
 *         self._ecx_contextt.manualstatechange = 0             # <<<<<<<<<<<<<<
 * 
 *         self.slaves = None
*/
  __pyx_v_self->_ecx_contextt.manualstatechange = 0;

  /* "pysoem/pysoem.pyx":280
 *         self._ecx_contextt.manualstatechange = 0
 * 
 *         self.slaves = None             # <<<<<<<<<<<<<<
 *         self.sdo_read_timeout = 700000
 *         self.sdo_write_timeout = 700000
*/
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_slaves, Py_None) < (0)) __PYX_ERR(0, 280, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":281
 * 
 *         self.slaves = None
 *         self.sdo_read_timeout = 700000             # <<<<<<<<<<<<<<
 *         self.sdo_write_timeout = 700000
 *         self.always_release_gil = settings.always_release_gil
*/
  __pyx_v_self->sdo_read_timeout = 0xAAE60;

  /* "pysoem/pysoem.pyx":282
 *         self.slaves = None
 *         self.sdo_read_timeout = 700000
 *         self.sdo_write_timeout = 700000             # <<<<<<<<<<<<<<
 *         self.always_release_gil = settings.always_release_gil
 *         self._settings.sdo_read_timeout = &self.sdo_read_timeout
*/
  __pyx_v_self->sdo_write_timeout = 0xAAE60;

  /* "pysoem/pysoem.pyx":283
 *         self.sdo_read_timeout = 700000
 *         self.sdo_write_timeout = 700000
 *         self.always_release_gil = settings.always_release_gil             # <<<<<<<<<<<<<<
 *         self._settings.sdo_read_timeout = &self.sdo_read_timeout
 *         self._settings.sdo_write_timeout = &self.sdo_write_timeout
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_settings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_always_release_gil); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyLong_As_int8_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->always_release_gil = __pyx_t_3;

  /* "pysoem/pysoem.pyx":284
 *         self.sdo_write_timeout = 700000
 *         self.always_release_gil = settings.always_release_gil
 *         self._settings.sdo_read_timeout = &self.sdo_read_timeout             # <<<<<<<<<<<<<<
 *         self._settings.sdo_write_timeout = &self.sdo_write_timeout
 *         self.context_initialized = False
*/
  __pyx_v_self->_settings.sdo_read_timeout = (&__pyx_v_self->sdo_read_timeout);

  /* "pysoem/pysoem.pyx":285
 *         self.always_release_gil = settings.always_release_gil
 *         self._settings.sdo_read_timeout = &self.sdo_read_timeout
 *         self._settings.sdo_write_timeout = &self.sdo_write_timeout             # <<<<<<<<<<<<<<
 *         self.context_initialized = False
 * 
*/
  __pyx_v_self->_settings.sdo_write_timeout = (&__pyx_v_self->sdo_write_timeout);

  /* "pysoem/pysoem.pyx":286
 *         self._settings.sdo_read_timeout = &self.sdo_read_timeout
 *         self._settings.sdo_write_timeout = &self.sdo_write_timeout
 *         self.context_initialized = False             # <<<<<<<<<<<<<<
 * 
 *     def open(self, ifname, ifname_red=None):
*/
  __pyx_v_self->context_initialized = 0;

  /* "pysoem/pysoem.pyx":257
 *     manual_state_change = property(_get_manual_state_change, _set_manual_state_change)
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._ecx_contextt.port = &self._ecx_port
 *         self._ecx_contextt.slavelist = &self._ec_slave[0]
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":288
 *         self.context_initialized = False
 * 
 *     def open(self, ifname, ifname_red=None):             # <<<<<<<<<<<<<<
 *         """Initialize and open network interface.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_3open(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_2open, "Initialize and open network interface.\n\n        On Linux the name of the interface is the same as usd by the system, e.g. ``eth0``, and as displayed by\n        ``ip addr``.\n\n        On Windows the names of the interfaces look like ``\\Device\\NPF_{1D123456-1E12-1C12-12F1-1234E123453B}``.\n        Finding the kind of name that SOEM expects is not straightforward. The most practical way is to use the\n        :func:`~find_adapters` method to find your available interfaces.\n\n        Args:\n            ifname(str): Interface name.\n            ifname_red(:obj:`str`, optional): Interface name of the second network interface card for redundancy.\n                Put to None if not used.\n        \n        Raises:\n            ConnectionError: When the specified interface dose not exist or\n                you have no permission to open the interface\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_3open = {"open", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_3open, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_2open};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_3open(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_ifname = 0;
  PyObject *__pyx_v_ifname_red = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("open (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_ifname,&__pyx_mstate_global->__pyx_n_u_ifname_red,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 288, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 288, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 288, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "open", 0) < (0)) __PYX_ERR(0, 288, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("open", 0, 1, 2, i); __PYX_ERR(0, 288, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 288, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 288, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_ifname = values[0];
    __pyx_v_ifname_red = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("open", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 288, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_2open(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v_ifname, __pyx_v_ifname_red);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_2open(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_ifname, PyObject *__pyx_v_ifname_red) {
  int __pyx_v_ret_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  char *__pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("open", 0);

  /* "pysoem/pysoem.pyx":307
 *                 you have no permission to open the interface
 *         """
 *         if ifname_red is None:             # <<<<<<<<<<<<<<
 *             ret_val = cpysoem.ecx_init(&self._ecx_contextt, ifname.encode('utf8'))
 *         else:
*/
  __pyx_t_1 = (__pyx_v_ifname_red == Py_None);
  if (__pyx_t_1) {

    /* "pysoem/pysoem.pyx":308
 *         """
 *         if ifname_red is None:
 *             ret_val = cpysoem.ecx_init(&self._ecx_contextt, ifname.encode('utf8'))             # <<<<<<<<<<<<<<
 *         else:
 *             ret_val = cpysoem.ecx_init_redundant(&self._ecx_contextt, &self._ecx_redport, ifname.encode('utf8'), ifname_red.encode('utf8'))
*/
    __pyx_t_3 = __pyx_v_ifname;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_n_u_utf8};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_t_2); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 308, __pyx_L1_error)
    __pyx_v_ret_val = ecx_init((&__pyx_v_self->_ecx_contextt), __pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pysoem/pysoem.pyx":307
 *                 you have no permission to open the interface
 *         """
 *         if ifname_red is None:             # <<<<<<<<<<<<<<
 *             ret_val = cpysoem.ecx_init(&self._ecx_contextt, ifname.encode('utf8'))
 *         else:
*/
    goto __pyx_L3;
  }

  /* "pysoem/pysoem.pyx":310
 *             ret_val = cpysoem.ecx_init(&self._ecx_contextt, ifname.encode('utf8'))
 *         else:
 *             ret_val = cpysoem.ecx_init_redundant(&self._ecx_contextt, &self._ecx_redport, ifname.encode('utf8'), ifname_red.encode('utf8'))             # <<<<<<<<<<<<<<
 *         if ret_val == 0:
 *             raise ConnectionError('could not open interface {}'.format(ifname))
*/
  /*else*/ {
    __pyx_t_3 = __pyx_v_ifname;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_n_u_utf8};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_6 = __Pyx_PyObject_AsString(__pyx_t_2); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 310, __pyx_L1_error)
    __pyx_t_7 = __pyx_v_ifname_red;
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_n_u_utf8};
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_t_3); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 310, __pyx_L1_error)
    __pyx_v_ret_val = ecx_init_redundant((&__pyx_v_self->_ecx_contextt), (&__pyx_v_self->_ecx_redport), __pyx_t_6, __pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "pysoem/pysoem.pyx":311
 *         else:
 *             ret_val = cpysoem.ecx_init_redundant(&self._ecx_contextt, &self._ecx_redport, ifname.encode('utf8'), ifname_red.encode('utf8'))
 *         if ret_val == 0:             # <<<<<<<<<<<<<<
 *             raise ConnectionError('could not open interface {}'.format(ifname))
 * 
*/
  __pyx_t_1 = (__pyx_v_ret_val == 0);
  if (unlikely(__pyx_t_1)) {

    /* "pysoem/pysoem.pyx":312
 *             ret_val = cpysoem.ecx_init_redundant(&self._ecx_contextt, &self._ecx_redport, ifname.encode('utf8'), ifname_red.encode('utf8'))
 *         if ret_val == 0:
 *             raise ConnectionError('could not open interface {}'.format(ifname))             # <<<<<<<<<<<<<<
 * 
 *         self.context_initialized = True
*/
    __pyx_t_2 = NULL;
    __pyx_t_8 = __pyx_mstate_global->__pyx_kp_u_could_not_open_interface;
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_ifname};
      __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
    }
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_7};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ConnectionError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 312, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":311
 *         else:
 *             ret_val = cpysoem.ecx_init_redundant(&self._ecx_contextt, &self._ecx_redport, ifname.encode('utf8'), ifname_red.encode('utf8'))
 *         if ret_val == 0:             # <<<<<<<<<<<<<<
 *             raise ConnectionError('could not open interface {}'.format(ifname))
 * 
*/
  }

  /* "pysoem/pysoem.pyx":314
 *             raise ConnectionError('could not open interface {}'.format(ifname))
 * 
 *         self.context_initialized = True             # <<<<<<<<<<<<<<
 * 
 *     def check_context_is_initialized(self):
*/
  __pyx_v_self->context_initialized = 1;

  /* "pysoem/pysoem.pyx":288
 *         self.context_initialized = False
 * 
 *     def open(self, ifname, ifname_red=None):             # <<<<<<<<<<<<<<
 *         """Initialize and open network interface.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":316
 *         self.context_initialized = True
 * 
 *     def check_context_is_initialized(self):             # <<<<<<<<<<<<<<
 *         if not self.context_initialized:
 *             raise NetworkInterfaceNotOpenError("SOEM Network interface is not initialized or has been closed. Call Master.open() first")
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_5check_context_is_initialized(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_5check_context_is_initialized = {"check_context_is_initialized", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_5check_context_is_initialized, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_5check_context_is_initialized(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_context_is_initialized (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("check_context_is_initialized", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("check_context_is_initialized", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_4check_context_is_initialized(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_4check_context_is_initialized(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_context_is_initialized", 0);

  /* "pysoem/pysoem.pyx":317
 * 
 *     def check_context_is_initialized(self):
 *         if not self.context_initialized:             # <<<<<<<<<<<<<<
 *             raise NetworkInterfaceNotOpenError("SOEM Network interface is not initialized or has been closed. Call Master.open() first")
 * 
*/
  __pyx_t_1 = (!(__pyx_v_self->context_initialized != 0));
  if (unlikely(__pyx_t_1)) {

    /* "pysoem/pysoem.pyx":318
 *     def check_context_is_initialized(self):
 *         if not self.context_initialized:
 *             raise NetworkInterfaceNotOpenError("SOEM Network interface is not initialized or has been closed. Call Master.open() first")             # <<<<<<<<<<<<<<
 * 
 *     cdef int __config_init_nogil(self, uint8_t usetable):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_NetworkInterfaceNotOpenError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_SOEM_Network_interface_is_not_in};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 318, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":317
 * 
 *     def check_context_is_initialized(self):
 *         if not self.context_initialized:             # <<<<<<<<<<<<<<
 *             raise NetworkInterfaceNotOpenError("SOEM Network interface is not initialized or has been closed. Call Master.open() first")
 * 
*/
  }

  /* "pysoem/pysoem.pyx":316
 *         self.context_initialized = True
 * 
 *     def check_context_is_initialized(self):             # <<<<<<<<<<<<<<
 *         if not self.context_initialized:
 *             raise NetworkInterfaceNotOpenError("SOEM Network interface is not initialized or has been closed. Call Master.open() first")
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.check_context_is_initialized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":320
 *             raise NetworkInterfaceNotOpenError("SOEM Network interface is not initialized or has been closed. Call Master.open() first")
 * 
 *     cdef int __config_init_nogil(self, uint8_t usetable):             # <<<<<<<<<<<<<<
 *         """Enumerate and init all slaves without GIL.
 * 
*/

static int __pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__config_init_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, uint8_t __pyx_v_usetable) {
  int __pyx_v_ret_val;
  int __pyx_r;

  /* "pysoem/pysoem.pyx":328
 *         cdef int ret_val
 * 
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             ret_val = cpysoem.ecx_config_init(&self._ecx_contextt, usetable)
*/
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":329
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             ret_val = cpysoem.ecx_config_init(&self._ecx_contextt, usetable)
 *         Py_DECREF(self)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pysoem/pysoem.pyx":330
 *         Py_INCREF(self)
 *         with nogil:
 *             ret_val = cpysoem.ecx_config_init(&self._ecx_contextt, usetable)             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
*/
        __pyx_v_ret_val = ecx_config_init((&__pyx_v_self->_ecx_contextt), __pyx_v_usetable);
      }

      /* "pysoem/pysoem.pyx":329
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             ret_val = cpysoem.ecx_config_init(&self._ecx_contextt, usetable)
 *         Py_DECREF(self)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pysoem/pysoem.pyx":331
 *         with nogil:
 *             ret_val = cpysoem.ecx_config_init(&self._ecx_contextt, usetable)
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *         return ret_val
*/
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":333
 *         Py_DECREF(self)
 * 
 *         return ret_val             # <<<<<<<<<<<<<<
 * 
 *     cpdef cpysoem.boolean check_release_gil(self, release_gil):
*/
  __pyx_r = __pyx_v_ret_val;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":320
 *             raise NetworkInterfaceNotOpenError("SOEM Network interface is not initialized or has been closed. Call Master.open() first")
 * 
 *     cdef int __config_init_nogil(self, uint8_t usetable):             # <<<<<<<<<<<<<<
 *         """Enumerate and init all slaves without GIL.
 * 
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":335
 *         return ret_val
 * 
 *     cpdef cpysoem.boolean check_release_gil(self, release_gil):             # <<<<<<<<<<<<<<
 *         """Checks if the GIL should be released.
 * 
*/

static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_7check_release_gil(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static boolean __pyx_f_6pysoem_6pysoem_10CdefMaster_check_release_gil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_release_gil, int __pyx_skip_dispatch) {
  boolean __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  boolean __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_release_gil", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_check_release_gil); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_6pysoem_6pysoem_10CdefMaster_7check_release_gil)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_release_gil};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_t_6 = __Pyx_PyLong_As_int8_t(__pyx_t_2); if (unlikely((__pyx_t_6 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pysoem/pysoem.pyx":341
 *             release_gil (boolean): True if the GIL should be released, False otherwise.
 *         """
 *         if release_gil is not None:             # <<<<<<<<<<<<<<
 *             return release_gil
 *         return self.always_release_gil
*/
  __pyx_t_7 = (__pyx_v_release_gil != Py_None);
  if (__pyx_t_7) {

    /* "pysoem/pysoem.pyx":342
 *         """
 *         if release_gil is not None:
 *             return release_gil             # <<<<<<<<<<<<<<
 *         return self.always_release_gil
 * 
*/
    __pyx_t_6 = __Pyx_PyLong_As_int8_t(__pyx_v_release_gil); if (unlikely((__pyx_t_6 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(0, 342, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "pysoem/pysoem.pyx":341
 *             release_gil (boolean): True if the GIL should be released, False otherwise.
 *         """
 *         if release_gil is not None:             # <<<<<<<<<<<<<<
 *             return release_gil
 *         return self.always_release_gil
*/
  }

  /* "pysoem/pysoem.pyx":343
 *         if release_gil is not None:
 *             return release_gil
 *         return self.always_release_gil             # <<<<<<<<<<<<<<
 * 
 *     def config_init(self, usetable=False, *, release_gil=None):
*/
  __pyx_r = __pyx_v_self->always_release_gil;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":335
 *         return ret_val
 * 
 *     cpdef cpysoem.boolean check_release_gil(self, release_gil):             # <<<<<<<<<<<<<<
 *         """Checks if the GIL should be released.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.check_release_gil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_7check_release_gil(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_6check_release_gil, "Checks if the GIL should be released.\n\n        Args:\n            release_gil (boolean): True if the GIL should be released, False otherwise.\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_7check_release_gil = {"check_release_gil", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_7check_release_gil, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_6check_release_gil};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_7check_release_gil(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_release_gil = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_release_gil (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_release_gil,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 335, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 335, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "check_release_gil", 0) < (0)) __PYX_ERR(0, 335, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("check_release_gil", 1, 1, 1, i); __PYX_ERR(0, 335, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 335, __pyx_L3_error)
    }
    __pyx_v_release_gil = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("check_release_gil", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 335, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.check_release_gil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_6check_release_gil(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v_release_gil);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_6check_release_gil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_release_gil) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  boolean __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_release_gil", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6pysoem_6pysoem_10CdefMaster_check_release_gil(__pyx_v_self, __pyx_v_release_gil, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 335, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.check_release_gil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":345
 *         return self.always_release_gil
 * 
 *     def config_init(self, usetable=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Enumerate and init all slaves.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_9config_init(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_8config_init, "Enumerate and init all slaves.\n        \n        Args:\n            usetable (bool): True when using configtable to init slaves, False otherwise.\n            release_gil (:obj:`bool`, optional): True to initialize the slaves releasing the GIL. Defaults to False.\n        \n        Returns:\n            int: Working counter of slave discover datagram = number of slaves found, -1 when no slave is connected\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_9config_init = {"config_init", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_9config_init, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_8config_init};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_9config_init(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_usetable = 0;
  PyObject *__pyx_v_release_gil = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("config_init (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_usetable,&__pyx_mstate_global->__pyx_n_u_release_gil,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 345, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 345, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "config_init", 0) < (0)) __PYX_ERR(0, 345, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 345, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_usetable = values[0];
    __pyx_v_release_gil = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("config_init", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 345, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.config_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_8config_init(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v_usetable, __pyx_v_release_gil);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_8config_init(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_usetable, PyObject *__pyx_v_release_gil) {
  int __pyx_v_ret_val;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  boolean __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  uint8_t __pyx_t_6;
  int __pyx_t_7;
  uint8 __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("config_init", 0);
  __Pyx_INCREF(__pyx_v_release_gil);

  /* "pysoem/pysoem.pyx":355
 *             int: Working counter of slave discover datagram = number of slaves found, -1 when no slave is connected
 *         """
 *         release_gil = self.check_release_gil(release_gil)             # <<<<<<<<<<<<<<
 *         self.check_context_is_initialized()
 *         self.slaves = []
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->__pyx_vtab)->check_release_gil(__pyx_v_self, __pyx_v_release_gil, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 355, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_release_gil, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":356
 *         """
 *         release_gil = self.check_release_gil(release_gil)
 *         self.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         self.slaves = []
 * 
*/
  __pyx_t_3 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":357
 *         release_gil = self.check_release_gil(release_gil)
 *         self.check_context_is_initialized()
 *         self.slaves = []             # <<<<<<<<<<<<<<
 * 
 *         cdef int ret_val
*/
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_slaves, __pyx_t_2) < (0)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":360
 * 
 *         cdef int ret_val
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             ret_val = self.__config_init_nogil(usetable)
 *         else:
*/
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_release_gil); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 360, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pysoem/pysoem.pyx":361
 *         cdef int ret_val
 *         if release_gil:
 *             ret_val = self.__config_init_nogil(usetable)             # <<<<<<<<<<<<<<
 *         else:
 *             ret_val = cpysoem.ecx_config_init(&self._ecx_contextt, usetable)
*/
    __pyx_t_6 = __Pyx_PyLong_As_uint8_t(__pyx_v_usetable); if (unlikely((__pyx_t_6 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 361, __pyx_L1_error)
    __pyx_t_7 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->__pyx_vtab)->_CdefMaster__config_init_nogil(__pyx_v_self, __pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 361, __pyx_L1_error)
    __pyx_v_ret_val = __pyx_t_7;

    /* "pysoem/pysoem.pyx":360
 * 
 *         cdef int ret_val
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             ret_val = self.__config_init_nogil(usetable)
 *         else:
*/
    goto __pyx_L3;
  }

  /* "pysoem/pysoem.pyx":363
 *             ret_val = self.__config_init_nogil(usetable)
 *         else:
 *             ret_val = cpysoem.ecx_config_init(&self._ecx_contextt, usetable)             # <<<<<<<<<<<<<<
 * 
 *         if ret_val > 0:
*/
  /*else*/ {
    __pyx_t_8 = __Pyx_PyLong_As_uint8_t(__pyx_v_usetable); if (unlikely((__pyx_t_8 == ((uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 363, __pyx_L1_error)
    __pyx_v_ret_val = ecx_config_init((&__pyx_v_self->_ecx_contextt), __pyx_t_8);
  }
  __pyx_L3:;

  /* "pysoem/pysoem.pyx":365
 *             ret_val = cpysoem.ecx_config_init(&self._ecx_contextt, usetable)
 * 
 *         if ret_val > 0:             # <<<<<<<<<<<<<<
 *           for i in range(self._ec_slavecount):
 *               self.slaves.append(self._get_slave(i))
*/
  __pyx_t_5 = (__pyx_v_ret_val > 0);
  if (__pyx_t_5) {

    /* "pysoem/pysoem.pyx":366
 * 
 *         if ret_val > 0:
 *           for i in range(self._ec_slavecount):             # <<<<<<<<<<<<<<
 *               self.slaves.append(self._get_slave(i))
 *         return ret_val
*/
    __pyx_t_7 = __pyx_v_self->_ec_slavecount;
    __pyx_t_9 = __pyx_t_7;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
      __pyx_v_i = __pyx_t_10;

      /* "pysoem/pysoem.pyx":367
 *         if ret_val > 0:
 *           for i in range(self._ec_slavecount):
 *               self.slaves.append(self._get_slave(i))             # <<<<<<<<<<<<<<
 *         return ret_val
 * 
*/
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_slaves); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_11 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyLong_From_int(__pyx_v_i); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_4 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_12};
        __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_slave, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 367, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __pyx_t_13 = __Pyx_PyObject_Append(__pyx_t_2, __pyx_t_3); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 367, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "pysoem/pysoem.pyx":365
 *             ret_val = cpysoem.ecx_config_init(&self._ecx_contextt, usetable)
 * 
 *         if ret_val > 0:             # <<<<<<<<<<<<<<
 *           for i in range(self._ec_slavecount):
 *               self.slaves.append(self._get_slave(i))
*/
  }

  /* "pysoem/pysoem.pyx":368
 *           for i in range(self._ec_slavecount):
 *               self.slaves.append(self._get_slave(i))
 *         return ret_val             # <<<<<<<<<<<<<<
 * 
 *     def config_map(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyLong_From_int(__pyx_v_ret_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":345
 *         return self.always_release_gil
 * 
 *     def config_init(self, usetable=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Enumerate and init all slaves.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.config_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_release_gil);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":370
 *         return ret_val
 * 
 *     def config_map(self):             # <<<<<<<<<<<<<<
 *         """Map all slaves PDOs in IO map.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_11config_map(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_10config_map, "Map all slaves PDOs in IO map.\n        \n        Returns:\n            int: IO map size (sum of all PDO in an out data)\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_11config_map = {"config_map", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_11config_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_10config_map};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_11config_map(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("config_map (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("config_map", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("config_map", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_10config_map(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_10config_map(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  struct __pyx_obj_6pysoem_6pysoem__CallbackData *__pyx_v_cd = 0;
  int __pyx_v_ret_val;
  PyObject *__pyx_v_slave = NULL;
  PyObject *__pyx_v_error_list = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("config_map", 0);

  /* "pysoem/pysoem.pyx":376
 *             int: IO map size (sum of all PDO in an out data)
 *         """
 *         self.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         cdef _CallbackData cd
 *         # ecx_config_map_group returns the actual IO map size (not an error value), expect the value to be less than EC_IOMAPSIZE
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":379
 *         cdef _CallbackData cd
 *         # ecx_config_map_group returns the actual IO map size (not an error value), expect the value to be less than EC_IOMAPSIZE
 *         ret_val = cpysoem.ecx_config_map_group(&self._ecx_contextt, &self.io_map, 0)             # <<<<<<<<<<<<<<
 *         # check for exceptions raised in the config functions
 *         for slave in self.slaves:
*/
  __pyx_v_ret_val = ecx_config_map_group((&__pyx_v_self->_ecx_contextt), (&__pyx_v_self->io_map), 0);

  /* "pysoem/pysoem.pyx":381
 *         ret_val = cpysoem.ecx_config_map_group(&self._ecx_contextt, &self.io_map, 0)
 *         # check for exceptions raised in the config functions
 *         for slave in self.slaves:             # <<<<<<<<<<<<<<
 *             cd = slave._cd
 *             if cd.exc_raised:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_slaves); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 381, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 381, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_2, __pyx_t_4, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_4;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 381, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4));
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_4);
        #endif
        ++__pyx_t_4;
      }
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 381, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_slave, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":382
 *         # check for exceptions raised in the config functions
 *         for slave in self.slaves:
 *             cd = slave._cd             # <<<<<<<<<<<<<<
 *             if cd.exc_raised:
 *                 raise cd.exc_info[0], cd.exc_info[1], cd.exc_info[2]
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_slave, __pyx_mstate_global->__pyx_n_u_cd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem__CallbackData))))) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_cd, ((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":383
 *         for slave in self.slaves:
 *             cd = slave._cd
 *             if cd.exc_raised:             # <<<<<<<<<<<<<<
 *                 raise cd.exc_info[0], cd.exc_info[1], cd.exc_info[2]
 *         logger.debug('io map size: {}'.format(ret_val))
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_cd->exc_raised); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 383, __pyx_L1_error)
    if (unlikely(__pyx_t_6)) {

      /* "pysoem/pysoem.pyx":384
 *             cd = slave._cd
 *             if cd.exc_raised:
 *                 raise cd.exc_info[0], cd.exc_info[1], cd.exc_info[2]             # <<<<<<<<<<<<<<
 *         logger.debug('io map size: {}'.format(ret_val))
 *         # raise an exception if one or more mailbox errors occured within ecx_config_map_group call
*/
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_cd->exc_info, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_cd->exc_info, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_cd->exc_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_1, __pyx_t_7, __pyx_t_8, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 384, __pyx_L1_error)

      /* "pysoem/pysoem.pyx":383
 *         for slave in self.slaves:
 *             cd = slave._cd
 *             if cd.exc_raised:             # <<<<<<<<<<<<<<
 *                 raise cd.exc_info[0], cd.exc_info[1], cd.exc_info[2]
 *         logger.debug('io map size: {}'.format(ret_val))
*/
    }

    /* "pysoem/pysoem.pyx":381
 *         ret_val = cpysoem.ecx_config_map_group(&self._ecx_contextt, &self.io_map, 0)
 *         # check for exceptions raised in the config functions
 *         for slave in self.slaves:             # <<<<<<<<<<<<<<
 *             cd = slave._cd
 *             if cd.exc_raised:
*/
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":385
 *             if cd.exc_raised:
 *                 raise cd.exc_info[0], cd.exc_info[1], cd.exc_info[2]
 *         logger.debug('io map size: {}'.format(ret_val))             # <<<<<<<<<<<<<<
 *         # raise an exception if one or more mailbox errors occured within ecx_config_map_group call
 *         error_list = self._collect_mailbox_errors()
*/
  __pyx_t_8 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_logger); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __pyx_mstate_global->__pyx_kp_u_io_map_size;
  __Pyx_INCREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyLong_From_int(__pyx_v_ret_val); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_10};
    __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  __pyx_t_3 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
    assert(__pyx_t_8);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
    __pyx_t_3 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":387
 *         logger.debug('io map size: {}'.format(ret_val))
 *         # raise an exception if one or more mailbox errors occured within ecx_config_map_group call
 *         error_list = self._collect_mailbox_errors()             # <<<<<<<<<<<<<<
 *         if len(error_list) > 0:
 *             raise ConfigMapError(error_list)
*/
  __pyx_t_1 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_collect_mailbox_errors, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_error_list = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":388
 *         # raise an exception if one or more mailbox errors occured within ecx_config_map_group call
 *         error_list = self._collect_mailbox_errors()
 *         if len(error_list) > 0:             # <<<<<<<<<<<<<<
 *             raise ConfigMapError(error_list)
 *         return ret_val
*/
  __pyx_t_4 = PyObject_Length(__pyx_v_error_list); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 388, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_4 > 0);
  if (unlikely(__pyx_t_6)) {

    /* "pysoem/pysoem.pyx":389
 *         error_list = self._collect_mailbox_errors()
 *         if len(error_list) > 0:
 *             raise ConfigMapError(error_list)             # <<<<<<<<<<<<<<
 *         return ret_val
 * 
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ConfigMapError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_error_list};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 389, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":388
 *         # raise an exception if one or more mailbox errors occured within ecx_config_map_group call
 *         error_list = self._collect_mailbox_errors()
 *         if len(error_list) > 0:             # <<<<<<<<<<<<<<
 *             raise ConfigMapError(error_list)
 *         return ret_val
*/
  }

  /* "pysoem/pysoem.pyx":390
 *         if len(error_list) > 0:
 *             raise ConfigMapError(error_list)
 *         return ret_val             # <<<<<<<<<<<<<<
 * 
 *     def config_overlap_map(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v_ret_val); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":370
 *         return ret_val
 * 
 *     def config_map(self):             # <<<<<<<<<<<<<<
 *         """Map all slaves PDOs in IO map.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.config_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cd);
  __Pyx_XDECREF(__pyx_v_slave);
  __Pyx_XDECREF(__pyx_v_error_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":392
 *         return ret_val
 * 
 *     def config_overlap_map(self):             # <<<<<<<<<<<<<<
 *         """Map all slaves PDOs to overlapping IO map.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_13config_overlap_map(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_12config_overlap_map, "Map all slaves PDOs to overlapping IO map.\n        \n        Returns:\n            int: IO map size (sum of all PDO in an out data)\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_13config_overlap_map = {"config_overlap_map", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_13config_overlap_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_12config_overlap_map};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_13config_overlap_map(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("config_overlap_map (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("config_overlap_map", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("config_overlap_map", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_12config_overlap_map(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_12config_overlap_map(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  struct __pyx_obj_6pysoem_6pysoem__CallbackData *__pyx_v_cd = 0;
  int __pyx_v_ret_val;
  PyObject *__pyx_v_slave = NULL;
  PyObject *__pyx_v_error_list = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("config_overlap_map", 0);

  /* "pysoem/pysoem.pyx":398
 *             int: IO map size (sum of all PDO in an out data)
 *         """
 *         self.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         cdef _CallbackData cd
 *         # ecx_config_map_group returns the actual IO map size (not an error value), expect the value to be less than EC_IOMAPSIZE
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":401
 *         cdef _CallbackData cd
 *         # ecx_config_map_group returns the actual IO map size (not an error value), expect the value to be less than EC_IOMAPSIZE
 *         ret_val = cpysoem.ecx_config_overlap_map_group(&self._ecx_contextt, &self.io_map, 0)             # <<<<<<<<<<<<<<
 *         # check for exceptions raised in the config functions
 *         for slave in self.slaves:
*/
  __pyx_v_ret_val = ecx_config_overlap_map_group((&__pyx_v_self->_ecx_contextt), (&__pyx_v_self->io_map), 0);

  /* "pysoem/pysoem.pyx":403
 *         ret_val = cpysoem.ecx_config_overlap_map_group(&self._ecx_contextt, &self.io_map, 0)
 *         # check for exceptions raised in the config functions
 *         for slave in self.slaves:             # <<<<<<<<<<<<<<
 *             cd = slave._cd
 *             if cd.exc_raised:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_slaves); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 403, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 403, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_2, __pyx_t_4, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_4;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 403, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4));
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_4);
        #endif
        ++__pyx_t_4;
      }
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 403, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_slave, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":404
 *         # check for exceptions raised in the config functions
 *         for slave in self.slaves:
 *             cd = slave._cd             # <<<<<<<<<<<<<<
 *             if cd.exc_raised:
 *                 raise cd.exc_info[0],cd.exc_info[1],cd.exc_info[2]
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_slave, __pyx_mstate_global->__pyx_n_u_cd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem__CallbackData))))) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_cd, ((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":405
 *         for slave in self.slaves:
 *             cd = slave._cd
 *             if cd.exc_raised:             # <<<<<<<<<<<<<<
 *                 raise cd.exc_info[0],cd.exc_info[1],cd.exc_info[2]
 *         logger.debug('io map size: {}'.format(ret_val))
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_cd->exc_raised); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 405, __pyx_L1_error)
    if (unlikely(__pyx_t_6)) {

      /* "pysoem/pysoem.pyx":406
 *             cd = slave._cd
 *             if cd.exc_raised:
 *                 raise cd.exc_info[0],cd.exc_info[1],cd.exc_info[2]             # <<<<<<<<<<<<<<
 *         logger.debug('io map size: {}'.format(ret_val))
 *         # raise an exception if one or more mailbox errors occured within ecx_config_overlap_map_group call
*/
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_cd->exc_info, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_cd->exc_info, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_cd->exc_info, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_1, __pyx_t_7, __pyx_t_8, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 406, __pyx_L1_error)

      /* "pysoem/pysoem.pyx":405
 *         for slave in self.slaves:
 *             cd = slave._cd
 *             if cd.exc_raised:             # <<<<<<<<<<<<<<
 *                 raise cd.exc_info[0],cd.exc_info[1],cd.exc_info[2]
 *         logger.debug('io map size: {}'.format(ret_val))
*/
    }

    /* "pysoem/pysoem.pyx":403
 *         ret_val = cpysoem.ecx_config_overlap_map_group(&self._ecx_contextt, &self.io_map, 0)
 *         # check for exceptions raised in the config functions
 *         for slave in self.slaves:             # <<<<<<<<<<<<<<
 *             cd = slave._cd
 *             if cd.exc_raised:
*/
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":407
 *             if cd.exc_raised:
 *                 raise cd.exc_info[0],cd.exc_info[1],cd.exc_info[2]
 *         logger.debug('io map size: {}'.format(ret_val))             # <<<<<<<<<<<<<<
 *         # raise an exception if one or more mailbox errors occured within ecx_config_overlap_map_group call
 *         error_list = self._collect_mailbox_errors()
*/
  __pyx_t_8 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_logger); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __pyx_mstate_global->__pyx_kp_u_io_map_size;
  __Pyx_INCREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyLong_From_int(__pyx_v_ret_val); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_10};
    __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  __pyx_t_3 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
    assert(__pyx_t_8);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
    __pyx_t_3 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":409
 *         logger.debug('io map size: {}'.format(ret_val))
 *         # raise an exception if one or more mailbox errors occured within ecx_config_overlap_map_group call
 *         error_list = self._collect_mailbox_errors()             # <<<<<<<<<<<<<<
 *         if len(error_list) > 0:
 *             raise ConfigMapError(error_list)
*/
  __pyx_t_1 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_collect_mailbox_errors, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_error_list = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":410
 *         # raise an exception if one or more mailbox errors occured within ecx_config_overlap_map_group call
 *         error_list = self._collect_mailbox_errors()
 *         if len(error_list) > 0:             # <<<<<<<<<<<<<<
 *             raise ConfigMapError(error_list)
 * 
*/
  __pyx_t_4 = PyObject_Length(__pyx_v_error_list); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 410, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_4 > 0);
  if (unlikely(__pyx_t_6)) {

    /* "pysoem/pysoem.pyx":411
 *         error_list = self._collect_mailbox_errors()
 *         if len(error_list) > 0:
 *             raise ConfigMapError(error_list)             # <<<<<<<<<<<<<<
 * 
 *         return ret_val
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ConfigMapError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_error_list};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 411, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":410
 *         # raise an exception if one or more mailbox errors occured within ecx_config_overlap_map_group call
 *         error_list = self._collect_mailbox_errors()
 *         if len(error_list) > 0:             # <<<<<<<<<<<<<<
 *             raise ConfigMapError(error_list)
 * 
*/
  }

  /* "pysoem/pysoem.pyx":413
 *             raise ConfigMapError(error_list)
 * 
 *         return ret_val             # <<<<<<<<<<<<<<
 * 
 *     def _collect_mailbox_errors(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v_ret_val); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":392
 *         return ret_val
 * 
 *     def config_overlap_map(self):             # <<<<<<<<<<<<<<
 *         """Map all slaves PDOs to overlapping IO map.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.config_overlap_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cd);
  __Pyx_XDECREF(__pyx_v_slave);
  __Pyx_XDECREF(__pyx_v_error_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":415
 *         return ret_val
 * 
 *     def _collect_mailbox_errors(self):             # <<<<<<<<<<<<<<
 *         # collect SDO or mailbox errors that occurred during PDO configuration read in ecx_config_map_group
 *         error_list = []
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_15_collect_mailbox_errors(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_15_collect_mailbox_errors = {"_collect_mailbox_errors", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_15_collect_mailbox_errors, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_15_collect_mailbox_errors(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_collect_mailbox_errors (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_collect_mailbox_errors", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_collect_mailbox_errors", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_14_collect_mailbox_errors(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_14_collect_mailbox_errors(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_v_error_list = NULL;
  ec_errort __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  char const *__pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  size_t __pyx_t_12;
  int __pyx_t_13;
  char *__pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_collect_mailbox_errors", 0);

  /* "pysoem/pysoem.pyx":417
 *     def _collect_mailbox_errors(self):
 *         # collect SDO or mailbox errors that occurred during PDO configuration read in ecx_config_map_group
 *         error_list = []             # <<<<<<<<<<<<<<
 *         cdef cpysoem.ec_errort err
 *         while cpysoem.ecx_poperror(&self._ecx_contextt, &err):
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_error_list = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":419
 *         error_list = []
 *         cdef cpysoem.ec_errort err
 *         while cpysoem.ecx_poperror(&self._ecx_contextt, &err):             # <<<<<<<<<<<<<<
 *             if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:
 *                 error_list.append(SdoError(err.Slave,
*/
  while (1) {
    __pyx_t_2 = (ecx_poperror((&__pyx_v_self->_ecx_contextt), (&__pyx_v_err)) != 0);
    if (!__pyx_t_2) break;

    /* "pysoem/pysoem.pyx":420
 *         cdef cpysoem.ec_errort err
 *         while cpysoem.ecx_poperror(&self._ecx_contextt, &err):
 *             if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:             # <<<<<<<<<<<<<<
 *                 error_list.append(SdoError(err.Slave,
 *                                            err.Index,
*/
    switch (__pyx_v_err.Etype) {
      case EC_ERR_TYPE_SDO_ERROR:

      /* "pysoem/pysoem.pyx":421
 *         while cpysoem.ecx_poperror(&self._ecx_contextt, &err):
 *             if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:
 *                 error_list.append(SdoError(err.Slave,             # <<<<<<<<<<<<<<
 *                                            err.Index,
 *                                            err.SubIdx,
*/
      __pyx_t_3 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_SdoError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyLong_From_uint16_t(__pyx_v_err.Slave); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "pysoem/pysoem.pyx":422
 *             if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:
 *                 error_list.append(SdoError(err.Slave,
 *                                            err.Index,             # <<<<<<<<<<<<<<
 *                                            err.SubIdx,
 *                                            err.AbortCode, cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8')))
*/
      __pyx_t_6 = __Pyx_PyLong_From_uint16_t(__pyx_v_err.Index); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 422, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "pysoem/pysoem.pyx":423
 *                 error_list.append(SdoError(err.Slave,
 *                                            err.Index,
 *                                            err.SubIdx,             # <<<<<<<<<<<<<<
 *                                            err.AbortCode, cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8')))
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:
*/
      __pyx_t_7 = __Pyx_PyLong_From_uint8_t(__pyx_v_err.SubIdx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 423, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "pysoem/pysoem.pyx":424
 *                                            err.Index,
 *                                            err.SubIdx,
 *                                            err.AbortCode, cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8')))             # <<<<<<<<<<<<<<
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:
 *                 error_list.append(MailboxError(err.Slave,
*/
      __pyx_t_8 = __Pyx_PyLong_From_int32_t(__pyx_v_err.AbortCode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 424, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = ec_sdoerror2string(__pyx_v_err.AbortCode);
      __pyx_t_10 = __Pyx_ssize_strlen(__pyx_t_9); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 424, __pyx_L1_error)
      __pyx_t_11 = __Pyx_decode_c_string(__pyx_t_9, 0, __pyx_t_10, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 424, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_3);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_12 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[6] = {__pyx_t_3, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_11};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_12, (6-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }

      /* "pysoem/pysoem.pyx":421
 *         while cpysoem.ecx_poperror(&self._ecx_contextt, &err):
 *             if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:
 *                 error_list.append(SdoError(err.Slave,             # <<<<<<<<<<<<<<
 *                                            err.Index,
 *                                            err.SubIdx,
*/
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_error_list, __pyx_t_1); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 421, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pysoem/pysoem.pyx":420
 *         cdef cpysoem.ec_errort err
 *         while cpysoem.ecx_poperror(&self._ecx_contextt, &err):
 *             if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:             # <<<<<<<<<<<<<<
 *                 error_list.append(SdoError(err.Slave,
 *                                            err.Index,
*/
      break;
      case EC_ERR_TYPE_MBX_ERROR:

      /* "pysoem/pysoem.pyx":426
 *                                            err.AbortCode, cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8')))
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:
 *                 error_list.append(MailboxError(err.Slave,             # <<<<<<<<<<<<<<
 *                                                err.ErrorCode,
 *                                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8')))
*/
      __pyx_t_4 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_MailboxError); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_8 = __Pyx_PyLong_From_uint16_t(__pyx_v_err.Slave); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "pysoem/pysoem.pyx":427
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:
 *                 error_list.append(MailboxError(err.Slave,
 *                                                err.ErrorCode,             # <<<<<<<<<<<<<<
 *                                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8')))
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:
*/
      __pyx_t_7 = __Pyx_PyLong_From_uint16_t(__pyx_v_err.ErrorCode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "pysoem/pysoem.pyx":428
 *                 error_list.append(MailboxError(err.Slave,
 *                                                err.ErrorCode,
 *                                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8')))             # <<<<<<<<<<<<<<
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:
 *                 error_list.append(PacketError(err.Slave,
*/
      __pyx_t_14 = ec_mbxerror2string(__pyx_v_err.ErrorCode);
      __pyx_t_10 = __Pyx_ssize_strlen(__pyx_t_14); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 428, __pyx_L1_error)
      __pyx_t_6 = __Pyx_decode_c_string(__pyx_t_14, 0, __pyx_t_10, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 428, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_12 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_11);
        assert(__pyx_t_4);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
        __pyx_t_12 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_t_8, __pyx_t_7, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_12, (4-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }

      /* "pysoem/pysoem.pyx":426
 *                                            err.AbortCode, cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8')))
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:
 *                 error_list.append(MailboxError(err.Slave,             # <<<<<<<<<<<<<<
 *                                                err.ErrorCode,
 *                                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8')))
*/
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_error_list, __pyx_t_1); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 426, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pysoem/pysoem.pyx":425
 *                                            err.SubIdx,
 *                                            err.AbortCode, cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8')))
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:             # <<<<<<<<<<<<<<
 *                 error_list.append(MailboxError(err.Slave,
 *                                                err.ErrorCode,
*/
      break;
      case EC_ERR_TYPE_PACKET_ERROR:

      /* "pysoem/pysoem.pyx":430
 *                                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8')))
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:
 *                 error_list.append(PacketError(err.Slave,             # <<<<<<<<<<<<<<
 *                                               err.ErrorCode))
 *             else:
*/
      __pyx_t_11 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_PacketError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyLong_From_uint16_t(__pyx_v_err.Slave); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "pysoem/pysoem.pyx":431
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:
 *                 error_list.append(PacketError(err.Slave,
 *                                               err.ErrorCode))             # <<<<<<<<<<<<<<
 *             else:
 *                 error_list.append(Exception('unexpected error'))
*/
      __pyx_t_8 = __Pyx_PyLong_From_uint16_t(__pyx_v_err.ErrorCode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 431, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_6);
        assert(__pyx_t_11);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
        __pyx_t_12 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_11, __pyx_t_7, __pyx_t_8};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 430, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }

      /* "pysoem/pysoem.pyx":430
 *                                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8')))
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:
 *                 error_list.append(PacketError(err.Slave,             # <<<<<<<<<<<<<<
 *                                               err.ErrorCode))
 *             else:
*/
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_error_list, __pyx_t_1); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 430, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pysoem/pysoem.pyx":429
 *                                                err.ErrorCode,
 *                                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8')))
 *             elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:             # <<<<<<<<<<<<<<
 *                 error_list.append(PacketError(err.Slave,
 *                                               err.ErrorCode))
*/
      break;
      default:

      /* "pysoem/pysoem.pyx":433
 *                                               err.ErrorCode))
 *             else:
 *                 error_list.append(Exception('unexpected error'))             # <<<<<<<<<<<<<<
 *         return error_list
 * 
*/
      __pyx_t_6 = NULL;
      __pyx_t_12 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_mstate_global->__pyx_kp_u_unexpected_error};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_Exception)), __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_error_list, __pyx_t_1); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      break;
    }
  }

  /* "pysoem/pysoem.pyx":434
 *             else:
 *                 error_list.append(Exception('unexpected error'))
 *         return error_list             # <<<<<<<<<<<<<<
 * 
 *     def config_dc(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_error_list);
  __pyx_r = __pyx_v_error_list;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":415
 *         return ret_val
 * 
 *     def _collect_mailbox_errors(self):             # <<<<<<<<<<<<<<
 *         # collect SDO or mailbox errors that occurred during PDO configuration read in ecx_config_map_group
 *         error_list = []
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster._collect_mailbox_errors", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_error_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":436
 *         return error_list
 * 
 *     def config_dc(self):             # <<<<<<<<<<<<<<
 *         """Locate DC slaves, measure propagation delays.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_17config_dc(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_16config_dc, "Locate DC slaves, measure propagation delays.\n        \n        Returns:\n            bool: if slaves are found with DC\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_17config_dc = {"config_dc", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_17config_dc, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_16config_dc};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_17config_dc(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("config_dc (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("config_dc", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("config_dc", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_16config_dc(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_16config_dc(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("config_dc", 0);

  /* "pysoem/pysoem.pyx":442
 *             bool: if slaves are found with DC
 *         """
 *         self.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_configdc(&self._ecx_contextt)
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":443
 *         """
 *         self.check_context_is_initialized()
 *         return cpysoem.ecx_configdc(&self._ecx_contextt)             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int8_t(ecx_configdc((&__pyx_v_self->_ecx_contextt))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":436
 *         return error_list
 * 
 *     def config_dc(self):             # <<<<<<<<<<<<<<
 *         """Locate DC slaves, measure propagation delays.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.config_dc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":445
 *         return cpysoem.ecx_configdc(&self._ecx_contextt)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the network interface.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_19close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_18close, "Close the network interface.\n        \n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_19close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_19close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_18close};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_19close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_18close(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_18close(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close", 0);

  /* "pysoem/pysoem.pyx":450
 *         """
 *         # ecx_close returns nothing
 *         self.context_initialized = False             # <<<<<<<<<<<<<<
 *         cpysoem.ecx_close(&self._ecx_contextt)
 * 
*/
  __pyx_v_self->context_initialized = 0;

  /* "pysoem/pysoem.pyx":451
 *         # ecx_close returns nothing
 *         self.context_initialized = False
 *         cpysoem.ecx_close(&self._ecx_contextt)             # <<<<<<<<<<<<<<
 * 
 *     def read_state(self):
*/
  ecx_close((&__pyx_v_self->_ecx_contextt));

  /* "pysoem/pysoem.pyx":445
 *         return cpysoem.ecx_configdc(&self._ecx_contextt)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the network interface.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":453
 *         cpysoem.ecx_close(&self._ecx_contextt)
 * 
 *     def read_state(self):             # <<<<<<<<<<<<<<
 *         """Read all slaves states.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_21read_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_20read_state, "Read all slaves states.\n        \n        Returns:\n            int: lowest state found\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_21read_state = {"read_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_21read_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_20read_state};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_21read_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_state (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_state", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_state", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_20read_state(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_20read_state(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_state", 0);

  /* "pysoem/pysoem.pyx":459
 *             int: lowest state found
 *         """
 *         self.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_readstate(&self._ecx_contextt)
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":460
 *         """
 *         self.check_context_is_initialized()
 *         return cpysoem.ecx_readstate(&self._ecx_contextt)             # <<<<<<<<<<<<<<
 * 
 *     def write_state(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(ecx_readstate((&__pyx_v_self->_ecx_contextt))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":453
 *         cpysoem.ecx_close(&self._ecx_contextt)
 * 
 *     def read_state(self):             # <<<<<<<<<<<<<<
 *         """Read all slaves states.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.read_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":462
 *         return cpysoem.ecx_readstate(&self._ecx_contextt)
 * 
 *     def write_state(self):             # <<<<<<<<<<<<<<
 *         """Write all slaves state.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_23write_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_22write_state, "Write all slaves state.\n        \n        The function does not check if the actual state is changed.\n        \n        Returns:\n            int: Working counter or EC_NOFRAME\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_23write_state = {"write_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_23write_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_22write_state};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_23write_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_state (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("write_state", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("write_state", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_22write_state(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_22write_state(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_state", 0);

  /* "pysoem/pysoem.pyx":470
 *             int: Working counter or EC_NOFRAME
 *         """
 *         self.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_writestate(&self._ecx_contextt, 0)
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":471
 *         """
 *         self.check_context_is_initialized()
 *         return cpysoem.ecx_writestate(&self._ecx_contextt, 0)             # <<<<<<<<<<<<<<
 * 
 *     def state_check(self, int expected_state, timeout=50000):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(ecx_writestate((&__pyx_v_self->_ecx_contextt), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":462
 *         return cpysoem.ecx_readstate(&self._ecx_contextt)
 * 
 *     def write_state(self):             # <<<<<<<<<<<<<<
 *         """Write all slaves state.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.write_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":473
 *         return cpysoem.ecx_writestate(&self._ecx_contextt, 0)
 * 
 *     def state_check(self, int expected_state, timeout=50000):             # <<<<<<<<<<<<<<
 *         """Check actual slave state.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_25state_check(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_24state_check, "Check actual slave state.\n        \n        This is a blocking function.\n        To refresh the state of all slaves read_state() should be called\n        \n        Args:\n            expected_state (int): Requested state\n            timeout (int): Timeout value in us\n        \n        Returns:\n            int: Requested state, or found state after timeout\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_25state_check = {"state_check", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_25state_check, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_24state_check};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_25state_check(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_expected_state;
  PyObject *__pyx_v_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("state_check (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_expected_state,&__pyx_mstate_global->__pyx_n_u_timeout,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 473, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 473, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 473, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "state_check", 0) < (0)) __PYX_ERR(0, 473, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_50000));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("state_check", 0, 1, 2, i); __PYX_ERR(0, 473, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 473, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 473, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_50000));
    }
    __pyx_v_expected_state = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_expected_state == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 473, __pyx_L3_error)
    __pyx_v_timeout = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("state_check", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 473, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.state_check", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_24state_check(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v_expected_state, __pyx_v_timeout);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_24state_check(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, int __pyx_v_expected_state, PyObject *__pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("state_check", 0);

  /* "pysoem/pysoem.pyx":486
 *             int: Requested state, or found state after timeout
 *         """
 *         self.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_statecheck(&self._ecx_contextt, 0, expected_state, timeout)
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":487
 *         """
 *         self.check_context_is_initialized()
 *         return cpysoem.ecx_statecheck(&self._ecx_contextt, 0, expected_state, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef int __send_processdata_nogil(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 487, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyLong_From_uint16_t(ecx_statecheck((&__pyx_v_self->_ecx_contextt), 0, __pyx_v_expected_state, __pyx_t_4)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":473
 *         return cpysoem.ecx_writestate(&self._ecx_contextt, 0)
 * 
 *     def state_check(self, int expected_state, timeout=50000):             # <<<<<<<<<<<<<<
 *         """Check actual slave state.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.state_check", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":489
 *         return cpysoem.ecx_statecheck(&self._ecx_contextt, 0, expected_state, timeout)
 * 
 *     cdef int __send_processdata_nogil(self):             # <<<<<<<<<<<<<<
 *         """Transmit processdata to slaves without GIL."""
 *         cdef int result
*/

static int __pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__send_processdata_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  int __pyx_v_result;
  int __pyx_r;

  /* "pysoem/pysoem.pyx":493
 *         cdef int result
 * 
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             result = cpysoem.ecx_send_processdata(&self._ecx_contextt)
*/
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":494
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_send_processdata(&self._ecx_contextt)
 *         Py_DECREF(self)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pysoem/pysoem.pyx":495
 *         Py_INCREF(self)
 *         with nogil:
 *             result = cpysoem.ecx_send_processdata(&self._ecx_contextt)             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
*/
        __pyx_v_result = ecx_send_processdata((&__pyx_v_self->_ecx_contextt));
      }

      /* "pysoem/pysoem.pyx":494
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_send_processdata(&self._ecx_contextt)
 *         Py_DECREF(self)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pysoem/pysoem.pyx":496
 *         with nogil:
 *             result = cpysoem.ecx_send_processdata(&self._ecx_contextt)
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *         return result
*/
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":498
 *         Py_DECREF(self)
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def send_processdata(self, *, release_gil=None):
*/
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":489
 *         return cpysoem.ecx_statecheck(&self._ecx_contextt, 0, expected_state, timeout)
 * 
 *     cdef int __send_processdata_nogil(self):             # <<<<<<<<<<<<<<
 *         """Transmit processdata to slaves without GIL."""
 *         cdef int result
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":500
 *         return result
 * 
 *     def send_processdata(self, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Transmit processdata to slaves.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_27send_processdata(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_26send_processdata, "Transmit processdata to slaves.\n        \n        Uses LRW, or LRD/LWR if LRW is not allowed (blockLRW).\n        Both the input and output processdata are transmitted.\n        The outputs with the actual data, the inputs have a placeholder.\n        The inputs are gathered with the receive processdata function.\n        In contrast to the base LRW function this function is non-blocking.\n        If the processdata does not fit in one datagram, multiple are used.\n        In order to recombine the slave response, a stack is used.\n\n        Args:\n            release_gil (:obj:`bool`, optional): True to transmit processdata releasing the GIL. Defaults to False.\n        \n        Returns:\n            int: >0 if processdata is transmitted, might only by 0 if config map is not configured properly\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_27send_processdata = {"send_processdata", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_27send_processdata, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_26send_processdata};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_27send_processdata(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_release_gil = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("send_processdata (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_release_gil,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 500, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, 0, __pyx_kwds_len, "send_processdata", 0) < (0)) __PYX_ERR(0, 500, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    } else if (unlikely(__pyx_nargs != 0)) {
      goto __pyx_L5_argtuple_error;
    } else {
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_release_gil = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("send_processdata", 1, 0, 0, __pyx_nargs); __PYX_ERR(0, 500, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.send_processdata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_26send_processdata(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v_release_gil);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_26send_processdata(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_release_gil) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  boolean __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("send_processdata", 0);
  __Pyx_INCREF(__pyx_v_release_gil);

  /* "pysoem/pysoem.pyx":517
 *             int: >0 if processdata is transmitted, might only by 0 if config map is not configured properly
 *         """
 *         release_gil = self.check_release_gil(release_gil)             # <<<<<<<<<<<<<<
 *         self.check_context_is_initialized()
 *         if release_gil:
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->__pyx_vtab)->check_release_gil(__pyx_v_self, __pyx_v_release_gil, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_release_gil, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":518
 *         """
 *         release_gil = self.check_release_gil(release_gil)
 *         self.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         if release_gil:
 *             return self.__send_processdata_nogil()
*/
  __pyx_t_3 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":519
 *         release_gil = self.check_release_gil(release_gil)
 *         self.check_context_is_initialized()
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             return self.__send_processdata_nogil()
 *         return cpysoem.ecx_send_processdata(&self._ecx_contextt)
*/
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_release_gil); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 519, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pysoem/pysoem.pyx":520
 *         self.check_context_is_initialized()
 *         if release_gil:
 *             return self.__send_processdata_nogil()             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_send_processdata(&self._ecx_contextt)
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->__pyx_vtab)->_CdefMaster__send_processdata_nogil(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 520, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pysoem/pysoem.pyx":519
 *         release_gil = self.check_release_gil(release_gil)
 *         self.check_context_is_initialized()
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             return self.__send_processdata_nogil()
 *         return cpysoem.ecx_send_processdata(&self._ecx_contextt)
*/
  }

  /* "pysoem/pysoem.pyx":521
 *         if release_gil:
 *             return self.__send_processdata_nogil()
 *         return cpysoem.ecx_send_processdata(&self._ecx_contextt)             # <<<<<<<<<<<<<<
 * 
 *     cdef int __send_overlap_processdata_nogil(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int(ecx_send_processdata((&__pyx_v_self->_ecx_contextt))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":500
 *         return result
 * 
 *     def send_processdata(self, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Transmit processdata to slaves.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.send_processdata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_release_gil);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":523
 *         return cpysoem.ecx_send_processdata(&self._ecx_contextt)
 * 
 *     cdef int __send_overlap_processdata_nogil(self):             # <<<<<<<<<<<<<<
 *         """Transmit overlap processdata to slaves without GIL."""
 *         cdef int result
*/

static int __pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__send_overlap_processdata_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  int __pyx_v_result;
  int __pyx_r;

  /* "pysoem/pysoem.pyx":526
 *         """Transmit overlap processdata to slaves without GIL."""
 *         cdef int result
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             result = cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)
*/
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":527
 *         cdef int result
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)
 *         Py_DECREF(self)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pysoem/pysoem.pyx":528
 *         Py_INCREF(self)
 *         with nogil:
 *             result = cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
*/
        __pyx_v_result = ecx_send_overlap_processdata((&__pyx_v_self->_ecx_contextt));
      }

      /* "pysoem/pysoem.pyx":527
 *         cdef int result
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)
 *         Py_DECREF(self)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pysoem/pysoem.pyx":529
 *         with nogil:
 *             result = cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *         return result
*/
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":531
 *         Py_DECREF(self)
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def send_overlap_processdata(self, *, release_gil=None):
*/
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":523
 *         return cpysoem.ecx_send_processdata(&self._ecx_contextt)
 * 
 *     cdef int __send_overlap_processdata_nogil(self):             # <<<<<<<<<<<<<<
 *         """Transmit overlap processdata to slaves without GIL."""
 *         cdef int result
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":533
 *         return result
 * 
 *     def send_overlap_processdata(self, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Transmit overlap processdata to slaves.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_29send_overlap_processdata(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_28send_overlap_processdata, "Transmit overlap processdata to slaves.\n        \n        Returns:\n            int: >0 if processdata is transmitted, might only by 0 if config map is not configured properly\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_29send_overlap_processdata = {"send_overlap_processdata", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_29send_overlap_processdata, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_28send_overlap_processdata};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_29send_overlap_processdata(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_release_gil = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("send_overlap_processdata (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_release_gil,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 533, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, 0, __pyx_kwds_len, "send_overlap_processdata", 0) < (0)) __PYX_ERR(0, 533, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    } else if (unlikely(__pyx_nargs != 0)) {
      goto __pyx_L5_argtuple_error;
    } else {
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_release_gil = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("send_overlap_processdata", 1, 0, 0, __pyx_nargs); __PYX_ERR(0, 533, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.send_overlap_processdata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_28send_overlap_processdata(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v_release_gil);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_28send_overlap_processdata(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_release_gil) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("send_overlap_processdata", 0);

  /* "pysoem/pysoem.pyx":539
 *             int: >0 if processdata is transmitted, might only by 0 if config map is not configured properly
 *         """
 *         self.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         if release_gil:
 *             return self.__send_overlap_processdata_nogil()
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":540
 *         """
 *         self.check_context_is_initialized()
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             return self.__send_overlap_processdata_nogil()
 *         return cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)
*/
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_release_gil); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 540, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "pysoem/pysoem.pyx":541
 *         self.check_context_is_initialized()
 *         if release_gil:
 *             return self.__send_overlap_processdata_nogil()             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->__pyx_vtab)->_CdefMaster__send_overlap_processdata_nogil(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 541, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pysoem/pysoem.pyx":540
 *         """
 *         self.check_context_is_initialized()
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             return self.__send_overlap_processdata_nogil()
 *         return cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)
*/
  }

  /* "pysoem/pysoem.pyx":542
 *         if release_gil:
 *             return self.__send_overlap_processdata_nogil()
 *         return cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)             # <<<<<<<<<<<<<<
 * 
 *     cdef int __receive_processdata_nogil(self, int timeout):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(ecx_send_overlap_processdata((&__pyx_v_self->_ecx_contextt))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":533
 *         return result
 * 
 *     def send_overlap_processdata(self, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Transmit overlap processdata to slaves.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.send_overlap_processdata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":544
 *         return cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)
 * 
 *     cdef int __receive_processdata_nogil(self, int timeout):             # <<<<<<<<<<<<<<
 *         """Receive processdata from slaves without GIL.
 * 
*/

static int __pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__receive_processdata_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, int __pyx_v_timeout) {
  int __pyx_v_result;
  int __pyx_r;

  /* "pysoem/pysoem.pyx":552
 *         cdef int result
 * 
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             result = cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)
*/
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":553
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)
 *         Py_DECREF(self)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pysoem/pysoem.pyx":554
 *         Py_INCREF(self)
 *         with nogil:
 *             result = cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
*/
        __pyx_v_result = ecx_receive_processdata((&__pyx_v_self->_ecx_contextt), __pyx_v_timeout);
      }

      /* "pysoem/pysoem.pyx":553
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)
 *         Py_DECREF(self)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pysoem/pysoem.pyx":555
 *         with nogil:
 *             result = cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *         return result
*/
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":557
 *         Py_DECREF(self)
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def receive_processdata(self, timeout=2000, *, release_gil=None):
*/
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":544
 *         return cpysoem.ecx_send_overlap_processdata(&self._ecx_contextt)
 * 
 *     cdef int __receive_processdata_nogil(self, int timeout):             # <<<<<<<<<<<<<<
 *         """Receive processdata from slaves without GIL.
 * 
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":559
 *         return result
 * 
 *     def receive_processdata(self, timeout=2000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Receive processdata from slaves.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_31receive_processdata(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_30receive_processdata, "Receive processdata from slaves.\n\n        Second part from send_processdata().\n        Received datagrams are recombined with the processdata with help from the stack.\n        If a datagram contains input processdata it copies it to the processdata structure.\n\n        Args:\n            timeout (int): Timeout in us.\n            release_gil (:obj:`bool`, optional): True to receive processdata releasing the GIL. Defaults to False.\n        Returns\n            int: Working Counter\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_31receive_processdata = {"receive_processdata", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_31receive_processdata, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_30receive_processdata};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_31receive_processdata(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_timeout = 0;
  PyObject *__pyx_v_release_gil = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("receive_processdata (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_timeout,&__pyx_mstate_global->__pyx_n_u_release_gil,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 559, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 559, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "receive_processdata", 0) < (0)) __PYX_ERR(0, 559, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 559, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_timeout = values[0];
    __pyx_v_release_gil = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("receive_processdata", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 559, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.receive_processdata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_30receive_processdata(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v_timeout, __pyx_v_release_gil);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_30receive_processdata(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_timeout, PyObject *__pyx_v_release_gil) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  boolean __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("receive_processdata", 0);
  __Pyx_INCREF(__pyx_v_release_gil);

  /* "pysoem/pysoem.pyx":572
 *             int: Working Counter
 *         """
 *         release_gil = self.check_release_gil(release_gil)             # <<<<<<<<<<<<<<
 *         self.check_context_is_initialized()
 *         if release_gil:
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->__pyx_vtab)->check_release_gil(__pyx_v_self, __pyx_v_release_gil, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_release_gil, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":573
 *         """
 *         release_gil = self.check_release_gil(release_gil)
 *         self.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         if release_gil:
 *             return self.__receive_processdata_nogil(timeout)
*/
  __pyx_t_3 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":574
 *         release_gil = self.check_release_gil(release_gil)
 *         self.check_context_is_initialized()
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             return self.__receive_processdata_nogil(timeout)
 *         return cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)
*/
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_release_gil); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 574, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pysoem/pysoem.pyx":575
 *         self.check_context_is_initialized()
 *         if release_gil:
 *             return self.__receive_processdata_nogil(timeout)             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 575, __pyx_L1_error)
    __pyx_t_7 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->__pyx_vtab)->_CdefMaster__receive_processdata_nogil(__pyx_v_self, __pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 575, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pysoem/pysoem.pyx":574
 *         release_gil = self.check_release_gil(release_gil)
 *         self.check_context_is_initialized()
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             return self.__receive_processdata_nogil(timeout)
 *         return cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)
*/
  }

  /* "pysoem/pysoem.pyx":576
 *         if release_gil:
 *             return self.__receive_processdata_nogil(timeout)
 *         return cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def _get_slave(self, int pos):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 576, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int(ecx_receive_processdata((&__pyx_v_self->_ecx_contextt), __pyx_t_7)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":559
 *         return result
 * 
 *     def receive_processdata(self, timeout=2000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Receive processdata from slaves.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.receive_processdata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_release_gil);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":578
 *         return cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)
 * 
 *     def _get_slave(self, int pos):             # <<<<<<<<<<<<<<
 *         if pos < 0:
 *             raise IndexError('requested slave device is not available')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_33_get_slave(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_33_get_slave = {"_get_slave", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_33_get_slave, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_33_get_slave(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_pos;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_slave (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pos,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 578, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 578, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_get_slave", 0) < (0)) __PYX_ERR(0, 578, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_get_slave", 1, 1, 1, i); __PYX_ERR(0, 578, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 578, __pyx_L3_error)
    }
    __pyx_v_pos = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_pos == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 578, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_slave", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 578, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster._get_slave", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_32_get_slave(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v_pos);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_32_get_slave(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, int __pyx_v_pos) {
  struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_ethercat_slave = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_slave", 0);

  /* "pysoem/pysoem.pyx":579
 * 
 *     def _get_slave(self, int pos):
 *         if pos < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError('requested slave device is not available')
 *         if pos >= self._ec_slavecount:
*/
  __pyx_t_1 = (__pyx_v_pos < 0);
  if (unlikely(__pyx_t_1)) {

    /* "pysoem/pysoem.pyx":580
 *     def _get_slave(self, int pos):
 *         if pos < 0:
 *             raise IndexError('requested slave device is not available')             # <<<<<<<<<<<<<<
 *         if pos >= self._ec_slavecount:
 *             raise IndexError('requested slave device is not available')
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_requested_slave_device_is_not_av};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_IndexError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 580, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":579
 * 
 *     def _get_slave(self, int pos):
 *         if pos < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError('requested slave device is not available')
 *         if pos >= self._ec_slavecount:
*/
  }

  /* "pysoem/pysoem.pyx":581
 *         if pos < 0:
 *             raise IndexError('requested slave device is not available')
 *         if pos >= self._ec_slavecount:             # <<<<<<<<<<<<<<
 *             raise IndexError('requested slave device is not available')
 *         ethercat_slave = CdefSlave(pos+1)
*/
  __pyx_t_1 = (__pyx_v_pos >= __pyx_v_self->_ec_slavecount);
  if (unlikely(__pyx_t_1)) {

    /* "pysoem/pysoem.pyx":582
 *             raise IndexError('requested slave device is not available')
 *         if pos >= self._ec_slavecount:
 *             raise IndexError('requested slave device is not available')             # <<<<<<<<<<<<<<
 *         ethercat_slave = CdefSlave(pos+1)
 *         ethercat_slave._master = self
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_requested_slave_device_is_not_av};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_IndexError)), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 582, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":581
 *         if pos < 0:
 *             raise IndexError('requested slave device is not available')
 *         if pos >= self._ec_slavecount:             # <<<<<<<<<<<<<<
 *             raise IndexError('requested slave device is not available')
 *         ethercat_slave = CdefSlave(pos+1)
*/
  }

  /* "pysoem/pysoem.pyx":583
 *         if pos >= self._ec_slavecount:
 *             raise IndexError('requested slave device is not available')
 *         ethercat_slave = CdefSlave(pos+1)             # <<<<<<<<<<<<<<
 *         ethercat_slave._master = self
 *         ethercat_slave._ecx_contextt = &self._ecx_contextt
*/
  __pyx_t_3 = NULL;
  __pyx_t_5 = __Pyx_PyLong_From_long((__pyx_v_pos + 1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_ethercat_slave = ((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":584
 *             raise IndexError('requested slave device is not available')
 *         ethercat_slave = CdefSlave(pos+1)
 *         ethercat_slave._master = self             # <<<<<<<<<<<<<<
 *         ethercat_slave._ecx_contextt = &self._ecx_contextt
 *         ethercat_slave._ec_slave = &self._ec_slave[pos+1] # +1 as _ec_slave[0] is reserved
*/
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  __Pyx_GOTREF((PyObject *)__pyx_v_ethercat_slave->_master);
  __Pyx_DECREF((PyObject *)__pyx_v_ethercat_slave->_master);
  __pyx_v_ethercat_slave->_master = __pyx_v_self;

  /* "pysoem/pysoem.pyx":585
 *         ethercat_slave = CdefSlave(pos+1)
 *         ethercat_slave._master = self
 *         ethercat_slave._ecx_contextt = &self._ecx_contextt             # <<<<<<<<<<<<<<
 *         ethercat_slave._ec_slave = &self._ec_slave[pos+1] # +1 as _ec_slave[0] is reserved
 *         ethercat_slave._the_masters_settings = &self._settings
*/
  __pyx_v_ethercat_slave->_ecx_contextt = (&__pyx_v_self->_ecx_contextt);

  /* "pysoem/pysoem.pyx":586
 *         ethercat_slave._master = self
 *         ethercat_slave._ecx_contextt = &self._ecx_contextt
 *         ethercat_slave._ec_slave = &self._ec_slave[pos+1] # +1 as _ec_slave[0] is reserved             # <<<<<<<<<<<<<<
 *         ethercat_slave._the_masters_settings = &self._settings
 *         return ethercat_slave
*/
  __pyx_v_ethercat_slave->_ec_slave = (&(__pyx_v_self->_ec_slave[(__pyx_v_pos + 1)]));

  /* "pysoem/pysoem.pyx":587
 *         ethercat_slave._ecx_contextt = &self._ecx_contextt
 *         ethercat_slave._ec_slave = &self._ec_slave[pos+1] # +1 as _ec_slave[0] is reserved
 *         ethercat_slave._the_masters_settings = &self._settings             # <<<<<<<<<<<<<<
 *         return ethercat_slave
 * 
*/
  __pyx_v_ethercat_slave->_the_masters_settings = (&__pyx_v_self->_settings);

  /* "pysoem/pysoem.pyx":588
 *         ethercat_slave._ec_slave = &self._ec_slave[pos+1] # +1 as _ec_slave[0] is reserved
 *         ethercat_slave._the_masters_settings = &self._settings
 *         return ethercat_slave             # <<<<<<<<<<<<<<
 * 
 *     def _get_state(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ethercat_slave);
  __pyx_r = ((PyObject *)__pyx_v_ethercat_slave);
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":578
 *         return cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)
 * 
 *     def _get_slave(self, int pos):             # <<<<<<<<<<<<<<
 *         if pos < 0:
 *             raise IndexError('requested slave device is not available')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster._get_slave", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ethercat_slave);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":590
 *         return ethercat_slave
 * 
 *     def _get_state(self):             # <<<<<<<<<<<<<<
 *         """Can be used to check if all slaves are in Operational state, or to request a new state for all slaves.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_35_get_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_34_get_state, "Can be used to check if all slaves are in Operational state, or to request a new state for all slaves.\n\n        Make sure to call write_state(), once a new state for all slaves was set.\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_35_get_state = {"_get_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_35_get_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_34_get_state};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_35_get_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_state (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_state", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_state", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_34_get_state(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_34_get_state(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_state", 0);

  /* "pysoem/pysoem.pyx":595
 *         Make sure to call write_state(), once a new state for all slaves was set.
 *         """
 *         return self._ec_slave[0].state             # <<<<<<<<<<<<<<
 * 
 *     def _set_state(self, value):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint16_t((__pyx_v_self->_ec_slave[0]).state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":590
 *         return ethercat_slave
 * 
 *     def _get_state(self):             # <<<<<<<<<<<<<<
 *         """Can be used to check if all slaves are in Operational state, or to request a new state for all slaves.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster._get_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":597
 *         return self._ec_slave[0].state
 * 
 *     def _set_state(self, value):             # <<<<<<<<<<<<<<
 *         self._ec_slave[0].state = value
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_37_set_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_37_set_state = {"_set_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_37_set_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_37_set_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_state (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_value,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 597, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 597, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_set_state", 0) < (0)) __PYX_ERR(0, 597, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_set_state", 1, 1, 1, i); __PYX_ERR(0, 597, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 597, __pyx_L3_error)
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_set_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 597, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster._set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_36_set_state(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_36_set_state(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16 __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_state", 0);

  /* "pysoem/pysoem.pyx":598
 * 
 *     def _set_state(self, value):
 *         self._ec_slave[0].state = value             # <<<<<<<<<<<<<<
 * 
 *     def _get_expected_wkc(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint16_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 598, __pyx_L1_error)
  (__pyx_v_self->_ec_slave[0]).state = __pyx_t_1;

  /* "pysoem/pysoem.pyx":597
 *         return self._ec_slave[0].state
 * 
 *     def _set_state(self, value):             # <<<<<<<<<<<<<<
 *         self._ec_slave[0].state = value
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster._set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":600
 *         self._ec_slave[0].state = value
 * 
 *     def _get_expected_wkc(self):             # <<<<<<<<<<<<<<
 *         """Calculates the expected Working Counter"""
 *         return (self._ec_group[0].outputsWKC * 2) + self._ec_group[0].inputsWKC
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_39_get_expected_wkc(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_38_get_expected_wkc, "Calculates the expected Working Counter");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_39_get_expected_wkc = {"_get_expected_wkc", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_39_get_expected_wkc, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_38_get_expected_wkc};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_39_get_expected_wkc(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_expected_wkc (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_expected_wkc", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_expected_wkc", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_38_get_expected_wkc(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_38_get_expected_wkc(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_expected_wkc", 0);

  /* "pysoem/pysoem.pyx":602
 *     def _get_expected_wkc(self):
 *         """Calculates the expected Working Counter"""
 *         return (self._ec_group[0].outputsWKC * 2) + self._ec_group[0].inputsWKC             # <<<<<<<<<<<<<<
 * 
 *     def _get_dc_time(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_long((((__pyx_v_self->_ec_group[0]).outputsWKC * 2) + (__pyx_v_self->_ec_group[0]).inputsWKC)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":600
 *         self._ec_slave[0].state = value
 * 
 *     def _get_expected_wkc(self):             # <<<<<<<<<<<<<<
 *         """Calculates the expected Working Counter"""
 *         return (self._ec_group[0].outputsWKC * 2) + self._ec_group[0].inputsWKC
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster._get_expected_wkc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":604
 *         return (self._ec_group[0].outputsWKC * 2) + self._ec_group[0].inputsWKC
 * 
 *     def _get_dc_time(self):             # <<<<<<<<<<<<<<
 *         """DC time in ns required to synchronize the EtherCAT cycle with SYNC0 cycles.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_41_get_dc_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_40_get_dc_time, "DC time in ns required to synchronize the EtherCAT cycle with SYNC0 cycles.\n\n        Note EtherCAT cycle here means the call of send_processdata and receive_processdata.");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_41_get_dc_time = {"_get_dc_time", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_41_get_dc_time, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_40_get_dc_time};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_41_get_dc_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_dc_time (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_dc_time", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_dc_time", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_40_get_dc_time(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_40_get_dc_time(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_dc_time", 0);

  /* "pysoem/pysoem.pyx":608
 * 
 *         Note EtherCAT cycle here means the call of send_processdata and receive_processdata."""
 *         return self._ec_DCtime             # <<<<<<<<<<<<<<
 * 
 *     def _set_manual_state_change(self, int manual_state_change):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_ec_DCtime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":604
 *         return (self._ec_group[0].outputsWKC * 2) + self._ec_group[0].inputsWKC
 * 
 *     def _get_dc_time(self):             # <<<<<<<<<<<<<<
 *         """DC time in ns required to synchronize the EtherCAT cycle with SYNC0 cycles.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster._get_dc_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":610
 *         return self._ec_DCtime
 * 
 *     def _set_manual_state_change(self, int manual_state_change):             # <<<<<<<<<<<<<<
 *         """Set manualstatechange variable in context.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_43_set_manual_state_change(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_10CdefMaster_42_set_manual_state_change, "Set manualstatechange variable in context.\n        \n        Flag to control legacy automatic state change or manual state change in functions\n        config_init() and config_map()\n        Flag value == 0 is legacy automatic state\n        Flag value != 0 and states must be handled manually\n        Args:\n            manual_state_change (int): The manual state change flag.\n\n        .. versionadded:: 1.0.5\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_43_set_manual_state_change = {"_set_manual_state_change", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_43_set_manual_state_change, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_42_set_manual_state_change};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_43_set_manual_state_change(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_manual_state_change;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_manual_state_change (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_manual_state_change,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 610, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 610, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_set_manual_state_change", 0) < (0)) __PYX_ERR(0, 610, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_set_manual_state_change", 1, 1, 1, i); __PYX_ERR(0, 610, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 610, __pyx_L3_error)
    }
    __pyx_v_manual_state_change = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_manual_state_change == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 610, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_set_manual_state_change", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 610, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster._set_manual_state_change", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_42_set_manual_state_change(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v_manual_state_change);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_42_set_manual_state_change(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, int __pyx_v_manual_state_change) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_manual_state_change", 0);

  /* "pysoem/pysoem.pyx":622
 *         .. versionadded:: 1.0.5
 *         """
 *         self._ecx_contextt.manualstatechange = manual_state_change             # <<<<<<<<<<<<<<
 * 
 *     def _get_manual_state_change(self):
*/
  __pyx_v_self->_ecx_contextt.manualstatechange = __pyx_v_manual_state_change;

  /* "pysoem/pysoem.pyx":610
 *         return self._ec_DCtime
 * 
 *     def _set_manual_state_change(self, int manual_state_change):             # <<<<<<<<<<<<<<
 *         """Set manualstatechange variable in context.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":624
 *         self._ecx_contextt.manualstatechange = manual_state_change
 * 
 *     def _get_manual_state_change(self):             # <<<<<<<<<<<<<<
 *         return self._ecx_contextt.manualstatechange
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_45_get_manual_state_change(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_45_get_manual_state_change = {"_get_manual_state_change", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_45_get_manual_state_change, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_45_get_manual_state_change(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_manual_state_change (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_manual_state_change", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_manual_state_change", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_44_get_manual_state_change(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_44_get_manual_state_change(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_manual_state_change", 0);

  /* "pysoem/pysoem.pyx":625
 * 
 *     def _get_manual_state_change(self):
 *         return self._ecx_contextt.manualstatechange             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->_ecx_contextt.manualstatechange); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":624
 *         self._ecx_contextt.manualstatechange = manual_state_change
 * 
 *     def _get_manual_state_change(self):             # <<<<<<<<<<<<<<
 *         return self._ecx_contextt.manualstatechange
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster._get_manual_state_change", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":247
 *     cdef char io_map[EC_IOMAPSIZE]
 *     cdef CdefMasterSettings _settings
 *     cdef public int sdo_read_timeout             # <<<<<<<<<<<<<<
 *     cdef public int sdo_write_timeout
 *     cdef public cpysoem.boolean always_release_gil
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout___get__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->sdo_read_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.sdo_read_timeout.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 247, __pyx_L1_error)
  __pyx_v_self->sdo_read_timeout = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.sdo_read_timeout.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":248
 *     cdef CdefMasterSettings _settings
 *     cdef public int sdo_read_timeout
 *     cdef public int sdo_write_timeout             # <<<<<<<<<<<<<<
 *     cdef public cpysoem.boolean always_release_gil
 *     cdef readonly cpysoem.boolean context_initialized
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout___get__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->sdo_write_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.sdo_write_timeout.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 248, __pyx_L1_error)
  __pyx_v_self->sdo_write_timeout = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.sdo_write_timeout.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":249
 *     cdef public int sdo_read_timeout
 *     cdef public int sdo_write_timeout
 *     cdef public cpysoem.boolean always_release_gil             # <<<<<<<<<<<<<<
 *     cdef readonly cpysoem.boolean context_initialized
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_18always_release_gil_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_18always_release_gil_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_18always_release_gil___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_18always_release_gil___get__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int8_t(__pyx_v_self->always_release_gil); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.always_release_gil.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_10CdefMaster_18always_release_gil_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_10CdefMaster_18always_release_gil_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_18always_release_gil_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_10CdefMaster_18always_release_gil_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  boolean __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyLong_As_int8_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(0, 249, __pyx_L1_error)
  __pyx_v_self->always_release_gil = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.always_release_gil.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":250
 *     cdef public int sdo_write_timeout
 *     cdef public cpysoem.boolean always_release_gil
 *     cdef readonly cpysoem.boolean context_initialized             # <<<<<<<<<<<<<<
 * 
 *     state = property(_get_state, _set_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_19context_initialized_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_19context_initialized_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_19context_initialized___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_19context_initialized___get__(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int8_t(__pyx_v_self->context_initialized); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.context_initialized.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_47__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_47__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_47__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_47__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_46__reduce_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_46__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_49__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10CdefMaster_49__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_49__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_10CdefMaster_49__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_10CdefMaster_48__setstate_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_10CdefMaster_48__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefMaster *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefMaster.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":638
 *     """
 * 
 *     def __init__(self, slave_pos, index, subindex, abort_code, desc):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.index = index
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_8SdoError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_8SdoError_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_8SdoError_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_8SdoError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_slave_pos = 0;
  PyObject *__pyx_v_index = 0;
  PyObject *__pyx_v_subindex = 0;
  PyObject *__pyx_v_abort_code = 0;
  PyObject *__pyx_v_desc = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_slave_pos,&__pyx_mstate_global->__pyx_n_u_index,&__pyx_mstate_global->__pyx_n_u_subindex,&__pyx_mstate_global->__pyx_n_u_abort_code,&__pyx_mstate_global->__pyx_n_u_desc,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 638, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 638, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 638, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 638, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 638, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 638, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 638, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 638, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 6; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, i); __PYX_ERR(0, 638, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 6)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 638, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 638, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 638, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 638, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 638, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 638, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_slave_pos = values[1];
    __pyx_v_index = values[2];
    __pyx_v_subindex = values[3];
    __pyx_v_abort_code = values[4];
    __pyx_v_desc = values[5];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 6, 6, __pyx_nargs); __PYX_ERR(0, 638, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.SdoError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_8SdoError___init__(__pyx_self, __pyx_v_self, __pyx_v_slave_pos, __pyx_v_index, __pyx_v_subindex, __pyx_v_abort_code, __pyx_v_desc);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_8SdoError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_slave_pos, PyObject *__pyx_v_index, PyObject *__pyx_v_subindex, PyObject *__pyx_v_abort_code, PyObject *__pyx_v_desc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pysoem/pysoem.pyx":639
 * 
 *     def __init__(self, slave_pos, index, subindex, abort_code, desc):
 *         self.slave_pos = slave_pos             # <<<<<<<<<<<<<<
 *         self.index = index
 *         self.subindex = subindex
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_slave_pos, __pyx_v_slave_pos) < (0)) __PYX_ERR(0, 639, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":640
 *     def __init__(self, slave_pos, index, subindex, abort_code, desc):
 *         self.slave_pos = slave_pos
 *         self.index = index             # <<<<<<<<<<<<<<
 *         self.subindex = subindex
 *         self.abort_code = abort_code
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_index, __pyx_v_index) < (0)) __PYX_ERR(0, 640, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":641
 *         self.slave_pos = slave_pos
 *         self.index = index
 *         self.subindex = subindex             # <<<<<<<<<<<<<<
 *         self.abort_code = abort_code
 *         self.desc = desc
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_subindex, __pyx_v_subindex) < (0)) __PYX_ERR(0, 641, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":642
 *         self.index = index
 *         self.subindex = subindex
 *         self.abort_code = abort_code             # <<<<<<<<<<<<<<
 *         self.desc = desc
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_abort_code, __pyx_v_abort_code) < (0)) __PYX_ERR(0, 642, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":643
 *         self.subindex = subindex
 *         self.abort_code = abort_code
 *         self.desc = desc             # <<<<<<<<<<<<<<
 * 
 * class Emergency(Exception):
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_desc, __pyx_v_desc) < (0)) __PYX_ERR(0, 643, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":638
 *     """
 * 
 *     def __init__(self, slave_pos, index, subindex, abort_code, desc):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.index = index
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.SdoError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":657
 *     """
 * 
 *     def __init__(self, slave_pos, error_code, error_reg, b1, w1, w2):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9Emergency_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9Emergency_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9Emergency_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9Emergency_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_slave_pos = 0;
  PyObject *__pyx_v_error_code = 0;
  PyObject *__pyx_v_error_reg = 0;
  PyObject *__pyx_v_b1 = 0;
  PyObject *__pyx_v_w1 = 0;
  PyObject *__pyx_v_w2 = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_slave_pos,&__pyx_mstate_global->__pyx_n_u_error_code,&__pyx_mstate_global->__pyx_n_u_error_reg,&__pyx_mstate_global->__pyx_n_u_b1,&__pyx_mstate_global->__pyx_n_u_w1,&__pyx_mstate_global->__pyx_n_u_w2,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 657, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 657, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 657, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 657, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 657, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 657, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 657, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 657, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 657, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 7; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, i); __PYX_ERR(0, 657, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 7)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 657, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 657, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 657, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 657, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 657, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 657, __pyx_L3_error)
      values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 657, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_slave_pos = values[1];
    __pyx_v_error_code = values[2];
    __pyx_v_error_reg = values[3];
    __pyx_v_b1 = values[4];
    __pyx_v_w1 = values[5];
    __pyx_v_w2 = values[6];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, __pyx_nargs); __PYX_ERR(0, 657, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.Emergency.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9Emergency___init__(__pyx_self, __pyx_v_self, __pyx_v_slave_pos, __pyx_v_error_code, __pyx_v_error_reg, __pyx_v_b1, __pyx_v_w1, __pyx_v_w2);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9Emergency___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_slave_pos, PyObject *__pyx_v_error_code, PyObject *__pyx_v_error_reg, PyObject *__pyx_v_b1, PyObject *__pyx_v_w1, PyObject *__pyx_v_w2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pysoem/pysoem.pyx":658
 * 
 *     def __init__(self, slave_pos, error_code, error_reg, b1, w1, w2):
 *         self.slave_pos = slave_pos             # <<<<<<<<<<<<<<
 *         self.error_code = error_code
 *         self.error_reg = error_reg
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_slave_pos, __pyx_v_slave_pos) < (0)) __PYX_ERR(0, 658, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":659
 *     def __init__(self, slave_pos, error_code, error_reg, b1, w1, w2):
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code             # <<<<<<<<<<<<<<
 *         self.error_reg = error_reg
 *         self.b1 = b1
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_error_code, __pyx_v_error_code) < (0)) __PYX_ERR(0, 659, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":660
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
 *         self.error_reg = error_reg             # <<<<<<<<<<<<<<
 *         self.b1 = b1
 *         self.w1 = w1
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_error_reg, __pyx_v_error_reg) < (0)) __PYX_ERR(0, 660, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":661
 *         self.error_code = error_code
 *         self.error_reg = error_reg
 *         self.b1 = b1             # <<<<<<<<<<<<<<
 *         self.w1 = w1
 *         self.w2 = w2
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_b1, __pyx_v_b1) < (0)) __PYX_ERR(0, 661, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":662
 *         self.error_reg = error_reg
 *         self.b1 = b1
 *         self.w1 = w1             # <<<<<<<<<<<<<<
 *         self.w2 = w2
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_w1, __pyx_v_w1) < (0)) __PYX_ERR(0, 662, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":663
 *         self.b1 = b1
 *         self.w1 = w1
 *         self.w2 = w2             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_w2, __pyx_v_w2) < (0)) __PYX_ERR(0, 663, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":657
 *     """
 * 
 *     def __init__(self, slave_pos, error_code, error_reg, b1, w1, w2):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.Emergency.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":665
 *         self.w2 = w2
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         b1w1w2_bytes = bytes([self.b1]) + self.w1.to_bytes(length=2, byteorder='little') + self.w2.to_bytes(length=2, byteorder='little')
 *         b1w1w2_str = ','.join(format(x, '02x') for x in b1w1w2_bytes)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9Emergency_3__str__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9Emergency_3__str__ = {"__str__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9Emergency_3__str__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9Emergency_3__str__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 665, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 665, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__str__", 0) < (0)) __PYX_ERR(0, 665, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__str__", 1, 1, 1, i); __PYX_ERR(0, 665, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 665, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__str__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 665, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.Emergency.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9Emergency_2__str__(__pyx_self, __pyx_v_self);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6pysoem_6pysoem_9Emergency_7__str___2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "pysoem/pysoem.pyx":667
 *     def __str__(self):
 *         b1w1w2_bytes = bytes([self.b1]) + self.w1.to_bytes(length=2, byteorder='little') + self.w2.to_bytes(length=2, byteorder='little')
 *         b1w1w2_str = ','.join(format(x, '02x') for x in b1w1w2_bytes)             # <<<<<<<<<<<<<<
 *         return f'Slave {self.slave_pos}:  {self.error_code:04x}, {self.error_reg:02x}, ({b1w1w2_str})'
 * 
*/

static PyObject *__pyx_pf_6pysoem_6pysoem_9Emergency_7__str___genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *)__pyx_tp_new_6pysoem_6pysoem___pyx_scope_struct_1_genexpr(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 667, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6pysoem_6pysoem_9Emergency_7__str___2generator1, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1]), (PyObject *) __pyx_cur_scope, __pyx_mstate_global->__pyx_n_u_genexpr, __pyx_mstate_global->__pyx_n_u_Emergency___str___locals_genexpr, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem); if (unlikely(!gen)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.Emergency.__str__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6pysoem_6pysoem_9Emergency_7__str___2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 667, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 667, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 667, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 667, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_2, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_2;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 667, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2));
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2);
        #endif
        ++__pyx_t_2;
      }
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 667, __pyx_L1_error)
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 667, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_x);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_x, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = PyObject_Format(__pyx_cur_scope->__pyx_v_x, __pyx_mstate_global->__pyx_kp_u_02x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  if (__Pyx_PyErr_Occurred()) {
    __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":665
 *         self.w2 = w2
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         b1w1w2_bytes = bytes([self.b1]) + self.w1.to_bytes(length=2, byteorder='little') + self.w2.to_bytes(length=2, byteorder='little')
 *         b1w1w2_str = ','.join(format(x, '02x') for x in b1w1w2_bytes)
*/

static PyObject *__pyx_pf_6pysoem_6pysoem_9Emergency_2__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_b1w1w2_bytes = NULL;
  PyObject *__pyx_v_b1w1w2_str = NULL;
  PyObject *__pyx_gb_6pysoem_6pysoem_9Emergency_7__str___2generator1 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7[9];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pysoem/pysoem.pyx":666
 * 
 *     def __str__(self):
 *         b1w1w2_bytes = bytes([self.b1]) + self.w1.to_bytes(length=2, byteorder='little') + self.w2.to_bytes(length=2, byteorder='little')             # <<<<<<<<<<<<<<
 *         b1w1w2_str = ','.join(format(x, '02x') for x in b1w1w2_bytes)
 *         return f'Slave {self.slave_pos}:  {self.error_code:04x}, {self.error_reg:02x}, ({b1w1w2_str})'
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_b1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_3) != (0)) __PYX_ERR(0, 666, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_w1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_2, NULL};
    __pyx_t_6 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_length, __pyx_mstate_global->__pyx_int_2, __pyx_t_6, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 666, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_byteorder, __pyx_mstate_global->__pyx_n_u_little, __pyx_t_6, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 666, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_to_bytes, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_w2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __pyx_t_6;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_1, NULL};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_length, __pyx_mstate_global->__pyx_int_2, __pyx_t_2, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 666, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_byteorder, __pyx_mstate_global->__pyx_n_u_little, __pyx_t_2, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 666, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_to_bytes, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_6 = PyNumber_Add(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_b1w1w2_bytes = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "pysoem/pysoem.pyx":667
 *     def __str__(self):
 *         b1w1w2_bytes = bytes([self.b1]) + self.w1.to_bytes(length=2, byteorder='little') + self.w2.to_bytes(length=2, byteorder='little')
 *         b1w1w2_str = ','.join(format(x, '02x') for x in b1w1w2_bytes)             # <<<<<<<<<<<<<<
 *         return f'Slave {self.slave_pos}:  {self.error_code:04x}, {self.error_reg:02x}, ({b1w1w2_str})'
 * 
*/
  __pyx_t_6 = __pyx_pf_6pysoem_6pysoem_9Emergency_7__str___genexpr(NULL, __pyx_v_b1w1w2_bytes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_Generator_GetInlinedResult(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyUnicode_Join(__pyx_mstate_global->__pyx_kp_u_, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_b1w1w2_str = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "pysoem/pysoem.pyx":668
 *         b1w1w2_bytes = bytes([self.b1]) + self.w1.to_bytes(length=2, byteorder='little') + self.w2.to_bytes(length=2, byteorder='little')
 *         b1w1w2_str = ','.join(format(x, '02x') for x in b1w1w2_bytes)
 *         return f'Slave {self.slave_pos}:  {self.error_code:04x}, {self.error_reg:02x}, ({b1w1w2_str})'             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_slave_pos); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_error_code); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_Format(__pyx_t_6, __pyx_mstate_global->__pyx_kp_u_04x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_error_reg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_Format(__pyx_t_6, __pyx_mstate_global->__pyx_kp_u_02x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_v_b1w1w2_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7[0] = __pyx_mstate_global->__pyx_kp_u_Slave;
  __pyx_t_7[1] = __pyx_t_4;
  __pyx_t_7[2] = __pyx_mstate_global->__pyx_kp_u__2;
  __pyx_t_7[3] = __pyx_t_3;
  __pyx_t_7[4] = __pyx_mstate_global->__pyx_kp_u__3;
  __pyx_t_7[5] = __pyx_t_2;
  __pyx_t_7[6] = __pyx_mstate_global->__pyx_kp_u__4;
  __pyx_t_7[7] = __pyx_t_6;
  __pyx_t_7[8] = __pyx_mstate_global->__pyx_kp_u__5;
  __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_7, 9, 6 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 3 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6));
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":665
 *         self.w2 = w2
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         b1w1w2_bytes = bytes([self.b1]) + self.w1.to_bytes(length=2, byteorder='little') + self.w2.to_bytes(length=2, byteorder='little')
 *         b1w1w2_str = ','.join(format(x, '02x') for x in b1w1w2_bytes)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pysoem.pysoem.Emergency.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_b1w1w2_bytes);
  __Pyx_XDECREF(__pyx_v_b1w1w2_str);
  __Pyx_XDECREF(__pyx_gb_6pysoem_6pysoem_9Emergency_7__str___2generator1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":678
 *     """
 * 
 *     def __init__(self, message):             # <<<<<<<<<<<<<<
 *         self.message = message
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12SdoInfoError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_12SdoInfoError_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_12SdoInfoError_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_12SdoInfoError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_message = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_message,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 678, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 678, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 678, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 678, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 678, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 678, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 678, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_message = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 678, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.SdoInfoError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12SdoInfoError___init__(__pyx_self, __pyx_v_self, __pyx_v_message);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12SdoInfoError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_message) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pysoem/pysoem.pyx":679
 * 
 *     def __init__(self, message):
 *         self.message = message             # <<<<<<<<<<<<<<
 * 
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_message, __pyx_v_message) < (0)) __PYX_ERR(0, 679, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":678
 *     """
 * 
 *     def __init__(self, message):             # <<<<<<<<<<<<<<
 *         self.message = message
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.SdoInfoError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":691
 *     """
 * 
 *     def __init__(self, slave_pos, error_code, desc):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_12MailboxError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_12MailboxError_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_12MailboxError_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_12MailboxError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_slave_pos = 0;
  PyObject *__pyx_v_error_code = 0;
  PyObject *__pyx_v_desc = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_slave_pos,&__pyx_mstate_global->__pyx_n_u_error_code,&__pyx_mstate_global->__pyx_n_u_desc,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 691, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 691, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 691, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 691, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 691, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 691, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, i); __PYX_ERR(0, 691, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 691, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 691, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 691, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 691, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_slave_pos = values[1];
    __pyx_v_error_code = values[2];
    __pyx_v_desc = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 691, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.MailboxError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_12MailboxError___init__(__pyx_self, __pyx_v_self, __pyx_v_slave_pos, __pyx_v_error_code, __pyx_v_desc);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_12MailboxError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_slave_pos, PyObject *__pyx_v_error_code, PyObject *__pyx_v_desc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pysoem/pysoem.pyx":692
 * 
 *     def __init__(self, slave_pos, error_code, desc):
 *         self.slave_pos = slave_pos             # <<<<<<<<<<<<<<
 *         self.error_code = error_code
 *         self.desc = desc
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_slave_pos, __pyx_v_slave_pos) < (0)) __PYX_ERR(0, 692, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":693
 *     def __init__(self, slave_pos, error_code, desc):
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code             # <<<<<<<<<<<<<<
 *         self.desc = desc
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_error_code, __pyx_v_error_code) < (0)) __PYX_ERR(0, 693, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":694
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
 *         self.desc = desc             # <<<<<<<<<<<<<<
 * 
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_desc, __pyx_v_desc) < (0)) __PYX_ERR(0, 694, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":691
 *     """
 * 
 *     def __init__(self, slave_pos, error_code, desc):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.MailboxError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":713
 *     }
 * 
 *     def __init__(self, slave_pos, error_code):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_11PacketError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_11PacketError_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_11PacketError_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_11PacketError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_slave_pos = 0;
  PyObject *__pyx_v_error_code = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_slave_pos,&__pyx_mstate_global->__pyx_n_u_error_code,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 713, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 713, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 713, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 713, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 713, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, i); __PYX_ERR(0, 713, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 713, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 713, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 713, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_slave_pos = values[1];
    __pyx_v_error_code = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 713, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.PacketError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_11PacketError___init__(__pyx_self, __pyx_v_self, __pyx_v_slave_pos, __pyx_v_error_code);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_11PacketError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_slave_pos, PyObject *__pyx_v_error_code) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pysoem/pysoem.pyx":714
 * 
 *     def __init__(self, slave_pos, error_code):
 *         self.slave_pos = slave_pos             # <<<<<<<<<<<<<<
 *         self.error_code = error_code
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_slave_pos, __pyx_v_slave_pos) < (0)) __PYX_ERR(0, 714, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":715
 *     def __init__(self, slave_pos, error_code):
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code             # <<<<<<<<<<<<<<
 * 
 *     def _get_desc(self):
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_error_code, __pyx_v_error_code) < (0)) __PYX_ERR(0, 715, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":713
 *     }
 * 
 *     def __init__(self, slave_pos, error_code):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.PacketError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":717
 *         self.error_code = error_code
 * 
 *     def _get_desc(self):             # <<<<<<<<<<<<<<
 *         return self._code_desc[self.error_code]
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_11PacketError_3_get_desc(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_11PacketError_3_get_desc = {"_get_desc", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_11PacketError_3_get_desc, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_11PacketError_3_get_desc(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_desc (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 717, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 717, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_get_desc", 0) < (0)) __PYX_ERR(0, 717, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_get_desc", 1, 1, 1, i); __PYX_ERR(0, 717, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 717, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_desc", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 717, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.PacketError._get_desc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_11PacketError_2_get_desc(__pyx_self, __pyx_v_self);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_11PacketError_2_get_desc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_desc", 0);

  /* "pysoem/pysoem.pyx":718
 * 
 *     def _get_desc(self):
 *         return self._code_desc[self.error_code]             # <<<<<<<<<<<<<<
 * 
 *     desc = property(_get_desc)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_code_desc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_error_code); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":717
 *         self.error_code = error_code
 * 
 *     def _get_desc(self):             # <<<<<<<<<<<<<<
 *         return self._code_desc[self.error_code]
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pysoem.pysoem.PacketError._get_desc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":730
 *     """
 * 
 *     def __init__(self, error_list):             # <<<<<<<<<<<<<<
 *         self.error_list = error_list
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_14ConfigMapError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_14ConfigMapError_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_14ConfigMapError_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_14ConfigMapError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_error_list = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_error_list,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 730, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 730, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 730, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 730, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 730, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 730, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 730, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_error_list = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 730, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.ConfigMapError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_14ConfigMapError___init__(__pyx_self, __pyx_v_self, __pyx_v_error_list);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_14ConfigMapError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_error_list) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pysoem/pysoem.pyx":731
 * 
 *     def __init__(self, error_list):
 *         self.error_list = error_list             # <<<<<<<<<<<<<<
 * 
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_error_list, __pyx_v_error_list) < (0)) __PYX_ERR(0, 731, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":730
 *     """
 * 
 *     def __init__(self, error_list):             # <<<<<<<<<<<<<<
 *         self.error_list = error_list
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.ConfigMapError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":741
 *     """
 * 
 *     def __init__(self, message):             # <<<<<<<<<<<<<<
 *         self.message = message
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_11EepromError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_11EepromError_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_11EepromError_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_11EepromError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_message = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_message,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 741, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 741, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 741, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 741, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 741, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 741, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 741, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_message = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 741, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.EepromError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_11EepromError___init__(__pyx_self, __pyx_v_self, __pyx_v_message);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_11EepromError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_message) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pysoem/pysoem.pyx":742
 * 
 *     def __init__(self, message):
 *         self.message = message             # <<<<<<<<<<<<<<
 * 
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_message, __pyx_v_message) < (0)) __PYX_ERR(0, 742, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":741
 *     """
 * 
 *     def __init__(self, message):             # <<<<<<<<<<<<<<
 *         self.message = message
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.EepromError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":753
 *     """
 * 
 *     def __init__(self, message=None, wkc=None):             # <<<<<<<<<<<<<<
 *         self.message = message
 *         self.wkc = wkc
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_8WkcError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_8WkcError_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_8WkcError_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_8WkcError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_message = 0;
  PyObject *__pyx_v_wkc = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_message,&__pyx_mstate_global->__pyx_n_u_wkc,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 753, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 753, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 753, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 753, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 753, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, i); __PYX_ERR(0, 753, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 753, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 753, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 753, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_self = values[0];
    __pyx_v_message = values[1];
    __pyx_v_wkc = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 753, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.WkcError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_8WkcError___init__(__pyx_self, __pyx_v_self, __pyx_v_message, __pyx_v_wkc);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_8WkcError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_wkc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pysoem/pysoem.pyx":754
 * 
 *     def __init__(self, message=None, wkc=None):
 *         self.message = message             # <<<<<<<<<<<<<<
 *         self.wkc = wkc
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_message, __pyx_v_message) < (0)) __PYX_ERR(0, 754, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":755
 *     def __init__(self, message=None, wkc=None):
 *         self.message = message
 *         self.wkc = wkc             # <<<<<<<<<<<<<<
 * 
 * class NetworkInterfaceNotOpenError(Exception):
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_wkc, __pyx_v_wkc) < (0)) __PYX_ERR(0, 755, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":753
 *     """
 * 
 *     def __init__(self, message=None, wkc=None):             # <<<<<<<<<<<<<<
 *         self.message = message
 *         self.wkc = wkc
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.WkcError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13_CallbackData_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13_CallbackData_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13_CallbackData_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13_CallbackData_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13_CallbackData___reduce_cython__(((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13_CallbackData___reduce_cython__(struct __pyx_obj_6pysoem_6pysoem__CallbackData *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.exc_info, self.exc_raised, self.func, self.slave)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
*/
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->exc_info);
  __Pyx_GIVEREF(__pyx_v_self->exc_info);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->exc_info) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->exc_raised);
  __Pyx_GIVEREF(__pyx_v_self->exc_raised);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->exc_raised) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->func);
  __Pyx_GIVEREF(__pyx_v_self->func);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->func) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->slave);
  __Pyx_GIVEREF(__pyx_v_self->slave);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->slave) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.exc_info, self.exc_raised, self.func, self.slave)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None and _dict:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.exc_info, self.exc_raised, self.func, self.slave)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v__dict); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None and _dict:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.exc_info is not None or self.exc_raised is not None or self.func is not None or self.slave is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.exc_info, self.exc_raised, self.func, self.slave)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None and _dict:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.exc_info is not None or self.exc_raised is not None or self.func is not None or self.slave is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, None), state
*/
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->exc_info != Py_None);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_3 = (__pyx_v_self->exc_raised != Py_None);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_3 = (__pyx_v_self->func != Py_None);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_3 = (__pyx_v_self->slave != Py_None);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L6_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.exc_info is not None or self.exc_raised is not None or self.func is not None or self.slave is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.exc_info is not None or self.exc_raised is not None or self.func is not None or self.slave is not None
 *     if use_setstate:
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__CallbackData); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_9788398);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_9788398);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_9788398) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.exc_info is not None or self.exc_raised is not None or self.func is not None or self.slave is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, None), state
 *     else:
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__CallbackData__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__CallbackData); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_9788398);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_9788398);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_9788398) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pysoem.pysoem._CallbackData.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__CallbackData__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13_CallbackData_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13_CallbackData_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13_CallbackData_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13_CallbackData_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem._CallbackData.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13_CallbackData_2__setstate_cython__(((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13_CallbackData_2__setstate_cython__(struct __pyx_obj_6pysoem_6pysoem__CallbackData *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__CallbackData__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __pyx_v___pyx_state;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 17, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
    __PYX_ERR(1, 17, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_6pysoem_6pysoem___pyx_unpickle__CallbackData__set_state(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__CallbackData__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem._CallbackData.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":817
 *     od = property(_get_od)
 * 
 *     def __init__(self, pos):             # <<<<<<<<<<<<<<
 *         self._pos = pos
 *         self._cd = _CallbackData()
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_9CdefSlave_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_9CdefSlave_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_pos = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pos,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 817, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 817, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 817, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 817, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 817, __pyx_L3_error)
    }
    __pyx_v_pos = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 817, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave___init__(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_pos);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_9CdefSlave___init__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_pos) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pysoem/pysoem.pyx":818
 * 
 *     def __init__(self, pos):
 *         self._pos = pos             # <<<<<<<<<<<<<<
 *         self._cd = _CallbackData()
 *         self._cd.slave = self
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_pos); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 818, __pyx_L1_error)
  __pyx_v_self->_pos = __pyx_t_1;

  /* "pysoem/pysoem.pyx":819
 *     def __init__(self, pos):
 *         self._pos = pos
 *         self._cd = _CallbackData()             # <<<<<<<<<<<<<<
 *         self._cd.slave = self
 *         self._emcy_callbacks = []
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem__CallbackData, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __Pyx_GIVEREF((PyObject *)__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_cd);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_cd);
  __pyx_v_self->_cd = ((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":820
 *         self._pos = pos
 *         self._cd = _CallbackData()
 *         self._cd.slave = self             # <<<<<<<<<<<<<<
 *         self._emcy_callbacks = []
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  __Pyx_GOTREF(__pyx_v_self->_cd->slave);
  __Pyx_DECREF(__pyx_v_self->_cd->slave);
  __pyx_v_self->_cd->slave = ((PyObject *)__pyx_v_self);

  /* "pysoem/pysoem.pyx":821
 *         self._cd = _CallbackData()
 *         self._cd.slave = self
 *         self._emcy_callbacks = []             # <<<<<<<<<<<<<<
 * 
 *     def dc_sync(self, act, sync0_cycle_time, sync0_shift_time=0, sync1_cycle_time=None):
*/
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_emcy_callbacks);
  __Pyx_DECREF(__pyx_v_self->_emcy_callbacks);
  __pyx_v_self->_emcy_callbacks = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":817
 *     od = property(_get_od)
 * 
 *     def __init__(self, pos):             # <<<<<<<<<<<<<<
 *         self._pos = pos
 *         self._cd = _CallbackData()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":823
 *         self._emcy_callbacks = []
 * 
 *     def dc_sync(self, act, sync0_cycle_time, sync0_shift_time=0, sync1_cycle_time=None):             # <<<<<<<<<<<<<<
 *         """Activate or deactivate SYNC pulses at the slave.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_3dc_sync(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_2dc_sync, "Activate or deactivate SYNC pulses at the slave.\n\n         Args:\n            act (bool): True = active, False = deactivate\n            sync0_cycle_time (int): Cycltime SYNC0 in ns\n            sync0_shift_time (int): Optional SYNC0 shift time in ns\n            sync1_cycle_time (int): Optional cycltime for SYNC1 in ns. This time is a delta time in relation to SYNC0.\n                                    If CylcTime1 = 0 then SYNC1 fires at the same time as SYNC0.\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_3dc_sync = {"dc_sync", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_3dc_sync, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_2dc_sync};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_3dc_sync(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_act = 0;
  PyObject *__pyx_v_sync0_cycle_time = 0;
  PyObject *__pyx_v_sync0_shift_time = 0;
  PyObject *__pyx_v_sync1_cycle_time = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dc_sync (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_act,&__pyx_mstate_global->__pyx_n_u_sync0_cycle_time,&__pyx_mstate_global->__pyx_n_u_sync0_shift_time,&__pyx_mstate_global->__pyx_n_u_sync1_cycle_time,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 823, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 823, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 823, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 823, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 823, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "dc_sync", 0) < (0)) __PYX_ERR(0, 823, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("dc_sync", 0, 2, 4, i); __PYX_ERR(0, 823, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 823, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 823, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 823, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 823, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_act = values[0];
    __pyx_v_sync0_cycle_time = values[1];
    __pyx_v_sync0_shift_time = values[2];
    __pyx_v_sync1_cycle_time = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dc_sync", 0, 2, 4, __pyx_nargs); __PYX_ERR(0, 823, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.dc_sync", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_2dc_sync(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_act, __pyx_v_sync0_cycle_time, __pyx_v_sync0_shift_time, __pyx_v_sync1_cycle_time);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_2dc_sync(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_act, PyObject *__pyx_v_sync0_cycle_time, PyObject *__pyx_v_sync0_shift_time, PyObject *__pyx_v_sync1_cycle_time) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  boolean __pyx_t_5;
  uint32 __pyx_t_6;
  int32 __pyx_t_7;
  uint32 __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dc_sync", 0);

  /* "pysoem/pysoem.pyx":833
 *                                     If CylcTime1 = 0 then SYNC1 fires at the same time as SYNC0.
 *         """
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 * 
 *         if sync1_cycle_time is None:
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":835
 *         self._master.check_context_is_initialized()
 * 
 *         if sync1_cycle_time is None:             # <<<<<<<<<<<<<<
 *             cpysoem.ecx_dcsync0(self._ecx_contextt, self._pos, act, sync0_cycle_time, sync0_shift_time)
 *         else:
*/
  __pyx_t_4 = (__pyx_v_sync1_cycle_time == Py_None);
  if (__pyx_t_4) {

    /* "pysoem/pysoem.pyx":836
 * 
 *         if sync1_cycle_time is None:
 *             cpysoem.ecx_dcsync0(self._ecx_contextt, self._pos, act, sync0_cycle_time, sync0_shift_time)             # <<<<<<<<<<<<<<
 *         else:
 *             cpysoem.ecx_dcsync01(self._ecx_contextt, self._pos, act, sync0_cycle_time, sync1_cycle_time, sync0_shift_time)
*/
    __pyx_t_5 = __Pyx_PyLong_As_int8_t(__pyx_v_act); if (unlikely((__pyx_t_5 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(0, 836, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyLong_As_uint32_t(__pyx_v_sync0_cycle_time); if (unlikely((__pyx_t_6 == ((uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 836, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyLong_As_int32_t(__pyx_v_sync0_shift_time); if (unlikely((__pyx_t_7 == ((int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 836, __pyx_L1_error)
    ecx_dcsync0(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_t_5, __pyx_t_6, __pyx_t_7);

    /* "pysoem/pysoem.pyx":835
 *         self._master.check_context_is_initialized()
 * 
 *         if sync1_cycle_time is None:             # <<<<<<<<<<<<<<
 *             cpysoem.ecx_dcsync0(self._ecx_contextt, self._pos, act, sync0_cycle_time, sync0_shift_time)
 *         else:
*/
    goto __pyx_L3;
  }

  /* "pysoem/pysoem.pyx":838
 *             cpysoem.ecx_dcsync0(self._ecx_contextt, self._pos, act, sync0_cycle_time, sync0_shift_time)
 *         else:
 *             cpysoem.ecx_dcsync01(self._ecx_contextt, self._pos, act, sync0_cycle_time, sync1_cycle_time, sync0_shift_time)             # <<<<<<<<<<<<<<
 * 
 *     cdef int __sdo_read_nogil(self, uint16_t index, uint8_t subindex, int8_t ca, int* size_inout, unsigned char* pbuf):
*/
  /*else*/ {
    __pyx_t_5 = __Pyx_PyLong_As_int8_t(__pyx_v_act); if (unlikely((__pyx_t_5 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyLong_As_uint32_t(__pyx_v_sync0_cycle_time); if (unlikely((__pyx_t_6 == ((uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyLong_As_uint32_t(__pyx_v_sync1_cycle_time); if (unlikely((__pyx_t_8 == ((uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyLong_As_int32_t(__pyx_v_sync0_shift_time); if (unlikely((__pyx_t_7 == ((int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L1_error)
    ecx_dcsync01(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_t_5, __pyx_t_6, __pyx_t_8, __pyx_t_7);
  }
  __pyx_L3:;

  /* "pysoem/pysoem.pyx":823
 *         self._emcy_callbacks = []
 * 
 *     def dc_sync(self, act, sync0_cycle_time, sync0_shift_time=0, sync1_cycle_time=None):             # <<<<<<<<<<<<<<
 *         """Activate or deactivate SYNC pulses at the slave.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.dc_sync", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":840
 *             cpysoem.ecx_dcsync01(self._ecx_contextt, self._pos, act, sync0_cycle_time, sync1_cycle_time, sync0_shift_time)
 * 
 *     cdef int __sdo_read_nogil(self, uint16_t index, uint8_t subindex, int8_t ca, int* size_inout, unsigned char* pbuf):             # <<<<<<<<<<<<<<
 *         """Read a CoE object without GIL.
 * 
*/

static int __pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__sdo_read_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, uint16_t __pyx_v_index, uint8_t __pyx_v_subindex, int8_t __pyx_v_ca, int *__pyx_v_size_inout, unsigned char *__pyx_v_pbuf) {
  int __pyx_v_result;
  int __pyx_r;

  /* "pysoem/pysoem.pyx":852
 *         cdef int result
 * 
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             result = cpysoem.ecx_SDOread(self._ecx_contextt, self._pos, index, subindex, ca, size_inout, pbuf, self._the_masters_settings.sdo_read_timeout[0])
*/
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":853
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_SDOread(self._ecx_contextt, self._pos, index, subindex, ca, size_inout, pbuf, self._the_masters_settings.sdo_read_timeout[0])
 *         Py_DECREF(self)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pysoem/pysoem.pyx":854
 *         Py_INCREF(self)
 *         with nogil:
 *             result = cpysoem.ecx_SDOread(self._ecx_contextt, self._pos, index, subindex, ca, size_inout, pbuf, self._the_masters_settings.sdo_read_timeout[0])             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
*/
        __pyx_v_result = ecx_SDOread(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_v_index, __pyx_v_subindex, __pyx_v_ca, __pyx_v_size_inout, __pyx_v_pbuf, (__pyx_v_self->_the_masters_settings->sdo_read_timeout[0]));
      }

      /* "pysoem/pysoem.pyx":853
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_SDOread(self._ecx_contextt, self._pos, index, subindex, ca, size_inout, pbuf, self._the_masters_settings.sdo_read_timeout[0])
 *         Py_DECREF(self)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pysoem/pysoem.pyx":855
 *         with nogil:
 *             result = cpysoem.ecx_SDOread(self._ecx_contextt, self._pos, index, subindex, ca, size_inout, pbuf, self._the_masters_settings.sdo_read_timeout[0])
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *         return result
*/
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":857
 *         Py_DECREF(self)
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def sdo_read(self, index, uint8_t subindex, int size=0, ca=False, *, release_gil=None):
*/
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":840
 *             cpysoem.ecx_dcsync01(self._ecx_contextt, self._pos, act, sync0_cycle_time, sync1_cycle_time, sync0_shift_time)
 * 
 *     cdef int __sdo_read_nogil(self, uint16_t index, uint8_t subindex, int8_t ca, int* size_inout, unsigned char* pbuf):             # <<<<<<<<<<<<<<
 *         """Read a CoE object without GIL.
 * 
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":859
 *         return result
 * 
 *     def sdo_read(self, index, uint8_t subindex, int size=0, ca=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Read a CoE object.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_5sdo_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_4sdo_read, "Read a CoE object.\n\n        When leaving out the size parameter, objects up to 256 bytes can be read.\n        If the size of the object is expected to be bigger, increase the size parameter.\n\n        Args:\n            index (int): Index of the object.\n            subindex (int): Subindex of the object.\n            size (:obj:`int`, optional): The size of the reading buffer.\n            ca (:obj:`bool`, optional): complete access.\n            release_gil (:obj:`bool`, optional): True to read a CoE object releasing the GIL. Defaults to False.\n\n        Returns:\n            bytes: The content of the sdo object.\n\n        Raises:\n            SdoError: if write fails, the exception includes the SDO abort code  \n            MailboxError: on errors in the mailbox protocol\n            PacketError: on packet level error\n            WkcError: if working counter is not higher than 0, the exception includes the working counter\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_5sdo_read = {"sdo_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_5sdo_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_4sdo_read};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_5sdo_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_index = 0;
  uint8_t __pyx_v_subindex;
  int __pyx_v_size;
  PyObject *__pyx_v_ca = 0;
  PyObject *__pyx_v_release_gil = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sdo_read (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_index,&__pyx_mstate_global->__pyx_n_u_subindex,&__pyx_mstate_global->__pyx_n_u_size,&__pyx_mstate_global->__pyx_n_u_ca,&__pyx_mstate_global->__pyx_n_u_release_gil,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 859, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 859, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 859, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 859, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 859, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "sdo_read", 0) < (0)) __PYX_ERR(0, 859, __pyx_L3_error)
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("sdo_read", 0, 2, 4, i); __PYX_ERR(0, 859, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 859, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 859, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 859, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 859, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_index = values[0];
    __pyx_v_subindex = __Pyx_PyLong_As_uint8_t(values[1]); if (unlikely((__pyx_v_subindex == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 859, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_size = __Pyx_PyLong_As_int(values[2]); if (unlikely((__pyx_v_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 859, __pyx_L3_error)
    } else {
      __pyx_v_size = ((int)0);
    }
    __pyx_v_ca = values[3];
    __pyx_v_release_gil = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sdo_read", 0, 2, 4, __pyx_nargs); __PYX_ERR(0, 859, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.sdo_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_4sdo_read(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_index, __pyx_v_subindex, __pyx_v_size, __pyx_v_ca, __pyx_v_release_gil);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_4sdo_read(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_index, uint8_t __pyx_v_subindex, int __pyx_v_size, PyObject *__pyx_v_ca, PyObject *__pyx_v_release_gil) {
  unsigned char *__pyx_v_pbuf;
  uint8_t __pyx_v_std_buffer[__pyx_e_6pysoem_6pysoem_STATIC_SDO_READ_BUFFER_SIZE];
  int __pyx_v_size_inout;
  int __pyx_v_result;
  ec_errort __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  boolean __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  uint16_t __pyx_t_6;
  int8_t __pyx_t_7;
  int __pyx_t_8;
  uint16 __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sdo_read", 0);
  __Pyx_INCREF(__pyx_v_release_gil);

  /* "pysoem/pysoem.pyx":881
 *             WkcError: if working counter is not higher than 0, the exception includes the working counter
 *         """
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)             # <<<<<<<<<<<<<<
 *         if self._ecx_contextt == NULL:
 *             raise UnboundLocalError()
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->_master->__pyx_vtab)->check_release_gil(__pyx_v_self->_master, __pyx_v_release_gil, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 881, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_release_gil, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":882
 *         """
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)
 *         if self._ecx_contextt == NULL:             # <<<<<<<<<<<<<<
 *             raise UnboundLocalError()
 * 
*/
  __pyx_t_3 = (__pyx_v_self->_ecx_contextt == NULL);
  if (unlikely(__pyx_t_3)) {

    /* "pysoem/pysoem.pyx":883
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)
 *         if self._ecx_contextt == NULL:
 *             raise UnboundLocalError()             # <<<<<<<<<<<<<<
 * 
 *         self._master.check_context_is_initialized()
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_UnboundLocalError)), __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 883, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":882
 *         """
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)
 *         if self._ecx_contextt == NULL:             # <<<<<<<<<<<<<<
 *             raise UnboundLocalError()
 * 
*/
  }

  /* "pysoem/pysoem.pyx":885
 *             raise UnboundLocalError()
 * 
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 * 
 *         cdef unsigned char* pbuf
*/
  __pyx_t_4 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":890
 *         cdef uint8_t std_buffer[STATIC_SDO_READ_BUFFER_SIZE]
 *         cdef int size_inout
 *         if size == 0:             # <<<<<<<<<<<<<<
 *             pbuf = std_buffer
 *             size_inout = STATIC_SDO_READ_BUFFER_SIZE
*/
  __pyx_t_3 = (__pyx_v_size == 0);
  if (__pyx_t_3) {

    /* "pysoem/pysoem.pyx":891
 *         cdef int size_inout
 *         if size == 0:
 *             pbuf = std_buffer             # <<<<<<<<<<<<<<
 *             size_inout = STATIC_SDO_READ_BUFFER_SIZE
 *         else:
*/
    __pyx_v_pbuf = __pyx_v_std_buffer;

    /* "pysoem/pysoem.pyx":892
 *         if size == 0:
 *             pbuf = std_buffer
 *             size_inout = STATIC_SDO_READ_BUFFER_SIZE             # <<<<<<<<<<<<<<
 *         else:
 *             pbuf = <unsigned char*>PyMem_Malloc((size)*sizeof(unsigned char))
*/
    __pyx_v_size_inout = __pyx_e_6pysoem_6pysoem_STATIC_SDO_READ_BUFFER_SIZE;

    /* "pysoem/pysoem.pyx":890
 *         cdef uint8_t std_buffer[STATIC_SDO_READ_BUFFER_SIZE]
 *         cdef int size_inout
 *         if size == 0:             # <<<<<<<<<<<<<<
 *             pbuf = std_buffer
 *             size_inout = STATIC_SDO_READ_BUFFER_SIZE
*/
    goto __pyx_L4;
  }

  /* "pysoem/pysoem.pyx":894
 *             size_inout = STATIC_SDO_READ_BUFFER_SIZE
 *         else:
 *             pbuf = <unsigned char*>PyMem_Malloc((size)*sizeof(unsigned char))             # <<<<<<<<<<<<<<
 *             size_inout = size
 * 
*/
  /*else*/ {
    __pyx_v_pbuf = ((unsigned char *)PyMem_Malloc((__pyx_v_size * (sizeof(unsigned char)))));

    /* "pysoem/pysoem.pyx":895
 *         else:
 *             pbuf = <unsigned char*>PyMem_Malloc((size)*sizeof(unsigned char))
 *             size_inout = size             # <<<<<<<<<<<<<<
 * 
 *         if pbuf == NULL:
*/
    __pyx_v_size_inout = __pyx_v_size;
  }
  __pyx_L4:;

  /* "pysoem/pysoem.pyx":897
 *             size_inout = size
 * 
 *         if pbuf == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
*/
  __pyx_t_3 = (__pyx_v_pbuf == NULL);
  if (unlikely(__pyx_t_3)) {

    /* "pysoem/pysoem.pyx":898
 * 
 *         if pbuf == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         cdef int result
*/
    PyErr_NoMemory(); __PYX_ERR(0, 898, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":897
 *             size_inout = size
 * 
 *         if pbuf == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
*/
  }

  /* "pysoem/pysoem.pyx":901
 * 
 *         cdef int result
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             result = self.__sdo_read_nogil(index, subindex, ca, &size_inout, pbuf)
 *         else:
*/
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_release_gil); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 901, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "pysoem/pysoem.pyx":902
 *         cdef int result
 *         if release_gil:
 *             result = self.__sdo_read_nogil(index, subindex, ca, &size_inout, pbuf)             # <<<<<<<<<<<<<<
 *         else:
 *             result = cpysoem.ecx_SDOread(self._ecx_contextt, self._pos, index, subindex, ca,
*/
    __pyx_t_6 = __Pyx_PyLong_As_uint16_t(__pyx_v_index); if (unlikely((__pyx_t_6 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 902, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyLong_As_int8_t(__pyx_v_ca); if (unlikely((__pyx_t_7 == ((int8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 902, __pyx_L1_error)
    __pyx_t_8 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_CdefSlave__sdo_read_nogil(__pyx_v_self, __pyx_t_6, __pyx_v_subindex, __pyx_t_7, (&__pyx_v_size_inout), __pyx_v_pbuf); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 902, __pyx_L1_error)
    __pyx_v_result = __pyx_t_8;

    /* "pysoem/pysoem.pyx":901
 * 
 *         cdef int result
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             result = self.__sdo_read_nogil(index, subindex, ca, &size_inout, pbuf)
 *         else:
*/
    goto __pyx_L6;
  }

  /* "pysoem/pysoem.pyx":904
 *             result = self.__sdo_read_nogil(index, subindex, ca, &size_inout, pbuf)
 *         else:
 *             result = cpysoem.ecx_SDOread(self._ecx_contextt, self._pos, index, subindex, ca,             # <<<<<<<<<<<<<<
 *                                               &size_inout, pbuf, self._the_masters_settings.sdo_read_timeout[0])
 * 
*/
  /*else*/ {
    __pyx_t_9 = __Pyx_PyLong_As_uint16_t(__pyx_v_index); if (unlikely((__pyx_t_9 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 904, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyLong_As_int8_t(__pyx_v_ca); if (unlikely((__pyx_t_1 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(0, 904, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":905
 *         else:
 *             result = cpysoem.ecx_SDOread(self._ecx_contextt, self._pos, index, subindex, ca,
 *                                               &size_inout, pbuf, self._the_masters_settings.sdo_read_timeout[0])             # <<<<<<<<<<<<<<
 * 
 *         cdef cpysoem.ec_errort err
*/
    __pyx_v_result = ecx_SDOread(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_t_9, __pyx_v_subindex, __pyx_t_1, (&__pyx_v_size_inout), __pyx_v_pbuf, (__pyx_v_self->_the_masters_settings->sdo_read_timeout[0]));
  }
  __pyx_L6:;

  /* "pysoem/pysoem.pyx":908
 * 
 *         cdef cpysoem.ec_errort err
 *         while cpysoem.ecx_poperror(self._ecx_contextt, &err):             # <<<<<<<<<<<<<<
 *             assert err.Slave == self._pos
 * 
*/
  while (1) {
    __pyx_t_3 = (ecx_poperror(__pyx_v_self->_ecx_contextt, (&__pyx_v_err)) != 0);
    if (!__pyx_t_3) break;

    /* "pysoem/pysoem.pyx":909
 *         cdef cpysoem.ec_errort err
 *         while cpysoem.ecx_poperror(self._ecx_contextt, &err):
 *             assert err.Slave == self._pos             # <<<<<<<<<<<<<<
 * 
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):
*/
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_3 = (__pyx_v_err.Slave == __pyx_v_self->_pos);
      if (unlikely(!__pyx_t_3)) {
        __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), 0, 0, 0);
        __PYX_ERR(0, 909, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 909, __pyx_L1_error)
    #endif

    /* "pysoem/pysoem.pyx":911
 *             assert err.Slave == self._pos
 * 
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):             # <<<<<<<<<<<<<<
 *                 self._on_emergency(&err)
 *             else:
*/
    __pyx_t_10 = (__pyx_v_err.Etype == EC_ERR_TYPE_EMERGENCY);
    if (__pyx_t_10) {
    } else {
      __pyx_t_3 = __pyx_t_10;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_2 = __pyx_v_self->_emcy_callbacks;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_11 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = (__pyx_t_11 > 0);
    __pyx_t_3 = __pyx_t_10;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_3) {

      /* "pysoem/pysoem.pyx":912
 * 
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):
 *                 self._on_emergency(&err)             # <<<<<<<<<<<<<<
 *             else:
 *                 if pbuf != std_buffer:
*/
      __pyx_t_2 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_on_emergency(__pyx_v_self, (&__pyx_v_err)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pysoem/pysoem.pyx":911
 *             assert err.Slave == self._pos
 * 
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):             # <<<<<<<<<<<<<<
 *                 self._on_emergency(&err)
 *             else:
*/
      goto __pyx_L9;
    }

    /* "pysoem/pysoem.pyx":914
 *                 self._on_emergency(&err)
 *             else:
 *                 if pbuf != std_buffer:             # <<<<<<<<<<<<<<
 *                     PyMem_Free(pbuf)
 *                 self._raise_exception(&err)
*/
    /*else*/ {
      __pyx_t_3 = (__pyx_v_pbuf != __pyx_v_std_buffer);
      if (__pyx_t_3) {

        /* "pysoem/pysoem.pyx":915
 *             else:
 *                 if pbuf != std_buffer:
 *                     PyMem_Free(pbuf)             # <<<<<<<<<<<<<<
 *                 self._raise_exception(&err)
 * 
*/
        PyMem_Free(__pyx_v_pbuf);

        /* "pysoem/pysoem.pyx":914
 *                 self._on_emergency(&err)
 *             else:
 *                 if pbuf != std_buffer:             # <<<<<<<<<<<<<<
 *                     PyMem_Free(pbuf)
 *                 self._raise_exception(&err)
*/
      }

      /* "pysoem/pysoem.pyx":916
 *                 if pbuf != std_buffer:
 *                     PyMem_Free(pbuf)
 *                 self._raise_exception(&err)             # <<<<<<<<<<<<<<
 * 
 *         if not result > 0:
*/
      __pyx_t_2 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_raise_exception(__pyx_v_self, (&__pyx_v_err)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 916, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L9:;
  }

  /* "pysoem/pysoem.pyx":918
 *                 self._raise_exception(&err)
 * 
 *         if not result > 0:             # <<<<<<<<<<<<<<
 *             if pbuf != std_buffer:
 *                     PyMem_Free(pbuf)
*/
  __pyx_t_3 = (!(__pyx_v_result > 0));
  if (__pyx_t_3) {

    /* "pysoem/pysoem.pyx":919
 * 
 *         if not result > 0:
 *             if pbuf != std_buffer:             # <<<<<<<<<<<<<<
 *                     PyMem_Free(pbuf)
 *             raise WkcError(wkc=result)
*/
    __pyx_t_3 = (__pyx_v_pbuf != __pyx_v_std_buffer);
    if (__pyx_t_3) {

      /* "pysoem/pysoem.pyx":920
 *         if not result > 0:
 *             if pbuf != std_buffer:
 *                     PyMem_Free(pbuf)             # <<<<<<<<<<<<<<
 *             raise WkcError(wkc=result)
 * 
*/
      PyMem_Free(__pyx_v_pbuf);

      /* "pysoem/pysoem.pyx":919
 * 
 *         if not result > 0:
 *             if pbuf != std_buffer:             # <<<<<<<<<<<<<<
 *                     PyMem_Free(pbuf)
 *             raise WkcError(wkc=result)
*/
    }

    /* "pysoem/pysoem.pyx":921
 *             if pbuf != std_buffer:
 *                     PyMem_Free(pbuf)
 *             raise WkcError(wkc=result)             # <<<<<<<<<<<<<<
 * 
 *         try:
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_WkcError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 921, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = __Pyx_PyLong_From_int(__pyx_v_result); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 921, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_12);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_12, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, NULL};
      __pyx_t_14 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 921, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_wkc, __pyx_t_13, __pyx_t_14, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 921, __pyx_L1_error)
      __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_14);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 921, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 921, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":918
 *                 self._raise_exception(&err)
 * 
 *         if not result > 0:             # <<<<<<<<<<<<<<
 *             if pbuf != std_buffer:
 *                     PyMem_Free(pbuf)
*/
  }

  /* "pysoem/pysoem.pyx":923
 *             raise WkcError(wkc=result)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             return PyBytes_FromStringAndSize(<char*>pbuf, size_inout)
 *         finally:
*/
  /*try:*/ {

    /* "pysoem/pysoem.pyx":924
 * 
 *         try:
 *             return PyBytes_FromStringAndSize(<char*>pbuf, size_inout)             # <<<<<<<<<<<<<<
 *         finally:
 *             if pbuf != std_buffer:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyBytes_FromStringAndSize(((char *)__pyx_v_pbuf), __pyx_v_size_inout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 924, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L15_return;
  }

  /* "pysoem/pysoem.pyx":926
 *             return PyBytes_FromStringAndSize(<char*>pbuf, size_inout)
 *         finally:
 *             if pbuf != std_buffer:             # <<<<<<<<<<<<<<
 *                 PyMem_Free(pbuf)
 * 
*/
  /*finally:*/ {
    __pyx_L16_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
      if ( unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_22);
      __pyx_t_8 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
      {
        __pyx_t_3 = (__pyx_v_pbuf != __pyx_v_std_buffer);
        if (__pyx_t_3) {

          /* "pysoem/pysoem.pyx":927
 *         finally:
 *             if pbuf != std_buffer:
 *                 PyMem_Free(pbuf)             # <<<<<<<<<<<<<<
 * 
 *     cdef int __sdo_write_nogil(self, uint16_t index, uint8_t subindex, int8_t ca, int size, bytes data):
*/
          PyMem_Free(__pyx_v_pbuf);

          /* "pysoem/pysoem.pyx":926
 *             return PyBytes_FromStringAndSize(<char*>pbuf, size_inout)
 *         finally:
 *             if pbuf != std_buffer:             # <<<<<<<<<<<<<<
 *                 PyMem_Free(pbuf)
 * 
*/
        }
      }
      __Pyx_XGIVEREF(__pyx_t_20);
      __Pyx_XGIVEREF(__pyx_t_21);
      __Pyx_XGIVEREF(__pyx_t_22);
      __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
      __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
      goto __pyx_L1_error;
    }
    __pyx_L15_return: {
      __pyx_t_22 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_3 = (__pyx_v_pbuf != __pyx_v_std_buffer);
      if (__pyx_t_3) {

        /* "pysoem/pysoem.pyx":927
 *         finally:
 *             if pbuf != std_buffer:
 *                 PyMem_Free(pbuf)             # <<<<<<<<<<<<<<
 * 
 *     cdef int __sdo_write_nogil(self, uint16_t index, uint8_t subindex, int8_t ca, int size, bytes data):
*/
        PyMem_Free(__pyx_v_pbuf);

        /* "pysoem/pysoem.pyx":926
 *             return PyBytes_FromStringAndSize(<char*>pbuf, size_inout)
 *         finally:
 *             if pbuf != std_buffer:             # <<<<<<<<<<<<<<
 *                 PyMem_Free(pbuf)
 * 
*/
      }
      __pyx_r = __pyx_t_22;
      __pyx_t_22 = 0;
      goto __pyx_L0;
    }
  }

  /* "pysoem/pysoem.pyx":859
 *         return result
 * 
 *     def sdo_read(self, index, uint8_t subindex, int size=0, ca=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Read a CoE object.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.sdo_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_release_gil);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":929
 *                 PyMem_Free(pbuf)
 * 
 *     cdef int __sdo_write_nogil(self, uint16_t index, uint8_t subindex, int8_t ca, int size, bytes data):             # <<<<<<<<<<<<<<
 *         """Write to a CoE object without GIL.
 * 
*/

static int __pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__sdo_write_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, uint16_t __pyx_v_index, uint8_t __pyx_v_subindex, int8_t __pyx_v_ca, int __pyx_v_size, PyObject *__pyx_v_data) {
  int __pyx_v_result;
  unsigned char *__pyx_v_c_data;
  int __pyx_r;
  unsigned char *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pysoem/pysoem.pyx":940
 *         """
 *         cdef int result
 *         cdef unsigned char* c_data = <unsigned char*> data             # <<<<<<<<<<<<<<
 * 
 *         Py_INCREF(self)
*/
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 940, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBytes_AsWritableUString(__pyx_v_data); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 940, __pyx_L1_error)
  __pyx_v_c_data = ((unsigned char *)__pyx_t_1);

  /* "pysoem/pysoem.pyx":942
 *         cdef unsigned char* c_data = <unsigned char*> data
 * 
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             result = cpysoem.ecx_SDOwrite(self._ecx_contextt, self._pos, index, subindex, ca, size, c_data, self._the_masters_settings.sdo_write_timeout[0])
*/
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":943
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_SDOwrite(self._ecx_contextt, self._pos, index, subindex, ca, size, c_data, self._the_masters_settings.sdo_write_timeout[0])
 *         Py_DECREF(self)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pysoem/pysoem.pyx":944
 *         Py_INCREF(self)
 *         with nogil:
 *             result = cpysoem.ecx_SDOwrite(self._ecx_contextt, self._pos, index, subindex, ca, size, c_data, self._the_masters_settings.sdo_write_timeout[0])             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
*/
        __pyx_v_result = ecx_SDOwrite(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_v_index, __pyx_v_subindex, __pyx_v_ca, __pyx_v_size, __pyx_v_c_data, (__pyx_v_self->_the_masters_settings->sdo_write_timeout[0]));
      }

      /* "pysoem/pysoem.pyx":943
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_SDOwrite(self._ecx_contextt, self._pos, index, subindex, ca, size, c_data, self._the_masters_settings.sdo_write_timeout[0])
 *         Py_DECREF(self)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pysoem/pysoem.pyx":945
 *         with nogil:
 *             result = cpysoem.ecx_SDOwrite(self._ecx_contextt, self._pos, index, subindex, ca, size, c_data, self._the_masters_settings.sdo_write_timeout[0])
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *         return result
*/
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":947
 *         Py_DECREF(self)
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def sdo_write(self, index, uint8_t subindex, bytes data, ca=False, *, release_gil=None):
*/
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":929
 *                 PyMem_Free(pbuf)
 * 
 *     cdef int __sdo_write_nogil(self, uint16_t index, uint8_t subindex, int8_t ca, int size, bytes data):             # <<<<<<<<<<<<<<
 *         """Write to a CoE object without GIL.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._CdefSlave__sdo_write_nogil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":949
 *         return result
 * 
 *     def sdo_write(self, index, uint8_t subindex, bytes data, ca=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Write to a CoE object.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_7sdo_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_6sdo_write, "Write to a CoE object.\n        \n        Args:\n            index (int): Index of the object.\n            subindex (int): Subindex of the object.\n            data (bytes): data to be written to the object.\n            ca (:obj:`bool`, optional): complete access.\n            release_gil (:obj:`bool`, optional): True to write to a CoE object releasing the GIL. Defaults to False.\n\n        Raises:\n            SdoError: if write fails, the exception includes the SDO abort code  \n            MailboxError: on errors in the mailbox protocol\n            PacketError: on packet level error\n            WkcError: if working counter is not higher than 0, the exception includes the working counter\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_7sdo_write = {"sdo_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_7sdo_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_6sdo_write};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_7sdo_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_index = 0;
  uint8_t __pyx_v_subindex;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_ca = 0;
  PyObject *__pyx_v_release_gil = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sdo_write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_index,&__pyx_mstate_global->__pyx_n_u_subindex,&__pyx_mstate_global->__pyx_n_u_data,&__pyx_mstate_global->__pyx_n_u_ca,&__pyx_mstate_global->__pyx_n_u_release_gil,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 949, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 949, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 949, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 949, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 949, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "sdo_write", 0) < (0)) __PYX_ERR(0, 949, __pyx_L3_error)
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("sdo_write", 0, 3, 4, i); __PYX_ERR(0, 949, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 949, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 949, __pyx_L3_error)
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 949, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 949, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_index = values[0];
    __pyx_v_subindex = __Pyx_PyLong_As_uint8_t(values[1]); if (unlikely((__pyx_v_subindex == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 949, __pyx_L3_error)
    __pyx_v_data = ((PyObject*)values[2]);
    __pyx_v_ca = values[3];
    __pyx_v_release_gil = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sdo_write", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 949, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.sdo_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), (&PyBytes_Type), 1, "data", 1))) __PYX_ERR(0, 949, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_6sdo_write(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_index, __pyx_v_subindex, __pyx_v_data, __pyx_v_ca, __pyx_v_release_gil);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_6sdo_write(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_index, uint8_t __pyx_v_subindex, PyObject *__pyx_v_data, PyObject *__pyx_v_ca, PyObject *__pyx_v_release_gil) {
  int __pyx_v_size;
  int __pyx_v_result;
  ec_errort __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  boolean __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  uint16_t __pyx_t_7;
  int8_t __pyx_t_8;
  int __pyx_t_9;
  uint16 __pyx_t_10;
  unsigned char *__pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sdo_write", 0);
  __Pyx_INCREF(__pyx_v_release_gil);

  /* "pysoem/pysoem.pyx":965
 *             WkcError: if working counter is not higher than 0, the exception includes the working counter
 *         """
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)             # <<<<<<<<<<<<<<
 *         self._master.check_context_is_initialized()
 * 
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->_master->__pyx_vtab)->check_release_gil(__pyx_v_self->_master, __pyx_v_release_gil, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 965, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_release_gil, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":966
 *         """
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 * 
 *         cdef int size = len(data)
*/
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":968
 *         self._master.check_context_is_initialized()
 * 
 *         cdef int size = len(data)             # <<<<<<<<<<<<<<
 *         cdef int result
 *         if release_gil:
*/
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 968, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyBytes_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 968, __pyx_L1_error)
  __pyx_v_size = __pyx_t_5;

  /* "pysoem/pysoem.pyx":970
 *         cdef int size = len(data)
 *         cdef int result
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             result = self.__sdo_write_nogil(index, subindex, ca, size, data)
 *         else:
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_release_gil); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 970, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "pysoem/pysoem.pyx":971
 *         cdef int result
 *         if release_gil:
 *             result = self.__sdo_write_nogil(index, subindex, ca, size, data)             # <<<<<<<<<<<<<<
 *         else:
 *             result = cpysoem.ecx_SDOwrite(self._ecx_contextt, self._pos, index, subindex, ca,
*/
    __pyx_t_7 = __Pyx_PyLong_As_uint16_t(__pyx_v_index); if (unlikely((__pyx_t_7 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 971, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyLong_As_int8_t(__pyx_v_ca); if (unlikely((__pyx_t_8 == ((int8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 971, __pyx_L1_error)
    __pyx_t_9 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_CdefSlave__sdo_write_nogil(__pyx_v_self, __pyx_t_7, __pyx_v_subindex, __pyx_t_8, __pyx_v_size, __pyx_v_data); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 971, __pyx_L1_error)
    __pyx_v_result = __pyx_t_9;

    /* "pysoem/pysoem.pyx":970
 *         cdef int size = len(data)
 *         cdef int result
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             result = self.__sdo_write_nogil(index, subindex, ca, size, data)
 *         else:
*/
    goto __pyx_L3;
  }

  /* "pysoem/pysoem.pyx":973
 *             result = self.__sdo_write_nogil(index, subindex, ca, size, data)
 *         else:
 *             result = cpysoem.ecx_SDOwrite(self._ecx_contextt, self._pos, index, subindex, ca,             # <<<<<<<<<<<<<<
 *                                                size, <unsigned char*>data, self._the_masters_settings.sdo_write_timeout[0])
 * 
*/
  /*else*/ {
    __pyx_t_10 = __Pyx_PyLong_As_uint16_t(__pyx_v_index); if (unlikely((__pyx_t_10 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 973, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyLong_As_int8_t(__pyx_v_ca); if (unlikely((__pyx_t_1 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(0, 973, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":974
 *         else:
 *             result = cpysoem.ecx_SDOwrite(self._ecx_contextt, self._pos, index, subindex, ca,
 *                                                size, <unsigned char*>data, self._the_masters_settings.sdo_write_timeout[0])             # <<<<<<<<<<<<<<
 * 
 *         cdef cpysoem.ec_errort err
*/
    if (unlikely(__pyx_v_data == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 974, __pyx_L1_error)
    }
    __pyx_t_11 = __Pyx_PyBytes_AsWritableUString(__pyx_v_data); if (unlikely((!__pyx_t_11) && PyErr_Occurred())) __PYX_ERR(0, 974, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":973
 *             result = self.__sdo_write_nogil(index, subindex, ca, size, data)
 *         else:
 *             result = cpysoem.ecx_SDOwrite(self._ecx_contextt, self._pos, index, subindex, ca,             # <<<<<<<<<<<<<<
 *                                                size, <unsigned char*>data, self._the_masters_settings.sdo_write_timeout[0])
 * 
*/
    __pyx_v_result = ecx_SDOwrite(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_t_10, __pyx_v_subindex, __pyx_t_1, __pyx_v_size, ((unsigned char *)__pyx_t_11), (__pyx_v_self->_the_masters_settings->sdo_write_timeout[0]));
  }
  __pyx_L3:;

  /* "pysoem/pysoem.pyx":977
 * 
 *         cdef cpysoem.ec_errort err
 *         while(cpysoem.ecx_poperror(self._ecx_contextt, &err)):             # <<<<<<<<<<<<<<
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):
 *                 self._on_emergency(&err)
*/
  while (1) {
    __pyx_t_6 = (ecx_poperror(__pyx_v_self->_ecx_contextt, (&__pyx_v_err)) != 0);
    if (!__pyx_t_6) break;

    /* "pysoem/pysoem.pyx":978
 *         cdef cpysoem.ec_errort err
 *         while(cpysoem.ecx_poperror(self._ecx_contextt, &err)):
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):             # <<<<<<<<<<<<<<
 *                 self._on_emergency(&err)
 *             else:
*/
    __pyx_t_12 = (__pyx_v_err.Etype == EC_ERR_TYPE_EMERGENCY);
    if (__pyx_t_12) {
    } else {
      __pyx_t_6 = __pyx_t_12;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_2 = __pyx_v_self->_emcy_callbacks;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 978, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_12 = (__pyx_t_5 > 0);
    __pyx_t_6 = __pyx_t_12;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_6) {

      /* "pysoem/pysoem.pyx":979
 *         while(cpysoem.ecx_poperror(self._ecx_contextt, &err)):
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):
 *                 self._on_emergency(&err)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._raise_exception(&err)
*/
      __pyx_t_2 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_on_emergency(__pyx_v_self, (&__pyx_v_err)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 979, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pysoem/pysoem.pyx":978
 *         cdef cpysoem.ec_errort err
 *         while(cpysoem.ecx_poperror(self._ecx_contextt, &err)):
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):             # <<<<<<<<<<<<<<
 *                 self._on_emergency(&err)
 *             else:
*/
      goto __pyx_L6;
    }

    /* "pysoem/pysoem.pyx":981
 *                 self._on_emergency(&err)
 *             else:
 *                 self._raise_exception(&err)             # <<<<<<<<<<<<<<
 * 
 *         if not result > 0:
*/
    /*else*/ {
      __pyx_t_2 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_raise_exception(__pyx_v_self, (&__pyx_v_err)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L6:;
  }

  /* "pysoem/pysoem.pyx":983
 *                 self._raise_exception(&err)
 * 
 *         if not result > 0:             # <<<<<<<<<<<<<<
 *             raise WkcError(wkc=result)
 * 
*/
  __pyx_t_6 = (!(__pyx_v_result > 0));
  if (unlikely(__pyx_t_6)) {

    /* "pysoem/pysoem.pyx":984
 * 
 *         if not result > 0:
 *             raise WkcError(wkc=result)             # <<<<<<<<<<<<<<
 * 
 *     def mbx_receive(self):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_WkcError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = __Pyx_PyLong_From_int(__pyx_v_result); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_13);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_13);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_13, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_3, NULL};
      __pyx_t_15 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 984, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_wkc, __pyx_t_14, __pyx_t_15, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 984, __pyx_L1_error)
      __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_13, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_15);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 984, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 984, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":983
 *                 self._raise_exception(&err)
 * 
 *         if not result > 0:             # <<<<<<<<<<<<<<
 *             raise WkcError(wkc=result)
 * 
*/
  }

  /* "pysoem/pysoem.pyx":949
 *         return result
 * 
 *     def sdo_write(self, index, uint8_t subindex, bytes data, ca=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Write to a CoE object.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.sdo_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_release_gil);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":986
 *             raise WkcError(wkc=result)
 * 
 *     def mbx_receive(self):             # <<<<<<<<<<<<<<
 *         """Read out the slaves out mailbox - to check for emergency messages.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_9mbx_receive(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_8mbx_receive, "Read out the slaves out mailbox - to check for emergency messages.\n\n        .. versionadded:: 1.0.4\n\n        :return: Work counter\n        :rtype: int\n        :raises Emergency: if an emergency message was received\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_9mbx_receive = {"mbx_receive", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_9mbx_receive, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_8mbx_receive};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_9mbx_receive(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mbx_receive (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("mbx_receive", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("mbx_receive", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_8mbx_receive(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_8mbx_receive(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  ec_mbxbuft __pyx_v_buf;
  int __pyx_v_wkt;
  ec_errort __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mbx_receive", 0);

  /* "pysoem/pysoem.pyx":995
 *         :raises Emergency: if an emergency message was received
 *         """
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 * 
 *         cdef cpysoem.ec_mbxbuft buf
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":998
 * 
 *         cdef cpysoem.ec_mbxbuft buf
 *         cpysoem.ec_clearmbx(&buf)             # <<<<<<<<<<<<<<
 *         cdef int wkt = cpysoem.ecx_mbxreceive(self._ecx_contextt, self._pos, &buf, 0)
 * 
*/
  ec_clearmbx((&__pyx_v_buf));

  /* "pysoem/pysoem.pyx":999
 *         cdef cpysoem.ec_mbxbuft buf
 *         cpysoem.ec_clearmbx(&buf)
 *         cdef int wkt = cpysoem.ecx_mbxreceive(self._ecx_contextt, self._pos, &buf, 0)             # <<<<<<<<<<<<<<
 * 
 *         cdef cpysoem.ec_errort err
*/
  __pyx_v_wkt = ecx_mbxreceive(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, (&__pyx_v_buf), 0);

  /* "pysoem/pysoem.pyx":1002
 * 
 *         cdef cpysoem.ec_errort err
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):             # <<<<<<<<<<<<<<
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):
 *                 self._on_emergency(&err)
*/
  __pyx_t_4 = (ecx_poperror(__pyx_v_self->_ecx_contextt, (&__pyx_v_err)) != 0);
  if (__pyx_t_4) {

    /* "pysoem/pysoem.pyx":1003
 *         cdef cpysoem.ec_errort err
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):             # <<<<<<<<<<<<<<
 *                 self._on_emergency(&err)
 *             else:
*/
    __pyx_t_5 = (__pyx_v_err.Etype == EC_ERR_TYPE_EMERGENCY);
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_1 = __pyx_v_self->_emcy_callbacks;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_6 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = (__pyx_t_6 > 0);
    __pyx_t_4 = __pyx_t_5;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_4) {

      /* "pysoem/pysoem.pyx":1004
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):
 *                 self._on_emergency(&err)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._raise_exception(&err)
*/
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_on_emergency(__pyx_v_self, (&__pyx_v_err)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1004, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pysoem/pysoem.pyx":1003
 *         cdef cpysoem.ec_errort err
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):             # <<<<<<<<<<<<<<
 *                 self._on_emergency(&err)
 *             else:
*/
      goto __pyx_L4;
    }

    /* "pysoem/pysoem.pyx":1006
 *                 self._on_emergency(&err)
 *             else:
 *                 self._raise_exception(&err)             # <<<<<<<<<<<<<<
 * 
 *         return wkt
*/
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_raise_exception(__pyx_v_self, (&__pyx_v_err)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L4:;

    /* "pysoem/pysoem.pyx":1002
 * 
 *         cdef cpysoem.ec_errort err
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):             # <<<<<<<<<<<<<<
 *             if (err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY) and (len(self._emcy_callbacks) > 0):
 *                 self._on_emergency(&err)
*/
  }

  /* "pysoem/pysoem.pyx":1008
 *                 self._raise_exception(&err)
 * 
 *         return wkt             # <<<<<<<<<<<<<<
 * 
 *     def write_state(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_wkt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":986
 *             raise WkcError(wkc=result)
 * 
 *     def mbx_receive(self):             # <<<<<<<<<<<<<<
 *         """Read out the slaves out mailbox - to check for emergency messages.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.mbx_receive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1010
 *         return wkt
 * 
 *     def write_state(self):             # <<<<<<<<<<<<<<
 *         """Write slave state.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_11write_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_10write_state, "Write slave state.\n\n        Note: The function does not check if the actual state is changed.\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_11write_state = {"write_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_11write_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_10write_state};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_11write_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_state (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("write_state", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("write_state", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_10write_state(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_10write_state(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_state", 0);

  /* "pysoem/pysoem.pyx":1015
 *         Note: The function does not check if the actual state is changed.
 *         """
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_writestate(self._ecx_contextt, self._pos)
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1016
 *         """
 *         self._master.check_context_is_initialized()
 *         return cpysoem.ecx_writestate(self._ecx_contextt, self._pos)             # <<<<<<<<<<<<<<
 * 
 *     def state_check(self, int expected_state, timeout=2000):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(ecx_writestate(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1010
 *         return wkt
 * 
 *     def write_state(self):             # <<<<<<<<<<<<<<
 *         """Write slave state.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.write_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1018
 *         return cpysoem.ecx_writestate(self._ecx_contextt, self._pos)
 * 
 *     def state_check(self, int expected_state, timeout=2000):             # <<<<<<<<<<<<<<
 *         """Wait for the slave to reach the state that was requested."""
 *         self._master.check_context_is_initialized()
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_13state_check(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_12state_check, "Wait for the slave to reach the state that was requested.");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_13state_check = {"state_check", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_13state_check, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_12state_check};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_13state_check(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_expected_state;
  PyObject *__pyx_v_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("state_check (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_expected_state,&__pyx_mstate_global->__pyx_n_u_timeout,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1018, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1018, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1018, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "state_check", 0) < (0)) __PYX_ERR(0, 1018, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("state_check", 0, 1, 2, i); __PYX_ERR(0, 1018, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1018, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1018, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
    }
    __pyx_v_expected_state = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_expected_state == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1018, __pyx_L3_error)
    __pyx_v_timeout = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("state_check", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1018, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.state_check", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_12state_check(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_expected_state, __pyx_v_timeout);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_12state_check(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, int __pyx_v_expected_state, PyObject *__pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("state_check", 0);

  /* "pysoem/pysoem.pyx":1020
 *     def state_check(self, int expected_state, timeout=2000):
 *         """Wait for the slave to reach the state that was requested."""
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_statecheck(self._ecx_contextt, self._pos, expected_state, timeout)
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1020, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1021
 *         """Wait for the slave to reach the state that was requested."""
 *         self._master.check_context_is_initialized()
 *         return cpysoem.ecx_statecheck(self._ecx_contextt, self._pos, expected_state, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def reconfig(self, timeout=500):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1021, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyLong_From_uint16_t(ecx_statecheck(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_v_expected_state, __pyx_t_4)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1018
 *         return cpysoem.ecx_writestate(self._ecx_contextt, self._pos)
 * 
 *     def state_check(self, int expected_state, timeout=2000):             # <<<<<<<<<<<<<<
 *         """Wait for the slave to reach the state that was requested."""
 *         self._master.check_context_is_initialized()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.state_check", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1023
 *         return cpysoem.ecx_statecheck(self._ecx_contextt, self._pos, expected_state, timeout)
 * 
 *     def reconfig(self, timeout=500):             # <<<<<<<<<<<<<<
 *         """Reconfigure slave.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_15reconfig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_14reconfig, "Reconfigure slave.\n\n        :param timeout: local timeout\n        :return: Slave state\n        :rtype: int\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_15reconfig = {"reconfig", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_15reconfig, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_14reconfig};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_15reconfig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reconfig (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_timeout,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1023, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1023, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "reconfig", 0) < (0)) __PYX_ERR(0, 1023, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_500));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1023, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_500));
    }
    __pyx_v_timeout = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("reconfig", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 1023, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.reconfig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_14reconfig(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_timeout);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_14reconfig(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reconfig", 0);

  /* "pysoem/pysoem.pyx":1030
 *         :rtype: int
 *         """
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_reconfig_slave(self._ecx_contextt, self._pos, timeout)
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1031
 *         """
 *         self._master.check_context_is_initialized()
 *         return cpysoem.ecx_reconfig_slave(self._ecx_contextt, self._pos, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def recover(self, timeout=500):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1031, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyLong_From_int(ecx_reconfig_slave(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_t_4)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1023
 *         return cpysoem.ecx_statecheck(self._ecx_contextt, self._pos, expected_state, timeout)
 * 
 *     def reconfig(self, timeout=500):             # <<<<<<<<<<<<<<
 *         """Reconfigure slave.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.reconfig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1033
 *         return cpysoem.ecx_reconfig_slave(self._ecx_contextt, self._pos, timeout)
 * 
 *     def recover(self, timeout=500):             # <<<<<<<<<<<<<<
 *         """Recover slave.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_17recover(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_16recover, "Recover slave.\n\n        :param timeout: local timeout\n        :return: >0 if successful\n        :rtype: int\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_17recover = {"recover", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_17recover, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_16recover};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_17recover(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("recover (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_timeout,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1033, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1033, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "recover", 0) < (0)) __PYX_ERR(0, 1033, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_500));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1033, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_500));
    }
    __pyx_v_timeout = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("recover", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 1033, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.recover", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_16recover(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_timeout);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_16recover(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recover", 0);

  /* "pysoem/pysoem.pyx":1040
 *         :rtype: int
 *         """
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         return cpysoem.ecx_recover_slave(self._ecx_contextt, self._pos, timeout)
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1041
 *         """
 *         self._master.check_context_is_initialized()
 *         return cpysoem.ecx_recover_slave(self._ecx_contextt, self._pos, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def eeprom_read(self, int word_address, timeout=20000):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1041, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyLong_From_int(ecx_recover_slave(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_t_4)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1033
 *         return cpysoem.ecx_reconfig_slave(self._ecx_contextt, self._pos, timeout)
 * 
 *     def recover(self, timeout=500):             # <<<<<<<<<<<<<<
 *         """Recover slave.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.recover", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1043
 *         return cpysoem.ecx_recover_slave(self._ecx_contextt, self._pos, timeout)
 * 
 *     def eeprom_read(self, int word_address, timeout=20000):             # <<<<<<<<<<<<<<
 *         """Read 4 byte from EEPROM
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_19eeprom_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_18eeprom_read, "Read 4 byte from EEPROM\n\n        Default timeout: 20000 us\n\n        Args:\n            word_address (int): EEPROM address to read from\n            timeout (:obj:`int`, optional): Timeout value in us\n\n        Returns:\n            bytes: EEPROM data\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_19eeprom_read = {"eeprom_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_19eeprom_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_18eeprom_read};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_19eeprom_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_word_address;
  PyObject *__pyx_v_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eeprom_read (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_word_address,&__pyx_mstate_global->__pyx_n_u_timeout,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1043, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1043, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1043, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "eeprom_read", 0) < (0)) __PYX_ERR(0, 1043, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_20000));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("eeprom_read", 0, 1, 2, i); __PYX_ERR(0, 1043, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1043, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1043, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_20000));
    }
    __pyx_v_word_address = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_word_address == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1043, __pyx_L3_error)
    __pyx_v_timeout = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("eeprom_read", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1043, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.eeprom_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_18eeprom_read(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_word_address, __pyx_v_timeout);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_18eeprom_read(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, int __pyx_v_word_address, PyObject *__pyx_v_timeout) {
  uint32_t __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eeprom_read", 0);

  /* "pysoem/pysoem.pyx":1055
 *             bytes: EEPROM data
 *         """
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         cdef uint32_t tmp = cpysoem.ecx_readeeprom(self._ecx_contextt, self._pos, word_address, timeout)
 *         return PyBytes_FromStringAndSize(<char*>&tmp, 4)
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1056
 *         """
 *         self._master.check_context_is_initialized()
 *         cdef uint32_t tmp = cpysoem.ecx_readeeprom(self._ecx_contextt, self._pos, word_address, timeout)             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(<char*>&tmp, 4)
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1056, __pyx_L1_error)
  __pyx_v_tmp = ecx_readeeprom(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_v_word_address, __pyx_t_4);

  /* "pysoem/pysoem.pyx":1057
 *         self._master.check_context_is_initialized()
 *         cdef uint32_t tmp = cpysoem.ecx_readeeprom(self._ecx_contextt, self._pos, word_address, timeout)
 *         return PyBytes_FromStringAndSize(<char*>&tmp, 4)             # <<<<<<<<<<<<<<
 * 
 *     def eeprom_write(self, int word_address, bytes data, timeout=20000):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(((char *)(&__pyx_v_tmp)), 4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1043
 *         return cpysoem.ecx_recover_slave(self._ecx_contextt, self._pos, timeout)
 * 
 *     def eeprom_read(self, int word_address, timeout=20000):             # <<<<<<<<<<<<<<
 *         """Read 4 byte from EEPROM
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.eeprom_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1059
 *         return PyBytes_FromStringAndSize(<char*>&tmp, 4)
 * 
 *     def eeprom_write(self, int word_address, bytes data, timeout=20000):             # <<<<<<<<<<<<<<
 *         """Write 2 byte (1 word) to EEPROM
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_21eeprom_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_20eeprom_write, "Write 2 byte (1 word) to EEPROM\n\n        Default timeout: 20000 us\n        \n        Args:\n            word_address (int): EEPROM address to write to\n            data (bytes): data (only 2 bytes are allowed)\n            timeout (:obj:`int`, optional): Timeout value in us\n\n        Raises:\n            EepromError: if write fails\n            AttributeError: if data size is not 2\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_21eeprom_write = {"eeprom_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_21eeprom_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_20eeprom_write};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_21eeprom_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_word_address;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eeprom_write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_word_address,&__pyx_mstate_global->__pyx_n_u_data,&__pyx_mstate_global->__pyx_n_u_timeout,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1059, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1059, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1059, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1059, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "eeprom_write", 0) < (0)) __PYX_ERR(0, 1059, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_20000));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("eeprom_write", 0, 2, 3, i); __PYX_ERR(0, 1059, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1059, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1059, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1059, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_20000));
    }
    __pyx_v_word_address = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_word_address == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1059, __pyx_L3_error)
    __pyx_v_data = ((PyObject*)values[1]);
    __pyx_v_timeout = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("eeprom_write", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 1059, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.eeprom_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), (&PyBytes_Type), 1, "data", 1))) __PYX_ERR(0, 1059, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_20eeprom_write(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_word_address, __pyx_v_data, __pyx_v_timeout);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_20eeprom_write(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, int __pyx_v_word_address, PyObject *__pyx_v_data, PyObject *__pyx_v_timeout) {
  uint16_t __pyx_v_tmp;
  int __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  unsigned char *__pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eeprom_write", 0);

  /* "pysoem/pysoem.pyx":1073
 *             AttributeError: if data size is not 2
 *         """
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         if not len(data) == 2:
 *             raise AttributeError()
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1074
 *         """
 *         self._master.check_context_is_initialized()
 *         if not len(data) == 2:             # <<<<<<<<<<<<<<
 *             raise AttributeError()
 *         cdef uint16_t tmp
*/
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1074, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1074, __pyx_L1_error)
  __pyx_t_5 = (!(__pyx_t_4 == 2));
  if (unlikely(__pyx_t_5)) {

    /* "pysoem/pysoem.pyx":1075
 *         self._master.check_context_is_initialized()
 *         if not len(data) == 2:
 *             raise AttributeError()             # <<<<<<<<<<<<<<
 *         cdef uint16_t tmp
 *         memcpy(<char*>&tmp, <unsigned char*>data, 2)
*/
    __pyx_t_2 = NULL;
    __pyx_t_3 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AttributeError)), __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1075, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1074
 *         """
 *         self._master.check_context_is_initialized()
 *         if not len(data) == 2:             # <<<<<<<<<<<<<<
 *             raise AttributeError()
 *         cdef uint16_t tmp
*/
  }

  /* "pysoem/pysoem.pyx":1077
 *             raise AttributeError()
 *         cdef uint16_t tmp
 *         memcpy(<char*>&tmp, <unsigned char*>data, 2)             # <<<<<<<<<<<<<<
 *         cdef int result = cpysoem.ecx_writeeeprom(self._ecx_contextt, self._pos, word_address, tmp, timeout)
 *         if not result > 0:
*/
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1077, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_AsWritableUString(__pyx_v_data); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1077, __pyx_L1_error)
  (void)(memcpy(((char *)(&__pyx_v_tmp)), ((unsigned char *)__pyx_t_6), 2));

  /* "pysoem/pysoem.pyx":1078
 *         cdef uint16_t tmp
 *         memcpy(<char*>&tmp, <unsigned char*>data, 2)
 *         cdef int result = cpysoem.ecx_writeeeprom(self._ecx_contextt, self._pos, word_address, tmp, timeout)             # <<<<<<<<<<<<<<
 *         if not result > 0:
 *             raise EepromError('EEPROM write error')
*/
  __pyx_t_7 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L1_error)
  __pyx_v_result = ecx_writeeeprom(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_v_word_address, __pyx_v_tmp, __pyx_t_7);

  /* "pysoem/pysoem.pyx":1079
 *         memcpy(<char*>&tmp, <unsigned char*>data, 2)
 *         cdef int result = cpysoem.ecx_writeeeprom(self._ecx_contextt, self._pos, word_address, tmp, timeout)
 *         if not result > 0:             # <<<<<<<<<<<<<<
 *             raise EepromError('EEPROM write error')
 * 
*/
  __pyx_t_5 = (!(__pyx_v_result > 0));
  if (unlikely(__pyx_t_5)) {

    /* "pysoem/pysoem.pyx":1080
 *         cdef int result = cpysoem.ecx_writeeeprom(self._ecx_contextt, self._pos, word_address, tmp, timeout)
 *         if not result > 0:
 *             raise EepromError('EEPROM write error')             # <<<<<<<<<<<<<<
 * 
 *     cdef int __foe_write_nogil(self, str filename, uint32_t password, int size, bytes data, int timeout):
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_EepromError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1080, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_EEPROM_write_error};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1080, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1080, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1079
 *         memcpy(<char*>&tmp, <unsigned char*>data, 2)
 *         cdef int result = cpysoem.ecx_writeeeprom(self._ecx_contextt, self._pos, word_address, tmp, timeout)
 *         if not result > 0:             # <<<<<<<<<<<<<<
 *             raise EepromError('EEPROM write error')
 * 
*/
  }

  /* "pysoem/pysoem.pyx":1059
 *         return PyBytes_FromStringAndSize(<char*>&tmp, 4)
 * 
 *     def eeprom_write(self, int word_address, bytes data, timeout=20000):             # <<<<<<<<<<<<<<
 *         """Write 2 byte (1 word) to EEPROM
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.eeprom_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1082
 *             raise EepromError('EEPROM write error')
 * 
 *     cdef int __foe_write_nogil(self, str filename, uint32_t password, int size, bytes data, int timeout):             # <<<<<<<<<<<<<<
 *         """Write given data to device using FoE without GIL.
 * 
*/

static int __pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__foe_write_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_filename, uint32_t __pyx_v_password, int __pyx_v_size, PyObject *__pyx_v_data, int __pyx_v_timeout) {
  int __pyx_v_result;
  PyObject *__pyx_v_encoded_filename = 0;
  char *__pyx_v_c_filename;
  unsigned char *__pyx_v_c_data;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  unsigned char *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_CdefSlave__foe_write_nogil", 0);

  /* "pysoem/pysoem.pyx":1093
 *         """
 *         cdef int result
 *         cdef bytes encoded_filename = filename.encode('utf-8')             # <<<<<<<<<<<<<<
 *         cdef char* c_filename = encoded_filename
 *         cdef unsigned char* c_data = <unsigned char*> data
*/
  if (unlikely(__pyx_v_filename == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 1093, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsUTF8String(__pyx_v_filename); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_encoded_filename = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1094
 *         cdef int result
 *         cdef bytes encoded_filename = filename.encode('utf-8')
 *         cdef char* c_filename = encoded_filename             # <<<<<<<<<<<<<<
 *         cdef unsigned char* c_data = <unsigned char*> data
 * 
*/
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_filename); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1094, __pyx_L1_error)
  __pyx_v_c_filename = __pyx_t_2;

  /* "pysoem/pysoem.pyx":1095
 *         cdef bytes encoded_filename = filename.encode('utf-8')
 *         cdef char* c_filename = encoded_filename
 *         cdef unsigned char* c_data = <unsigned char*> data             # <<<<<<<<<<<<<<
 * 
 *         Py_INCREF(self)
*/
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1095, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_AsWritableUString(__pyx_v_data); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1095, __pyx_L1_error)
  __pyx_v_c_data = ((unsigned char *)__pyx_t_3);

  /* "pysoem/pysoem.pyx":1097
 *         cdef unsigned char* c_data = <unsigned char*> data
 * 
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             result = cpysoem.ecx_FOEwrite(self._ecx_contextt, self._pos, c_filename, password, size, c_data, timeout)
*/
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":1098
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_FOEwrite(self._ecx_contextt, self._pos, c_filename, password, size, c_data, timeout)
 *         Py_DECREF(self)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pysoem/pysoem.pyx":1099
 *         Py_INCREF(self)
 *         with nogil:
 *             result = cpysoem.ecx_FOEwrite(self._ecx_contextt, self._pos, c_filename, password, size, c_data, timeout)             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
*/
        __pyx_v_result = ecx_FOEwrite(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_v_c_filename, __pyx_v_password, __pyx_v_size, __pyx_v_c_data, __pyx_v_timeout);
      }

      /* "pysoem/pysoem.pyx":1098
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_FOEwrite(self._ecx_contextt, self._pos, c_filename, password, size, c_data, timeout)
 *         Py_DECREF(self)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pysoem/pysoem.pyx":1100
 *         with nogil:
 *             result = cpysoem.ecx_FOEwrite(self._ecx_contextt, self._pos, c_filename, password, size, c_data, timeout)
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *         return result
*/
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":1102
 *         Py_DECREF(self)
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def foe_write(self, filename, password, bytes data, timeout = 200000, *, release_gil=None):
*/
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1082
 *             raise EepromError('EEPROM write error')
 * 
 *     cdef int __foe_write_nogil(self, str filename, uint32_t password, int size, bytes data, int timeout):             # <<<<<<<<<<<<<<
 *         """Write given data to device using FoE without GIL.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._CdefSlave__foe_write_nogil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_filename);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1104
 *         return result
 * 
 *     def foe_write(self, filename, password, bytes data, timeout = 200000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """ Write given data to device using FoE
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_23foe_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_22foe_write, " Write given data to device using FoE\n\n        Args:\n            filename (string): name of the target file\n            password (int): password for the target file, accepted range: 0 to 2^32 - 1\n            data (bytes): data\n            timeout (int): Timeout value in us\n            release_gil (:obj:`bool`, optional): True to FoE write releasing the GIL. Defaults to False.\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_23foe_write = {"foe_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_23foe_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_22foe_write};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_23foe_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_filename = 0;
  PyObject *__pyx_v_password = 0;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_timeout = 0;
  PyObject *__pyx_v_release_gil = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("foe_write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_filename,&__pyx_mstate_global->__pyx_n_u_password,&__pyx_mstate_global->__pyx_n_u_data,&__pyx_mstate_global->__pyx_n_u_timeout,&__pyx_mstate_global->__pyx_n_u_release_gil,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1104, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "foe_write", 0) < (0)) __PYX_ERR(0, 1104, __pyx_L3_error)
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_200000));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("foe_write", 0, 3, 4, i); __PYX_ERR(0, 1104, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1104, __pyx_L3_error)
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1104, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1104, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_200000));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_filename = values[0];
    __pyx_v_password = values[1];
    __pyx_v_data = ((PyObject*)values[2]);
    __pyx_v_timeout = values[3];
    __pyx_v_release_gil = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("foe_write", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 1104, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.foe_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), (&PyBytes_Type), 1, "data", 1))) __PYX_ERR(0, 1104, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_22foe_write(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_filename, __pyx_v_password, __pyx_v_data, __pyx_v_timeout, __pyx_v_release_gil);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_22foe_write(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_password, PyObject *__pyx_v_data, PyObject *__pyx_v_timeout, PyObject *__pyx_v_release_gil) {
  int __pyx_v_result;
  int __pyx_v_size;
  ec_errort __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  boolean __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  uint32_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  char *__pyx_t_10;
  uint32 __pyx_t_11;
  unsigned char *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("foe_write", 0);
  __Pyx_INCREF(__pyx_v_release_gil);

  /* "pysoem/pysoem.pyx":1114
 *             release_gil (:obj:`bool`, optional): True to FoE write releasing the GIL. Defaults to False.
 *         """
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)             # <<<<<<<<<<<<<<
 *         # error handling
 *         if self._ecx_contextt == NULL:
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->_master->__pyx_vtab)->check_release_gil(__pyx_v_self->_master, __pyx_v_release_gil, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1114, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_release_gil, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":1116
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)
 *         # error handling
 *         if self._ecx_contextt == NULL:             # <<<<<<<<<<<<<<
 *             raise UnboundLocalError()
 * 
*/
  __pyx_t_3 = (__pyx_v_self->_ecx_contextt == NULL);
  if (unlikely(__pyx_t_3)) {

    /* "pysoem/pysoem.pyx":1117
 *         # error handling
 *         if self._ecx_contextt == NULL:
 *             raise UnboundLocalError()             # <<<<<<<<<<<<<<
 * 
 *         self._master.check_context_is_initialized()
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_UnboundLocalError)), __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1117, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1116
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)
 *         # error handling
 *         if self._ecx_contextt == NULL:             # <<<<<<<<<<<<<<
 *             raise UnboundLocalError()
 * 
*/
  }

  /* "pysoem/pysoem.pyx":1119
 *             raise UnboundLocalError()
 * 
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 * 
 *         cdef int result
*/
  __pyx_t_4 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":1122
 * 
 *         cdef int result
 *         cdef int size = len(data)             # <<<<<<<<<<<<<<
 * 
 *         if release_gil:
*/
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1122, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1122, __pyx_L1_error)
  __pyx_v_size = __pyx_t_6;

  /* "pysoem/pysoem.pyx":1124
 *         cdef int size = len(data)
 * 
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             result = self.__foe_write_nogil(filename, password, size, data, timeout)
 *         else:
*/
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_release_gil); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1124, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "pysoem/pysoem.pyx":1125
 * 
 *         if release_gil:
 *             result = self.__foe_write_nogil(filename, password, size, data, timeout)             # <<<<<<<<<<<<<<
 *         else:
 *             result = cpysoem.ecx_FOEwrite(self._ecx_contextt, self._pos, filename.encode('utf8'), password, size, <unsigned char*>data, timeout)
*/
    __pyx_t_2 = __pyx_v_filename;
    __Pyx_INCREF(__pyx_t_2);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 1125, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyLong_As_uint32_t(__pyx_v_password); if (unlikely((__pyx_t_7 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1125, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1125, __pyx_L1_error)
    __pyx_t_9 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_CdefSlave__foe_write_nogil(__pyx_v_self, ((PyObject*)__pyx_t_2), __pyx_t_7, __pyx_v_size, __pyx_v_data, __pyx_t_8); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1125, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_result = __pyx_t_9;

    /* "pysoem/pysoem.pyx":1124
 *         cdef int size = len(data)
 * 
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             result = self.__foe_write_nogil(filename, password, size, data, timeout)
 *         else:
*/
    goto __pyx_L4;
  }

  /* "pysoem/pysoem.pyx":1127
 *             result = self.__foe_write_nogil(filename, password, size, data, timeout)
 *         else:
 *             result = cpysoem.ecx_FOEwrite(self._ecx_contextt, self._pos, filename.encode('utf8'), password, size, <unsigned char*>data, timeout)             # <<<<<<<<<<<<<<
 * 
 *         # error handling
*/
  /*else*/ {
    __pyx_t_4 = __pyx_v_filename;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_utf8};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_10 = __Pyx_PyObject_AsWritableString(__pyx_t_2); if (unlikely((!__pyx_t_10) && PyErr_Occurred())) __PYX_ERR(0, 1127, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyLong_As_uint32_t(__pyx_v_password); if (unlikely((__pyx_t_11 == ((uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1127, __pyx_L1_error)
    if (unlikely(__pyx_v_data == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1127, __pyx_L1_error)
    }
    __pyx_t_12 = __Pyx_PyBytes_AsWritableUString(__pyx_v_data); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 1127, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1127, __pyx_L1_error)
    __pyx_v_result = ecx_FOEwrite(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_t_10, __pyx_t_11, __pyx_v_size, ((unsigned char *)__pyx_t_12), __pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "pysoem/pysoem.pyx":1131
 *         # error handling
 *         cdef cpysoem.ec_errort err
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):             # <<<<<<<<<<<<<<
 *             assert err.Slave == self._pos
 *             self._raise_exception(&err)
*/
  __pyx_t_3 = (ecx_poperror(__pyx_v_self->_ecx_contextt, (&__pyx_v_err)) != 0);
  if (__pyx_t_3) {

    /* "pysoem/pysoem.pyx":1132
 *         cdef cpysoem.ec_errort err
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):
 *             assert err.Slave == self._pos             # <<<<<<<<<<<<<<
 *             self._raise_exception(&err)
 * 
*/
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_3 = (__pyx_v_err.Slave == __pyx_v_self->_pos);
      if (unlikely(!__pyx_t_3)) {
        __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), 0, 0, 0);
        __PYX_ERR(0, 1132, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 1132, __pyx_L1_error)
    #endif

    /* "pysoem/pysoem.pyx":1133
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):
 *             assert err.Slave == self._pos
 *             self._raise_exception(&err)             # <<<<<<<<<<<<<<
 * 
 *         return result
*/
    __pyx_t_2 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_raise_exception(__pyx_v_self, (&__pyx_v_err)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pysoem/pysoem.pyx":1131
 *         # error handling
 *         cdef cpysoem.ec_errort err
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):             # <<<<<<<<<<<<<<
 *             assert err.Slave == self._pos
 *             self._raise_exception(&err)
*/
  }

  /* "pysoem/pysoem.pyx":1135
 *             self._raise_exception(&err)
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef int __foe_read_nogil(self, str filename, uint32_t password, int size_inout, unsigned char* pbuf, int timeout):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v_result); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1104
 *         return result
 * 
 *     def foe_write(self, filename, password, bytes data, timeout = 200000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """ Write given data to device using FoE
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.foe_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_release_gil);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1137
 *         return result
 * 
 *     cdef int __foe_read_nogil(self, str filename, uint32_t password, int size_inout, unsigned char* pbuf, int timeout):             # <<<<<<<<<<<<<<
 *         """Read given filename from device using FoE without GIL.
 * 
*/

static int __pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__foe_read_nogil(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_filename, uint32_t __pyx_v_password, int __pyx_v_size_inout, unsigned char *__pyx_v_pbuf, int __pyx_v_timeout) {
  int __pyx_v_result;
  PyObject *__pyx_v_encoded_filename = 0;
  char *__pyx_v_c_filename;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_CdefSlave__foe_read_nogil", 0);

  /* "pysoem/pysoem.pyx":1148
 *         """
 *         cdef int result
 *         cdef bytes encoded_filename = filename.encode('utf-8')             # <<<<<<<<<<<<<<
 *         cdef char* c_filename = encoded_filename
 * 
*/
  if (unlikely(__pyx_v_filename == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 1148, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsUTF8String(__pyx_v_filename); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_encoded_filename = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1149
 *         cdef int result
 *         cdef bytes encoded_filename = filename.encode('utf-8')
 *         cdef char* c_filename = encoded_filename             # <<<<<<<<<<<<<<
 * 
 *         Py_INCREF(self)
*/
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_filename); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1149, __pyx_L1_error)
  __pyx_v_c_filename = __pyx_t_2;

  /* "pysoem/pysoem.pyx":1151
 *         cdef char* c_filename = encoded_filename
 * 
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             result = cpysoem.ecx_FOEread(self._ecx_contextt, self._pos, c_filename, password, &size_inout, pbuf, timeout)
*/
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":1152
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_FOEread(self._ecx_contextt, self._pos, c_filename, password, &size_inout, pbuf, timeout)
 *         Py_DECREF(self)
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pysoem/pysoem.pyx":1153
 *         Py_INCREF(self)
 *         with nogil:
 *             result = cpysoem.ecx_FOEread(self._ecx_contextt, self._pos, c_filename, password, &size_inout, pbuf, timeout)             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
*/
        __pyx_v_result = ecx_FOEread(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_v_c_filename, __pyx_v_password, (&__pyx_v_size_inout), __pyx_v_pbuf, __pyx_v_timeout);
      }

      /* "pysoem/pysoem.pyx":1152
 * 
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_FOEread(self._ecx_contextt, self._pos, c_filename, password, &size_inout, pbuf, timeout)
 *         Py_DECREF(self)
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pysoem/pysoem.pyx":1154
 *         with nogil:
 *             result = cpysoem.ecx_FOEread(self._ecx_contextt, self._pos, c_filename, password, &size_inout, pbuf, timeout)
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *         return result
*/
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "pysoem/pysoem.pyx":1156
 *         Py_DECREF(self)
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def foe_read(self, filename, password, size, timeout = 200000, *, release_gil=None):
*/
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1137
 *         return result
 * 
 *     cdef int __foe_read_nogil(self, str filename, uint32_t password, int size_inout, unsigned char* pbuf, int timeout):             # <<<<<<<<<<<<<<
 *         """Read given filename from device using FoE without GIL.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._CdefSlave__foe_read_nogil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_filename);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1158
 *         return result
 * 
 *     def foe_read(self, filename, password, size, timeout = 200000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Read given filename from device using FoE
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_25foe_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_24foe_read, "Read given filename from device using FoE\n\n        Args:\n            filename (string): name of the target file\n            password (int): password for target file\n            size (int): maximum file size\n            timeout (int): Timeout value in us\n            release_gil (:obj:`bool`, optional): True to FoE write releasing the GIL. Defaults to False.\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_25foe_read = {"foe_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_25foe_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_24foe_read};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_25foe_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_filename = 0;
  PyObject *__pyx_v_password = 0;
  PyObject *__pyx_v_size = 0;
  PyObject *__pyx_v_timeout = 0;
  PyObject *__pyx_v_release_gil = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("foe_read (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_filename,&__pyx_mstate_global->__pyx_n_u_password,&__pyx_mstate_global->__pyx_n_u_size,&__pyx_mstate_global->__pyx_n_u_timeout,&__pyx_mstate_global->__pyx_n_u_release_gil,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1158, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1158, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1158, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1158, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1158, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "foe_read", 0) < (0)) __PYX_ERR(0, 1158, __pyx_L3_error)
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_200000));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("foe_read", 0, 3, 4, i); __PYX_ERR(0, 1158, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1158, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1158, __pyx_L3_error)
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1158, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1158, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_200000));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_filename = values[0];
    __pyx_v_password = values[1];
    __pyx_v_size = values[2];
    __pyx_v_timeout = values[3];
    __pyx_v_release_gil = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("foe_read", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 1158, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.foe_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_24foe_read(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_filename, __pyx_v_password, __pyx_v_size, __pyx_v_timeout, __pyx_v_release_gil);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_24foe_read(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_password, PyObject *__pyx_v_size, PyObject *__pyx_v_timeout, PyObject *__pyx_v_release_gil) {
  unsigned char *__pyx_v_pbuf;
  int __pyx_v_size_inout;
  CYTHON_UNUSED int __pyx_v_result;
  ec_errort __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  boolean __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  uint32_t __pyx_t_7;
  int __pyx_t_8;
  char *__pyx_t_9;
  uint32 __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("foe_read", 0);
  __Pyx_INCREF(__pyx_v_release_gil);

  /* "pysoem/pysoem.pyx":1168
 *             release_gil (:obj:`bool`, optional): True to FoE write releasing the GIL. Defaults to False.
 *         """
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)             # <<<<<<<<<<<<<<
 *         if self._ecx_contextt == NULL:
 *             raise UnboundLocalError()
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster *)__pyx_v_self->_master->__pyx_vtab)->check_release_gil(__pyx_v_self->_master, __pyx_v_release_gil, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1168, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_release_gil, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":1169
 *         """
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)
 *         if self._ecx_contextt == NULL:             # <<<<<<<<<<<<<<
 *             raise UnboundLocalError()
 * 
*/
  __pyx_t_3 = (__pyx_v_self->_ecx_contextt == NULL);
  if (unlikely(__pyx_t_3)) {

    /* "pysoem/pysoem.pyx":1170
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)
 *         if self._ecx_contextt == NULL:
 *             raise UnboundLocalError()             # <<<<<<<<<<<<<<
 * 
 *         self._master.check_context_is_initialized()
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_UnboundLocalError)), __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1170, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1169
 *         """
 *         release_gil = self._master.check_release_gil(release_gil=release_gil)
 *         if self._ecx_contextt == NULL:             # <<<<<<<<<<<<<<
 *             raise UnboundLocalError()
 * 
*/
  }

  /* "pysoem/pysoem.pyx":1172
 *             raise UnboundLocalError()
 * 
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 * 
 *         # prepare call of c function
*/
  __pyx_t_4 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":1177
 *         cdef unsigned char* pbuf
 *         cdef int size_inout
 *         pbuf = <unsigned char*>PyMem_Malloc((size)*sizeof(unsigned char))             # <<<<<<<<<<<<<<
 *         size_inout = size
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_FromSize_t((sizeof(unsigned char))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyNumber_Multiply(__pyx_v_size, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyLong_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_pbuf = ((unsigned char *)PyMem_Malloc(__pyx_t_5));

  /* "pysoem/pysoem.pyx":1178
 *         cdef int size_inout
 *         pbuf = <unsigned char*>PyMem_Malloc((size)*sizeof(unsigned char))
 *         size_inout = size             # <<<<<<<<<<<<<<
 * 
 *         cdef int result
*/
  __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_size); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1178, __pyx_L1_error)
  __pyx_v_size_inout = __pyx_t_6;

  /* "pysoem/pysoem.pyx":1181
 * 
 *         cdef int result
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             result = self.__foe_read_nogil(filename, password, size_inout, pbuf, timeout)
 *         else:
*/
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_release_gil); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1181, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "pysoem/pysoem.pyx":1182
 *         cdef int result
 *         if release_gil:
 *             result = self.__foe_read_nogil(filename, password, size_inout, pbuf, timeout)             # <<<<<<<<<<<<<<
 *         else:
 *             result = cpysoem.ecx_FOEread(self._ecx_contextt, self._pos, filename.encode('utf8'), password, &size_inout, pbuf, timeout)
*/
    __pyx_t_4 = __pyx_v_filename;
    __Pyx_INCREF(__pyx_t_4);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_4))) __PYX_ERR(0, 1182, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyLong_As_uint32_t(__pyx_v_password); if (unlikely((__pyx_t_7 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1182, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1182, __pyx_L1_error)
    __pyx_t_8 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_CdefSlave__foe_read_nogil(__pyx_v_self, ((PyObject*)__pyx_t_4), __pyx_t_7, __pyx_v_size_inout, __pyx_v_pbuf, __pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1182, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_result = __pyx_t_8;

    /* "pysoem/pysoem.pyx":1181
 * 
 *         cdef int result
 *         if release_gil:             # <<<<<<<<<<<<<<
 *             result = self.__foe_read_nogil(filename, password, size_inout, pbuf, timeout)
 *         else:
*/
    goto __pyx_L4;
  }

  /* "pysoem/pysoem.pyx":1184
 *             result = self.__foe_read_nogil(filename, password, size_inout, pbuf, timeout)
 *         else:
 *             result = cpysoem.ecx_FOEread(self._ecx_contextt, self._pos, filename.encode('utf8'), password, &size_inout, pbuf, timeout)             # <<<<<<<<<<<<<<
 * 
 *         # error handling
*/
  /*else*/ {
    __pyx_t_2 = __pyx_v_filename;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_utf8};
      __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_t_9 = __Pyx_PyObject_AsWritableString(__pyx_t_4); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 1184, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyLong_As_uint32_t(__pyx_v_password); if (unlikely((__pyx_t_10 == ((uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1184, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyLong_As_int(__pyx_v_timeout); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1184, __pyx_L1_error)
    __pyx_v_result = ecx_FOEread(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, __pyx_t_9, __pyx_t_10, (&__pyx_v_size_inout), __pyx_v_pbuf, __pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L4:;

  /* "pysoem/pysoem.pyx":1188
 *         # error handling
 *         cdef cpysoem.ec_errort err
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):             # <<<<<<<<<<<<<<
 *             PyMem_Free(pbuf)
 *             assert err.Slave == self._pos
*/
  __pyx_t_3 = (ecx_poperror(__pyx_v_self->_ecx_contextt, (&__pyx_v_err)) != 0);
  if (__pyx_t_3) {

    /* "pysoem/pysoem.pyx":1189
 *         cdef cpysoem.ec_errort err
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):
 *             PyMem_Free(pbuf)             # <<<<<<<<<<<<<<
 *             assert err.Slave == self._pos
 *             self._raise_exception(&err)
*/
    PyMem_Free(__pyx_v_pbuf);

    /* "pysoem/pysoem.pyx":1190
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):
 *             PyMem_Free(pbuf)
 *             assert err.Slave == self._pos             # <<<<<<<<<<<<<<
 *             self._raise_exception(&err)
 * 
*/
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_3 = (__pyx_v_err.Slave == __pyx_v_self->_pos);
      if (unlikely(!__pyx_t_3)) {
        __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), 0, 0, 0);
        __PYX_ERR(0, 1190, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 1190, __pyx_L1_error)
    #endif

    /* "pysoem/pysoem.pyx":1191
 *             PyMem_Free(pbuf)
 *             assert err.Slave == self._pos
 *             self._raise_exception(&err)             # <<<<<<<<<<<<<<
 * 
 *         # return data
*/
    __pyx_t_4 = ((struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave *)__pyx_v_self->__pyx_vtab)->_raise_exception(__pyx_v_self, (&__pyx_v_err)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pysoem/pysoem.pyx":1188
 *         # error handling
 *         cdef cpysoem.ec_errort err
 *         if cpysoem.ecx_poperror(self._ecx_contextt, &err):             # <<<<<<<<<<<<<<
 *             PyMem_Free(pbuf)
 *             assert err.Slave == self._pos
*/
  }

  /* "pysoem/pysoem.pyx":1194
 * 
 *         # return data
 *         try:             # <<<<<<<<<<<<<<
 *             return PyBytes_FromStringAndSize(<char*>pbuf, size_inout)
 *         finally:
*/
  /*try:*/ {

    /* "pysoem/pysoem.pyx":1195
 *         # return data
 *         try:
 *             return PyBytes_FromStringAndSize(<char*>pbuf, size_inout)             # <<<<<<<<<<<<<<
 *         finally:
 *             PyMem_Free(pbuf)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = PyBytes_FromStringAndSize(((char *)__pyx_v_pbuf), __pyx_v_size_inout); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1195, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L6_return;
  }

  /* "pysoem/pysoem.pyx":1197
 *             return PyBytes_FromStringAndSize(<char*>pbuf, size_inout)
 *         finally:
 *             PyMem_Free(pbuf)             # <<<<<<<<<<<<<<
 * 
 *     def amend_mbx(self, mailbox, start_address, size):
*/
  /*finally:*/ {
    __pyx_L7_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      if ( unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_8 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_pbuf);
      }
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_11;
      goto __pyx_L1_error;
    }
    __pyx_L6_return: {
      __pyx_t_17 = __pyx_r;
      __pyx_r = 0;
      PyMem_Free(__pyx_v_pbuf);
      __pyx_r = __pyx_t_17;
      __pyx_t_17 = 0;
      goto __pyx_L0;
    }
  }

  /* "pysoem/pysoem.pyx":1158
 *         return result
 * 
 *     def foe_read(self, filename, password, size, timeout = 200000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Read given filename from device using FoE
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.foe_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_release_gil);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1199
 *             PyMem_Free(pbuf)
 * 
 *     def amend_mbx(self, mailbox, start_address, size):             # <<<<<<<<<<<<<<
 *         """Change the start address and size of a mailbox.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_27amend_mbx(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_26amend_mbx, "Change the start address and size of a mailbox.\n\n        Note that the slave must me in INIT state to do that.\n\n        :param str mailbox: Ether 'out', or 'in' to specify which mailbox to update.\n        :param int start_address: New start address for the mailbox.\n        :param int size: New size of the mailbox.\n\n        .. versionadded:: 1.0.6\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_27amend_mbx = {"amend_mbx", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_27amend_mbx, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_26amend_mbx};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_27amend_mbx(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_mailbox = 0;
  PyObject *__pyx_v_start_address = 0;
  PyObject *__pyx_v_size = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("amend_mbx (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mailbox,&__pyx_mstate_global->__pyx_n_u_start_address,&__pyx_mstate_global->__pyx_n_u_size,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1199, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1199, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1199, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1199, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "amend_mbx", 0) < (0)) __PYX_ERR(0, 1199, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("amend_mbx", 1, 3, 3, i); __PYX_ERR(0, 1199, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1199, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1199, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1199, __pyx_L3_error)
    }
    __pyx_v_mailbox = values[0];
    __pyx_v_start_address = values[1];
    __pyx_v_size = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("amend_mbx", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1199, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.amend_mbx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_26amend_mbx(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_mailbox, __pyx_v_start_address, __pyx_v_size);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_26amend_mbx(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_mailbox, PyObject *__pyx_v_start_address, PyObject *__pyx_v_size) {
  long __pyx_v_fpwr_timeout_us;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  uint16 __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("amend_mbx", 0);

  /* "pysoem/pysoem.pyx":1210
 *         .. versionadded:: 1.0.6
 *         """
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 * 
 *         fpwr_timeout_us = 4000
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1212
 *         self._master.check_context_is_initialized()
 * 
 *         fpwr_timeout_us = 4000             # <<<<<<<<<<<<<<
 *         if mailbox == 'out':
 *             # Clear the slaves mailbox configuration.
*/
  __pyx_v_fpwr_timeout_us = 0xFA0;

  /* "pysoem/pysoem.pyx":1213
 * 
 *         fpwr_timeout_us = 4000
 *         if mailbox == 'out':             # <<<<<<<<<<<<<<
 *             # Clear the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM0, bytes(sizeof(self._ec_slave.SM[0])))
*/
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_mailbox, __pyx_mstate_global->__pyx_n_u_out, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1213, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "pysoem/pysoem.pyx":1215
 *         if mailbox == 'out':
 *             # Clear the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM0, bytes(sizeof(self._ec_slave.SM[0])))             # <<<<<<<<<<<<<<
 *             self._ec_slave.SM[0].StartAddr = start_address
 *             self._ec_slave.SM[0].SMlength = size
*/
    __pyx_t_2 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ECT_REG_SM0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = __Pyx_PyLong_FromSize_t((sizeof((__pyx_v_self->_ec_slave->SM[0])))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_8};
      __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __pyx_t_3 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_5, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_fpwr, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":1216
 *             # Clear the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM0, bytes(sizeof(self._ec_slave.SM[0])))
 *             self._ec_slave.SM[0].StartAddr = start_address             # <<<<<<<<<<<<<<
 *             self._ec_slave.SM[0].SMlength = size
 *             self._ec_slave.mbx_wo = start_address
*/
    __pyx_t_9 = __Pyx_PyLong_As_uint16_t(__pyx_v_start_address); if (unlikely((__pyx_t_9 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1216, __pyx_L1_error)
    (__pyx_v_self->_ec_slave->SM[0]).StartAddr = __pyx_t_9;

    /* "pysoem/pysoem.pyx":1217
 *             self._fpwr(ECT_REG_SM0, bytes(sizeof(self._ec_slave.SM[0])))
 *             self._ec_slave.SM[0].StartAddr = start_address
 *             self._ec_slave.SM[0].SMlength = size             # <<<<<<<<<<<<<<
 *             self._ec_slave.mbx_wo = start_address
 *             self._ec_slave.mbx_l = size
*/
    __pyx_t_9 = __Pyx_PyLong_As_uint16_t(__pyx_v_size); if (unlikely((__pyx_t_9 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1217, __pyx_L1_error)
    (__pyx_v_self->_ec_slave->SM[0]).SMlength = __pyx_t_9;

    /* "pysoem/pysoem.pyx":1218
 *             self._ec_slave.SM[0].StartAddr = start_address
 *             self._ec_slave.SM[0].SMlength = size
 *             self._ec_slave.mbx_wo = start_address             # <<<<<<<<<<<<<<
 *             self._ec_slave.mbx_l = size
 *             # Update the slaves mailbox configuration.
*/
    __pyx_t_9 = __Pyx_PyLong_As_uint16_t(__pyx_v_start_address); if (unlikely((__pyx_t_9 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1218, __pyx_L1_error)
    __pyx_v_self->_ec_slave->mbx_wo = __pyx_t_9;

    /* "pysoem/pysoem.pyx":1219
 *             self._ec_slave.SM[0].SMlength = size
 *             self._ec_slave.mbx_wo = start_address
 *             self._ec_slave.mbx_l = size             # <<<<<<<<<<<<<<
 *             # Update the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM0,
*/
    __pyx_t_9 = __Pyx_PyLong_As_uint16_t(__pyx_v_size); if (unlikely((__pyx_t_9 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1219, __pyx_L1_error)
    __pyx_v_self->_ec_slave->mbx_l = __pyx_t_9;

    /* "pysoem/pysoem.pyx":1221
 *             self._ec_slave.mbx_l = size
 *             # Update the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM0,             # <<<<<<<<<<<<<<
 *                        PyBytes_FromStringAndSize(<char*>&self._ec_slave.SM[0], sizeof(self._ec_slave.SM[0])),
 *                        fpwr_timeout_us)
*/
    __pyx_t_6 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ECT_REG_SM0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "pysoem/pysoem.pyx":1222
 *             # Update the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM0,
 *                        PyBytes_FromStringAndSize(<char*>&self._ec_slave.SM[0], sizeof(self._ec_slave.SM[0])),             # <<<<<<<<<<<<<<
 *                        fpwr_timeout_us)
 *         elif mailbox == 'in':
*/
    __pyx_t_2 = PyBytes_FromStringAndSize(((char *)(&(__pyx_v_self->_ec_slave->SM[0]))), (sizeof((__pyx_v_self->_ec_slave->SM[0])))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "pysoem/pysoem.pyx":1223
 *             self._fpwr(ECT_REG_SM0,
 *                        PyBytes_FromStringAndSize(<char*>&self._ec_slave.SM[0], sizeof(self._ec_slave.SM[0])),
 *                        fpwr_timeout_us)             # <<<<<<<<<<<<<<
 *         elif mailbox == 'in':
 *             # Clear the slaves mailbox configuration.
*/
    __pyx_t_8 = __Pyx_PyLong_From_long(__pyx_v_fpwr_timeout_us); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = 0;
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_6, __pyx_t_5, __pyx_t_2, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_fpwr, __pyx_callargs+__pyx_t_3, (4-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1221, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":1213
 * 
 *         fpwr_timeout_us = 4000
 *         if mailbox == 'out':             # <<<<<<<<<<<<<<
 *             # Clear the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM0, bytes(sizeof(self._ec_slave.SM[0])))
*/
    goto __pyx_L3;
  }

  /* "pysoem/pysoem.pyx":1224
 *                        PyBytes_FromStringAndSize(<char*>&self._ec_slave.SM[0], sizeof(self._ec_slave.SM[0])),
 *                        fpwr_timeout_us)
 *         elif mailbox == 'in':             # <<<<<<<<<<<<<<
 *             # Clear the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM1, bytes(sizeof(self._ec_slave.SM[1])))
*/
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_mailbox, __pyx_mstate_global->__pyx_n_u_in, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1224, __pyx_L1_error)
  if (likely(__pyx_t_4)) {

    /* "pysoem/pysoem.pyx":1226
 *         elif mailbox == 'in':
 *             # Clear the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM1, bytes(sizeof(self._ec_slave.SM[1])))             # <<<<<<<<<<<<<<
 *             self._ec_slave.SM[1].StartAddr = start_address
 *             self._ec_slave.SM[1].SMlength = size
*/
    __pyx_t_8 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ECT_REG_SM1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    __pyx_t_7 = __Pyx_PyLong_FromSize_t((sizeof((__pyx_v_self->_ec_slave->SM[1])))); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_7};
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __pyx_t_3 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_t_2, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_fpwr, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":1227
 *             # Clear the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM1, bytes(sizeof(self._ec_slave.SM[1])))
 *             self._ec_slave.SM[1].StartAddr = start_address             # <<<<<<<<<<<<<<
 *             self._ec_slave.SM[1].SMlength = size
 *             self._ec_slave.mbx_ro = start_address
*/
    __pyx_t_9 = __Pyx_PyLong_As_uint16_t(__pyx_v_start_address); if (unlikely((__pyx_t_9 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1227, __pyx_L1_error)
    (__pyx_v_self->_ec_slave->SM[1]).StartAddr = __pyx_t_9;

    /* "pysoem/pysoem.pyx":1228
 *             self._fpwr(ECT_REG_SM1, bytes(sizeof(self._ec_slave.SM[1])))
 *             self._ec_slave.SM[1].StartAddr = start_address
 *             self._ec_slave.SM[1].SMlength = size             # <<<<<<<<<<<<<<
 *             self._ec_slave.mbx_ro = start_address
 *             self._ec_slave.mbx_rl = size
*/
    __pyx_t_9 = __Pyx_PyLong_As_uint16_t(__pyx_v_size); if (unlikely((__pyx_t_9 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1228, __pyx_L1_error)
    (__pyx_v_self->_ec_slave->SM[1]).SMlength = __pyx_t_9;

    /* "pysoem/pysoem.pyx":1229
 *             self._ec_slave.SM[1].StartAddr = start_address
 *             self._ec_slave.SM[1].SMlength = size
 *             self._ec_slave.mbx_ro = start_address             # <<<<<<<<<<<<<<
 *             self._ec_slave.mbx_rl = size
 *             # Update the slaves mailbox configuration.
*/
    __pyx_t_9 = __Pyx_PyLong_As_uint16_t(__pyx_v_start_address); if (unlikely((__pyx_t_9 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1229, __pyx_L1_error)
    __pyx_v_self->_ec_slave->mbx_ro = __pyx_t_9;

    /* "pysoem/pysoem.pyx":1230
 *             self._ec_slave.SM[1].SMlength = size
 *             self._ec_slave.mbx_ro = start_address
 *             self._ec_slave.mbx_rl = size             # <<<<<<<<<<<<<<
 *             # Update the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM1,
*/
    __pyx_t_9 = __Pyx_PyLong_As_uint16_t(__pyx_v_size); if (unlikely((__pyx_t_9 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1230, __pyx_L1_error)
    __pyx_v_self->_ec_slave->mbx_rl = __pyx_t_9;

    /* "pysoem/pysoem.pyx":1232
 *             self._ec_slave.mbx_rl = size
 *             # Update the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM1,             # <<<<<<<<<<<<<<
 *                        PyBytes_FromStringAndSize(<char*>&self._ec_slave.SM[1], sizeof(self._ec_slave.SM[1])),
 *                        fpwr_timeout_us)
*/
    __pyx_t_5 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ECT_REG_SM1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "pysoem/pysoem.pyx":1233
 *             # Update the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM1,
 *                        PyBytes_FromStringAndSize(<char*>&self._ec_slave.SM[1], sizeof(self._ec_slave.SM[1])),             # <<<<<<<<<<<<<<
 *                        fpwr_timeout_us)
 *         else:
*/
    __pyx_t_8 = PyBytes_FromStringAndSize(((char *)(&(__pyx_v_self->_ec_slave->SM[1]))), (sizeof((__pyx_v_self->_ec_slave->SM[1])))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "pysoem/pysoem.pyx":1234
 *             self._fpwr(ECT_REG_SM1,
 *                        PyBytes_FromStringAndSize(<char*>&self._ec_slave.SM[1], sizeof(self._ec_slave.SM[1])),
 *                        fpwr_timeout_us)             # <<<<<<<<<<<<<<
 *         else:
 *             raise AttributeError()
*/
    __pyx_t_7 = __Pyx_PyLong_From_long(__pyx_v_fpwr_timeout_us); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = 0;
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_t_2, __pyx_t_8, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_fpwr, __pyx_callargs+__pyx_t_3, (4-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":1224
 *                        PyBytes_FromStringAndSize(<char*>&self._ec_slave.SM[0], sizeof(self._ec_slave.SM[0])),
 *                        fpwr_timeout_us)
 *         elif mailbox == 'in':             # <<<<<<<<<<<<<<
 *             # Clear the slaves mailbox configuration.
 *             self._fpwr(ECT_REG_SM1, bytes(sizeof(self._ec_slave.SM[1])))
*/
    goto __pyx_L3;
  }

  /* "pysoem/pysoem.pyx":1236
 *                        fpwr_timeout_us)
 *         else:
 *             raise AttributeError()             # <<<<<<<<<<<<<<
 * 
 *     def set_watchdog(self, wd_type, wd_time_ms):
*/
  /*else*/ {
    __pyx_t_7 = NULL;
    __pyx_t_3 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AttributeError)), __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1236, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pysoem/pysoem.pyx":1199
 *             PyMem_Free(pbuf)
 * 
 *     def amend_mbx(self, mailbox, start_address, size):             # <<<<<<<<<<<<<<
 *         """Change the start address and size of a mailbox.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.amend_mbx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1238
 *             raise AttributeError()
 * 
 *     def set_watchdog(self, wd_type, wd_time_ms):             # <<<<<<<<<<<<<<
 *         """Change the watchdog time of the PDI or Process Data watchdog.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_29set_watchdog(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_28set_watchdog, "Change the watchdog time of the PDI or Process Data watchdog.\n\n        .. warning:: This is experimental.\n\n        :param str wd_type: Ether 'pdi', or 'processdata' to specify the watchdog time to be updated.\n        :param float wd_time_ms: Watchdog time in ms.\n\n        At the default watchdog time divider the precision is 0.1 ms.\n\n        .. versionadded:: 1.0.6\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_29set_watchdog = {"set_watchdog", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_29set_watchdog, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_28set_watchdog};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_29set_watchdog(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_wd_type = 0;
  PyObject *__pyx_v_wd_time_ms = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_watchdog (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_wd_type,&__pyx_mstate_global->__pyx_n_u_wd_time_ms,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1238, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1238, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1238, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_watchdog", 0) < (0)) __PYX_ERR(0, 1238, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_watchdog", 1, 2, 2, i); __PYX_ERR(0, 1238, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1238, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1238, __pyx_L3_error)
    }
    __pyx_v_wd_type = values[0];
    __pyx_v_wd_time_ms = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_watchdog", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1238, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.set_watchdog", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_28set_watchdog(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_wd_type, __pyx_v_wd_time_ms);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_28set_watchdog(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_wd_type, PyObject *__pyx_v_wd_time_ms) {
  long __pyx_v_fprd_fpwr_timeout_us;
  PyObject *__pyx_v_wd_type_to_reg_map = NULL;
  PyObject *__pyx_v_wd_time_ms_limit = NULL;
  PyObject *__pyx_v_wd_div_ns = NULL;
  PyObject *__pyx_v_wd_time_reg = NULL;
  PyObject *__pyx_v_actual_wd_time_ms = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9[5];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_watchdog", 0);

  /* "pysoem/pysoem.pyx":1250
 *         .. versionadded:: 1.0.6
 *         """
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 * 
 *         fprd_fpwr_timeout_us = 4000
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1252
 *         self._master.check_context_is_initialized()
 * 
 *         fprd_fpwr_timeout_us = 4000             # <<<<<<<<<<<<<<
 *         wd_type_to_reg_map = {
 *             'pdi': ECT_REG_WD_TIME_PDI,
*/
  __pyx_v_fprd_fpwr_timeout_us = 0xFA0;

  /* "pysoem/pysoem.pyx":1254
 *         fprd_fpwr_timeout_us = 4000
 *         wd_type_to_reg_map = {
 *             'pdi': ECT_REG_WD_TIME_PDI,             # <<<<<<<<<<<<<<
 *             'processdata': ECT_REG_WD_TIME_PROCESSDATA,
 *         }
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ECT_REG_WD_TIME_PDI); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pdi, __pyx_t_2) < (0)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":1255
 *         wd_type_to_reg_map = {
 *             'pdi': ECT_REG_WD_TIME_PDI,
 *             'processdata': ECT_REG_WD_TIME_PROCESSDATA,             # <<<<<<<<<<<<<<
 *         }
 *         if wd_type not in wd_type_to_reg_map.keys():
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ECT_REG_WD_TIME_PROCESSDATA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_processdata, __pyx_t_2) < (0)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_wd_type_to_reg_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1257
 *             'processdata': ECT_REG_WD_TIME_PROCESSDATA,
 *         }
 *         if wd_type not in wd_type_to_reg_map.keys():             # <<<<<<<<<<<<<<
 *             raise AttributeError()
 *         wd_time_ms_limit = self.get_max_watchdog_time()
*/
  __pyx_t_1 = __Pyx_PyDict_Keys(__pyx_v_wd_type_to_reg_map); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_wd_type, __pyx_t_1, Py_NE)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "pysoem/pysoem.pyx":1258
 *         }
 *         if wd_type not in wd_type_to_reg_map.keys():
 *             raise AttributeError()             # <<<<<<<<<<<<<<
 *         wd_time_ms_limit = self.get_max_watchdog_time()
 *         if wd_time_ms > wd_time_ms_limit:
*/
    __pyx_t_2 = NULL;
    __pyx_t_3 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AttributeError)), __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1258, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1257
 *             'processdata': ECT_REG_WD_TIME_PROCESSDATA,
 *         }
 *         if wd_type not in wd_type_to_reg_map.keys():             # <<<<<<<<<<<<<<
 *             raise AttributeError()
 *         wd_time_ms_limit = self.get_max_watchdog_time()
*/
  }

  /* "pysoem/pysoem.pyx":1259
 *         if wd_type not in wd_type_to_reg_map.keys():
 *             raise AttributeError()
 *         wd_time_ms_limit = self.get_max_watchdog_time()             # <<<<<<<<<<<<<<
 *         if wd_time_ms > wd_time_ms_limit:
 *             raise AttributeError('wd_time_ms is limited to {} ms'.format(wd_time_ms_limit))
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_max_watchdog_time, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_wd_time_ms_limit = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1260
 *             raise AttributeError()
 *         wd_time_ms_limit = self.get_max_watchdog_time()
 *         if wd_time_ms > wd_time_ms_limit:             # <<<<<<<<<<<<<<
 *             raise AttributeError('wd_time_ms is limited to {} ms'.format(wd_time_ms_limit))
 * 
*/
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_wd_time_ms, __pyx_v_wd_time_ms_limit, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "pysoem/pysoem.pyx":1261
 *         wd_time_ms_limit = self.get_max_watchdog_time()
 *         if wd_time_ms > wd_time_ms_limit:
 *             raise AttributeError('wd_time_ms is limited to {} ms'.format(wd_time_ms_limit))             # <<<<<<<<<<<<<<
 * 
 *         wd_div_ns = self._get_watchdog_divider_ns()
*/
    __pyx_t_2 = NULL;
    __pyx_t_6 = __pyx_mstate_global->__pyx_kp_u_wd_time_ms_is_limited_to_ms;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_3 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_wd_time_ms_limit};
      __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __pyx_t_3 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AttributeError)), __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1261, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1260
 *             raise AttributeError()
 *         wd_time_ms_limit = self.get_max_watchdog_time()
 *         if wd_time_ms > wd_time_ms_limit:             # <<<<<<<<<<<<<<
 *             raise AttributeError('wd_time_ms is limited to {} ms'.format(wd_time_ms_limit))
 * 
*/
  }

  /* "pysoem/pysoem.pyx":1263
 *             raise AttributeError('wd_time_ms is limited to {} ms'.format(wd_time_ms_limit))
 * 
 *         wd_div_ns = self._get_watchdog_divider_ns()             # <<<<<<<<<<<<<<
 *         wd_time_reg = int((wd_time_ms*1000000.0) / wd_div_ns)
 * 
*/
  __pyx_t_5 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_watchdog_divider_ns, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_wd_div_ns = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1264
 * 
 *         wd_div_ns = self._get_watchdog_divider_ns()
 *         wd_time_reg = int((wd_time_ms*1000000.0) / wd_div_ns)             # <<<<<<<<<<<<<<
 * 
 *         self._fpwr(wd_type_to_reg_map[wd_type],
*/
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_wd_time_ms, __pyx_mstate_global->__pyx_float_1000000_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_v_wd_div_ns); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_wd_time_reg = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1266
 *         wd_time_reg = int((wd_time_ms*1000000.0) / wd_div_ns)
 * 
 *         self._fpwr(wd_type_to_reg_map[wd_type],             # <<<<<<<<<<<<<<
 *                    wd_time_reg.to_bytes(2, byteorder='little', signed=False),
 *                    fprd_fpwr_timeout_us)
*/
  __pyx_t_5 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_wd_type_to_reg_map, __pyx_v_wd_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "pysoem/pysoem.pyx":1267
 * 
 *         self._fpwr(wd_type_to_reg_map[wd_type],
 *                    wd_time_reg.to_bytes(2, byteorder='little', signed=False),             # <<<<<<<<<<<<<<
 *                    fprd_fpwr_timeout_us)
 * 
*/
  __pyx_t_7 = __pyx_v_wd_time_reg;
  __Pyx_INCREF(__pyx_t_7);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_7, __pyx_mstate_global->__pyx_int_2};
    __pyx_t_8 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_byteorder, __pyx_mstate_global->__pyx_n_u_little, __pyx_t_8, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1267, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_signed, Py_False, __pyx_t_8, __pyx_callargs+2, 1) < (0)) __PYX_ERR(0, 1267, __pyx_L1_error)
    __pyx_t_6 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_to_bytes, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }

  /* "pysoem/pysoem.pyx":1268
 *         self._fpwr(wd_type_to_reg_map[wd_type],
 *                    wd_time_reg.to_bytes(2, byteorder='little', signed=False),
 *                    fprd_fpwr_timeout_us)             # <<<<<<<<<<<<<<
 * 
 *         actual_wd_time_ms = wd_time_reg * wd_div_ns / 1000000.0
*/
  __pyx_t_8 = __Pyx_PyLong_From_long(__pyx_v_fprd_fpwr_timeout_us); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_t_2, __pyx_t_6, __pyx_t_8};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_fpwr, __pyx_callargs+__pyx_t_3, (4-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1270
 *                    fprd_fpwr_timeout_us)
 * 
 *         actual_wd_time_ms = wd_time_reg * wd_div_ns / 1000000.0             # <<<<<<<<<<<<<<
 * 
 *         if actual_wd_time_ms != wd_time_ms:
*/
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_wd_time_reg, __pyx_v_wd_div_ns); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyFloat_DivideObjC(__pyx_t_1, __pyx_mstate_global->__pyx_float_1000000_0, 1000000.0, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_actual_wd_time_ms = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "pysoem/pysoem.pyx":1272
 *         actual_wd_time_ms = wd_time_reg * wd_div_ns / 1000000.0
 * 
 *         if actual_wd_time_ms != wd_time_ms:             # <<<<<<<<<<<<<<
 *             warnings.warn(f'The actual set watchdog time ({actual_wd_time_ms} ms) differs from the requested watchdog time ({wd_time_ms} ms) due to resolution limits of the hardware!', UserWarning)
 * 
*/
  __pyx_t_8 = PyObject_RichCompare(__pyx_v_actual_wd_time_ms, __pyx_v_wd_time_ms, Py_NE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__pyx_t_4) {

    /* "pysoem/pysoem.pyx":1273
 * 
 *         if actual_wd_time_ms != wd_time_ms:
 *             warnings.warn(f'The actual set watchdog time ({actual_wd_time_ms} ms) differs from the requested watchdog time ({wd_time_ms} ms) due to resolution limits of the hardware!', UserWarning)             # <<<<<<<<<<<<<<
 * 
 *     def get_watchdog(self, wd_type):
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_warnings); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_actual_wd_time_ms, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_v_wd_time_ms, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9[0] = __pyx_mstate_global->__pyx_kp_u_The_actual_set_watchdog_time;
    __pyx_t_9[1] = __pyx_t_6;
    __pyx_t_9[2] = __pyx_mstate_global->__pyx_kp_u_ms_differs_from_the_requested_w;
    __pyx_t_9[3] = __pyx_t_5;
    __pyx_t_9[4] = __pyx_mstate_global->__pyx_kp_u_ms_due_to_resolution_limits_of;
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_9, 5, 30 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + 47 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5) + 46, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5));
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_t_7, ((PyObject *)(((PyTypeObject*)PyExc_UserWarning)))};
      __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "pysoem/pysoem.pyx":1272
 *         actual_wd_time_ms = wd_time_reg * wd_div_ns / 1000000.0
 * 
 *         if actual_wd_time_ms != wd_time_ms:             # <<<<<<<<<<<<<<
 *             warnings.warn(f'The actual set watchdog time ({actual_wd_time_ms} ms) differs from the requested watchdog time ({wd_time_ms} ms) due to resolution limits of the hardware!', UserWarning)
 * 
*/
  }

  /* "pysoem/pysoem.pyx":1238
 *             raise AttributeError()
 * 
 *     def set_watchdog(self, wd_type, wd_time_ms):             # <<<<<<<<<<<<<<
 *         """Change the watchdog time of the PDI or Process Data watchdog.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.set_watchdog", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wd_type_to_reg_map);
  __Pyx_XDECREF(__pyx_v_wd_time_ms_limit);
  __Pyx_XDECREF(__pyx_v_wd_div_ns);
  __Pyx_XDECREF(__pyx_v_wd_time_reg);
  __Pyx_XDECREF(__pyx_v_actual_wd_time_ms);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1275
 *             warnings.warn(f'The actual set watchdog time ({actual_wd_time_ms} ms) differs from the requested watchdog time ({wd_time_ms} ms) due to resolution limits of the hardware!', UserWarning)
 * 
 *     def get_watchdog(self, wd_type):             # <<<<<<<<<<<<<<
 *         """Get the watchdog time of the PDI or Process Data watchdog.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_31get_watchdog(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_30get_watchdog, "Get the watchdog time of the PDI or Process Data watchdog.\n\n        :param str wd_type: Either 'pdi', or 'processdata' to specify the watchdog time to be read.\n\n        Returns:\n            float: The watchdog time in ms.\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_31get_watchdog = {"get_watchdog", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_31get_watchdog, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_30get_watchdog};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_31get_watchdog(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_wd_type = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_watchdog (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_wd_type,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1275, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1275, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_watchdog", 0) < (0)) __PYX_ERR(0, 1275, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_watchdog", 1, 1, 1, i); __PYX_ERR(0, 1275, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1275, __pyx_L3_error)
    }
    __pyx_v_wd_type = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_watchdog", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1275, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.get_watchdog", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_30get_watchdog(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_wd_type);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_30get_watchdog(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_wd_type) {
  long __pyx_v_fprd_fpwr_timeout_us;
  PyObject *__pyx_v_wd_type_to_reg_map = NULL;
  PyObject *__pyx_v_wd_div_ns = NULL;
  PyObject *__pyx_v_wd_time_reg = NULL;
  PyObject *__pyx_v_wd_time_ms = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_watchdog", 0);

  /* "pysoem/pysoem.pyx":1283
 *             float: The watchdog time in ms.
 *         """
 *         fprd_fpwr_timeout_us = 4000             # <<<<<<<<<<<<<<
 *         wd_type_to_reg_map = {
 *             'pdi': ECT_REG_WD_TIME_PDI,
*/
  __pyx_v_fprd_fpwr_timeout_us = 0xFA0;

  /* "pysoem/pysoem.pyx":1285
 *         fprd_fpwr_timeout_us = 4000
 *         wd_type_to_reg_map = {
 *             'pdi': ECT_REG_WD_TIME_PDI,             # <<<<<<<<<<<<<<
 *             'processdata': ECT_REG_WD_TIME_PROCESSDATA,
 *         }
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ECT_REG_WD_TIME_PDI); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pdi, __pyx_t_2) < (0)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":1286
 *         wd_type_to_reg_map = {
 *             'pdi': ECT_REG_WD_TIME_PDI,
 *             'processdata': ECT_REG_WD_TIME_PROCESSDATA,             # <<<<<<<<<<<<<<
 *         }
 *         if wd_type not in wd_type_to_reg_map.keys():
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ECT_REG_WD_TIME_PROCESSDATA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_processdata, __pyx_t_2) < (0)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_wd_type_to_reg_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1288
 *             'processdata': ECT_REG_WD_TIME_PROCESSDATA,
 *         }
 *         if wd_type not in wd_type_to_reg_map.keys():             # <<<<<<<<<<<<<<
 *             raise AttributeError()
 *         wd_div_ns = self._get_watchdog_divider_ns()
*/
  __pyx_t_1 = __Pyx_PyDict_Keys(__pyx_v_wd_type_to_reg_map); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_wd_type, __pyx_t_1, Py_NE)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "pysoem/pysoem.pyx":1289
 *         }
 *         if wd_type not in wd_type_to_reg_map.keys():
 *             raise AttributeError()             # <<<<<<<<<<<<<<
 *         wd_div_ns = self._get_watchdog_divider_ns()
 *         wd_time_reg = int.from_bytes(self._fprd(wd_type_to_reg_map[wd_type], 2, fprd_fpwr_timeout_us),
*/
    __pyx_t_2 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_AttributeError)), __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1289, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1288
 *             'processdata': ECT_REG_WD_TIME_PROCESSDATA,
 *         }
 *         if wd_type not in wd_type_to_reg_map.keys():             # <<<<<<<<<<<<<<
 *             raise AttributeError()
 *         wd_div_ns = self._get_watchdog_divider_ns()
*/
  }

  /* "pysoem/pysoem.pyx":1290
 *         if wd_type not in wd_type_to_reg_map.keys():
 *             raise AttributeError()
 *         wd_div_ns = self._get_watchdog_divider_ns()             # <<<<<<<<<<<<<<
 *         wd_time_reg = int.from_bytes(self._fprd(wd_type_to_reg_map[wd_type], 2, fprd_fpwr_timeout_us),
 *                                         byteorder='little',
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_watchdog_divider_ns, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_wd_div_ns = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1291
 *             raise AttributeError()
 *         wd_div_ns = self._get_watchdog_divider_ns()
 *         wd_time_reg = int.from_bytes(self._fprd(wd_type_to_reg_map[wd_type], 2, fprd_fpwr_timeout_us),             # <<<<<<<<<<<<<<
 *                                         byteorder='little',
 *                                         signed=False)
*/
  __pyx_t_2 = ((PyObject *)(&PyLong_Type));
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_6 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_wd_type_to_reg_map, __pyx_v_wd_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyLong_From_long(__pyx_v_fprd_fpwr_timeout_us); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_6, __pyx_t_7, __pyx_mstate_global->__pyx_int_2, __pyx_t_8};
    __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_fprd, __pyx_callargs+__pyx_t_4, (4-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }

  /* "pysoem/pysoem.pyx":1293
 *         wd_time_reg = int.from_bytes(self._fprd(wd_type_to_reg_map[wd_type], 2, fprd_fpwr_timeout_us),
 *                                         byteorder='little',
 *                                         signed=False)             # <<<<<<<<<<<<<<
 *         wd_time_ms = wd_time_reg * wd_div_ns / 1000000.0
 *         return wd_time_ms
*/
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_2, __pyx_t_5};
    __pyx_t_8 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_byteorder, __pyx_mstate_global->__pyx_n_u_little, __pyx_t_8, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1291, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_signed, Py_False, __pyx_t_8, __pyx_callargs+2, 1) < (0)) __PYX_ERR(0, 1291, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_from_bytes, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_wd_time_reg = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1294
 *                                         byteorder='little',
 *                                         signed=False)
 *         wd_time_ms = wd_time_reg * wd_div_ns / 1000000.0             # <<<<<<<<<<<<<<
 *         return wd_time_ms
 * 
*/
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_wd_time_reg, __pyx_v_wd_div_ns); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyFloat_DivideObjC(__pyx_t_1, __pyx_mstate_global->__pyx_float_1000000_0, 1000000.0, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_wd_time_ms = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "pysoem/pysoem.pyx":1295
 *                                         signed=False)
 *         wd_time_ms = wd_time_reg * wd_div_ns / 1000000.0
 *         return wd_time_ms             # <<<<<<<<<<<<<<
 * 
 *     def _get_watchdog_divider_ns(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wd_time_ms);
  __pyx_r = __pyx_v_wd_time_ms;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1275
 *             warnings.warn(f'The actual set watchdog time ({actual_wd_time_ms} ms) differs from the requested watchdog time ({wd_time_ms} ms) due to resolution limits of the hardware!', UserWarning)
 * 
 *     def get_watchdog(self, wd_type):             # <<<<<<<<<<<<<<
 *         """Get the watchdog time of the PDI or Process Data watchdog.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.get_watchdog", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wd_type_to_reg_map);
  __Pyx_XDECREF(__pyx_v_wd_div_ns);
  __Pyx_XDECREF(__pyx_v_wd_time_reg);
  __Pyx_XDECREF(__pyx_v_wd_time_ms);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1297
 *         return wd_time_ms
 * 
 *     def _get_watchdog_divider_ns(self):             # <<<<<<<<<<<<<<
 *         fprd_fpwr_timeout_us = 4000
 *         wd_div_reg = int.from_bytes(self._fprd(ECT_REG_WD_DIV, 2, fprd_fpwr_timeout_us),
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_33_get_watchdog_divider_ns(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_33_get_watchdog_divider_ns = {"_get_watchdog_divider_ns", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_33_get_watchdog_divider_ns, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_33_get_watchdog_divider_ns(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_watchdog_divider_ns (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_watchdog_divider_ns", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_watchdog_divider_ns", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_32_get_watchdog_divider_ns(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_32_get_watchdog_divider_ns(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  long __pyx_v_fprd_fpwr_timeout_us;
  PyObject *__pyx_v_wd_div_reg = NULL;
  PyObject *__pyx_v_wd_div_ns = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_watchdog_divider_ns", 0);

  /* "pysoem/pysoem.pyx":1298
 * 
 *     def _get_watchdog_divider_ns(self):
 *         fprd_fpwr_timeout_us = 4000             # <<<<<<<<<<<<<<
 *         wd_div_reg = int.from_bytes(self._fprd(ECT_REG_WD_DIV, 2, fprd_fpwr_timeout_us),
 *                                     byteorder='little',
*/
  __pyx_v_fprd_fpwr_timeout_us = 0xFA0;

  /* "pysoem/pysoem.pyx":1299
 *     def _get_watchdog_divider_ns(self):
 *         fprd_fpwr_timeout_us = 4000
 *         wd_div_reg = int.from_bytes(self._fprd(ECT_REG_WD_DIV, 2, fprd_fpwr_timeout_us),             # <<<<<<<<<<<<<<
 *                                     byteorder='little',
 *                                     signed=False)
*/
  __pyx_t_2 = ((PyObject *)(&PyLong_Type));
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ECT_REG_WD_DIV); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyLong_From_long(__pyx_v_fprd_fpwr_timeout_us); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = 0;
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_int_2, __pyx_t_6};
    __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_fprd, __pyx_callargs+__pyx_t_7, (4-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }

  /* "pysoem/pysoem.pyx":1301
 *         wd_div_reg = int.from_bytes(self._fprd(ECT_REG_WD_DIV, 2, fprd_fpwr_timeout_us),
 *                                     byteorder='little',
 *                                     signed=False)             # <<<<<<<<<<<<<<
 *         wd_div_ns = 40 * (wd_div_reg + 2)
 *         return wd_div_ns
*/
  __pyx_t_7 = 0;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_2, __pyx_t_3};
    __pyx_t_6 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_byteorder, __pyx_mstate_global->__pyx_n_u_little, __pyx_t_6, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1299, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_signed, Py_False, __pyx_t_6, __pyx_callargs+2, 1) < (0)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_from_bytes, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_wd_div_reg = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1302
 *                                     byteorder='little',
 *                                     signed=False)
 *         wd_div_ns = 40 * (wd_div_reg + 2)             # <<<<<<<<<<<<<<
 *         return wd_div_ns
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_AddObjC(__pyx_v_wd_div_reg, __pyx_mstate_global->__pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyLong_MultiplyCObj(__pyx_mstate_global->__pyx_int_40, __pyx_t_1, 40, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_wd_div_ns = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "pysoem/pysoem.pyx":1303
 *                                     signed=False)
 *         wd_div_ns = 40 * (wd_div_reg + 2)
 *         return wd_div_ns             # <<<<<<<<<<<<<<
 * 
 *     def get_max_watchdog_time(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wd_div_ns);
  __pyx_r = __pyx_v_wd_div_ns;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1297
 *         return wd_time_ms
 * 
 *     def _get_watchdog_divider_ns(self):             # <<<<<<<<<<<<<<
 *         fprd_fpwr_timeout_us = 4000
 *         wd_div_reg = int.from_bytes(self._fprd(ECT_REG_WD_DIV, 2, fprd_fpwr_timeout_us),
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_watchdog_divider_ns", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wd_div_reg);
  __Pyx_XDECREF(__pyx_v_wd_div_ns);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1305
 *         return wd_div_ns
 * 
 *     def get_max_watchdog_time(self):             # <<<<<<<<<<<<<<
 *         """Get the maximum watchdog timeout.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_35get_max_watchdog_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_34get_max_watchdog_time, "Get the maximum watchdog timeout.\n        \n        Returns:\n            float: The maximum watchdog timeout in ms.\n        \n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_35get_max_watchdog_time = {"get_max_watchdog_time", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_35get_max_watchdog_time, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_34get_max_watchdog_time};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_35get_max_watchdog_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_max_watchdog_time (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_max_watchdog_time", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_max_watchdog_time", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_34get_max_watchdog_time(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_34get_max_watchdog_time(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_v_wd_div_ns = NULL;
  PyObject *__pyx_v_wd_time_ms_limit = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_max_watchdog_time", 0);

  /* "pysoem/pysoem.pyx":1312
 * 
 *         """
 *         wd_div_ns = self._get_watchdog_divider_ns()             # <<<<<<<<<<<<<<
 *         wd_time_ms_limit = 0xFFFF * wd_div_ns / 1000000.0
 *         return wd_time_ms_limit
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_watchdog_divider_ns, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_wd_div_ns = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1313
 *         """
 *         wd_div_ns = self._get_watchdog_divider_ns()
 *         wd_time_ms_limit = 0xFFFF * wd_div_ns / 1000000.0             # <<<<<<<<<<<<<<
 *         return wd_time_ms_limit
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_MultiplyCObj(__pyx_mstate_global->__pyx_int_65535, __pyx_v_wd_div_ns, 0xFFFF, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyFloat_DivideObjC(__pyx_t_1, __pyx_mstate_global->__pyx_float_1000000_0, 1000000.0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_wd_time_ms_limit = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":1314
 *         wd_div_ns = self._get_watchdog_divider_ns()
 *         wd_time_ms_limit = 0xFFFF * wd_div_ns / 1000000.0
 *         return wd_time_ms_limit             # <<<<<<<<<<<<<<
 * 
 *     def add_emergency_callback(self, callback):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wd_time_ms_limit);
  __pyx_r = __pyx_v_wd_time_ms_limit;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1305
 *         return wd_div_ns
 * 
 *     def get_max_watchdog_time(self):             # <<<<<<<<<<<<<<
 *         """Get the maximum watchdog timeout.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.get_max_watchdog_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wd_div_ns);
  __Pyx_XDECREF(__pyx_v_wd_time_ms_limit);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1316
 *         return wd_time_ms_limit
 * 
 *     def add_emergency_callback(self, callback):             # <<<<<<<<<<<<<<
 *         """Get notified on EMCY messages from this slave.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_37add_emergency_callback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_36add_emergency_callback, "Get notified on EMCY messages from this slave.\n\n        :param callback:\n            Callable which must take one argument of an\n            :class:`~Emergency` instance.\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_37add_emergency_callback = {"add_emergency_callback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_37add_emergency_callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_36add_emergency_callback};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_37add_emergency_callback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_callback = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_emergency_callback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_callback,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1316, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1316, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "add_emergency_callback", 0) < (0)) __PYX_ERR(0, 1316, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("add_emergency_callback", 1, 1, 1, i); __PYX_ERR(0, 1316, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1316, __pyx_L3_error)
    }
    __pyx_v_callback = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_emergency_callback", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1316, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.add_emergency_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_36add_emergency_callback(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_callback);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_36add_emergency_callback(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_callback) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_emergency_callback", 0);

  /* "pysoem/pysoem.pyx":1323
 *             :class:`~Emergency` instance.
 *         """
 *         self._master.check_context_is_initialized()             # <<<<<<<<<<<<<<
 *         self._emcy_callbacks.append(callback)
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_master);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1324
 *         """
 *         self._master.check_context_is_initialized()
 *         self._emcy_callbacks.append(callback)             # <<<<<<<<<<<<<<
 * 
 *     cdef _on_emergency(self, cpysoem.ec_errort* emcy):
*/
  __pyx_t_4 = __Pyx_PyObject_Append(__pyx_v_self->_emcy_callbacks, __pyx_v_callback); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1324, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":1316
 *         return wd_time_ms_limit
 * 
 *     def add_emergency_callback(self, callback):             # <<<<<<<<<<<<<<
 *         """Get notified on EMCY messages from this slave.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.add_emergency_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1326
 *         self._emcy_callbacks.append(callback)
 * 
 *     cdef _on_emergency(self, cpysoem.ec_errort* emcy):             # <<<<<<<<<<<<<<
 *         """Notify all emergency callbacks that an emergency message
 *         was received.
*/

static PyObject *__pyx_f_6pysoem_6pysoem_9CdefSlave__on_emergency(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, ec_errort *__pyx_v_emcy) {
  PyObject *__pyx_v_emergency_msg = NULL;
  PyObject *__pyx_v_callback = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_emergency", 0);

  /* "pysoem/pysoem.pyx":1332
 *         :param emcy: Emergency object.
 *         """
 *         emergency_msg = Emergency(emcy.Slave,             # <<<<<<<<<<<<<<
 *                                emcy.ErrorCode,
 *                                emcy.ErrorReg,
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Emergency); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyLong_From_uint16_t(__pyx_v_emcy->Slave); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pysoem/pysoem.pyx":1333
 *         """
 *         emergency_msg = Emergency(emcy.Slave,
 *                                emcy.ErrorCode,             # <<<<<<<<<<<<<<
 *                                emcy.ErrorReg,
 *                                emcy.b1,
*/
  __pyx_t_5 = __Pyx_PyLong_From_uint16_t(__pyx_v_emcy->ErrorCode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pysoem/pysoem.pyx":1334
 *         emergency_msg = Emergency(emcy.Slave,
 *                                emcy.ErrorCode,
 *                                emcy.ErrorReg,             # <<<<<<<<<<<<<<
 *                                emcy.b1,
 *                                emcy.w1,
*/
  __pyx_t_6 = __Pyx_PyLong_From_uint8_t(__pyx_v_emcy->ErrorReg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "pysoem/pysoem.pyx":1335
 *                                emcy.ErrorCode,
 *                                emcy.ErrorReg,
 *                                emcy.b1,             # <<<<<<<<<<<<<<
 *                                emcy.w1,
 *                                emcy.w2)
*/
  __pyx_t_7 = __Pyx_PyLong_From_uint8_t(__pyx_v_emcy->b1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "pysoem/pysoem.pyx":1336
 *                                emcy.ErrorReg,
 *                                emcy.b1,
 *                                emcy.w1,             # <<<<<<<<<<<<<<
 *                                emcy.w2)
 *         for callback in self._emcy_callbacks:
*/
  __pyx_t_8 = __Pyx_PyLong_From_uint16_t(__pyx_v_emcy->w1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pysoem/pysoem.pyx":1337
 *                                emcy.b1,
 *                                emcy.w1,
 *                                emcy.w2)             # <<<<<<<<<<<<<<
 *         for callback in self._emcy_callbacks:
 *             callback(emergency_msg)
*/
  __pyx_t_9 = __Pyx_PyLong_From_uint16_t(__pyx_v_emcy->w2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_10 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[7] = {__pyx_t_2, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_10, (7-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_emergency_msg = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1338
 *                                emcy.w1,
 *                                emcy.w2)
 *         for callback in self._emcy_callbacks:             # <<<<<<<<<<<<<<
 *             callback(emergency_msg)
 * 
*/
  if (likely(PyList_CheckExact(__pyx_v_self->_emcy_callbacks)) || PyTuple_CheckExact(__pyx_v_self->_emcy_callbacks)) {
    __pyx_t_1 = __pyx_v_self->_emcy_callbacks; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_self->_emcy_callbacks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1338, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1338, __pyx_L1_error)
          #endif
          if (__pyx_t_11 >= __pyx_temp) break;
        }
        __pyx_t_3 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_11, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_11;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1338, __pyx_L1_error)
          #endif
          if (__pyx_t_11 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_11));
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_11);
        #endif
        ++__pyx_t_11;
      }
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1338, __pyx_L1_error)
    } else {
      __pyx_t_3 = __pyx_t_12(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1338, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_callback, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pysoem/pysoem.pyx":1339
 *                                emcy.w2)
 *         for callback in self._emcy_callbacks:
 *             callback(emergency_msg)             # <<<<<<<<<<<<<<
 * 
 *     def _disable_complete_access(self):
*/
    __pyx_t_9 = NULL;
    __Pyx_INCREF(__pyx_v_callback);
    __pyx_t_8 = __pyx_v_callback; 
    __pyx_t_10 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      assert(__pyx_t_9);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
      __pyx_t_10 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_emergency_msg};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pysoem/pysoem.pyx":1338
 *                                emcy.w1,
 *                                emcy.w2)
 *         for callback in self._emcy_callbacks:             # <<<<<<<<<<<<<<
 *             callback(emergency_msg)
 * 
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1326
 *         self._emcy_callbacks.append(callback)
 * 
 *     cdef _on_emergency(self, cpysoem.ec_errort* emcy):             # <<<<<<<<<<<<<<
 *         """Notify all emergency callbacks that an emergency message
 *         was received.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._on_emergency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_emergency_msg);
  __Pyx_XDECREF(__pyx_v_callback);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1341
 *             callback(emergency_msg)
 * 
 *     def _disable_complete_access(self):             # <<<<<<<<<<<<<<
 *         """Helper function that stops config_map() from using "complete access" for SDO requests for this device.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_39_disable_complete_access(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_38_disable_complete_access, "Helper function that stops config_map() from using \"complete access\" for SDO requests for this device.\n\n        This should only be used if your device has issues handling complete access requests but the CoE details of the\n        SII tells that SDO complete access is supported by the device. If you need this function something is wrong\n        with your device and you should contact the manufacturer about this issue.\n\n        .. warning:: This is experimental.\n\n        .. versionadded:: 1.1.3\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_39_disable_complete_access = {"_disable_complete_access", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_39_disable_complete_access, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_38_disable_complete_access};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_39_disable_complete_access(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_disable_complete_access (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_disable_complete_access", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_disable_complete_access", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_38_disable_complete_access(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_38_disable_complete_access(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8 __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_disable_complete_access", 0);

  /* "pysoem/pysoem.pyx":1352
 *         .. versionadded:: 1.1.3
 *         """
 *         self._ec_slave.CoEdetails &= ~ECT_COEDET_SDOCA             # <<<<<<<<<<<<<<
 * 
 *     def _fprd(self, int address, int size, timeout_us=2000):
*/
  __pyx_t_1 = __Pyx_PyLong_From_uint8_t(__pyx_v_self->_ec_slave->CoEdetails); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ECT_COEDET_SDOCA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Invert(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_InPlaceAnd(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyLong_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_4 == ((uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->_ec_slave->CoEdetails = __pyx_t_4;

  /* "pysoem/pysoem.pyx":1341
 *             callback(emergency_msg)
 * 
 *     def _disable_complete_access(self):             # <<<<<<<<<<<<<<
 *         """Helper function that stops config_map() from using "complete access" for SDO requests for this device.
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._disable_complete_access", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1354
 *         self._ec_slave.CoEdetails &= ~ECT_COEDET_SDOCA
 * 
 *     def _fprd(self, int address, int size, timeout_us=2000):             # <<<<<<<<<<<<<<
 *         """Send and receive of the FPRD cmd primitive (Configured Address Physical Read)."""
 *         cdef unsigned char* data
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_41_fprd(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_40_fprd, "Send and receive of the FPRD cmd primitive (Configured Address Physical Read).");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_41_fprd = {"_fprd", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_41_fprd, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_40_fprd};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_41_fprd(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_address;
  int __pyx_v_size;
  PyObject *__pyx_v_timeout_us = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_fprd (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_address,&__pyx_mstate_global->__pyx_n_u_size,&__pyx_mstate_global->__pyx_n_u_timeout_us,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1354, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1354, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1354, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1354, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_fprd", 0) < (0)) __PYX_ERR(0, 1354, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_fprd", 0, 2, 3, i); __PYX_ERR(0, 1354, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1354, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1354, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1354, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
    }
    __pyx_v_address = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_address == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1354, __pyx_L3_error)
    __pyx_v_size = __Pyx_PyLong_As_int(values[1]); if (unlikely((__pyx_v_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1354, __pyx_L3_error)
    __pyx_v_timeout_us = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fprd", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 1354, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._fprd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_40_fprd(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_address, __pyx_v_size, __pyx_v_timeout_us);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_40_fprd(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, int __pyx_v_address, int __pyx_v_size, PyObject *__pyx_v_timeout_us) {
  unsigned char *__pyx_v_data;
  int __pyx_v_wkc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_fprd", 0);

  /* "pysoem/pysoem.pyx":1357
 *         """Send and receive of the FPRD cmd primitive (Configured Address Physical Read)."""
 *         cdef unsigned char* data
 *         data = <unsigned char*>PyMem_Malloc(size)             # <<<<<<<<<<<<<<
 *         cdef int wkc = cpysoem.ecx_FPRD(self._ecx_contextt.port, self._ec_slave.configadr, address, size, data, timeout_us)
 *         if wkc != 1:
*/
  __pyx_v_data = ((unsigned char *)PyMem_Malloc(__pyx_v_size));

  /* "pysoem/pysoem.pyx":1358
 *         cdef unsigned char* data
 *         data = <unsigned char*>PyMem_Malloc(size)
 *         cdef int wkc = cpysoem.ecx_FPRD(self._ecx_contextt.port, self._ec_slave.configadr, address, size, data, timeout_us)             # <<<<<<<<<<<<<<
 *         if wkc != 1:
 *             PyMem_Free(data)
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_timeout_us); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1358, __pyx_L1_error)
  __pyx_v_wkc = ecx_FPRD(__pyx_v_self->_ecx_contextt->port, __pyx_v_self->_ec_slave->configadr, __pyx_v_address, __pyx_v_size, __pyx_v_data, __pyx_t_1);

  /* "pysoem/pysoem.pyx":1359
 *         data = <unsigned char*>PyMem_Malloc(size)
 *         cdef int wkc = cpysoem.ecx_FPRD(self._ecx_contextt.port, self._ec_slave.configadr, address, size, data, timeout_us)
 *         if wkc != 1:             # <<<<<<<<<<<<<<
 *             PyMem_Free(data)
 *             raise WkcError()
*/
  __pyx_t_2 = (__pyx_v_wkc != 1);
  if (unlikely(__pyx_t_2)) {

    /* "pysoem/pysoem.pyx":1360
 *         cdef int wkc = cpysoem.ecx_FPRD(self._ecx_contextt.port, self._ec_slave.configadr, address, size, data, timeout_us)
 *         if wkc != 1:
 *             PyMem_Free(data)             # <<<<<<<<<<<<<<
 *             raise WkcError()
 *         try:
*/
    PyMem_Free(__pyx_v_data);

    /* "pysoem/pysoem.pyx":1361
 *         if wkc != 1:
 *             PyMem_Free(data)
 *             raise WkcError()             # <<<<<<<<<<<<<<
 *         try:
 *             return PyBytes_FromStringAndSize(<char*>data, size)
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_WkcError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1361, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1361, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1359
 *         data = <unsigned char*>PyMem_Malloc(size)
 *         cdef int wkc = cpysoem.ecx_FPRD(self._ecx_contextt.port, self._ec_slave.configadr, address, size, data, timeout_us)
 *         if wkc != 1:             # <<<<<<<<<<<<<<
 *             PyMem_Free(data)
 *             raise WkcError()
*/
  }

  /* "pysoem/pysoem.pyx":1362
 *             PyMem_Free(data)
 *             raise WkcError()
 *         try:             # <<<<<<<<<<<<<<
 *             return PyBytes_FromStringAndSize(<char*>data, size)
 *         finally:
*/
  /*try:*/ {

    /* "pysoem/pysoem.pyx":1363
 *             raise WkcError()
 *         try:
 *             return PyBytes_FromStringAndSize(<char*>data, size)             # <<<<<<<<<<<<<<
 *         finally:
 *             PyMem_Free(data)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyBytes_FromStringAndSize(((char *)__pyx_v_data), __pyx_v_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1363, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L4_return;
  }

  /* "pysoem/pysoem.pyx":1365
 *             return PyBytes_FromStringAndSize(<char*>data, size)
 *         finally:
 *             PyMem_Free(data)             # <<<<<<<<<<<<<<
 * 
 *     def _fpwr(self, int address, bytes data, timeout_us=2000):
*/
  /*finally:*/ {
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ( unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_1 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_data);
      }
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_1; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
    }
    __pyx_L4_return: {
      __pyx_t_14 = __pyx_r;
      __pyx_r = 0;
      PyMem_Free(__pyx_v_data);
      __pyx_r = __pyx_t_14;
      __pyx_t_14 = 0;
      goto __pyx_L0;
    }
  }

  /* "pysoem/pysoem.pyx":1354
 *         self._ec_slave.CoEdetails &= ~ECT_COEDET_SDOCA
 * 
 *     def _fprd(self, int address, int size, timeout_us=2000):             # <<<<<<<<<<<<<<
 *         """Send and receive of the FPRD cmd primitive (Configured Address Physical Read)."""
 *         cdef unsigned char* data
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._fprd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1367
 *             PyMem_Free(data)
 * 
 *     def _fpwr(self, int address, bytes data, timeout_us=2000):             # <<<<<<<<<<<<<<
 *         """Send and receive of the FPWR cmd primitive (Configured Address Physical Write)."""
 *         cdef int wkc = cpysoem.ecx_FPWR(self._ecx_contextt.port, self._ec_slave.configadr, address, <int>len(data), <unsigned char*>data, timeout_us)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_43_fpwr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_42_fpwr, "Send and receive of the FPWR cmd primitive (Configured Address Physical Write).");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_43_fpwr = {"_fpwr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_43_fpwr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_42_fpwr};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_43_fpwr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_address;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_timeout_us = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_fpwr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_address,&__pyx_mstate_global->__pyx_n_u_data,&__pyx_mstate_global->__pyx_n_u_timeout_us,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1367, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1367, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1367, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1367, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_fpwr", 0) < (0)) __PYX_ERR(0, 1367, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_fpwr", 0, 2, 3, i); __PYX_ERR(0, 1367, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1367, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1367, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1367, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
    }
    __pyx_v_address = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_address == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1367, __pyx_L3_error)
    __pyx_v_data = ((PyObject*)values[1]);
    __pyx_v_timeout_us = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_fpwr", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 1367, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._fpwr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), (&PyBytes_Type), 1, "data", 1))) __PYX_ERR(0, 1367, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_42_fpwr(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_address, __pyx_v_data, __pyx_v_timeout_us);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_42_fpwr(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, int __pyx_v_address, PyObject *__pyx_v_data, PyObject *__pyx_v_timeout_us) {
  int __pyx_v_wkc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  unsigned char *__pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_fpwr", 0);

  /* "pysoem/pysoem.pyx":1369
 *     def _fpwr(self, int address, bytes data, timeout_us=2000):
 *         """Send and receive of the FPWR cmd primitive (Configured Address Physical Write)."""
 *         cdef int wkc = cpysoem.ecx_FPWR(self._ecx_contextt.port, self._ec_slave.configadr, address, <int>len(data), <unsigned char*>data, timeout_us)             # <<<<<<<<<<<<<<
 *         if wkc != 1:
 *             raise WkcError()
*/
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1369, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBytes_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1369, __pyx_L1_error)
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1369, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableUString(__pyx_v_data); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1369, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyLong_As_int(__pyx_v_timeout_us); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1369, __pyx_L1_error)
  __pyx_v_wkc = ecx_FPWR(__pyx_v_self->_ecx_contextt->port, __pyx_v_self->_ec_slave->configadr, __pyx_v_address, ((int)__pyx_t_1), ((unsigned char *)__pyx_t_2), __pyx_t_3);

  /* "pysoem/pysoem.pyx":1370
 *         """Send and receive of the FPWR cmd primitive (Configured Address Physical Write)."""
 *         cdef int wkc = cpysoem.ecx_FPWR(self._ecx_contextt.port, self._ec_slave.configadr, address, <int>len(data), <unsigned char*>data, timeout_us)
 *         if wkc != 1:             # <<<<<<<<<<<<<<
 *             raise WkcError()
 * 
*/
  __pyx_t_4 = (__pyx_v_wkc != 1);
  if (unlikely(__pyx_t_4)) {

    /* "pysoem/pysoem.pyx":1371
 *         cdef int wkc = cpysoem.ecx_FPWR(self._ecx_contextt.port, self._ec_slave.configadr, address, <int>len(data), <unsigned char*>data, timeout_us)
 *         if wkc != 1:
 *             raise WkcError()             # <<<<<<<<<<<<<<
 * 
 *     cdef _raise_exception(self, cpysoem.ec_errort* err):
*/
    __pyx_t_6 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_WkcError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_6);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_8 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1371, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1370
 *         """Send and receive of the FPWR cmd primitive (Configured Address Physical Write)."""
 *         cdef int wkc = cpysoem.ecx_FPWR(self._ecx_contextt.port, self._ec_slave.configadr, address, <int>len(data), <unsigned char*>data, timeout_us)
 *         if wkc != 1:             # <<<<<<<<<<<<<<
 *             raise WkcError()
 * 
*/
  }

  /* "pysoem/pysoem.pyx":1367
 *             PyMem_Free(data)
 * 
 *     def _fpwr(self, int address, bytes data, timeout_us=2000):             # <<<<<<<<<<<<<<
 *         """Send and receive of the FPWR cmd primitive (Configured Address Physical Write)."""
 *         cdef int wkc = cpysoem.ecx_FPWR(self._ecx_contextt.port, self._ec_slave.configadr, address, <int>len(data), <unsigned char*>data, timeout_us)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._fpwr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1373
 *             raise WkcError()
 * 
 *     cdef _raise_exception(self, cpysoem.ec_errort* err):             # <<<<<<<<<<<<<<
 *         if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:
 *             raise SdoError(err.Slave,
*/

static PyObject *__pyx_f_6pysoem_6pysoem_9CdefSlave__raise_exception(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, ec_errort *__pyx_v_err) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  char const *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  size_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  char *__pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_raise_exception", 0);

  /* "pysoem/pysoem.pyx":1374
 * 
 *     cdef _raise_exception(self, cpysoem.ec_errort* err):
 *         if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:             # <<<<<<<<<<<<<<
 *             raise SdoError(err.Slave,
 *                            err.Index,
*/
  switch (__pyx_v_err->Etype) {
    case EC_ERR_TYPE_SDO_ERROR:

    /* "pysoem/pysoem.pyx":1375
 *     cdef _raise_exception(self, cpysoem.ec_errort* err):
 *         if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:
 *             raise SdoError(err.Slave,             # <<<<<<<<<<<<<<
 *                            err.Index,
 *                            err.SubIdx,
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_SdoError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyLong_From_uint16_t(__pyx_v_err->Slave); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "pysoem/pysoem.pyx":1376
 *         if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:
 *             raise SdoError(err.Slave,
 *                            err.Index,             # <<<<<<<<<<<<<<
 *                            err.SubIdx,
 *                            err.AbortCode,
*/
    __pyx_t_5 = __Pyx_PyLong_From_uint16_t(__pyx_v_err->Index); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "pysoem/pysoem.pyx":1377
 *             raise SdoError(err.Slave,
 *                            err.Index,
 *                            err.SubIdx,             # <<<<<<<<<<<<<<
 *                            err.AbortCode,
 *                            cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8'))
*/
    __pyx_t_6 = __Pyx_PyLong_From_uint8_t(__pyx_v_err->SubIdx); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "pysoem/pysoem.pyx":1378
 *                            err.Index,
 *                            err.SubIdx,
 *                            err.AbortCode,             # <<<<<<<<<<<<<<
 *                            cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8'))
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY:
*/
    __pyx_t_7 = __Pyx_PyLong_From_int32_t(__pyx_v_err->AbortCode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "pysoem/pysoem.pyx":1379
 *                            err.SubIdx,
 *                            err.AbortCode,
 *                            cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8'))             # <<<<<<<<<<<<<<
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY:
 *             warnings.warn('This way of catching emergency messages is deprecated, use the add_emergency_callback() function!', FutureWarning)
*/
    __pyx_t_8 = ec_sdoerror2string(__pyx_v_err->AbortCode);
    __pyx_t_9 = __Pyx_ssize_strlen(__pyx_t_8); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1379, __pyx_L1_error)
    __pyx_t_10 = __Pyx_decode_c_string(__pyx_t_8, 0, __pyx_t_9, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_11 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[6] = {__pyx_t_2, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_10};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_11, (6-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1375, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1375, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1374
 * 
 *     cdef _raise_exception(self, cpysoem.ec_errort* err):
 *         if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:             # <<<<<<<<<<<<<<
 *             raise SdoError(err.Slave,
 *                            err.Index,
*/
    break;
    case EC_ERR_TYPE_EMERGENCY:

    /* "pysoem/pysoem.pyx":1381
 *                            cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8'))
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY:
 *             warnings.warn('This way of catching emergency messages is deprecated, use the add_emergency_callback() function!', FutureWarning)             # <<<<<<<<<<<<<<
 *             raise Emergency(err.Slave,
 *                             err.ErrorCode,
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_warnings); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_warn); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_11 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_11 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_This_way_of_catching_emergency_m, ((PyObject *)(((PyTypeObject*)PyExc_FutureWarning)))};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_11, (3-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":1382
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY:
 *             warnings.warn('This way of catching emergency messages is deprecated, use the add_emergency_callback() function!', FutureWarning)
 *             raise Emergency(err.Slave,             # <<<<<<<<<<<<<<
 *                             err.ErrorCode,
 *                             err.ErrorReg,
*/
    __pyx_t_7 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Emergency); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyLong_From_uint16_t(__pyx_v_err->Slave); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);

    /* "pysoem/pysoem.pyx":1383
 *             warnings.warn('This way of catching emergency messages is deprecated, use the add_emergency_callback() function!', FutureWarning)
 *             raise Emergency(err.Slave,
 *                             err.ErrorCode,             # <<<<<<<<<<<<<<
 *                             err.ErrorReg,
 *                             err.b1,
*/
    __pyx_t_6 = __Pyx_PyLong_From_uint16_t(__pyx_v_err->ErrorCode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "pysoem/pysoem.pyx":1384
 *             raise Emergency(err.Slave,
 *                             err.ErrorCode,
 *                             err.ErrorReg,             # <<<<<<<<<<<<<<
 *                             err.b1,
 *                             err.w1,
*/
    __pyx_t_5 = __Pyx_PyLong_From_uint8_t(__pyx_v_err->ErrorReg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "pysoem/pysoem.pyx":1385
 *                             err.ErrorCode,
 *                             err.ErrorReg,
 *                             err.b1,             # <<<<<<<<<<<<<<
 *                             err.w1,
 *                             err.w2)
*/
    __pyx_t_4 = __Pyx_PyLong_From_uint8_t(__pyx_v_err->b1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "pysoem/pysoem.pyx":1386
 *                             err.ErrorReg,
 *                             err.b1,
 *                             err.w1,             # <<<<<<<<<<<<<<
 *                             err.w2)
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:
*/
    __pyx_t_2 = __Pyx_PyLong_From_uint16_t(__pyx_v_err->w1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "pysoem/pysoem.pyx":1387
 *                             err.b1,
 *                             err.w1,
 *                             err.w2)             # <<<<<<<<<<<<<<
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:
 *             raise MailboxError(err.Slave,
*/
    __pyx_t_12 = __Pyx_PyLong_From_uint16_t(__pyx_v_err->w2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_11 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_7);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_11 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[7] = {__pyx_t_7, __pyx_t_10, __pyx_t_6, __pyx_t_5, __pyx_t_4, __pyx_t_2, __pyx_t_12};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_11, (7-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1382, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1380
 *                            err.AbortCode,
 *                            cpysoem.ec_sdoerror2string(err.AbortCode).decode('utf8'))
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_EMERGENCY:             # <<<<<<<<<<<<<<
 *             warnings.warn('This way of catching emergency messages is deprecated, use the add_emergency_callback() function!', FutureWarning)
 *             raise Emergency(err.Slave,
*/
    break;
    case EC_ERR_TYPE_MBX_ERROR:

    /* "pysoem/pysoem.pyx":1389
 *                             err.w2)
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:
 *             raise MailboxError(err.Slave,             # <<<<<<<<<<<<<<
 *                                err.ErrorCode,
 *                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8'))
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_MailboxError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_2 = __Pyx_PyLong_From_uint16_t(__pyx_v_err->Slave); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "pysoem/pysoem.pyx":1390
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:
 *             raise MailboxError(err.Slave,
 *                                err.ErrorCode,             # <<<<<<<<<<<<<<
 *                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8'))
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:
*/
    __pyx_t_4 = __Pyx_PyLong_From_uint16_t(__pyx_v_err->ErrorCode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "pysoem/pysoem.pyx":1391
 *             raise MailboxError(err.Slave,
 *                                err.ErrorCode,
 *                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8'))             # <<<<<<<<<<<<<<
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:
 *             raise PacketError(err.Slave,
*/
    __pyx_t_13 = ec_mbxerror2string(__pyx_v_err->ErrorCode);
    __pyx_t_9 = __Pyx_ssize_strlen(__pyx_t_13); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1391, __pyx_L1_error)
    __pyx_t_5 = __Pyx_decode_c_string(__pyx_t_13, 0, __pyx_t_9, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_12);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_12);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_12, __pyx__function);
      __pyx_t_11 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_t_2, __pyx_t_4, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_11, (4-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1389, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1388
 *                             err.w1,
 *                             err.w2)
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_MBX_ERROR:             # <<<<<<<<<<<<<<
 *             raise MailboxError(err.Slave,
 *                                err.ErrorCode,
*/
    break;
    case EC_ERR_TYPE_PACKET_ERROR:

    /* "pysoem/pysoem.pyx":1393
 *                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8'))
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:
 *             raise PacketError(err.Slave,             # <<<<<<<<<<<<<<
 *                               err.ErrorCode)
 *         else:
*/
    __pyx_t_12 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PacketError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyLong_From_uint16_t(__pyx_v_err->Slave); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "pysoem/pysoem.pyx":1394
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:
 *             raise PacketError(err.Slave,
 *                               err.ErrorCode)             # <<<<<<<<<<<<<<
 *         else:
 *             raise Exception('unexpected error, Etype: {}'.format(err.Etype))
*/
    __pyx_t_2 = __Pyx_PyLong_From_uint16_t(__pyx_v_err->ErrorCode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_12);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_11 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_12, __pyx_t_4, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_11, (3-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1393, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1392
 *                                err.ErrorCode,
 *                                cpysoem.ec_mbxerror2string(err.ErrorCode).decode('utf8'))
 *         elif err.Etype == cpysoem.EC_ERR_TYPE_PACKET_ERROR:             # <<<<<<<<<<<<<<
 *             raise PacketError(err.Slave,
 *                               err.ErrorCode)
*/
    break;
    default:

    /* "pysoem/pysoem.pyx":1396
 *                               err.ErrorCode)
 *         else:
 *             raise Exception('unexpected error, Etype: {}'.format(err.Etype))             # <<<<<<<<<<<<<<
 * 
 *     def _get_name(self):
*/
    __pyx_t_5 = NULL;
    __pyx_t_4 = __pyx_mstate_global->__pyx_kp_u_unexpected_error_Etype;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_12 = __Pyx_PyLong_From_ec_err_type(__pyx_v_err->Etype); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_11 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_12};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_11 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_Exception)), __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1396, __pyx_L1_error)
    break;
  }

  /* "pysoem/pysoem.pyx":1373
 *             raise WkcError()
 * 
 *     cdef _raise_exception(self, cpysoem.ec_errort* err):             # <<<<<<<<<<<<<<
 *         if err.Etype == cpysoem.EC_ERR_TYPE_SDO_ERROR:
 *             raise SdoError(err.Slave,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._raise_exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1398
 *             raise Exception('unexpected error, Etype: {}'.format(err.Etype))
 * 
 *     def _get_name(self):             # <<<<<<<<<<<<<<
 *         """Name of the slave, read out from the slaves SII during config_init."""
 *         return (<bytes>self._ec_slave.name).decode('utf8')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_45_get_name(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_44_get_name, "Name of the slave, read out from the slaves SII during config_init.");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_45_get_name = {"_get_name", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_45_get_name, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_44_get_name};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_45_get_name(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_name (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_name", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_name", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_44_get_name(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_44_get_name(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_name", 0);

  /* "pysoem/pysoem.pyx":1400
 *     def _get_name(self):
 *         """Name of the slave, read out from the slaves SII during config_init."""
 *         return (<bytes>self._ec_slave.name).decode('utf8')             # <<<<<<<<<<<<<<
 * 
 *     def _get_eep_man(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->_ec_slave->name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 1400, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_decode_bytes(((PyObject*)__pyx_t_1), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1398
 *             raise Exception('unexpected error, Etype: {}'.format(err.Etype))
 * 
 *     def _get_name(self):             # <<<<<<<<<<<<<<
 *         """Name of the slave, read out from the slaves SII during config_init."""
 *         return (<bytes>self._ec_slave.name).decode('utf8')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1402
 *         return (<bytes>self._ec_slave.name).decode('utf8')
 * 
 *     def _get_eep_man(self):             # <<<<<<<<<<<<<<
 *         """Vendor ID of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_man
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_47_get_eep_man(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_46_get_eep_man, "Vendor ID of the slave, read out from the slaves SII during config_init.");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_47_get_eep_man = {"_get_eep_man", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_47_get_eep_man, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_46_get_eep_man};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_47_get_eep_man(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_eep_man (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_eep_man", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_eep_man", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_46_get_eep_man(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_46_get_eep_man(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_eep_man", 0);

  /* "pysoem/pysoem.pyx":1404
 *     def _get_eep_man(self):
 *         """Vendor ID of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_man             # <<<<<<<<<<<<<<
 * 
 *     def _get_eep_id(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->_ec_slave->eep_man); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1402
 *         return (<bytes>self._ec_slave.name).decode('utf8')
 * 
 *     def _get_eep_man(self):             # <<<<<<<<<<<<<<
 *         """Vendor ID of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_man
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_eep_man", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1406
 *         return self._ec_slave.eep_man
 * 
 *     def _get_eep_id(self):             # <<<<<<<<<<<<<<
 *         """Product Code of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_id
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_49_get_eep_id(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_48_get_eep_id, "Product Code of the slave, read out from the slaves SII during config_init.");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_49_get_eep_id = {"_get_eep_id", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_49_get_eep_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_48_get_eep_id};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_49_get_eep_id(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_eep_id (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_eep_id", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_eep_id", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_48_get_eep_id(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_48_get_eep_id(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_eep_id", 0);

  /* "pysoem/pysoem.pyx":1408
 *     def _get_eep_id(self):
 *         """Product Code of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_id             # <<<<<<<<<<<<<<
 * 
 *     def _get_eep_rev(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->_ec_slave->eep_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1406
 *         return self._ec_slave.eep_man
 * 
 *     def _get_eep_id(self):             # <<<<<<<<<<<<<<
 *         """Product Code of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_id
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_eep_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1410
 *         return self._ec_slave.eep_id
 * 
 *     def _get_eep_rev(self):             # <<<<<<<<<<<<<<
 *         """Revision Number of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_rev
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_51_get_eep_rev(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_50_get_eep_rev, "Revision Number of the slave, read out from the slaves SII during config_init.");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_51_get_eep_rev = {"_get_eep_rev", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_51_get_eep_rev, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_50_get_eep_rev};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_51_get_eep_rev(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_eep_rev (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_eep_rev", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_eep_rev", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_50_get_eep_rev(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_50_get_eep_rev(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_eep_rev", 0);

  /* "pysoem/pysoem.pyx":1412
 *     def _get_eep_rev(self):
 *         """Revision Number of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_rev             # <<<<<<<<<<<<<<
 * 
 *     def _get_PO2SOconfig(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->_ec_slave->eep_rev); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1410
 *         return self._ec_slave.eep_id
 * 
 *     def _get_eep_rev(self):             # <<<<<<<<<<<<<<
 *         """Revision Number of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_rev
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_eep_rev", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1414
 *         return self._ec_slave.eep_rev
 * 
 *     def _get_PO2SOconfig(self):             # <<<<<<<<<<<<<<
 *         """Slaves callback function that is called during config_map.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_53_get_PO2SOconfig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_52_get_PO2SOconfig, "Slaves callback function that is called during config_map.\n        \n        When the state changes from Pre-Operational state to Operational state.");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_53_get_PO2SOconfig = {"_get_PO2SOconfig", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_53_get_PO2SOconfig, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_52_get_PO2SOconfig};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_53_get_PO2SOconfig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_PO2SOconfig (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_PO2SOconfig", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_PO2SOconfig", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_52_get_PO2SOconfig(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_52_get_PO2SOconfig(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_get_PO2SOconfig", 0);

  /* "pysoem/pysoem.pyx":1418
 * 
 *         When the state changes from Pre-Operational state to Operational state."""
 *         if not self._ec_slave.user:             # <<<<<<<<<<<<<<
 *             return None
 * 
*/
  __pyx_t_1 = (!(__pyx_v_self->_ec_slave->user != 0));
  if (__pyx_t_1) {

    /* "pysoem/pysoem.pyx":1419
 *         When the state changes from Pre-Operational state to Operational state."""
 *         if not self._ec_slave.user:
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         return <object>self._ec_slave.user
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pysoem/pysoem.pyx":1418
 * 
 *         When the state changes from Pre-Operational state to Operational state."""
 *         if not self._ec_slave.user:             # <<<<<<<<<<<<<<
 *             return None
 * 
*/
  }

  /* "pysoem/pysoem.pyx":1421
 *             return None
 * 
 *         return <object>self._ec_slave.user             # <<<<<<<<<<<<<<
 * 
 *     def _get_PO2SOconfigEx(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_ec_slave->user));
  __pyx_r = ((PyObject *)__pyx_v_self->_ec_slave->user);
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1414
 *         return self._ec_slave.eep_rev
 * 
 *     def _get_PO2SOconfig(self):             # <<<<<<<<<<<<<<
 *         """Slaves callback function that is called during config_map.
 * 
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1423
 *         return <object>self._ec_slave.user
 * 
 *     def _get_PO2SOconfigEx(self):             # <<<<<<<<<<<<<<
 *         """Alternative callback function that is called during config_map.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_55_get_PO2SOconfigEx(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_54_get_PO2SOconfigEx, "Alternative callback function that is called during config_map.\n\n        More precisely the function is called during the transition from Pre-Operational to Safe-Operational state.\n        Use this instead of the config_func. The difference is that the callbacks signature is fn(CdefSlave: slave).\n\n        .. versionadded:: 1.1.0\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_55_get_PO2SOconfigEx = {"_get_PO2SOconfigEx", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_55_get_PO2SOconfigEx, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_54_get_PO2SOconfigEx};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_55_get_PO2SOconfigEx(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_PO2SOconfigEx (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_PO2SOconfigEx", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_PO2SOconfigEx", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_54_get_PO2SOconfigEx(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_54_get_PO2SOconfigEx(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_get_PO2SOconfigEx", 0);

  /* "pysoem/pysoem.pyx":1431
 *         .. versionadded:: 1.1.0
 *         """
 *         if not self._ec_slave.user:             # <<<<<<<<<<<<<<
 *             return None
 * 
*/
  __pyx_t_1 = (!(__pyx_v_self->_ec_slave->user != 0));
  if (__pyx_t_1) {

    /* "pysoem/pysoem.pyx":1432
 *         """
 *         if not self._ec_slave.user:
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         return <object>self._ec_slave.user
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pysoem/pysoem.pyx":1431
 *         .. versionadded:: 1.1.0
 *         """
 *         if not self._ec_slave.user:             # <<<<<<<<<<<<<<
 *             return None
 * 
*/
  }

  /* "pysoem/pysoem.pyx":1434
 *             return None
 * 
 *         return <object>self._ec_slave.user             # <<<<<<<<<<<<<<
 * 
 *     def _set_PO2SOconfig(self, value):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_ec_slave->user));
  __pyx_r = ((PyObject *)__pyx_v_self->_ec_slave->user);
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1423
 *         return <object>self._ec_slave.user
 * 
 *     def _get_PO2SOconfigEx(self):             # <<<<<<<<<<<<<<
 *         """Alternative callback function that is called during config_map.
 * 
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1436
 *         return <object>self._ec_slave.user
 * 
 *     def _set_PO2SOconfig(self, value):             # <<<<<<<<<<<<<<
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_57_set_PO2SOconfig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_57_set_PO2SOconfig = {"_set_PO2SOconfig", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_57_set_PO2SOconfig, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_57_set_PO2SOconfig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_PO2SOconfig (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_value,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1436, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1436, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_set_PO2SOconfig", 0) < (0)) __PYX_ERR(0, 1436, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_set_PO2SOconfig", 1, 1, 1, i); __PYX_ERR(0, 1436, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1436, __pyx_L3_error)
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_set_PO2SOconfig", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1436, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._set_PO2SOconfig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_56_set_PO2SOconfig(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_56_set_PO2SOconfig(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_set_PO2SOconfig", 0);

  /* "pysoem/pysoem.pyx":1437
 * 
 *     def _set_PO2SOconfig(self, value):
 *         self._cd.func = value             # <<<<<<<<<<<<<<
 *         self._ec_slave.user = <void*>self._cd
 *         if value is None:
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_cd->func);
  __Pyx_DECREF(__pyx_v_self->_cd->func);
  __pyx_v_self->_cd->func = __pyx_v_value;

  /* "pysoem/pysoem.pyx":1438
 *     def _set_PO2SOconfig(self, value):
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd             # <<<<<<<<<<<<<<
 *         if value is None:
 *             self._ec_slave.PO2SOconfig = NULL
*/
  __pyx_v_self->_ec_slave->user = ((void *)__pyx_v_self->_cd);

  /* "pysoem/pysoem.pyx":1439
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd
 *         if value is None:             # <<<<<<<<<<<<<<
 *             self._ec_slave.PO2SOconfig = NULL
 *         else:
*/
  __pyx_t_1 = (__pyx_v_value == Py_None);
  if (__pyx_t_1) {

    /* "pysoem/pysoem.pyx":1440
 *         self._ec_slave.user = <void*>self._cd
 *         if value is None:
 *             self._ec_slave.PO2SOconfig = NULL             # <<<<<<<<<<<<<<
 *         else:
 *             self._ec_slave.PO2SOconfig = _xPO2SOconfig
*/
    __pyx_v_self->_ec_slave->PO2SOconfig = NULL;

    /* "pysoem/pysoem.pyx":1439
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd
 *         if value is None:             # <<<<<<<<<<<<<<
 *             self._ec_slave.PO2SOconfig = NULL
 *         else:
*/
    goto __pyx_L3;
  }

  /* "pysoem/pysoem.pyx":1442
 *             self._ec_slave.PO2SOconfig = NULL
 *         else:
 *             self._ec_slave.PO2SOconfig = _xPO2SOconfig             # <<<<<<<<<<<<<<
 * 
 *     def _set_PO2SOconfigEx(self, value):
*/
  /*else*/ {
    __pyx_v_self->_ec_slave->PO2SOconfig = __pyx_f_6pysoem_6pysoem__xPO2SOconfig;
  }
  __pyx_L3:;

  /* "pysoem/pysoem.pyx":1436
 *         return <object>self._ec_slave.user
 * 
 *     def _set_PO2SOconfig(self, value):             # <<<<<<<<<<<<<<
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1444
 *             self._ec_slave.PO2SOconfig = _xPO2SOconfig
 * 
 *     def _set_PO2SOconfigEx(self, value):             # <<<<<<<<<<<<<<
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_59_set_PO2SOconfigEx(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_59_set_PO2SOconfigEx = {"_set_PO2SOconfigEx", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_59_set_PO2SOconfigEx, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_59_set_PO2SOconfigEx(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_PO2SOconfigEx (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_value,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1444, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1444, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_set_PO2SOconfigEx", 0) < (0)) __PYX_ERR(0, 1444, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_set_PO2SOconfigEx", 1, 1, 1, i); __PYX_ERR(0, 1444, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1444, __pyx_L3_error)
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_set_PO2SOconfigEx", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1444, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._set_PO2SOconfigEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_58_set_PO2SOconfigEx(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_58_set_PO2SOconfigEx(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_set_PO2SOconfigEx", 0);

  /* "pysoem/pysoem.pyx":1445
 * 
 *     def _set_PO2SOconfigEx(self, value):
 *         self._cd.func = value             # <<<<<<<<<<<<<<
 *         self._ec_slave.user = <void*>self._cd
 *         if value is None:
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_cd->func);
  __Pyx_DECREF(__pyx_v_self->_cd->func);
  __pyx_v_self->_cd->func = __pyx_v_value;

  /* "pysoem/pysoem.pyx":1446
 *     def _set_PO2SOconfigEx(self, value):
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd             # <<<<<<<<<<<<<<
 *         if value is None:
 *             self._ec_slave.PO2SOconfig = NULL
*/
  __pyx_v_self->_ec_slave->user = ((void *)__pyx_v_self->_cd);

  /* "pysoem/pysoem.pyx":1447
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd
 *         if value is None:             # <<<<<<<<<<<<<<
 *             self._ec_slave.PO2SOconfig = NULL
 *         else:
*/
  __pyx_t_1 = (__pyx_v_value == Py_None);
  if (__pyx_t_1) {

    /* "pysoem/pysoem.pyx":1448
 *         self._ec_slave.user = <void*>self._cd
 *         if value is None:
 *             self._ec_slave.PO2SOconfig = NULL             # <<<<<<<<<<<<<<
 *         else:
 *             self._ec_slave.PO2SOconfig = _xPO2SOconfigEx
*/
    __pyx_v_self->_ec_slave->PO2SOconfig = NULL;

    /* "pysoem/pysoem.pyx":1447
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd
 *         if value is None:             # <<<<<<<<<<<<<<
 *             self._ec_slave.PO2SOconfig = NULL
 *         else:
*/
    goto __pyx_L3;
  }

  /* "pysoem/pysoem.pyx":1450
 *             self._ec_slave.PO2SOconfig = NULL
 *         else:
 *             self._ec_slave.PO2SOconfig = _xPO2SOconfigEx             # <<<<<<<<<<<<<<
 * 
 *     def _get_state(self):
*/
  /*else*/ {
    __pyx_v_self->_ec_slave->PO2SOconfig = __pyx_f_6pysoem_6pysoem__xPO2SOconfigEx;
  }
  __pyx_L3:;

  /* "pysoem/pysoem.pyx":1444
 *             self._ec_slave.PO2SOconfig = _xPO2SOconfig
 * 
 *     def _set_PO2SOconfigEx(self, value):             # <<<<<<<<<<<<<<
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1452
 *             self._ec_slave.PO2SOconfig = _xPO2SOconfigEx
 * 
 *     def _get_state(self):             # <<<<<<<<<<<<<<
 *         """Request a new state.
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_61_get_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6pysoem_6pysoem_9CdefSlave_60_get_state, "Request a new state.\n\n        After a new state has been set, `write_state` must be called.\n        ");
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_61_get_state = {"_get_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_61_get_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_60_get_state};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_61_get_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_state (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_state", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_state", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_60_get_state(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_60_get_state(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_state", 0);

  /* "pysoem/pysoem.pyx":1457
 *         After a new state has been set, `write_state` must be called.
 *         """
 *         return self._ec_slave.state             # <<<<<<<<<<<<<<
 * 
 *     def _set_state(self, value):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint16_t(__pyx_v_self->_ec_slave->state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1452
 *             self._ec_slave.PO2SOconfig = _xPO2SOconfigEx
 * 
 *     def _get_state(self):             # <<<<<<<<<<<<<<
 *         """Request a new state.
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1459
 *         return self._ec_slave.state
 * 
 *     def _set_state(self, value):             # <<<<<<<<<<<<<<
 *         self._ec_slave.state = value
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_63_set_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_63_set_state = {"_set_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_63_set_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_63_set_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_state (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_value,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1459, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1459, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_set_state", 0) < (0)) __PYX_ERR(0, 1459, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_set_state", 1, 1, 1, i); __PYX_ERR(0, 1459, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1459, __pyx_L3_error)
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_set_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1459, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_62_set_state(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_62_set_state(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16 __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_state", 0);

  /* "pysoem/pysoem.pyx":1460
 * 
 *     def _set_state(self, value):
 *         self._ec_slave.state = value             # <<<<<<<<<<<<<<
 * 
 *     def _get_input(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint16_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((uint16)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1460, __pyx_L1_error)
  __pyx_v_self->_ec_slave->state = __pyx_t_1;

  /* "pysoem/pysoem.pyx":1459
 *         return self._ec_slave.state
 * 
 *     def _set_state(self, value):             # <<<<<<<<<<<<<<
 *         self._ec_slave.state = value
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1462
 *         self._ec_slave.state = value
 * 
 *     def _get_input(self):             # <<<<<<<<<<<<<<
 *         num_bytes = self._ec_slave.Ibytes
 *         if (self._ec_slave.Ibytes == 0 and self._ec_slave.Ibits > 0):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_65_get_input(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_65_get_input = {"_get_input", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_65_get_input, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_65_get_input(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_input (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_input", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_input", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_64_get_input(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_64_get_input(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  long __pyx_v_num_bytes;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32 __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_input", 0);

  /* "pysoem/pysoem.pyx":1463
 * 
 *     def _get_input(self):
 *         num_bytes = self._ec_slave.Ibytes             # <<<<<<<<<<<<<<
 *         if (self._ec_slave.Ibytes == 0 and self._ec_slave.Ibits > 0):
 *             num_bytes = 1
*/
  __pyx_t_1 = __pyx_v_self->_ec_slave->Ibytes;
  __pyx_v_num_bytes = __pyx_t_1;

  /* "pysoem/pysoem.pyx":1464
 *     def _get_input(self):
 *         num_bytes = self._ec_slave.Ibytes
 *         if (self._ec_slave.Ibytes == 0 and self._ec_slave.Ibits > 0):             # <<<<<<<<<<<<<<
 *             num_bytes = 1
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.inputs, num_bytes)
*/
  __pyx_t_3 = (__pyx_v_self->_ec_slave->Ibytes == 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_self->_ec_slave->Ibits > 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pysoem/pysoem.pyx":1465
 *         num_bytes = self._ec_slave.Ibytes
 *         if (self._ec_slave.Ibytes == 0 and self._ec_slave.Ibits > 0):
 *             num_bytes = 1             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.inputs, num_bytes)
 * 
*/
    __pyx_v_num_bytes = 1;

    /* "pysoem/pysoem.pyx":1464
 *     def _get_input(self):
 *         num_bytes = self._ec_slave.Ibytes
 *         if (self._ec_slave.Ibytes == 0 and self._ec_slave.Ibits > 0):             # <<<<<<<<<<<<<<
 *             num_bytes = 1
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.inputs, num_bytes)
*/
  }

  /* "pysoem/pysoem.pyx":1466
 *         if (self._ec_slave.Ibytes == 0 and self._ec_slave.Ibits > 0):
 *             num_bytes = 1
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.inputs, num_bytes)             # <<<<<<<<<<<<<<
 * 
 *     def _get_output(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyBytes_FromStringAndSize(((char *)__pyx_v_self->_ec_slave->inputs), __pyx_v_num_bytes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1462
 *         self._ec_slave.state = value
 * 
 *     def _get_input(self):             # <<<<<<<<<<<<<<
 *         num_bytes = self._ec_slave.Ibytes
 *         if (self._ec_slave.Ibytes == 0 and self._ec_slave.Ibits > 0):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_input", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1468
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.inputs, num_bytes)
 * 
 *     def _get_output(self):             # <<<<<<<<<<<<<<
 *         num_bytes = self._ec_slave.Obytes
 *         if (self._ec_slave.Obytes == 0 and self._ec_slave.Obits > 0):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_67_get_output(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_67_get_output = {"_get_output", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_67_get_output, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_67_get_output(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_output (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_output", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_output", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_66_get_output(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_66_get_output(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  long __pyx_v_num_bytes;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32 __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_output", 0);

  /* "pysoem/pysoem.pyx":1469
 * 
 *     def _get_output(self):
 *         num_bytes = self._ec_slave.Obytes             # <<<<<<<<<<<<<<
 *         if (self._ec_slave.Obytes == 0 and self._ec_slave.Obits > 0):
 *             num_bytes = 1
*/
  __pyx_t_1 = __pyx_v_self->_ec_slave->Obytes;
  __pyx_v_num_bytes = __pyx_t_1;

  /* "pysoem/pysoem.pyx":1470
 *     def _get_output(self):
 *         num_bytes = self._ec_slave.Obytes
 *         if (self._ec_slave.Obytes == 0 and self._ec_slave.Obits > 0):             # <<<<<<<<<<<<<<
 *             num_bytes = 1
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.outputs, num_bytes)
*/
  __pyx_t_3 = (__pyx_v_self->_ec_slave->Obytes == 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_self->_ec_slave->Obits > 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pysoem/pysoem.pyx":1471
 *         num_bytes = self._ec_slave.Obytes
 *         if (self._ec_slave.Obytes == 0 and self._ec_slave.Obits > 0):
 *             num_bytes = 1             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.outputs, num_bytes)
 * 
*/
    __pyx_v_num_bytes = 1;

    /* "pysoem/pysoem.pyx":1470
 *     def _get_output(self):
 *         num_bytes = self._ec_slave.Obytes
 *         if (self._ec_slave.Obytes == 0 and self._ec_slave.Obits > 0):             # <<<<<<<<<<<<<<
 *             num_bytes = 1
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.outputs, num_bytes)
*/
  }

  /* "pysoem/pysoem.pyx":1472
 *         if (self._ec_slave.Obytes == 0 and self._ec_slave.Obits > 0):
 *             num_bytes = 1
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.outputs, num_bytes)             # <<<<<<<<<<<<<<
 * 
 *     def _set_output(self, bytes value):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyBytes_FromStringAndSize(((char *)__pyx_v_self->_ec_slave->outputs), __pyx_v_num_bytes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1468
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.inputs, num_bytes)
 * 
 *     def _get_output(self):             # <<<<<<<<<<<<<<
 *         num_bytes = self._ec_slave.Obytes
 *         if (self._ec_slave.Obytes == 0 and self._ec_slave.Obits > 0):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_output", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1474
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.outputs, num_bytes)
 * 
 *     def _set_output(self, bytes value):             # <<<<<<<<<<<<<<
 *         memcpy(<char*>self._ec_slave.outputs, <char*>value, len(value))
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_69_set_output(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_69_set_output = {"_set_output", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_69_set_output, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_69_set_output(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_output (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_value,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1474, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1474, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_set_output", 0) < (0)) __PYX_ERR(0, 1474, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_set_output", 1, 1, 1, i); __PYX_ERR(0, 1474, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1474, __pyx_L3_error)
    }
    __pyx_v_value = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_set_output", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1474, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._set_output", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyBytes_Type), 1, "value", 1))) __PYX_ERR(0, 1474, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_68_set_output(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_68_set_output(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_output", 0);

  /* "pysoem/pysoem.pyx":1475
 * 
 *     def _set_output(self, bytes value):
 *         memcpy(<char*>self._ec_slave.outputs, <char*>value, len(value))             # <<<<<<<<<<<<<<
 * 
 *     def _get_al_status(self):
*/
  if (unlikely(__pyx_v_value == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1475, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBytes_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 1475, __pyx_L1_error)
  if (unlikely(__pyx_v_value == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1475, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_GET_SIZE(__pyx_v_value); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1475, __pyx_L1_error)
  (void)(memcpy(((char *)__pyx_v_self->_ec_slave->outputs), ((char *)__pyx_t_1), __pyx_t_2));

  /* "pysoem/pysoem.pyx":1474
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.outputs, num_bytes)
 * 
 *     def _set_output(self, bytes value):             # <<<<<<<<<<<<<<
 *         memcpy(<char*>self._ec_slave.outputs, <char*>value, len(value))
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._set_output", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1477
 *         memcpy(<char*>self._ec_slave.outputs, <char*>value, len(value))
 * 
 *     def _get_al_status(self):             # <<<<<<<<<<<<<<
 *         return self._ec_slave.ALstatuscode
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_71_get_al_status(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_71_get_al_status = {"_get_al_status", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_71_get_al_status, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_71_get_al_status(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_al_status (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_al_status", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_al_status", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_70_get_al_status(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_70_get_al_status(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_al_status", 0);

  /* "pysoem/pysoem.pyx":1478
 * 
 *     def _get_al_status(self):
 *         return self._ec_slave.ALstatuscode             # <<<<<<<<<<<<<<
 * 
 *     def _get_is_lost(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint16_t(__pyx_v_self->_ec_slave->ALstatuscode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1477
 *         memcpy(<char*>self._ec_slave.outputs, <char*>value, len(value))
 * 
 *     def _get_al_status(self):             # <<<<<<<<<<<<<<
 *         return self._ec_slave.ALstatuscode
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_al_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1480
 *         return self._ec_slave.ALstatuscode
 * 
 *     def _get_is_lost(self):             # <<<<<<<<<<<<<<
 *         return self._ec_slave.islost
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_73_get_is_lost(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_73_get_is_lost = {"_get_is_lost", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_73_get_is_lost, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_73_get_is_lost(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_is_lost (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_is_lost", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_is_lost", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_72_get_is_lost(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_72_get_is_lost(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_is_lost", 0);

  /* "pysoem/pysoem.pyx":1481
 * 
 *     def _get_is_lost(self):
 *         return self._ec_slave.islost             # <<<<<<<<<<<<<<
 * 
 *     def _set_is_lost(self, value):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int8_t(__pyx_v_self->_ec_slave->islost); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1480
 *         return self._ec_slave.ALstatuscode
 * 
 *     def _get_is_lost(self):             # <<<<<<<<<<<<<<
 *         return self._ec_slave.islost
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_is_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1483
 *         return self._ec_slave.islost
 * 
 *     def _set_is_lost(self, value):             # <<<<<<<<<<<<<<
 *         self._ec_slave.islost = value
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_75_set_is_lost(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_75_set_is_lost = {"_set_is_lost", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_75_set_is_lost, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_75_set_is_lost(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_is_lost (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_value,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1483, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1483, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_set_is_lost", 0) < (0)) __PYX_ERR(0, 1483, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_set_is_lost", 1, 1, 1, i); __PYX_ERR(0, 1483, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1483, __pyx_L3_error)
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_set_is_lost", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1483, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._set_is_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_74_set_is_lost(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_74_set_is_lost(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  boolean __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_is_lost", 0);

  /* "pysoem/pysoem.pyx":1484
 * 
 *     def _set_is_lost(self, value):
 *         self._ec_slave.islost = value             # <<<<<<<<<<<<<<
 * 
 *     def _get_od(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_int8_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1484, __pyx_L1_error)
  __pyx_v_self->_ec_slave->islost = __pyx_t_1;

  /* "pysoem/pysoem.pyx":1483
 *         return self._ec_slave.islost
 * 
 *     def _set_is_lost(self, value):             # <<<<<<<<<<<<<<
 *         self._ec_slave.islost = value
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._set_is_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1486
 *         self._ec_slave.islost = value
 * 
 *     def _get_od(self):             # <<<<<<<<<<<<<<
 *         logger.debug('ecx_readODlist()')
 *         cdef int result = cpysoem.ecx_readODlist(self._ecx_contextt, self._pos, &self._ex_odlist)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_77_get_od(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_77_get_od = {"_get_od", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_77_get_od, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_77_get_od(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_od (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_od", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_od", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_76_get_od(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_76_get_od(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  int __pyx_v_result;
  PyObject *__pyx_v_coe_objects = NULL;
  uint16 __pyx_v_i;
  struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_coe_object = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  uint16 __pyx_t_7;
  uint16 __pyx_t_8;
  uint16 __pyx_t_9;
  ecx_contextt *__pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_od", 0);

  /* "pysoem/pysoem.pyx":1487
 * 
 *     def _get_od(self):
 *         logger.debug('ecx_readODlist()')             # <<<<<<<<<<<<<<
 *         cdef int result = cpysoem.ecx_readODlist(self._ecx_contextt, self._pos, &self._ex_odlist)
 *         if not result > 0:
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_ecx_readODlist};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1488
 *     def _get_od(self):
 *         logger.debug('ecx_readODlist()')
 *         cdef int result = cpysoem.ecx_readODlist(self._ecx_contextt, self._pos, &self._ex_odlist)             # <<<<<<<<<<<<<<
 *         if not result > 0:
 *             raise SdoInfoError('Sdo List Info read failed')
*/
  __pyx_v_result = ecx_readODlist(__pyx_v_self->_ecx_contextt, __pyx_v_self->_pos, (&__pyx_v_self->_ex_odlist));

  /* "pysoem/pysoem.pyx":1489
 *         logger.debug('ecx_readODlist()')
 *         cdef int result = cpysoem.ecx_readODlist(self._ecx_contextt, self._pos, &self._ex_odlist)
 *         if not result > 0:             # <<<<<<<<<<<<<<
 *             raise SdoInfoError('Sdo List Info read failed')
 * 
*/
  __pyx_t_6 = (!(__pyx_v_result > 0));
  if (unlikely(__pyx_t_6)) {

    /* "pysoem/pysoem.pyx":1490
 *         cdef int result = cpysoem.ecx_readODlist(self._ecx_contextt, self._pos, &self._ex_odlist)
 *         if not result > 0:
 *             raise SdoInfoError('Sdo List Info read failed')             # <<<<<<<<<<<<<<
 * 
 *         coe_objects = []
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_SdoInfoError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Sdo_List_Info_read_failed};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1490, __pyx_L1_error)

    /* "pysoem/pysoem.pyx":1489
 *         logger.debug('ecx_readODlist()')
 *         cdef int result = cpysoem.ecx_readODlist(self._ecx_contextt, self._pos, &self._ex_odlist)
 *         if not result > 0:             # <<<<<<<<<<<<<<
 *             raise SdoInfoError('Sdo List Info read failed')
 * 
*/
  }

  /* "pysoem/pysoem.pyx":1492
 *             raise SdoInfoError('Sdo List Info read failed')
 * 
 *         coe_objects = []             # <<<<<<<<<<<<<<
 *         for i in range(self._ex_odlist.Entries):
 *             coe_object = CdefCoeObject(i)
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_coe_objects = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1493
 * 
 *         coe_objects = []
 *         for i in range(self._ex_odlist.Entries):             # <<<<<<<<<<<<<<
 *             coe_object = CdefCoeObject(i)
 *             coe_object._ecx_context = self._ecx_contextt
*/
  __pyx_t_7 = __pyx_v_self->_ex_odlist.Entries;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "pysoem/pysoem.pyx":1494
 *         coe_objects = []
 *         for i in range(self._ex_odlist.Entries):
 *             coe_object = CdefCoeObject(i)             # <<<<<<<<<<<<<<
 *             coe_object._ecx_context = self._ecx_contextt
 *             coe_object._ex_odlist = &self._ex_odlist
*/
    __pyx_t_2 = NULL;
    __pyx_t_4 = __Pyx_PyLong_From_uint16_t(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1494, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_coe_object, ((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":1495
 *         for i in range(self._ex_odlist.Entries):
 *             coe_object = CdefCoeObject(i)
 *             coe_object._ecx_context = self._ecx_contextt             # <<<<<<<<<<<<<<
 *             coe_object._ex_odlist = &self._ex_odlist
 *             coe_objects.append(coe_object)
*/
    __pyx_t_10 = __pyx_v_self->_ecx_contextt;
    __pyx_v_coe_object->_ecx_context = __pyx_t_10;

    /* "pysoem/pysoem.pyx":1496
 *             coe_object = CdefCoeObject(i)
 *             coe_object._ecx_context = self._ecx_contextt
 *             coe_object._ex_odlist = &self._ex_odlist             # <<<<<<<<<<<<<<
 *             coe_objects.append(coe_object)
 * 
*/
    __pyx_v_coe_object->_ex_odlist = (&__pyx_v_self->_ex_odlist);

    /* "pysoem/pysoem.pyx":1497
 *             coe_object._ecx_context = self._ecx_contextt
 *             coe_object._ex_odlist = &self._ex_odlist
 *             coe_objects.append(coe_object)             # <<<<<<<<<<<<<<
 * 
 *         return coe_objects
*/
    __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_coe_objects, ((PyObject *)__pyx_v_coe_object)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 1497, __pyx_L1_error)
  }

  /* "pysoem/pysoem.pyx":1499
 *             coe_objects.append(coe_object)
 * 
 *         return coe_objects             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_coe_objects);
  __pyx_r = __pyx_v_coe_objects;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1486
 *         self._ec_slave.islost = value
 * 
 *     def _get_od(self):             # <<<<<<<<<<<<<<
 *         logger.debug('ecx_readODlist()')
 *         cdef int result = cpysoem.ecx_readODlist(self._ecx_contextt, self._pos, &self._ex_odlist)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._get_od", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_coe_objects);
  __Pyx_XDECREF((PyObject *)__pyx_v_coe_object);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":795
 *     obtained by slaves list
 *     """
 *     cdef readonly CdefMaster _master             # <<<<<<<<<<<<<<
 *     cdef cpysoem.ecx_contextt* _ecx_contextt
 *     cdef cpysoem.ec_slavet* _ec_slave
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_7_master_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_7_master_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_7_master___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_7_master___get__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_master);
  __pyx_r = ((PyObject *)__pyx_v_self->_master);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":800
 *     cdef CdefMasterSettings* _the_masters_settings
 *     cdef int _pos # keep in mind that first slave has pos 1
 *     cdef public _CallbackData _cd             # <<<<<<<<<<<<<<
 *     cdef cpysoem.ec_ODlistt _ex_odlist
 *     cdef public _emcy_callbacks
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_3_cd_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_3_cd_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_3_cd___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_3_cd___get__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_cd);
  __pyx_r = ((PyObject *)__pyx_v_self->_cd);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_9CdefSlave_3_cd_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_9CdefSlave_3_cd_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_3_cd_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_9CdefSlave_3_cd_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem__CallbackData))))) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_cd);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_cd);
  __pyx_v_self->_cd = ((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave._cd.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_9CdefSlave_3_cd_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_9CdefSlave_3_cd_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_3_cd_4__del__(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_9CdefSlave_3_cd_4__del__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_cd);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_cd);
  __pyx_v_self->_cd = ((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":802
 *     cdef public _CallbackData _cd
 *     cdef cpysoem.ec_ODlistt _ex_odlist
 *     cdef public _emcy_callbacks             # <<<<<<<<<<<<<<
 * 
 *     name = property(_get_name)
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks___get__(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks___get__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_emcy_callbacks);
  __pyx_r = __pyx_v_self->_emcy_callbacks;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_2__set__(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_2__set__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_emcy_callbacks);
  __Pyx_DECREF(__pyx_v_self->_emcy_callbacks);
  __pyx_v_self->_emcy_callbacks = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_4__del__(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_4__del__(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_emcy_callbacks);
  __Pyx_DECREF(__pyx_v_self->_emcy_callbacks);
  __pyx_v_self->_emcy_callbacks = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_79__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_79__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_79__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_79__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_78__reduce_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_78__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self__ec_slave_self__ecx_context, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_81__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_9CdefSlave_81__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_81__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_9CdefSlave_81__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9CdefSlave_80__setstate_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9CdefSlave_80__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefSlave *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self__ec_slave_self__ecx_context, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefSlave.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1523
 *     obj_access = property(_get_obj_access)
 * 
 *     def __init__(self, int item):             # <<<<<<<<<<<<<<
 *         self._item = item
 *         self._is_description_read = False
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_13CdefCoeObject_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_13CdefCoeObject_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_item;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_item,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1523, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1523, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 1523, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 1523, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1523, __pyx_L3_error)
    }
    __pyx_v_item = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_item == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1523, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1523, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject___init__(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self), __pyx_v_item);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_13CdefCoeObject___init__(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self, int __pyx_v_item) {
  int __pyx_r;

  /* "pysoem/pysoem.pyx":1524
 * 
 *     def __init__(self, int item):
 *         self._item = item             # <<<<<<<<<<<<<<
 *         self._is_description_read = False
 *         self._are_entries_read = False
*/
  __pyx_v_self->_item = __pyx_v_item;

  /* "pysoem/pysoem.pyx":1525
 *     def __init__(self, int item):
 *         self._item = item
 *         self._is_description_read = False             # <<<<<<<<<<<<<<
 *         self._are_entries_read = False
 * 
*/
  __pyx_v_self->_is_description_read = 0;

  /* "pysoem/pysoem.pyx":1526
 *         self._item = item
 *         self._is_description_read = False
 *         self._are_entries_read = False             # <<<<<<<<<<<<<<
 * 
 *     def _read_description(self):
*/
  __pyx_v_self->_are_entries_read = 0;

  /* "pysoem/pysoem.pyx":1523
 *     obj_access = property(_get_obj_access)
 * 
 *     def __init__(self, int item):             # <<<<<<<<<<<<<<
 *         self._item = item
 *         self._is_description_read = False
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1528
 *         self._are_entries_read = False
 * 
 *     def _read_description(self):             # <<<<<<<<<<<<<<
 *         cdef int result
 *         if not self._is_description_read:
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_3_read_description(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_3_read_description = {"_read_description", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_3_read_description, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_3_read_description(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read_description (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_read_description", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_read_description", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_2_read_description(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_2_read_description(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self) {
  int __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_read_description", 0);

  /* "pysoem/pysoem.pyx":1530
 *     def _read_description(self):
 *         cdef int result
 *         if not self._is_description_read:             # <<<<<<<<<<<<<<
 *           logger.debug('ecx_readODdescription()')
 *           result = cpysoem.ecx_readODdescription(self._ecx_context, self._item, self._ex_odlist)
*/
  __pyx_t_1 = (!(__pyx_v_self->_is_description_read != 0));
  if (__pyx_t_1) {

    /* "pysoem/pysoem.pyx":1531
 *         cdef int result
 *         if not self._is_description_read:
 *           logger.debug('ecx_readODdescription()')             # <<<<<<<<<<<<<<
 *           result = cpysoem.ecx_readODdescription(self._ecx_context, self._item, self._ex_odlist)
 *           if not result > 0:
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_logger); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_ecx_readODdescription};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1531, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pysoem/pysoem.pyx":1532
 *         if not self._is_description_read:
 *           logger.debug('ecx_readODdescription()')
 *           result = cpysoem.ecx_readODdescription(self._ecx_context, self._item, self._ex_odlist)             # <<<<<<<<<<<<<<
 *           if not result > 0:
 *               raise SdoInfoError('Sdo Object Info read failed')
*/
    __pyx_v_result = ecx_readODdescription(__pyx_v_self->_ecx_context, __pyx_v_self->_item, __pyx_v_self->_ex_odlist);

    /* "pysoem/pysoem.pyx":1533
 *           logger.debug('ecx_readODdescription()')
 *           result = cpysoem.ecx_readODdescription(self._ecx_context, self._item, self._ex_odlist)
 *           if not result > 0:             # <<<<<<<<<<<<<<
 *               raise SdoInfoError('Sdo Object Info read failed')
 *           self._is_description_read = True
*/
    __pyx_t_1 = (!(__pyx_v_result > 0));
    if (unlikely(__pyx_t_1)) {

      /* "pysoem/pysoem.pyx":1534
 *           result = cpysoem.ecx_readODdescription(self._ecx_context, self._item, self._ex_odlist)
 *           if not result > 0:
 *               raise SdoInfoError('Sdo Object Info read failed')             # <<<<<<<<<<<<<<
 *           self._is_description_read = True
 * 
*/
      __pyx_t_5 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_SdoInfoError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        assert(__pyx_t_5);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
        __pyx_t_6 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_Sdo_Object_Info_read_failed};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1534, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1534, __pyx_L1_error)

      /* "pysoem/pysoem.pyx":1533
 *           logger.debug('ecx_readODdescription()')
 *           result = cpysoem.ecx_readODdescription(self._ecx_context, self._item, self._ex_odlist)
 *           if not result > 0:             # <<<<<<<<<<<<<<
 *               raise SdoInfoError('Sdo Object Info read failed')
 *           self._is_description_read = True
*/
    }

    /* "pysoem/pysoem.pyx":1535
 *           if not result > 0:
 *               raise SdoInfoError('Sdo Object Info read failed')
 *           self._is_description_read = True             # <<<<<<<<<<<<<<
 * 
 *     def _read_entries(self):
*/
    __pyx_v_self->_is_description_read = 1;

    /* "pysoem/pysoem.pyx":1530
 *     def _read_description(self):
 *         cdef int result
 *         if not self._is_description_read:             # <<<<<<<<<<<<<<
 *           logger.debug('ecx_readODdescription()')
 *           result = cpysoem.ecx_readODdescription(self._ecx_context, self._item, self._ex_odlist)
*/
  }

  /* "pysoem/pysoem.pyx":1528
 *         self._are_entries_read = False
 * 
 *     def _read_description(self):             # <<<<<<<<<<<<<<
 *         cdef int result
 *         if not self._is_description_read:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject._read_description", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1537
 *           self._is_description_read = True
 * 
 *     def _read_entries(self):             # <<<<<<<<<<<<<<
 *         cdef int result
 *         if not self._are_entries_read:
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_5_read_entries(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_5_read_entries = {"_read_entries", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_5_read_entries, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_5_read_entries(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read_entries (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_read_entries", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_read_entries", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_4_read_entries(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_4_read_entries(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self) {
  int __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_read_entries", 0);

  /* "pysoem/pysoem.pyx":1539
 *     def _read_entries(self):
 *         cdef int result
 *         if not self._are_entries_read:             # <<<<<<<<<<<<<<
 *             logger.debug('ecx_readOE()')
 *             result = cpysoem.ecx_readOE(self._ecx_context, self._item, self._ex_odlist, &self._ex_oelist)
*/
  __pyx_t_1 = (!(__pyx_v_self->_are_entries_read != 0));
  if (__pyx_t_1) {

    /* "pysoem/pysoem.pyx":1540
 *         cdef int result
 *         if not self._are_entries_read:
 *             logger.debug('ecx_readOE()')             # <<<<<<<<<<<<<<
 *             result = cpysoem.ecx_readOE(self._ecx_context, self._item, self._ex_odlist, &self._ex_oelist)
 *             if not result > 0:
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_logger); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_ecx_readOE};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pysoem/pysoem.pyx":1541
 *         if not self._are_entries_read:
 *             logger.debug('ecx_readOE()')
 *             result = cpysoem.ecx_readOE(self._ecx_context, self._item, self._ex_odlist, &self._ex_oelist)             # <<<<<<<<<<<<<<
 *             if not result > 0:
 *                 raise SdoInfoError('Sdo ObjectEntry Info read failed')
*/
    __pyx_v_result = ecx_readOE(__pyx_v_self->_ecx_context, __pyx_v_self->_item, __pyx_v_self->_ex_odlist, (&__pyx_v_self->_ex_oelist));

    /* "pysoem/pysoem.pyx":1542
 *             logger.debug('ecx_readOE()')
 *             result = cpysoem.ecx_readOE(self._ecx_context, self._item, self._ex_odlist, &self._ex_oelist)
 *             if not result > 0:             # <<<<<<<<<<<<<<
 *                 raise SdoInfoError('Sdo ObjectEntry Info read failed')
 *             self._are_entries_read = True
*/
    __pyx_t_1 = (!(__pyx_v_result > 0));
    if (unlikely(__pyx_t_1)) {

      /* "pysoem/pysoem.pyx":1543
 *             result = cpysoem.ecx_readOE(self._ecx_context, self._item, self._ex_odlist, &self._ex_oelist)
 *             if not result > 0:
 *                 raise SdoInfoError('Sdo ObjectEntry Info read failed')             # <<<<<<<<<<<<<<
 *             self._are_entries_read = True
 * 
*/
      __pyx_t_5 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_SdoInfoError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1543, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        assert(__pyx_t_5);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
        __pyx_t_6 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_Sdo_ObjectEntry_Info_read_failed};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1543, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1543, __pyx_L1_error)

      /* "pysoem/pysoem.pyx":1542
 *             logger.debug('ecx_readOE()')
 *             result = cpysoem.ecx_readOE(self._ecx_context, self._item, self._ex_odlist, &self._ex_oelist)
 *             if not result > 0:             # <<<<<<<<<<<<<<
 *                 raise SdoInfoError('Sdo ObjectEntry Info read failed')
 *             self._are_entries_read = True
*/
    }

    /* "pysoem/pysoem.pyx":1544
 *             if not result > 0:
 *                 raise SdoInfoError('Sdo ObjectEntry Info read failed')
 *             self._are_entries_read = True             # <<<<<<<<<<<<<<
 * 
 *     def _get_index(self):
*/
    __pyx_v_self->_are_entries_read = 1;

    /* "pysoem/pysoem.pyx":1539
 *     def _read_entries(self):
 *         cdef int result
 *         if not self._are_entries_read:             # <<<<<<<<<<<<<<
 *             logger.debug('ecx_readOE()')
 *             result = cpysoem.ecx_readOE(self._ecx_context, self._item, self._ex_odlist, &self._ex_oelist)
*/
  }

  /* "pysoem/pysoem.pyx":1537
 *           self._is_description_read = True
 * 
 *     def _read_entries(self):             # <<<<<<<<<<<<<<
 *         cdef int result
 *         if not self._are_entries_read:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject._read_entries", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1546
 *             self._are_entries_read = True
 * 
 *     def _get_index(self):             # <<<<<<<<<<<<<<
 *         return self._ex_odlist.Index[self._item]
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_7_get_index(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_7_get_index = {"_get_index", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_7_get_index, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_7_get_index(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_index (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_index", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_index", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_6_get_index(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_6_get_index(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_index", 0);

  /* "pysoem/pysoem.pyx":1547
 * 
 *     def _get_index(self):
 *         return self._ex_odlist.Index[self._item]             # <<<<<<<<<<<<<<
 * 
 *     def _get_data_type(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint16_t((__pyx_v_self->_ex_odlist->Index[__pyx_v_self->_item])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1546
 *             self._are_entries_read = True
 * 
 *     def _get_index(self):             # <<<<<<<<<<<<<<
 *         return self._ex_odlist.Index[self._item]
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject._get_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1549
 *         return self._ex_odlist.Index[self._item]
 * 
 *     def _get_data_type(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         return self._ex_odlist.DataType[self._item]
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_9_get_data_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_9_get_data_type = {"_get_data_type", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_9_get_data_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_9_get_data_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_data_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_data_type", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_data_type", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_8_get_data_type(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_8_get_data_type(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_data_type", 0);

  /* "pysoem/pysoem.pyx":1550
 * 
 *     def _get_data_type(self):
 *         self._read_description()             # <<<<<<<<<<<<<<
 *         return self._ex_odlist.DataType[self._item]
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read_description, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1551
 *     def _get_data_type(self):
 *         self._read_description()
 *         return self._ex_odlist.DataType[self._item]             # <<<<<<<<<<<<<<
 * 
 *     def _get_object_code(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint16_t((__pyx_v_self->_ex_odlist->DataType[__pyx_v_self->_item])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1549
 *         return self._ex_odlist.Index[self._item]
 * 
 *     def _get_data_type(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         return self._ex_odlist.DataType[self._item]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject._get_data_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1553
 *         return self._ex_odlist.DataType[self._item]
 * 
 *     def _get_object_code(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         return self._ex_odlist.ObjectCode[self._item]
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_11_get_object_code(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_11_get_object_code = {"_get_object_code", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_11_get_object_code, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_11_get_object_code(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_object_code (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_object_code", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_object_code", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_10_get_object_code(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_10_get_object_code(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_object_code", 0);

  /* "pysoem/pysoem.pyx":1554
 * 
 *     def _get_object_code(self):
 *         self._read_description()             # <<<<<<<<<<<<<<
 *         return self._ex_odlist.ObjectCode[self._item]
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read_description, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1555
 *     def _get_object_code(self):
 *         self._read_description()
 *         return self._ex_odlist.ObjectCode[self._item]             # <<<<<<<<<<<<<<
 * 
 *     def _get_name(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint8_t((__pyx_v_self->_ex_odlist->ObjectCode[__pyx_v_self->_item])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1553
 *         return self._ex_odlist.DataType[self._item]
 * 
 *     def _get_object_code(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         return self._ex_odlist.ObjectCode[self._item]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject._get_object_code", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1557
 *         return self._ex_odlist.ObjectCode[self._item]
 * 
 *     def _get_name(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         return self._ex_odlist.Name[self._item]
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_13_get_name(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_13_get_name = {"_get_name", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_13_get_name, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_13_get_name(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_name (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_name", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_name", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_12_get_name(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_12_get_name(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_name", 0);

  /* "pysoem/pysoem.pyx":1558
 * 
 *     def _get_name(self):
 *         self._read_description()             # <<<<<<<<<<<<<<
 *         return self._ex_odlist.Name[self._item]
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read_description, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1559
 *     def _get_name(self):
 *         self._read_description()
 *         return self._ex_odlist.Name[self._item]             # <<<<<<<<<<<<<<
 * 
 *     def _get_entries(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString((__pyx_v_self->_ex_odlist->Name[__pyx_v_self->_item])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1557
 *         return self._ex_odlist.ObjectCode[self._item]
 * 
 *     def _get_name(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         return self._ex_odlist.Name[self._item]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject._get_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1561
 *         return self._ex_odlist.Name[self._item]
 * 
 *     def _get_entries(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         self._read_entries()
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_15_get_entries(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_15_get_entries = {"_get_entries", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_15_get_entries, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_15_get_entries(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_entries (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_entries", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_entries", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_14_get_entries(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_14_get_entries(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self) {
  PyObject *__pyx_v_entries = NULL;
  long __pyx_v_i;
  struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_entry = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  long __pyx_t_5;
  long __pyx_t_6;
  long __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_entries", 0);

  /* "pysoem/pysoem.pyx":1562
 * 
 *     def _get_entries(self):
 *         self._read_description()             # <<<<<<<<<<<<<<
 *         self._read_entries()
 * 
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read_description, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1563
 *     def _get_entries(self):
 *         self._read_description()
 *         self._read_entries()             # <<<<<<<<<<<<<<
 * 
 *         if self._ex_odlist.MaxSub[self._item] == 0:
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read_entries, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1565
 *         self._read_entries()
 * 
 *         if self._ex_odlist.MaxSub[self._item] == 0:             # <<<<<<<<<<<<<<
 *             return []
 *         else:
*/
  __pyx_t_4 = ((__pyx_v_self->_ex_odlist->MaxSub[__pyx_v_self->_item]) == 0);
  if (__pyx_t_4) {

    /* "pysoem/pysoem.pyx":1566
 * 
 *         if self._ex_odlist.MaxSub[self._item] == 0:
 *             return []             # <<<<<<<<<<<<<<
 *         else:
 *             entries = []
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pysoem/pysoem.pyx":1565
 *         self._read_entries()
 * 
 *         if self._ex_odlist.MaxSub[self._item] == 0:             # <<<<<<<<<<<<<<
 *             return []
 *         else:
*/
  }

  /* "pysoem/pysoem.pyx":1568
 *             return []
 *         else:
 *             entries = []             # <<<<<<<<<<<<<<
 *             for i in range(self._ex_odlist.MaxSub[self._item]+1):
 *                 entry = CdefCoeObjectEntry(i)
*/
  /*else*/ {
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_entries = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "pysoem/pysoem.pyx":1569
 *         else:
 *             entries = []
 *             for i in range(self._ex_odlist.MaxSub[self._item]+1):             # <<<<<<<<<<<<<<
 *                 entry = CdefCoeObjectEntry(i)
 *                 entry._ex_oelist = &self._ex_oelist
*/
    __pyx_t_5 = ((__pyx_v_self->_ex_odlist->MaxSub[__pyx_v_self->_item]) + 1);
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "pysoem/pysoem.pyx":1570
 *             entries = []
 *             for i in range(self._ex_odlist.MaxSub[self._item]+1):
 *                 entry = CdefCoeObjectEntry(i)             # <<<<<<<<<<<<<<
 *                 entry._ex_oelist = &self._ex_oelist
 *                 entries.append(entry)
*/
      __pyx_t_2 = NULL;
      __pyx_t_8 = __Pyx_PyLong_From_long(__pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_8};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1570, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_entry, ((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "pysoem/pysoem.pyx":1571
 *             for i in range(self._ex_odlist.MaxSub[self._item]+1):
 *                 entry = CdefCoeObjectEntry(i)
 *                 entry._ex_oelist = &self._ex_oelist             # <<<<<<<<<<<<<<
 *                 entries.append(entry)
 *             return entries
*/
      __pyx_v_entry->_ex_oelist = (&__pyx_v_self->_ex_oelist);

      /* "pysoem/pysoem.pyx":1572
 *                 entry = CdefCoeObjectEntry(i)
 *                 entry._ex_oelist = &self._ex_oelist
 *                 entries.append(entry)             # <<<<<<<<<<<<<<
 *             return entries
 * 
*/
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_entries, ((PyObject *)__pyx_v_entry)); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 1572, __pyx_L1_error)
    }

    /* "pysoem/pysoem.pyx":1573
 *                 entry._ex_oelist = &self._ex_oelist
 *                 entries.append(entry)
 *             return entries             # <<<<<<<<<<<<<<
 * 
 *     def _get_bit_length(self):
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_entries);
    __pyx_r = __pyx_v_entries;
    goto __pyx_L0;
  }

  /* "pysoem/pysoem.pyx":1561
 *         return self._ex_odlist.Name[self._item]
 * 
 *     def _get_entries(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         self._read_entries()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject._get_entries", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_entries);
  __Pyx_XDECREF((PyObject *)__pyx_v_entry);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1575
 *             return entries
 * 
 *     def _get_bit_length(self):             # <<<<<<<<<<<<<<
 *         cdef int sum = 0
 *         self._read_description()
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_17_get_bit_length(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_17_get_bit_length = {"_get_bit_length", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_17_get_bit_length, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_17_get_bit_length(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_bit_length (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_bit_length", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_bit_length", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_16_get_bit_length(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_16_get_bit_length(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self) {
  int __pyx_v_sum;
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  long __pyx_t_5;
  long __pyx_t_6;
  long __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_bit_length", 0);

  /* "pysoem/pysoem.pyx":1576
 * 
 *     def _get_bit_length(self):
 *         cdef int sum = 0             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         self._read_entries()
*/
  __pyx_v_sum = 0;

  /* "pysoem/pysoem.pyx":1577
 *     def _get_bit_length(self):
 *         cdef int sum = 0
 *         self._read_description()             # <<<<<<<<<<<<<<
 *         self._read_entries()
 *         if self._ex_odlist.MaxSub[self._item] == 0:
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read_description, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1578
 *         cdef int sum = 0
 *         self._read_description()
 *         self._read_entries()             # <<<<<<<<<<<<<<
 *         if self._ex_odlist.MaxSub[self._item] == 0:
 *             return self._ex_oelist.BitLength[0]
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read_entries, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1579
 *         self._read_description()
 *         self._read_entries()
 *         if self._ex_odlist.MaxSub[self._item] == 0:             # <<<<<<<<<<<<<<
 *             return self._ex_oelist.BitLength[0]
 *         else:
*/
  __pyx_t_4 = ((__pyx_v_self->_ex_odlist->MaxSub[__pyx_v_self->_item]) == 0);
  if (__pyx_t_4) {

    /* "pysoem/pysoem.pyx":1580
 *         self._read_entries()
 *         if self._ex_odlist.MaxSub[self._item] == 0:
 *             return self._ex_oelist.BitLength[0]             # <<<<<<<<<<<<<<
 *         else:
 *             for i in range(self._ex_odlist.MaxSub[self._item]+1):
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyLong_From_uint16_t((__pyx_v_self->_ex_oelist.BitLength[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pysoem/pysoem.pyx":1579
 *         self._read_description()
 *         self._read_entries()
 *         if self._ex_odlist.MaxSub[self._item] == 0:             # <<<<<<<<<<<<<<
 *             return self._ex_oelist.BitLength[0]
 *         else:
*/
  }

  /* "pysoem/pysoem.pyx":1582
 *             return self._ex_oelist.BitLength[0]
 *         else:
 *             for i in range(self._ex_odlist.MaxSub[self._item]+1):             # <<<<<<<<<<<<<<
 *                 sum += self._ex_oelist.BitLength[i]
 *             return sum
*/
  /*else*/ {
    __pyx_t_5 = ((__pyx_v_self->_ex_odlist->MaxSub[__pyx_v_self->_item]) + 1);
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "pysoem/pysoem.pyx":1583
 *         else:
 *             for i in range(self._ex_odlist.MaxSub[self._item]+1):
 *                 sum += self._ex_oelist.BitLength[i]             # <<<<<<<<<<<<<<
 *             return sum
 * 
*/
      __pyx_v_sum = (__pyx_v_sum + (__pyx_v_self->_ex_oelist.BitLength[__pyx_v_i]));
    }

    /* "pysoem/pysoem.pyx":1584
 *             for i in range(self._ex_odlist.MaxSub[self._item]+1):
 *                 sum += self._ex_oelist.BitLength[i]
 *             return sum             # <<<<<<<<<<<<<<
 * 
 *     def _get_obj_access(self):
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_sum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "pysoem/pysoem.pyx":1575
 *             return entries
 * 
 *     def _get_bit_length(self):             # <<<<<<<<<<<<<<
 *         cdef int sum = 0
 *         self._read_description()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject._get_bit_length", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1586
 *             return sum
 * 
 *     def _get_obj_access(self):             # <<<<<<<<<<<<<<
 *         # read functions to return correct value if no previous access to other variables is done
 *         self._read_description()
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_19_get_obj_access(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_19_get_obj_access = {"_get_obj_access", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_19_get_obj_access, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_19_get_obj_access(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_obj_access (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_obj_access", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_obj_access", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_18_get_obj_access(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_18_get_obj_access(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_obj_access", 0);

  /* "pysoem/pysoem.pyx":1588
 *     def _get_obj_access(self):
 *         # read functions to return correct value if no previous access to other variables is done
 *         self._read_description()             # <<<<<<<<<<<<<<
 *         self._read_entries()
 *         if self._ex_odlist.MaxSub[self._item] == 0:
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read_description, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1589
 *         # read functions to return correct value if no previous access to other variables is done
 *         self._read_description()
 *         self._read_entries()             # <<<<<<<<<<<<<<
 *         if self._ex_odlist.MaxSub[self._item] == 0:
 *             return self._ex_oelist.ObjAccess[0]
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_read_entries, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1589, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1590
 *         self._read_description()
 *         self._read_entries()
 *         if self._ex_odlist.MaxSub[self._item] == 0:             # <<<<<<<<<<<<<<
 *             return self._ex_oelist.ObjAccess[0]
 *         else:
*/
  __pyx_t_4 = ((__pyx_v_self->_ex_odlist->MaxSub[__pyx_v_self->_item]) == 0);
  if (__pyx_t_4) {

    /* "pysoem/pysoem.pyx":1591
 *         self._read_entries()
 *         if self._ex_odlist.MaxSub[self._item] == 0:
 *             return self._ex_oelist.ObjAccess[0]             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyLong_From_uint16_t((__pyx_v_self->_ex_oelist.ObjAccess[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pysoem/pysoem.pyx":1590
 *         self._read_description()
 *         self._read_entries()
 *         if self._ex_odlist.MaxSub[self._item] == 0:             # <<<<<<<<<<<<<<
 *             return self._ex_oelist.ObjAccess[0]
 *         else:
*/
  }

  /* "pysoem/pysoem.pyx":1593
 *             return self._ex_oelist.ObjAccess[0]
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 * 
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
    __pyx_r = __pyx_mstate_global->__pyx_int_0;
    goto __pyx_L0;
  }

  /* "pysoem/pysoem.pyx":1586
 *             return sum
 * 
 *     def _get_obj_access(self):             # <<<<<<<<<<<<<<
 *         # read functions to return correct value if no previous access to other variables is done
 *         self._read_description()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject._get_obj_access", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_21__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_20__reduce_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self__ecx_context_self__ex_odlis, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_23__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_13CdefCoeObject_22__setstate_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_13CdefCoeObject_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self__ecx_context_self__ex_odlis, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObject.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1605
 *     obj_access = property(_get_obj_access)
 * 
 *     def __init__(self, int item):             # <<<<<<<<<<<<<<
 *         self._item = item
 * 
*/

/* Python wrapper */
static int __pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_item;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_item,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1605, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1605, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 1605, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 1605, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1605, __pyx_L3_error)
    }
    __pyx_v_item = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_item == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1605, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1605, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObjectEntry.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry___init__(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *)__pyx_v_self), __pyx_v_item);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry___init__(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self, int __pyx_v_item) {
  int __pyx_r;

  /* "pysoem/pysoem.pyx":1606
 * 
 *     def __init__(self, int item):
 *         self._item = item             # <<<<<<<<<<<<<<
 * 
 *     def _get_name(self):
*/
  __pyx_v_self->_item = __pyx_v_item;

  /* "pysoem/pysoem.pyx":1605
 *     obj_access = property(_get_obj_access)
 * 
 *     def __init__(self, int item):             # <<<<<<<<<<<<<<
 *         self._item = item
 * 
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1608
 *         self._item = item
 * 
 *     def _get_name(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.Name[self._item]
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_3_get_name(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_3_get_name = {"_get_name", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_3_get_name, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_3_get_name(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_name (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_name", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_name", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_2_get_name(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_2_get_name(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_name", 0);

  /* "pysoem/pysoem.pyx":1609
 * 
 *     def _get_name(self):
 *         return self._ex_oelist.Name[self._item]             # <<<<<<<<<<<<<<
 * 
 *     def _get_data_type(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString((__pyx_v_self->_ex_oelist->Name[__pyx_v_self->_item])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1608
 *         self._item = item
 * 
 *     def _get_name(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.Name[self._item]
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObjectEntry._get_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1611
 *         return self._ex_oelist.Name[self._item]
 * 
 *     def _get_data_type(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.DataType[self._item]
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_5_get_data_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_5_get_data_type = {"_get_data_type", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_5_get_data_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_5_get_data_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_data_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_data_type", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_data_type", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_4_get_data_type(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_4_get_data_type(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_data_type", 0);

  /* "pysoem/pysoem.pyx":1612
 * 
 *     def _get_data_type(self):
 *         return self._ex_oelist.DataType[self._item]             # <<<<<<<<<<<<<<
 * 
 *     def _get_bit_length(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint16_t((__pyx_v_self->_ex_oelist->DataType[__pyx_v_self->_item])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1611
 *         return self._ex_oelist.Name[self._item]
 * 
 *     def _get_data_type(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.DataType[self._item]
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObjectEntry._get_data_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1614
 *         return self._ex_oelist.DataType[self._item]
 * 
 *     def _get_bit_length(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.BitLength[self._item]
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_7_get_bit_length(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_7_get_bit_length = {"_get_bit_length", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_7_get_bit_length, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_7_get_bit_length(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_bit_length (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_bit_length", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_bit_length", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_6_get_bit_length(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_6_get_bit_length(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_bit_length", 0);

  /* "pysoem/pysoem.pyx":1615
 * 
 *     def _get_bit_length(self):
 *         return self._ex_oelist.BitLength[self._item]             # <<<<<<<<<<<<<<
 * 
 *     def _get_obj_access(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint16_t((__pyx_v_self->_ex_oelist->BitLength[__pyx_v_self->_item])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1614
 *         return self._ex_oelist.DataType[self._item]
 * 
 *     def _get_bit_length(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.BitLength[self._item]
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObjectEntry._get_bit_length", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1617
 *         return self._ex_oelist.BitLength[self._item]
 * 
 *     def _get_obj_access(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.ObjAccess[self._item]
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_9_get_obj_access(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_9_get_obj_access = {"_get_obj_access", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_9_get_obj_access, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_9_get_obj_access(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_obj_access (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("_get_obj_access", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("_get_obj_access", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_8_get_obj_access(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_8_get_obj_access(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_obj_access", 0);

  /* "pysoem/pysoem.pyx":1618
 * 
 *     def _get_obj_access(self):
 *         return self._ex_oelist.ObjAccess[self._item]             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint16_t((__pyx_v_self->_ex_oelist->ObjAccess[__pyx_v_self->_item])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pysoem/pysoem.pyx":1617
 *         return self._ex_oelist.BitLength[self._item]
 * 
 *     def _get_obj_access(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.ObjAccess[self._item]
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObjectEntry._get_obj_access", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_10__reduce_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self__ex_oelist_cannot_be_conver, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObjectEntry.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObjectEntry.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_12__setstate_cython__(((struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_18CdefCoeObjectEntry_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self__ex_oelist_cannot_be_conver, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pysoem.pysoem.CdefCoeObjectEntry.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1621
 * 
 * 
 * cdef int _xPO2SOconfig(cpysoem.uint16 slave, void* user) noexcept:             # <<<<<<<<<<<<<<
 *     cdef _CallbackData cd
 *     cd = <object>user
*/

static int __pyx_f_6pysoem_6pysoem__xPO2SOconfig(uint16 __pyx_v_slave, void *__pyx_v_user) {
  struct __pyx_obj_6pysoem_6pysoem__CallbackData *__pyx_v_cd = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_xPO2SOconfig", 0);

  /* "pysoem/pysoem.pyx":1623
 * cdef int _xPO2SOconfig(cpysoem.uint16 slave, void* user) noexcept:
 *     cdef _CallbackData cd
 *     cd = <object>user             # <<<<<<<<<<<<<<
 *     cd.exc_raised = False
 *     try:
*/
  __pyx_t_1 = ((PyObject *)__pyx_v_user);
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem__CallbackData))))) __PYX_ERR(0, 1623, __pyx_L1_error)
  __pyx_v_cd = ((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1624
 *     cdef _CallbackData cd
 *     cd = <object>user
 *     cd.exc_raised = False             # <<<<<<<<<<<<<<
 *     try:
 *         (<object>cd.func)(slave-1)
*/
  __Pyx_INCREF(Py_False);
  __Pyx_GIVEREF(Py_False);
  __Pyx_GOTREF(__pyx_v_cd->exc_raised);
  __Pyx_DECREF(__pyx_v_cd->exc_raised);
  __pyx_v_cd->exc_raised = Py_False;

  /* "pysoem/pysoem.pyx":1625
 *     cd = <object>user
 *     cd.exc_raised = False
 *     try:             # <<<<<<<<<<<<<<
 *         (<object>cd.func)(slave-1)
 *     except:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "pysoem/pysoem.pyx":1626
 *     cd.exc_raised = False
 *     try:
 *         (<object>cd.func)(slave-1)             # <<<<<<<<<<<<<<
 *     except:
 *         cd.exc_raised = True
*/
      __pyx_t_5 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_cd->func));
      __pyx_t_6 = __pyx_v_cd->func; 
      __pyx_t_7 = __Pyx_PyLong_From_long((__pyx_v_slave - 1)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1626, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        assert(__pyx_t_5);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
        __pyx_t_8 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_7};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1626, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pysoem/pysoem.pyx":1625
 *     cd = <object>user
 *     cd.exc_raised = False
 *     try:             # <<<<<<<<<<<<<<
 *         (<object>cd.func)(slave-1)
 *     except:
*/
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "pysoem/pysoem.pyx":1627
 *     try:
 *         (<object>cd.func)(slave-1)
 *     except:             # <<<<<<<<<<<<<<
 *         cd.exc_raised = True
 *         cd.exc_info = sys.exc_info()
*/
    /*except:*/ {
      __Pyx_AddTraceback("pysoem.pysoem._xPO2SOconfig", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 1627, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "pysoem/pysoem.pyx":1628
 *         (<object>cd.func)(slave-1)
 *     except:
 *         cd.exc_raised = True             # <<<<<<<<<<<<<<
 *         cd.exc_info = sys.exc_info()
 * 
*/
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      __Pyx_GOTREF(__pyx_v_cd->exc_raised);
      __Pyx_DECREF(__pyx_v_cd->exc_raised);
      __pyx_v_cd->exc_raised = Py_True;

      /* "pysoem/pysoem.pyx":1629
 *     except:
 *         cd.exc_raised = True
 *         cd.exc_info = sys.exc_info()             # <<<<<<<<<<<<<<
 * 
 * 
*/
      __pyx_t_9 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_sys); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1629, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_exc_info); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1629, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_8 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_11);
        assert(__pyx_t_9);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
        __pyx_t_8 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
        __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1629, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_v_cd->exc_info);
      __Pyx_DECREF(__pyx_v_cd->exc_info);
      __pyx_v_cd->exc_info = __pyx_t_5;
      __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }

    /* "pysoem/pysoem.pyx":1625
 *     cd = <object>user
 *     cd.exc_raised = False
 *     try:             # <<<<<<<<<<<<<<
 *         (<object>cd.func)(slave-1)
 *     except:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L8_try_end:;
  }

  /* "pysoem/pysoem.pyx":1621
 * 
 * 
 * cdef int _xPO2SOconfig(cpysoem.uint16 slave, void* user) noexcept:             # <<<<<<<<<<<<<<
 *     cdef _CallbackData cd
 *     cd = <object>user
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_WriteUnraisable("pysoem.pysoem._xPO2SOconfig", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cd);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pysoem/pysoem.pyx":1632
 * 
 * 
 * cdef int _xPO2SOconfigEx(cpysoem.uint16 slave, void* user) noexcept:             # <<<<<<<<<<<<<<
 *     cdef _CallbackData cd
 *     cd = <object>user
*/

static int __pyx_f_6pysoem_6pysoem__xPO2SOconfigEx(CYTHON_UNUSED uint16 __pyx_v_slave, void *__pyx_v_user) {
  struct __pyx_obj_6pysoem_6pysoem__CallbackData *__pyx_v_cd = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_xPO2SOconfigEx", 0);

  /* "pysoem/pysoem.pyx":1634
 * cdef int _xPO2SOconfigEx(cpysoem.uint16 slave, void* user) noexcept:
 *     cdef _CallbackData cd
 *     cd = <object>user             # <<<<<<<<<<<<<<
 *     cd.exc_raised = False
 *     try:
*/
  __pyx_t_1 = ((PyObject *)__pyx_v_user);
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem__CallbackData))))) __PYX_ERR(0, 1634, __pyx_L1_error)
  __pyx_v_cd = ((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pysoem/pysoem.pyx":1635
 *     cdef _CallbackData cd
 *     cd = <object>user
 *     cd.exc_raised = False             # <<<<<<<<<<<<<<
 *     try:
 *         (<object>cd.func)(cd.slave)
*/
  __Pyx_INCREF(Py_False);
  __Pyx_GIVEREF(Py_False);
  __Pyx_GOTREF(__pyx_v_cd->exc_raised);
  __Pyx_DECREF(__pyx_v_cd->exc_raised);
  __pyx_v_cd->exc_raised = Py_False;

  /* "pysoem/pysoem.pyx":1636
 *     cd = <object>user
 *     cd.exc_raised = False
 *     try:             # <<<<<<<<<<<<<<
 *         (<object>cd.func)(cd.slave)
 *     except:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "pysoem/pysoem.pyx":1637
 *     cd.exc_raised = False
 *     try:
 *         (<object>cd.func)(cd.slave)             # <<<<<<<<<<<<<<
 *     except:
 *         cd.exc_raised = True
*/
      __pyx_t_5 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_cd->func));
      __pyx_t_6 = __pyx_v_cd->func; 
      __pyx_t_7 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        assert(__pyx_t_5);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
        __pyx_t_7 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_cd->slave};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1637, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pysoem/pysoem.pyx":1636
 *     cd = <object>user
 *     cd.exc_raised = False
 *     try:             # <<<<<<<<<<<<<<
 *         (<object>cd.func)(cd.slave)
 *     except:
*/
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "pysoem/pysoem.pyx":1638
 *     try:
 *         (<object>cd.func)(cd.slave)
 *     except:             # <<<<<<<<<<<<<<
 *         cd.exc_raised = True
 *         cd.exc_info = sys.exc_info()
*/
    /*except:*/ {
      __Pyx_AddTraceback("pysoem.pysoem._xPO2SOconfigEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 1638, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);

      /* "pysoem/pysoem.pyx":1639
 *         (<object>cd.func)(cd.slave)
 *     except:
 *         cd.exc_raised = True             # <<<<<<<<<<<<<<
 *         cd.exc_info = sys.exc_info()
*/
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      __Pyx_GOTREF(__pyx_v_cd->exc_raised);
      __Pyx_DECREF(__pyx_v_cd->exc_raised);
      __pyx_v_cd->exc_raised = Py_True;

      /* "pysoem/pysoem.pyx":1640
 *     except:
 *         cd.exc_raised = True
 *         cd.exc_info = sys.exc_info()             # <<<<<<<<<<<<<<
*/
      __pyx_t_9 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_sys); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1640, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_exc_info); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1640, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_7 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_11);
        assert(__pyx_t_9);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
        __pyx_t_7 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1640, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_GIVEREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_v_cd->exc_info);
      __Pyx_DECREF(__pyx_v_cd->exc_info);
      __pyx_v_cd->exc_info = __pyx_t_8;
      __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L4_exception_handled;
    }

    /* "pysoem/pysoem.pyx":1636
 *     cd = <object>user
 *     cd.exc_raised = False
 *     try:             # <<<<<<<<<<<<<<
 *         (<object>cd.func)(cd.slave)
 *     except:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L8_try_end:;
  }

  /* "pysoem/pysoem.pyx":1632
 * 
 * 
 * cdef int _xPO2SOconfigEx(cpysoem.uint16 slave, void* user) noexcept:             # <<<<<<<<<<<<<<
 *     cdef _CallbackData cd
 *     cd = <object>user
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_WriteUnraisable("pysoem.pysoem._xPO2SOconfigEx", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cd);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_CdefSettings(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe6b07d6, 0xdac0a2c, 0x9cc8369, b'always_release_gil, timeouts')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_8__pyx_unpickle_CdefSettings(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_8__pyx_unpickle_CdefSettings = {"__pyx_unpickle_CdefSettings", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_8__pyx_unpickle_CdefSettings, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_8__pyx_unpickle_CdefSettings(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CdefSettings (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_CdefSettings", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CdefSettings", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CdefSettings", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.__pyx_unpickle_CdefSettings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_7__pyx_unpickle_CdefSettings(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_7__pyx_unpickle_CdefSettings(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CdefSettings", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle_CdefSettings(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe6b07d6, 0xdac0a2c, 0x9cc8369, b'always_release_gil, timeouts')             # <<<<<<<<<<<<<<
 *     __pyx_result = CdefSettings.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0xe6b07d6, 0xdac0a2c, 0x9cc8369, __pyx_k_always_release_gil_timeouts); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe6b07d6, 0xdac0a2c, 0x9cc8369, b'always_release_gil, timeouts')
 *     __pyx_result = CdefSettings.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CdefSettings__set_state(<CdefSettings> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSettings);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe6b07d6, 0xdac0a2c, 0x9cc8369, b'always_release_gil, timeouts')
 *     __pyx_result = CdefSettings.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CdefSettings__set_state(<CdefSettings> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = CdefSettings.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CdefSettings__set_state(<CdefSettings> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CdefSettings__set_state(CdefSettings __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_6pysoem_6pysoem___pyx_unpickle_CdefSettings__set_state(((struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe6b07d6, 0xdac0a2c, 0x9cc8369, b'always_release_gil, timeouts')
 *     __pyx_result = CdefSettings.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CdefSettings__set_state(<CdefSettings> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CdefSettings__set_state(<CdefSettings> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CdefSettings__set_state(CdefSettings __pyx_result, __pyx_state: tuple):
 *     __pyx_result.always_release_gil = __pyx_state[0]; __pyx_result.timeouts = __pyx_state[1]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_CdefSettings(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe6b07d6, 0xdac0a2c, 0x9cc8369, b'always_release_gil, timeouts')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pysoem.pysoem.__pyx_unpickle_CdefSettings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CdefSettings__set_state(<CdefSettings> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CdefSettings__set_state(CdefSettings __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.always_release_gil = __pyx_state[0]; __pyx_result.timeouts = __pyx_state[1]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 2)
*/

static PyObject *__pyx_f_6pysoem_6pysoem___pyx_unpickle_CdefSettings__set_state(struct __pyx_obj_6pysoem_6pysoem_CdefSettings *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  boolean __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CdefSettings__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CdefSettings__set_state(CdefSettings __pyx_result, __pyx_state: tuple):
 *     __pyx_result.always_release_gil = __pyx_state[0]; __pyx_result.timeouts = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 2)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_int8_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((boolean)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->always_release_gil = __pyx_t_2;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->timeouts);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->timeouts);
  __pyx_v___pyx_result->timeouts = ((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CdefSettings__set_state(CdefSettings __pyx_result, __pyx_state: tuple):
 *     __pyx_result.always_release_gil = __pyx_state[0]; __pyx_result.timeouts = __pyx_state[1]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 2)             # <<<<<<<<<<<<<<
*/
  __pyx_t_3 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle_CdefSettings__set_state(<CdefSettings> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CdefSettings__set_state(CdefSettings __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.always_release_gil = __pyx_state[0]; __pyx_result.timeouts = __pyx_state[1]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 2)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.__pyx_unpickle_CdefSettings__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__CallbackData(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x0955bee, 0x383c86a, 0xa1aa8ca, b'exc_info, exc_raised, func, slave')
*/

/* Python wrapper */
static PyObject *__pyx_pw_6pysoem_6pysoem_10__pyx_unpickle__CallbackData(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6pysoem_6pysoem_10__pyx_unpickle__CallbackData = {"__pyx_unpickle__CallbackData", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10__pyx_unpickle__CallbackData, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pysoem_6pysoem_10__pyx_unpickle__CallbackData(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__CallbackData (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 4, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle__CallbackData", 0) < (0)) __PYX_ERR(1, 4, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__CallbackData", 1, 3, 3, i); __PYX_ERR(1, 4, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 4, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 4, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L3_error)
    __pyx_v___pyx_state = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__CallbackData", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pysoem.pysoem.__pyx_unpickle__CallbackData", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v___pyx_state), (&PyTuple_Type), 1, "__pyx_state", 1))) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pysoem_6pysoem_9__pyx_unpickle__CallbackData(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pysoem_6pysoem_9__pyx_unpickle__CallbackData(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__CallbackData", 0);

  /* "(tree fragment)":6
 * def __pyx_unpickle__CallbackData(__pyx_type, long __pyx_checksum, tuple __pyx_state):
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x0955bee, 0x383c86a, 0xa1aa8ca, b'exc_info, exc_raised, func, slave')             # <<<<<<<<<<<<<<
 *     __pyx_result = _CallbackData.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
  __pyx_t_1 = __Pyx_CheckUnpickleChecksum(__pyx_v___pyx_checksum, 0x0955bee, 0x383c86a, 0xa1aa8ca, __pyx_k_exc_info_exc_raised_func_slave); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 6, __pyx_L1_error)

  /* "(tree fragment)":7
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x0955bee, 0x383c86a, 0xa1aa8ca, b'exc_info, exc_raised, func, slave')
 *     __pyx_result = _CallbackData.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__CallbackData__set_state(<_CallbackData> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem__CallbackData);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x0955bee, 0x383c86a, 0xa1aa8ca, b'exc_info, exc_raised, func, slave')
 *     __pyx_result = _CallbackData.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__CallbackData__set_state(<_CallbackData> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_5 = (__pyx_v___pyx_state != ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "(tree fragment)":9
 *     __pyx_result = _CallbackData.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__CallbackData__set_state(<_CallbackData> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__CallbackData__set_state(_CallbackData __pyx_result, __pyx_state: tuple):
*/
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "cannot pass None into a C function argument that is declared 'not None'");
      __PYX_ERR(1, 9, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_f_6pysoem_6pysoem___pyx_unpickle__CallbackData__set_state(((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)__pyx_v___pyx_result), __pyx_v___pyx_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x0955bee, 0x383c86a, 0xa1aa8ca, b'exc_info, exc_raised, func, slave')
 *     __pyx_result = _CallbackData.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__CallbackData__set_state(<_CallbackData> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__CallbackData__set_state(<_CallbackData> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__CallbackData__set_state(_CallbackData __pyx_result, __pyx_state: tuple):
 *     __pyx_result.exc_info = __pyx_state[0]; __pyx_result.exc_raised = __pyx_state[1]; __pyx_result.func = __pyx_state[2]; __pyx_result.slave = __pyx_state[3]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle__CallbackData(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0x0955bee, 0x383c86a, 0xa1aa8ca, b'exc_info, exc_raised, func, slave')
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pysoem.pysoem.__pyx_unpickle__CallbackData", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__CallbackData__set_state(<_CallbackData> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__CallbackData__set_state(_CallbackData __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.exc_info = __pyx_state[0]; __pyx_result.exc_raised = __pyx_state[1]; __pyx_result.func = __pyx_state[2]; __pyx_result.slave = __pyx_state[3]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 4)
*/

static PyObject *__pyx_f_6pysoem_6pysoem___pyx_unpickle__CallbackData__set_state(struct __pyx_obj_6pysoem_6pysoem__CallbackData *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__CallbackData__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__CallbackData__set_state(_CallbackData __pyx_result, __pyx_state: tuple):
 *     __pyx_result.exc_info = __pyx_state[0]; __pyx_result.exc_raised = __pyx_state[1]; __pyx_result.func = __pyx_state[2]; __pyx_result.slave = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 4)
*/
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->exc_info);
  __Pyx_DECREF(__pyx_v___pyx_result->exc_info);
  __pyx_v___pyx_result->exc_info = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->exc_raised);
  __Pyx_DECREF(__pyx_v___pyx_result->exc_raised);
  __pyx_v___pyx_result->exc_raised = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->func);
  __Pyx_DECREF(__pyx_v___pyx_result->func);
  __pyx_v___pyx_result->func = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->slave);
  __Pyx_DECREF(__pyx_v___pyx_result->slave);
  __pyx_v___pyx_result->slave = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__CallbackData__set_state(_CallbackData __pyx_result, __pyx_state: tuple):
 *     __pyx_result.exc_info = __pyx_state[0]; __pyx_result.exc_raised = __pyx_state[1]; __pyx_result.func = __pyx_state[2]; __pyx_result.slave = __pyx_state[3]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 4)             # <<<<<<<<<<<<<<
*/
  __pyx_t_2 = __Pyx_UpdateUnpickledDict(((PyObject *)__pyx_v___pyx_result), __pyx_v___pyx_state, 4); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

  /* "(tree fragment)":11
 *         __pyx_unpickle__CallbackData__set_state(<_CallbackData> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__CallbackData__set_state(_CallbackData __pyx_result, __pyx_state: tuple):             # <<<<<<<<<<<<<<
 *     __pyx_result.exc_info = __pyx_state[0]; __pyx_result.exc_raised = __pyx_state[1]; __pyx_result.func = __pyx_state[2]; __pyx_result.slave = __pyx_state[3]
 *     __Pyx_UpdateUnpickledDict(__pyx_result, __pyx_state, 4)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pysoem.pysoem.__pyx_unpickle__CallbackData__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */

static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefTimeouts(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6pysoem_6pysoem_CdefTimeouts(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pysoem_6pysoem_CdefTimeouts) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_ret(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_3ret_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_ret(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_3ret_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_safe(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_4safe_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_safe(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_4safe_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_eeprom(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_6eeprom_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_eeprom(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_6eeprom_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_tx_mailbox(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_tx_mailbox(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10tx_mailbox_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_rx_mailbox(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_rx_mailbox(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_10rx_mailbox_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_5state_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_state(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_12CdefTimeouts_5state_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6pysoem_6pysoem_CdefTimeouts[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_12CdefTimeouts_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pysoem_6pysoem_CdefTimeouts[] = {
  {"ret", __pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_ret, __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_ret, 0, 0},
  {"safe", __pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_safe, __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_safe, 0, 0},
  {"eeprom", __pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_eeprom, __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_eeprom, 0, 0},
  {"tx_mailbox", __pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_tx_mailbox, __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_tx_mailbox, 0, 0},
  {"rx_mailbox", __pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_rx_mailbox, __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_rx_mailbox, 0, 0},
  {"state", __pyx_getprop_6pysoem_6pysoem_12CdefTimeouts_state, __pyx_setprop_6pysoem_6pysoem_12CdefTimeouts_state, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pysoem_6pysoem_CdefTimeouts_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pysoem_6pysoem_CdefTimeouts},
  {Py_tp_methods, (void *)__pyx_methods_6pysoem_6pysoem_CdefTimeouts},
  {Py_tp_getset, (void *)__pyx_getsets_6pysoem_6pysoem_CdefTimeouts},
  {Py_tp_new, (void *)__pyx_tp_new_6pysoem_6pysoem_CdefTimeouts},
  {0, 0},
};
static PyType_Spec __pyx_type_6pysoem_6pysoem_CdefTimeouts_spec = {
  "pysoem.pysoem.CdefTimeouts",
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_6pysoem_6pysoem_CdefTimeouts_slots,
};
#else

static PyTypeObject __pyx_type_6pysoem_6pysoem_CdefTimeouts = {
  PyVarObject_HEAD_INIT(0, 0)
  "pysoem.pysoem.""CdefTimeouts", /*tp_name*/
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pysoem_6pysoem_CdefTimeouts, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pysoem_6pysoem_CdefTimeouts, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pysoem_6pysoem_CdefTimeouts, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pysoem_6pysoem_CdefTimeouts, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefSettings(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pysoem_6pysoem_CdefSettings *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)o);
  p->timeouts = ((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pysoem_6pysoem_CdefSettings(PyObject *o) {
  struct __pyx_obj_6pysoem_6pysoem_CdefSettings *p = (struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pysoem_6pysoem_CdefSettings) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->timeouts);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6pysoem_6pysoem_CdefSettings(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pysoem_6pysoem_CdefSettings *p = (struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->timeouts) {
    e = (*v)(((PyObject *)p->timeouts), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pysoem_6pysoem_CdefSettings(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pysoem_6pysoem_CdefSettings *p = (struct __pyx_obj_6pysoem_6pysoem_CdefSettings *)o;
  tmp = ((PyObject*)p->timeouts);
  p->timeouts = ((struct __pyx_obj_6pysoem_6pysoem_CdefTimeouts *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_12CdefSettings_timeouts(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_12CdefSettings_8timeouts_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_12CdefSettings_timeouts(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_12CdefSettings_8timeouts_3__set__(o, v);
  }
  else {
    return __pyx_pw_6pysoem_6pysoem_12CdefSettings_8timeouts_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_12CdefSettings_always_release_gil(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_12CdefSettings_18always_release_gil_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_12CdefSettings_always_release_gil(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_12CdefSettings_18always_release_gil_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6pysoem_6pysoem_CdefSettings[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_12CdefSettings_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_12CdefSettings_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pysoem_6pysoem_CdefSettings[] = {
  {"timeouts", __pyx_getprop_6pysoem_6pysoem_12CdefSettings_timeouts, __pyx_setprop_6pysoem_6pysoem_12CdefSettings_timeouts, 0, 0},
  {"always_release_gil", __pyx_getprop_6pysoem_6pysoem_12CdefSettings_always_release_gil, __pyx_setprop_6pysoem_6pysoem_12CdefSettings_always_release_gil, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pysoem_6pysoem_CdefSettings_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pysoem_6pysoem_CdefSettings},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pysoem_6pysoem_CdefSettings},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pysoem_6pysoem_CdefSettings},
  {Py_tp_methods, (void *)__pyx_methods_6pysoem_6pysoem_CdefSettings},
  {Py_tp_getset, (void *)__pyx_getsets_6pysoem_6pysoem_CdefSettings},
  {Py_tp_init, (void *)__pyx_pw_6pysoem_6pysoem_12CdefSettings_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pysoem_6pysoem_CdefSettings},
  {0, 0},
};
static PyType_Spec __pyx_type_6pysoem_6pysoem_CdefSettings_spec = {
  "pysoem.pysoem.CdefSettings",
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefSettings),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pysoem_6pysoem_CdefSettings_slots,
};
#else

static PyTypeObject __pyx_type_6pysoem_6pysoem_CdefSettings = {
  PyVarObject_HEAD_INIT(0, 0)
  "pysoem.pysoem.""CdefSettings", /*tp_name*/
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefSettings), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pysoem_6pysoem_CdefSettings, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pysoem_6pysoem_CdefSettings, /*tp_traverse*/
  __pyx_tp_clear_6pysoem_6pysoem_CdefSettings, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pysoem_6pysoem_CdefSettings, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pysoem_6pysoem_CdefSettings, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pysoem_6pysoem_12CdefSettings_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pysoem_6pysoem_CdefSettings, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pysoem_6pysoem_CdefMaster __pyx_vtable_6pysoem_6pysoem_CdefMaster;

static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefMaster(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pysoem_6pysoem_CdefMaster *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pysoem_6pysoem_CdefMaster;
  if (unlikely(__pyx_pw_6pysoem_6pysoem_10CdefMaster_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6pysoem_6pysoem_CdefMaster(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pysoem_6pysoem_CdefMaster) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_10CdefMaster_sdo_read_timeout(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_10CdefMaster_sdo_read_timeout(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_10CdefMaster_16sdo_read_timeout_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_10CdefMaster_sdo_write_timeout(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_10CdefMaster_sdo_write_timeout(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_10CdefMaster_17sdo_write_timeout_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_10CdefMaster_always_release_gil(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_10CdefMaster_18always_release_gil_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_10CdefMaster_always_release_gil(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_10CdefMaster_18always_release_gil_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_10CdefMaster_context_initialized(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_10CdefMaster_19context_initialized_1__get__(o);
}

static PyMethodDef __pyx_methods_6pysoem_6pysoem_CdefMaster[] = {
  {"open", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_3open, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_2open},
  {"check_context_is_initialized", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_5check_context_is_initialized, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"config_init", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_9config_init, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_8config_init},
  {"config_map", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_11config_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_10config_map},
  {"config_overlap_map", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_13config_overlap_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_12config_overlap_map},
  {"_collect_mailbox_errors", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_15_collect_mailbox_errors, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"config_dc", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_17config_dc, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_16config_dc},
  {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_19close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_18close},
  {"read_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_21read_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_20read_state},
  {"write_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_23write_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_22write_state},
  {"state_check", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_25state_check, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_24state_check},
  {"send_processdata", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_27send_processdata, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_26send_processdata},
  {"send_overlap_processdata", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_29send_overlap_processdata, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_28send_overlap_processdata},
  {"receive_processdata", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_31receive_processdata, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_30receive_processdata},
  {"_get_slave", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_33_get_slave, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_35_get_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_34_get_state},
  {"_set_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_37_set_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_expected_wkc", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_39_get_expected_wkc, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_38_get_expected_wkc},
  {"_get_dc_time", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_41_get_dc_time, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_40_get_dc_time},
  {"_set_manual_state_change", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_43_set_manual_state_change, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_10CdefMaster_42_set_manual_state_change},
  {"_get_manual_state_change", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_45_get_manual_state_change, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_47__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_10CdefMaster_49__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pysoem_6pysoem_CdefMaster[] = {
  {"sdo_read_timeout", __pyx_getprop_6pysoem_6pysoem_10CdefMaster_sdo_read_timeout, __pyx_setprop_6pysoem_6pysoem_10CdefMaster_sdo_read_timeout, 0, 0},
  {"sdo_write_timeout", __pyx_getprop_6pysoem_6pysoem_10CdefMaster_sdo_write_timeout, __pyx_setprop_6pysoem_6pysoem_10CdefMaster_sdo_write_timeout, 0, 0},
  {"always_release_gil", __pyx_getprop_6pysoem_6pysoem_10CdefMaster_always_release_gil, __pyx_setprop_6pysoem_6pysoem_10CdefMaster_always_release_gil, 0, 0},
  {"context_initialized", __pyx_getprop_6pysoem_6pysoem_10CdefMaster_context_initialized, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pysoem_6pysoem_CdefMaster_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pysoem_6pysoem_CdefMaster},
  {Py_tp_doc, (void *)PyDoc_STR("Representing a logical EtherCAT master device.\n    \n    Please do not use this class directly, but the class Master instead.\n    Master is a typical Python object, with all it's benefits over\n    cdef classes. For example you can add new attributes dynamically.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_6pysoem_6pysoem_CdefMaster},
  {Py_tp_getset, (void *)__pyx_getsets_6pysoem_6pysoem_CdefMaster},
  {Py_tp_new, (void *)__pyx_tp_new_6pysoem_6pysoem_CdefMaster},
  {0, 0},
};
static PyType_Spec __pyx_type_6pysoem_6pysoem_CdefMaster_spec = {
  "pysoem.pysoem.CdefMaster",
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefMaster),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_6pysoem_6pysoem_CdefMaster_slots,
};
#else

static PyTypeObject __pyx_type_6pysoem_6pysoem_CdefMaster = {
  PyVarObject_HEAD_INIT(0, 0)
  "pysoem.pysoem.""CdefMaster", /*tp_name*/
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefMaster), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pysoem_6pysoem_CdefMaster, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Representing a logical EtherCAT master device.\n    \n    Please do not use this class directly, but the class Master instead.\n    Master is a typical Python object, with all it's benefits over\n    cdef classes. For example you can add new attributes dynamically.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pysoem_6pysoem_CdefMaster, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pysoem_6pysoem_CdefMaster, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pysoem_6pysoem_CdefMaster, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_6pysoem_6pysoem__CallbackData(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pysoem_6pysoem__CallbackData *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)o);
  p->slave = Py_None; Py_INCREF(Py_None);
  p->func = Py_None; Py_INCREF(Py_None);
  p->exc_raised = Py_None; Py_INCREF(Py_None);
  p->exc_info = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pysoem_6pysoem__CallbackData(PyObject *o) {
  struct __pyx_obj_6pysoem_6pysoem__CallbackData *p = (struct __pyx_obj_6pysoem_6pysoem__CallbackData *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pysoem_6pysoem__CallbackData) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->slave);
  Py_CLEAR(p->func);
  Py_CLEAR(p->exc_raised);
  Py_CLEAR(p->exc_info);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6pysoem_6pysoem__CallbackData(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pysoem_6pysoem__CallbackData *p = (struct __pyx_obj_6pysoem_6pysoem__CallbackData *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->slave) {
    e = (*v)(p->slave, a); if (e) return e;
  }
  if (p->func) {
    e = (*v)(p->func, a); if (e) return e;
  }
  if (p->exc_raised) {
    e = (*v)(p->exc_raised, a); if (e) return e;
  }
  if (p->exc_info) {
    e = (*v)(p->exc_info, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pysoem_6pysoem__CallbackData(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pysoem_6pysoem__CallbackData *p = (struct __pyx_obj_6pysoem_6pysoem__CallbackData *)o;
  tmp = ((PyObject*)p->slave);
  p->slave = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->func);
  p->func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->exc_raised);
  p->exc_raised = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->exc_info);
  p->exc_info = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6pysoem_6pysoem__CallbackData[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13_CallbackData_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13_CallbackData_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pysoem_6pysoem__CallbackData_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pysoem_6pysoem__CallbackData},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pysoem_6pysoem__CallbackData},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pysoem_6pysoem__CallbackData},
  {Py_tp_methods, (void *)__pyx_methods_6pysoem_6pysoem__CallbackData},
  {Py_tp_new, (void *)__pyx_tp_new_6pysoem_6pysoem__CallbackData},
  {0, 0},
};
static PyType_Spec __pyx_type_6pysoem_6pysoem__CallbackData_spec = {
  "pysoem.pysoem._CallbackData",
  sizeof(struct __pyx_obj_6pysoem_6pysoem__CallbackData),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pysoem_6pysoem__CallbackData_slots,
};
#else

static PyTypeObject __pyx_type_6pysoem_6pysoem__CallbackData = {
  PyVarObject_HEAD_INIT(0, 0)
  "pysoem.pysoem.""_CallbackData", /*tp_name*/
  sizeof(struct __pyx_obj_6pysoem_6pysoem__CallbackData), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pysoem_6pysoem__CallbackData, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pysoem_6pysoem__CallbackData, /*tp_traverse*/
  __pyx_tp_clear_6pysoem_6pysoem__CallbackData, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pysoem_6pysoem__CallbackData, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pysoem_6pysoem__CallbackData, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6pysoem_6pysoem_CdefSlave __pyx_vtable_6pysoem_6pysoem_CdefSlave;

static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefSlave(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pysoem_6pysoem_CdefSlave *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pysoem_6pysoem_CdefSlave;
  p->_master = ((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)Py_None); Py_INCREF(Py_None);
  p->_cd = ((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)Py_None); Py_INCREF(Py_None);
  p->_emcy_callbacks = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pysoem_6pysoem_CdefSlave(PyObject *o) {
  struct __pyx_obj_6pysoem_6pysoem_CdefSlave *p = (struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pysoem_6pysoem_CdefSlave) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_master);
  Py_CLEAR(p->_cd);
  Py_CLEAR(p->_emcy_callbacks);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_6pysoem_6pysoem_CdefSlave(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pysoem_6pysoem_CdefSlave *p = (struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->_master) {
    e = (*v)(((PyObject *)p->_master), a); if (e) return e;
  }
  if (p->_cd) {
    e = (*v)(((PyObject *)p->_cd), a); if (e) return e;
  }
  if (p->_emcy_callbacks) {
    e = (*v)(p->_emcy_callbacks, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pysoem_6pysoem_CdefSlave(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pysoem_6pysoem_CdefSlave *p = (struct __pyx_obj_6pysoem_6pysoem_CdefSlave *)o;
  tmp = ((PyObject*)p->_master);
  p->_master = ((struct __pyx_obj_6pysoem_6pysoem_CdefMaster *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_cd);
  p->_cd = ((struct __pyx_obj_6pysoem_6pysoem__CallbackData *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_emcy_callbacks);
  p->_emcy_callbacks = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_9CdefSlave__master(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_9CdefSlave_7_master_1__get__(o);
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_9CdefSlave__cd(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_9CdefSlave_3_cd_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_9CdefSlave__cd(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_9CdefSlave_3_cd_3__set__(o, v);
  }
  else {
    return __pyx_pw_6pysoem_6pysoem_9CdefSlave_3_cd_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6pysoem_6pysoem_9CdefSlave__emcy_callbacks(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_1__get__(o);
}

static int __pyx_setprop_6pysoem_6pysoem_9CdefSlave__emcy_callbacks(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_3__set__(o, v);
  }
  else {
    return __pyx_pw_6pysoem_6pysoem_9CdefSlave_15_emcy_callbacks_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_6pysoem_6pysoem_CdefSlave[] = {
  {"dc_sync", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_3dc_sync, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_2dc_sync},
  {"sdo_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_5sdo_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_4sdo_read},
  {"sdo_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_7sdo_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_6sdo_write},
  {"mbx_receive", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_9mbx_receive, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_8mbx_receive},
  {"write_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_11write_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_10write_state},
  {"state_check", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_13state_check, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_12state_check},
  {"reconfig", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_15reconfig, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_14reconfig},
  {"recover", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_17recover, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_16recover},
  {"eeprom_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_19eeprom_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_18eeprom_read},
  {"eeprom_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_21eeprom_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_20eeprom_write},
  {"foe_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_23foe_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_22foe_write},
  {"foe_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_25foe_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_24foe_read},
  {"amend_mbx", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_27amend_mbx, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_26amend_mbx},
  {"set_watchdog", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_29set_watchdog, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_28set_watchdog},
  {"get_watchdog", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_31get_watchdog, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_30get_watchdog},
  {"_get_watchdog_divider_ns", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_33_get_watchdog_divider_ns, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_max_watchdog_time", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_35get_max_watchdog_time, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_34get_max_watchdog_time},
  {"add_emergency_callback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_37add_emergency_callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_36add_emergency_callback},
  {"_disable_complete_access", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_39_disable_complete_access, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_38_disable_complete_access},
  {"_fprd", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_41_fprd, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_40_fprd},
  {"_fpwr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_43_fpwr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_42_fpwr},
  {"_get_name", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_45_get_name, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_44_get_name},
  {"_get_eep_man", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_47_get_eep_man, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_46_get_eep_man},
  {"_get_eep_id", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_49_get_eep_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_48_get_eep_id},
  {"_get_eep_rev", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_51_get_eep_rev, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_50_get_eep_rev},
  {"_get_PO2SOconfig", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_53_get_PO2SOconfig, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_52_get_PO2SOconfig},
  {"_get_PO2SOconfigEx", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_55_get_PO2SOconfigEx, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_54_get_PO2SOconfigEx},
  {"_set_PO2SOconfig", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_57_set_PO2SOconfig, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_set_PO2SOconfigEx", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_59_set_PO2SOconfigEx, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_61_get_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6pysoem_6pysoem_9CdefSlave_60_get_state},
  {"_set_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_63_set_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_input", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_65_get_input, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_output", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_67_get_output, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_set_output", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_69_set_output, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_al_status", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_71_get_al_status, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_is_lost", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_73_get_is_lost, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_set_is_lost", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_75_set_is_lost, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_od", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_77_get_od, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_79__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_9CdefSlave_81__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pysoem_6pysoem_CdefSlave[] = {
  {"_master", __pyx_getprop_6pysoem_6pysoem_9CdefSlave__master, 0, 0, 0},
  {"_cd", __pyx_getprop_6pysoem_6pysoem_9CdefSlave__cd, __pyx_setprop_6pysoem_6pysoem_9CdefSlave__cd, 0, 0},
  {"_emcy_callbacks", __pyx_getprop_6pysoem_6pysoem_9CdefSlave__emcy_callbacks, __pyx_setprop_6pysoem_6pysoem_9CdefSlave__emcy_callbacks, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pysoem_6pysoem_CdefSlave_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pysoem_6pysoem_CdefSlave},
  {Py_tp_doc, (void *)PyDoc_STR("Represents a slave device\n\n    Do not use this class in application code. Instances are created\n    by a Master instance on a successful config_init(). They then can be \n    obtained by slaves list\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pysoem_6pysoem_CdefSlave},
  {Py_tp_clear, (void *)__pyx_tp_clear_6pysoem_6pysoem_CdefSlave},
  {Py_tp_methods, (void *)__pyx_methods_6pysoem_6pysoem_CdefSlave},
  {Py_tp_getset, (void *)__pyx_getsets_6pysoem_6pysoem_CdefSlave},
  {Py_tp_init, (void *)__pyx_pw_6pysoem_6pysoem_9CdefSlave_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pysoem_6pysoem_CdefSlave},
  {0, 0},
};
static PyType_Spec __pyx_type_6pysoem_6pysoem_CdefSlave_spec = {
  "pysoem.pysoem.CdefSlave",
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefSlave),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pysoem_6pysoem_CdefSlave_slots,
};
#else

static PyTypeObject __pyx_type_6pysoem_6pysoem_CdefSlave = {
  PyVarObject_HEAD_INIT(0, 0)
  "pysoem.pysoem.""CdefSlave", /*tp_name*/
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefSlave), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pysoem_6pysoem_CdefSlave, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Represents a slave device\n\n    Do not use this class in application code. Instances are created\n    by a Master instance on a successful config_init(). They then can be \n    obtained by slaves list\n    "), /*tp_doc*/
  __pyx_tp_traverse_6pysoem_6pysoem_CdefSlave, /*tp_traverse*/
  __pyx_tp_clear_6pysoem_6pysoem_CdefSlave, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pysoem_6pysoem_CdefSlave, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pysoem_6pysoem_CdefSlave, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pysoem_6pysoem_9CdefSlave_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pysoem_6pysoem_CdefSlave, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefCoeObject(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_6pysoem_6pysoem_CdefCoeObject(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pysoem_6pysoem_CdefCoeObject) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyMethodDef __pyx_methods_6pysoem_6pysoem_CdefCoeObject[] = {
  {"_read_description", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_3_read_description, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_read_entries", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_5_read_entries, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_index", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_7_get_index, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_data_type", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_9_get_data_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_object_code", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_11_get_object_code, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_name", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_13_get_name, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_entries", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_15_get_entries, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_bit_length", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_17_get_bit_length, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_obj_access", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_19_get_obj_access, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pysoem_6pysoem_CdefCoeObject_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pysoem_6pysoem_CdefCoeObject},
  {Py_tp_doc, (void *)PyDoc_STR("Object info for objects in the object dictionary.\n\n    Do not create instances of this class, you get instances of this type by the CdefSlave.od property.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_6pysoem_6pysoem_CdefCoeObject},
  {Py_tp_init, (void *)__pyx_pw_6pysoem_6pysoem_13CdefCoeObject_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pysoem_6pysoem_CdefCoeObject},
  {0, 0},
};
static PyType_Spec __pyx_type_6pysoem_6pysoem_CdefCoeObject_spec = {
  "pysoem.pysoem.CdefCoeObject",
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_6pysoem_6pysoem_CdefCoeObject_slots,
};
#else

static PyTypeObject __pyx_type_6pysoem_6pysoem_CdefCoeObject = {
  PyVarObject_HEAD_INIT(0, 0)
  "pysoem.pysoem.""CdefCoeObject", /*tp_name*/
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObject), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pysoem_6pysoem_CdefCoeObject, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Object info for objects in the object dictionary.\n\n    Do not create instances of this class, you get instances of this type by the CdefSlave.od property.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pysoem_6pysoem_CdefCoeObject, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pysoem_6pysoem_13CdefCoeObject_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pysoem_6pysoem_CdefCoeObject, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_6pysoem_6pysoem_CdefCoeObjectEntry(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_6pysoem_6pysoem_CdefCoeObjectEntry(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pysoem_6pysoem_CdefCoeObjectEntry) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyMethodDef __pyx_methods_6pysoem_6pysoem_CdefCoeObjectEntry[] = {
  {"_get_name", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_3_get_name, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_data_type", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_5_get_data_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_bit_length", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_7_get_bit_length, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_obj_access", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_9_get_obj_access, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pysoem_6pysoem_CdefCoeObjectEntry_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pysoem_6pysoem_CdefCoeObjectEntry},
  {Py_tp_methods, (void *)__pyx_methods_6pysoem_6pysoem_CdefCoeObjectEntry},
  {Py_tp_init, (void *)__pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6pysoem_6pysoem_CdefCoeObjectEntry},
  {0, 0},
};
static PyType_Spec __pyx_type_6pysoem_6pysoem_CdefCoeObjectEntry_spec = {
  "pysoem.pysoem.CdefCoeObjectEntry",
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_6pysoem_6pysoem_CdefCoeObjectEntry_slots,
};
#else

static PyTypeObject __pyx_type_6pysoem_6pysoem_CdefCoeObjectEntry = {
  PyVarObject_HEAD_INIT(0, 0)
  "pysoem.pysoem.""CdefCoeObjectEntry", /*tp_name*/
  sizeof(struct __pyx_obj_6pysoem_6pysoem_CdefCoeObjectEntry), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pysoem_6pysoem_CdefCoeObjectEntry, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pysoem_6pysoem_CdefCoeObjectEntry, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6pysoem_6pysoem_18CdefCoeObjectEntry_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pysoem_6pysoem_CdefCoeObjectEntry, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_6pysoem_6pysoem___pyx_scope_struct__open(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_6pysoem_6pysoem___pyx_scope_struct__open > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open, sizeof(struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_6pysoem_6pysoem___pyx_scope_struct__open[--__pyx_mstate_global->__pyx_freecount_6pysoem_6pysoem___pyx_scope_struct__open];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6pysoem_6pysoem___pyx_scope_struct__open(PyObject *o) {
  struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *p = (struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pysoem_6pysoem___pyx_scope_struct__open) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_ifname);
  Py_CLEAR(p->__pyx_v_master);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_6pysoem_6pysoem___pyx_scope_struct__open < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open, sizeof(struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open))))
  {
    __pyx_mstate_global->__pyx_freelist_6pysoem_6pysoem___pyx_scope_struct__open[__pyx_mstate_global->__pyx_freecount_6pysoem_6pysoem___pyx_scope_struct__open++] = ((struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_6pysoem_6pysoem___pyx_scope_struct__open(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *p = (struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_v_ifname) {
    e = (*v)(p->__pyx_v_ifname, a); if (e) return e;
  }
  if (p->__pyx_v_master) {
    e = (*v)(p->__pyx_v_master, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pysoem_6pysoem___pyx_scope_struct__open_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pysoem_6pysoem___pyx_scope_struct__open},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pysoem_6pysoem___pyx_scope_struct__open},
  {Py_tp_new, (void *)__pyx_tp_new_6pysoem_6pysoem___pyx_scope_struct__open},
  {0, 0},
};
static PyType_Spec __pyx_type_6pysoem_6pysoem___pyx_scope_struct__open_spec = {
  "pysoem.pysoem.__pyx_scope_struct__open",
  sizeof(struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pysoem_6pysoem___pyx_scope_struct__open_slots,
};
#else

static PyTypeObject __pyx_type_6pysoem_6pysoem___pyx_scope_struct__open = {
  PyVarObject_HEAD_INIT(0, 0)
  "pysoem.pysoem.""__pyx_scope_struct__open", /*tp_name*/
  sizeof(struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct__open), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pysoem_6pysoem___pyx_scope_struct__open, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pysoem_6pysoem___pyx_scope_struct__open, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pysoem_6pysoem___pyx_scope_struct__open, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_6pysoem_6pysoem___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_6pysoem_6pysoem___pyx_scope_struct_1_genexpr > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr, sizeof(struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_6pysoem_6pysoem___pyx_scope_struct_1_genexpr[--__pyx_mstate_global->__pyx_freecount_6pysoem_6pysoem___pyx_scope_struct_1_genexpr];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6pysoem_6pysoem___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6pysoem_6pysoem___pyx_scope_struct_1_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_x);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_6pysoem_6pysoem___pyx_scope_struct_1_genexpr < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr, sizeof(struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr))))
  {
    __pyx_mstate_global->__pyx_freelist_6pysoem_6pysoem___pyx_scope_struct_1_genexpr[__pyx_mstate_global->__pyx_freecount_6pysoem_6pysoem___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_6pysoem_6pysoem___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_x) {
    e = (*v)(p->__pyx_v_x, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6pysoem_6pysoem___pyx_scope_struct_1_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6pysoem_6pysoem___pyx_scope_struct_1_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6pysoem_6pysoem___pyx_scope_struct_1_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_6pysoem_6pysoem___pyx_scope_struct_1_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_6pysoem_6pysoem___pyx_scope_struct_1_genexpr_spec = {
  "pysoem.pysoem.__pyx_scope_struct_1_genexpr",
  sizeof(struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6pysoem_6pysoem___pyx_scope_struct_1_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_6pysoem_6pysoem___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "pysoem.pysoem.""__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6pysoem_6pysoem___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pysoem_6pysoem___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6pysoem_6pysoem___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pysoem_6pysoem___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_EnumBase = Py_None; Py_INCREF(Py_None);
  __Pyx_FlagBase = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pysoem_6pysoem_CdefTimeouts_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts)) __PYX_ERR(0, 57, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pysoem_6pysoem_CdefTimeouts_spec, __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts) < (0)) __PYX_ERR(0, 57, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts = &__pyx_type_6pysoem_6pysoem_CdefTimeouts;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts) < (0)) __PYX_ERR(0, 57, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts->tp_dictoffset && __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CdefTimeouts, (PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts) < (0)) __PYX_ERR(0, 57, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefTimeouts) < (0)) __PYX_ERR(0, 57, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pysoem_6pysoem_CdefSettings_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings)) __PYX_ERR(0, 112, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pysoem_6pysoem_CdefSettings_spec, __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings) < (0)) __PYX_ERR(0, 112, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings = &__pyx_type_6pysoem_6pysoem_CdefSettings;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings) < (0)) __PYX_ERR(0, 112, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings->tp_dictoffset && __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CdefSettings, (PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings) < (0)) __PYX_ERR(0, 112, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSettings) < (0)) __PYX_ERR(0, 112, __pyx_L1_error)
  __pyx_vtabptr_6pysoem_6pysoem_CdefMaster = &__pyx_vtable_6pysoem_6pysoem_CdefMaster;
  __pyx_vtable_6pysoem_6pysoem_CdefMaster._CdefMaster__config_init_nogil = (int (*)(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *, uint8_t))__pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__config_init_nogil;
  __pyx_vtable_6pysoem_6pysoem_CdefMaster.check_release_gil = (boolean (*)(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6pysoem_6pysoem_10CdefMaster_check_release_gil;
  __pyx_vtable_6pysoem_6pysoem_CdefMaster._CdefMaster__send_processdata_nogil = (int (*)(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *))__pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__send_processdata_nogil;
  __pyx_vtable_6pysoem_6pysoem_CdefMaster._CdefMaster__send_overlap_processdata_nogil = (int (*)(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *))__pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__send_overlap_processdata_nogil;
  __pyx_vtable_6pysoem_6pysoem_CdefMaster._CdefMaster__receive_processdata_nogil = (int (*)(struct __pyx_obj_6pysoem_6pysoem_CdefMaster *, int))__pyx_f_6pysoem_6pysoem_10CdefMaster__CdefMaster__receive_processdata_nogil;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pysoem_6pysoem_CdefMaster_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster)) __PYX_ERR(0, 219, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pysoem_6pysoem_CdefMaster_spec, __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster) < (0)) __PYX_ERR(0, 219, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster = &__pyx_type_6pysoem_6pysoem_CdefMaster;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster) < (0)) __PYX_ERR(0, 219, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster->tp_dictoffset && __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_vtabptr_6pysoem_6pysoem_CdefMaster) < (0)) __PYX_ERR(0, 219, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster) < (0)) __PYX_ERR(0, 219, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CdefMaster, (PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster) < (0)) __PYX_ERR(0, 219, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefMaster) < (0)) __PYX_ERR(0, 219, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pysoem_6pysoem__CallbackData_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData)) __PYX_ERR(0, 762, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pysoem_6pysoem__CallbackData_spec, __pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData) < (0)) __PYX_ERR(0, 762, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData = &__pyx_type_6pysoem_6pysoem__CallbackData;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData) < (0)) __PYX_ERR(0, 762, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData->tp_dictoffset && __pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CallbackData, (PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData) < (0)) __PYX_ERR(0, 762, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem__CallbackData) < (0)) __PYX_ERR(0, 762, __pyx_L1_error)
  __pyx_vtabptr_6pysoem_6pysoem_CdefSlave = &__pyx_vtable_6pysoem_6pysoem_CdefSlave;
  __pyx_vtable_6pysoem_6pysoem_CdefSlave._CdefSlave__sdo_read_nogil = (int (*)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, uint16_t, uint8_t, int8_t, int *, unsigned char *))__pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__sdo_read_nogil;
  __pyx_vtable_6pysoem_6pysoem_CdefSlave._CdefSlave__sdo_write_nogil = (int (*)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, uint16_t, uint8_t, int8_t, int, PyObject *))__pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__sdo_write_nogil;
  __pyx_vtable_6pysoem_6pysoem_CdefSlave._CdefSlave__foe_write_nogil = (int (*)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, PyObject *, uint32_t, int, PyObject *, int))__pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__foe_write_nogil;
  __pyx_vtable_6pysoem_6pysoem_CdefSlave._CdefSlave__foe_read_nogil = (int (*)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, PyObject *, uint32_t, int, unsigned char *, int))__pyx_f_6pysoem_6pysoem_9CdefSlave__CdefSlave__foe_read_nogil;
  __pyx_vtable_6pysoem_6pysoem_CdefSlave._on_emergency = (PyObject *(*)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, ec_errort *))__pyx_f_6pysoem_6pysoem_9CdefSlave__on_emergency;
  __pyx_vtable_6pysoem_6pysoem_CdefSlave._raise_exception = (PyObject *(*)(struct __pyx_obj_6pysoem_6pysoem_CdefSlave *, ec_errort *))__pyx_f_6pysoem_6pysoem_9CdefSlave__raise_exception;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pysoem_6pysoem_CdefSlave_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave)) __PYX_ERR(0, 788, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pysoem_6pysoem_CdefSlave_spec, __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave) < (0)) __PYX_ERR(0, 788, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave = &__pyx_type_6pysoem_6pysoem_CdefSlave;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave) < (0)) __PYX_ERR(0, 788, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave->tp_dictoffset && __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_vtabptr_6pysoem_6pysoem_CdefSlave) < (0)) __PYX_ERR(0, 788, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave) < (0)) __PYX_ERR(0, 788, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CdefSlave, (PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave) < (0)) __PYX_ERR(0, 788, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefSlave) < (0)) __PYX_ERR(0, 788, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pysoem_6pysoem_CdefCoeObject_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject)) __PYX_ERR(0, 1503, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pysoem_6pysoem_CdefCoeObject_spec, __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject) < (0)) __PYX_ERR(0, 1503, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject = &__pyx_type_6pysoem_6pysoem_CdefCoeObject;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject) < (0)) __PYX_ERR(0, 1503, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject->tp_dictoffset && __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CdefCoeObject, (PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject) < (0)) __PYX_ERR(0, 1503, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject) < (0)) __PYX_ERR(0, 1503, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pysoem_6pysoem_CdefCoeObjectEntry_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry)) __PYX_ERR(0, 1596, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pysoem_6pysoem_CdefCoeObjectEntry_spec, __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry) < (0)) __PYX_ERR(0, 1596, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry = &__pyx_type_6pysoem_6pysoem_CdefCoeObjectEntry;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry) < (0)) __PYX_ERR(0, 1596, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry->tp_dictoffset && __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CdefCoeObjectEntry, (PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry) < (0)) __PYX_ERR(0, 1596, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry) < (0)) __PYX_ERR(0, 1596, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pysoem_6pysoem___pyx_scope_struct__open_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open)) __PYX_ERR(0, 172, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pysoem_6pysoem___pyx_scope_struct__open_spec, __pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open) < (0)) __PYX_ERR(0, 172, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open = &__pyx_type_6pysoem_6pysoem___pyx_scope_struct__open;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open) < (0)) __PYX_ERR(0, 172, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open->tp_dictoffset && __pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct__open->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6pysoem_6pysoem___pyx_scope_struct_1_genexpr_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr)) __PYX_ERR(0, 667, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6pysoem_6pysoem___pyx_scope_struct_1_genexpr_spec, __pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr) < (0)) __PYX_ERR(0, 667, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr = &__pyx_type_6pysoem_6pysoem___pyx_scope_struct_1_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr) < (0)) __PYX_ERR(0, 667, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr->tp_dictoffset && __pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_6pysoem_6pysoem___pyx_scope_struct_1_genexpr->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_pysoem(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_pysoem},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "pysoem",
      __pyx_k_PySOEM_is_a_Cython_wrapper_for_t, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_pysoem(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_pysoem(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_pysoem(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  size_t __pyx_t_34;
  int __pyx_t_35;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'pysoem' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "pysoem" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_pysoem", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_pysoem__pysoem) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name_2, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pysoem.pysoem")) {
      if (unlikely((PyDict_SetItemString(modules, "pysoem.pysoem", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "EnumBase":10
 * 
 * cdef object __Pyx_EnumBase
 * from enum import IntEnum as __Pyx_EnumBase             # <<<<<<<<<<<<<<
 * 
 * cdef object __Pyx_FlagBase
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntEnum};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_enum, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntEnum};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 10, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      switch (__pyx_t_3) {
        case 0:
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_XGOTREF(__Pyx_EnumBase);
        __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        break;
        default:;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":13
 * 
 * cdef object __Pyx_FlagBase
 * from enum import IntFlag as __Pyx_FlagBase             # <<<<<<<<<<<<<<
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntFlag};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_enum, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntFlag};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      switch (__pyx_t_3) {
        case 0:
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_XGOTREF(__Pyx_FlagBase);
        __Pyx_DECREF_SET(__Pyx_FlagBase, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        break;
        default:;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":23
 * 
 * 
 * ec_datatype = __Pyx_FlagBase('ec_datatype', [             # <<<<<<<<<<<<<<
 *     ('ECT_BOOLEAN', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BOOLEAN)),
 *     ('ECT_INTEGER8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER8)),
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_5 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ec_datatype = __Pyx_FlagBase('ec_datatype', [
 *     ('ECT_BOOLEAN', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BOOLEAN)),             # <<<<<<<<<<<<<<
 *     ('ECT_INTEGER8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER8)),
 *     ('ECT_INTEGER16', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER16)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_BOOLEAN); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_BOOLEAN, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":25
 * ec_datatype = __Pyx_FlagBase('ec_datatype', [
 *     ('ECT_BOOLEAN', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BOOLEAN)),
 *     ('ECT_INTEGER8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER8)),             # <<<<<<<<<<<<<<
 *     ('ECT_INTEGER16', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER16)),
 *     ('ECT_INTEGER32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER32)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_INTEGER8); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_INTEGER8, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":26
 *     ('ECT_BOOLEAN', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BOOLEAN)),
 *     ('ECT_INTEGER8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER8)),
 *     ('ECT_INTEGER16', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER16)),             # <<<<<<<<<<<<<<
 *     ('ECT_INTEGER32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER32)),
 *     ('ECT_UNSIGNED8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED8)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_INTEGER16); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_INTEGER16, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":27
 *     ('ECT_INTEGER8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER8)),
 *     ('ECT_INTEGER16', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER16)),
 *     ('ECT_INTEGER32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER32)),             # <<<<<<<<<<<<<<
 *     ('ECT_UNSIGNED8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED8)),
 *     ('ECT_UNSIGNED16', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED16)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_INTEGER32); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_INTEGER32, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":28
 *     ('ECT_INTEGER16', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER16)),
 *     ('ECT_INTEGER32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER32)),
 *     ('ECT_UNSIGNED8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED8)),             # <<<<<<<<<<<<<<
 *     ('ECT_UNSIGNED16', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED16)),
 *     ('ECT_UNSIGNED32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED32)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_UNSIGNED8); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_UNSIGNED8, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":29
 *     ('ECT_INTEGER32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER32)),
 *     ('ECT_UNSIGNED8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED8)),
 *     ('ECT_UNSIGNED16', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED16)),             # <<<<<<<<<<<<<<
 *     ('ECT_UNSIGNED32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED32)),
 *     ('ECT_REAL32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_REAL32)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_UNSIGNED16); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_UNSIGNED16, __pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":30
 *     ('ECT_UNSIGNED8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED8)),
 *     ('ECT_UNSIGNED16', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED16)),
 *     ('ECT_UNSIGNED32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED32)),             # <<<<<<<<<<<<<<
 *     ('ECT_REAL32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_REAL32)),
 *     ('ECT_VISIBLE_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_VISIBLE_STRING)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_UNSIGNED32); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_UNSIGNED32, __pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":31
 *     ('ECT_UNSIGNED16', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED16)),
 *     ('ECT_UNSIGNED32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED32)),
 *     ('ECT_REAL32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_REAL32)),             # <<<<<<<<<<<<<<
 *     ('ECT_VISIBLE_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_VISIBLE_STRING)),
 *     ('ECT_OCTET_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_OCTET_STRING)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_REAL32); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_REAL32, __pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":32
 *     ('ECT_UNSIGNED32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED32)),
 *     ('ECT_REAL32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_REAL32)),
 *     ('ECT_VISIBLE_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_VISIBLE_STRING)),             # <<<<<<<<<<<<<<
 *     ('ECT_OCTET_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_OCTET_STRING)),
 *     ('ECT_UNICODE_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNICODE_STRING)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_VISIBLE_STRING); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_VISIBLE_STRING, __pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":33
 *     ('ECT_REAL32', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_REAL32)),
 *     ('ECT_VISIBLE_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_VISIBLE_STRING)),
 *     ('ECT_OCTET_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_OCTET_STRING)),             # <<<<<<<<<<<<<<
 *     ('ECT_UNICODE_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNICODE_STRING)),
 *     ('ECT_TIME_OF_DAY', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_TIME_OF_DAY)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_OCTET_STRING); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_OCTET_STRING, __pyx_t_6); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":34
 *     ('ECT_VISIBLE_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_VISIBLE_STRING)),
 *     ('ECT_OCTET_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_OCTET_STRING)),
 *     ('ECT_UNICODE_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNICODE_STRING)),             # <<<<<<<<<<<<<<
 *     ('ECT_TIME_OF_DAY', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_TIME_OF_DAY)),
 *     ('ECT_TIME_DIFFERENCE', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_TIME_DIFFERENCE)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_UNICODE_STRING); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_UNICODE_STRING, __pyx_t_6); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":35
 *     ('ECT_OCTET_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_OCTET_STRING)),
 *     ('ECT_UNICODE_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNICODE_STRING)),
 *     ('ECT_TIME_OF_DAY', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_TIME_OF_DAY)),             # <<<<<<<<<<<<<<
 *     ('ECT_TIME_DIFFERENCE', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_TIME_DIFFERENCE)),
 *     ('ECT_DOMAIN', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_DOMAIN)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_TIME_OF_DAY); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_TIME_OF_DAY, __pyx_t_6); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":36
 *     ('ECT_UNICODE_STRING', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNICODE_STRING)),
 *     ('ECT_TIME_OF_DAY', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_TIME_OF_DAY)),
 *     ('ECT_TIME_DIFFERENCE', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_TIME_DIFFERENCE)),             # <<<<<<<<<<<<<<
 *     ('ECT_DOMAIN', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_DOMAIN)),
 *     ('ECT_INTEGER24', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER24)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_TIME_DIFFERENCE); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_TIME_DIFFERENCE, __pyx_t_6); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":37
 *     ('ECT_TIME_OF_DAY', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_TIME_OF_DAY)),
 *     ('ECT_TIME_DIFFERENCE', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_TIME_DIFFERENCE)),
 *     ('ECT_DOMAIN', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_DOMAIN)),             # <<<<<<<<<<<<<<
 *     ('ECT_INTEGER24', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER24)),
 *     ('ECT_REAL64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_REAL64)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_DOMAIN); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_DOMAIN, __pyx_t_6); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":38
 *     ('ECT_TIME_DIFFERENCE', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_TIME_DIFFERENCE)),
 *     ('ECT_DOMAIN', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_DOMAIN)),
 *     ('ECT_INTEGER24', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER24)),             # <<<<<<<<<<<<<<
 *     ('ECT_REAL64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_REAL64)),
 *     ('ECT_INTEGER64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER64)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_INTEGER24); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_INTEGER24, __pyx_t_6); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":39
 *     ('ECT_DOMAIN', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_DOMAIN)),
 *     ('ECT_INTEGER24', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER24)),
 *     ('ECT_REAL64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_REAL64)),             # <<<<<<<<<<<<<<
 *     ('ECT_INTEGER64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER64)),
 *     ('ECT_UNSIGNED24', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED24)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_REAL64); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_REAL64, __pyx_t_6); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":40
 *     ('ECT_INTEGER24', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER24)),
 *     ('ECT_REAL64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_REAL64)),
 *     ('ECT_INTEGER64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER64)),             # <<<<<<<<<<<<<<
 *     ('ECT_UNSIGNED24', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED24)),
 *     ('ECT_UNSIGNED64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED64)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_INTEGER64); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_INTEGER64, __pyx_t_6); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":41
 *     ('ECT_REAL64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_REAL64)),
 *     ('ECT_INTEGER64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER64)),
 *     ('ECT_UNSIGNED24', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED24)),             # <<<<<<<<<<<<<<
 *     ('ECT_UNSIGNED64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED64)),
 *     ('ECT_BIT1', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT1)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_UNSIGNED24); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_UNSIGNED24, __pyx_t_6); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":42
 *     ('ECT_INTEGER64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER64)),
 *     ('ECT_UNSIGNED24', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED24)),
 *     ('ECT_UNSIGNED64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED64)),             # <<<<<<<<<<<<<<
 *     ('ECT_BIT1', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT1)),
 *     ('ECT_BIT2', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT2)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_UNSIGNED64); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_UNSIGNED64, __pyx_t_6); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":43
 *     ('ECT_UNSIGNED24', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED24)),
 *     ('ECT_UNSIGNED64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED64)),
 *     ('ECT_BIT1', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT1)),             # <<<<<<<<<<<<<<
 *     ('ECT_BIT2', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT2)),
 *     ('ECT_BIT3', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT3)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_BIT1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_BIT1, __pyx_t_6); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":44
 *     ('ECT_UNSIGNED64', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_UNSIGNED64)),
 *     ('ECT_BIT1', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT1)),
 *     ('ECT_BIT2', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT2)),             # <<<<<<<<<<<<<<
 *     ('ECT_BIT3', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT3)),
 *     ('ECT_BIT4', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT4)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_BIT2); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_BIT2, __pyx_t_6); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":45
 *     ('ECT_BIT1', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT1)),
 *     ('ECT_BIT2', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT2)),
 *     ('ECT_BIT3', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT3)),             # <<<<<<<<<<<<<<
 *     ('ECT_BIT4', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT4)),
 *     ('ECT_BIT5', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT5)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_BIT3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_28 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_BIT3, __pyx_t_6); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":46
 *     ('ECT_BIT2', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT2)),
 *     ('ECT_BIT3', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT3)),
 *     ('ECT_BIT4', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT4)),             # <<<<<<<<<<<<<<
 *     ('ECT_BIT5', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT5)),
 *     ('ECT_BIT6', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT6)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_BIT4); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_29 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_BIT4, __pyx_t_6); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":47
 *     ('ECT_BIT3', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT3)),
 *     ('ECT_BIT4', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT4)),
 *     ('ECT_BIT5', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT5)),             # <<<<<<<<<<<<<<
 *     ('ECT_BIT6', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT6)),
 *     ('ECT_BIT7', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT7)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_BIT5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_BIT5, __pyx_t_6); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":48
 *     ('ECT_BIT4', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT4)),
 *     ('ECT_BIT5', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT5)),
 *     ('ECT_BIT6', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT6)),             # <<<<<<<<<<<<<<
 *     ('ECT_BIT7', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT7)),
 *     ('ECT_BIT8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT8)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_BIT6); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_BIT6, __pyx_t_6); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":49
 *     ('ECT_BIT5', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT5)),
 *     ('ECT_BIT6', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT6)),
 *     ('ECT_BIT7', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT7)),             # <<<<<<<<<<<<<<
 *     ('ECT_BIT8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT8)),
 * 
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_BIT7); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_BIT7, __pyx_t_6); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":50
 *     ('ECT_BIT6', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT6)),
 *     ('ECT_BIT7', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT7)),
 *     ('ECT_BIT8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT8)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pysoem.pysoem'))
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(__pyx_e_6pysoem_6pysoem_ECT_BIT8); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_33 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_ECT_BIT8, __pyx_t_6); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":23
 * 
 * 
 * ec_datatype = __Pyx_FlagBase('ec_datatype', [             # <<<<<<<<<<<<<<
 *     ('ECT_BOOLEAN', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BOOLEAN)),
 *     ('ECT_INTEGER8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_INTEGER8)),
*/
  __pyx_t_6 = __Pyx_PyList_Pack(27, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "EnumType":52
 *     ('ECT_BIT8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT8)),
 * 
 * ], module=globals().get("__module__", 'pysoem.pysoem'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_33 = __Pyx_Globals(); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_33, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __Pyx_PyObject_Call(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ec_datatype, __pyx_t_6};
    __pyx_t_32 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_33, __pyx_t_32, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_34, (3-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_32);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ec_datatype, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":54
 * ], module=globals().get("__module__", 'pysoem.pysoem'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_35 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_35) {

    /* "EnumType":59
 * 
 * 
 *     ec_datatype._member_names_ = list(ec_datatype.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ec_datatype); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ec_datatype); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "EnumType":54
 * ], module=globals().get("__module__", 'pysoem.pysoem'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "pysoem/pysoem.pyx":20
 * cimport cpysoem
 * 
 * import sys             # <<<<<<<<<<<<<<
 * import logging
 * import collections
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_sys, 0, 0, __pyx_mstate_global->__pyx_kp_u_pysoem_sys, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_sys, __pyx_t_5) < (0)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":21
 * 
 * import sys
 * import logging             # <<<<<<<<<<<<<<
 * import collections
 * import time
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_logging, 0, 0, __pyx_mstate_global->__pyx_kp_u_pysoem_logging, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_logging, __pyx_t_5) < (0)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":22
 * import sys
 * import logging
 * import collections             # <<<<<<<<<<<<<<
 * import time
 * import contextlib
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_collections, 0, 0, __pyx_mstate_global->__pyx_kp_u_pysoem_collections, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_collections, __pyx_t_5) < (0)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":23
 * import logging
 * import collections
 * import time             # <<<<<<<<<<<<<<
 * import contextlib
 * import warnings
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_time, 0, 0, __pyx_mstate_global->__pyx_kp_u_pysoem_time, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_time, __pyx_t_5) < (0)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":24
 * import collections
 * import time
 * import contextlib             # <<<<<<<<<<<<<<
 * import warnings
 * 
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_contextlib, 0, 0, __pyx_mstate_global->__pyx_kp_u_pysoem_contextlib, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_contextlib, __pyx_t_5) < (0)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":25
 * import time
 * import contextlib
 * import warnings             # <<<<<<<<<<<<<<
 * 
 * from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, __pyx_mstate_global->__pyx_kp_u_pysoem_warnings, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_warnings, __pyx_t_5) < (0)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":33
 * from cpython.ref cimport Py_INCREF, Py_DECREF
 * 
 * logger = logging.getLogger(__name__)             # <<<<<<<<<<<<<<
 * 
 * NONE_STATE = cpysoem.EC_STATE_NONE
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_logging); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_33 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_getLogger); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_name_2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_32};
    __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_33, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_logger, __pyx_t_5) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":35
 * logger = logging.getLogger(__name__)
 * 
 * NONE_STATE = cpysoem.EC_STATE_NONE             # <<<<<<<<<<<<<<
 * INIT_STATE = cpysoem.EC_STATE_INIT
 * PREOP_STATE = cpysoem.EC_STATE_PRE_OP
*/
  __pyx_t_5 = __Pyx_PyLong_From_ec_state(EC_STATE_NONE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_NONE_STATE, __pyx_t_5) < (0)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":36
 * 
 * NONE_STATE = cpysoem.EC_STATE_NONE
 * INIT_STATE = cpysoem.EC_STATE_INIT             # <<<<<<<<<<<<<<
 * PREOP_STATE = cpysoem.EC_STATE_PRE_OP
 * BOOT_STATE = cpysoem.EC_STATE_BOOT
*/
  __pyx_t_5 = __Pyx_PyLong_From_ec_state(EC_STATE_INIT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_INIT_STATE, __pyx_t_5) < (0)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":37
 * NONE_STATE = cpysoem.EC_STATE_NONE
 * INIT_STATE = cpysoem.EC_STATE_INIT
 * PREOP_STATE = cpysoem.EC_STATE_PRE_OP             # <<<<<<<<<<<<<<
 * BOOT_STATE = cpysoem.EC_STATE_BOOT
 * SAFEOP_STATE = cpysoem.EC_STATE_SAFE_OP
*/
  __pyx_t_5 = __Pyx_PyLong_From_ec_state(EC_STATE_PRE_OP); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_PREOP_STATE, __pyx_t_5) < (0)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":38
 * INIT_STATE = cpysoem.EC_STATE_INIT
 * PREOP_STATE = cpysoem.EC_STATE_PRE_OP
 * BOOT_STATE = cpysoem.EC_STATE_BOOT             # <<<<<<<<<<<<<<
 * SAFEOP_STATE = cpysoem.EC_STATE_SAFE_OP
 * OP_STATE = cpysoem.EC_STATE_OPERATIONAL
*/
  __pyx_t_5 = __Pyx_PyLong_From_ec_state(EC_STATE_BOOT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_BOOT_STATE, __pyx_t_5) < (0)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":39
 * PREOP_STATE = cpysoem.EC_STATE_PRE_OP
 * BOOT_STATE = cpysoem.EC_STATE_BOOT
 * SAFEOP_STATE = cpysoem.EC_STATE_SAFE_OP             # <<<<<<<<<<<<<<
 * OP_STATE = cpysoem.EC_STATE_OPERATIONAL
 * STATE_ACK = cpysoem.EC_STATE_ACK
*/
  __pyx_t_5 = __Pyx_PyLong_From_ec_state(EC_STATE_SAFE_OP); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_SAFEOP_STATE, __pyx_t_5) < (0)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":40
 * BOOT_STATE = cpysoem.EC_STATE_BOOT
 * SAFEOP_STATE = cpysoem.EC_STATE_SAFE_OP
 * OP_STATE = cpysoem.EC_STATE_OPERATIONAL             # <<<<<<<<<<<<<<
 * STATE_ACK = cpysoem.EC_STATE_ACK
 * STATE_ERROR = cpysoem.EC_STATE_ERROR
*/
  __pyx_t_5 = __Pyx_PyLong_From_ec_state(EC_STATE_OPERATIONAL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_OP_STATE, __pyx_t_5) < (0)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":41
 * SAFEOP_STATE = cpysoem.EC_STATE_SAFE_OP
 * OP_STATE = cpysoem.EC_STATE_OPERATIONAL
 * STATE_ACK = cpysoem.EC_STATE_ACK             # <<<<<<<<<<<<<<
 * STATE_ERROR = cpysoem.EC_STATE_ERROR
 * 
*/
  __pyx_t_5 = __Pyx_PyLong_From_ec_state(EC_STATE_ACK); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_STATE_ACK, __pyx_t_5) < (0)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":42
 * OP_STATE = cpysoem.EC_STATE_OPERATIONAL
 * STATE_ACK = cpysoem.EC_STATE_ACK
 * STATE_ERROR = cpysoem.EC_STATE_ERROR             # <<<<<<<<<<<<<<
 * 
 * ECT_REG_WD_DIV = 0x0400
*/
  __pyx_t_5 = __Pyx_PyLong_From_ec_state(EC_STATE_ERROR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_STATE_ERROR, __pyx_t_5) < (0)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":44
 * STATE_ERROR = cpysoem.EC_STATE_ERROR
 * 
 * ECT_REG_WD_DIV = 0x0400             # <<<<<<<<<<<<<<
 * ECT_REG_WD_TIME_PDI = 0x0410
 * ECT_REG_WD_TIME_PROCESSDATA = 0x0420
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_REG_WD_DIV, __pyx_mstate_global->__pyx_int_1024) < (0)) __PYX_ERR(0, 44, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":45
 * 
 * ECT_REG_WD_DIV = 0x0400
 * ECT_REG_WD_TIME_PDI = 0x0410             # <<<<<<<<<<<<<<
 * ECT_REG_WD_TIME_PROCESSDATA = 0x0420
 * ECT_REG_SM0 = 0x0800
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_REG_WD_TIME_PDI, __pyx_mstate_global->__pyx_int_1040) < (0)) __PYX_ERR(0, 45, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":46
 * ECT_REG_WD_DIV = 0x0400
 * ECT_REG_WD_TIME_PDI = 0x0410
 * ECT_REG_WD_TIME_PROCESSDATA = 0x0420             # <<<<<<<<<<<<<<
 * ECT_REG_SM0 = 0x0800
 * ECT_REG_SM1 = ECT_REG_SM0 + 0x08
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_REG_WD_TIME_PROCESSDATA, __pyx_mstate_global->__pyx_int_1056) < (0)) __PYX_ERR(0, 46, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":47
 * ECT_REG_WD_TIME_PDI = 0x0410
 * ECT_REG_WD_TIME_PROCESSDATA = 0x0420
 * ECT_REG_SM0 = 0x0800             # <<<<<<<<<<<<<<
 * ECT_REG_SM1 = ECT_REG_SM0 + 0x08
 * 
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_REG_SM0, __pyx_mstate_global->__pyx_int_2048) < (0)) __PYX_ERR(0, 47, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":48
 * ECT_REG_WD_TIME_PROCESSDATA = 0x0420
 * ECT_REG_SM0 = 0x0800
 * ECT_REG_SM1 = ECT_REG_SM0 + 0x08             # <<<<<<<<<<<<<<
 * 
 * ECT_COEDET_SDO       = 0x01
*/
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ECT_REG_SM0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_33 = __Pyx_PyLong_AddObjC(__pyx_t_5, __pyx_mstate_global->__pyx_int_8, 0x08, 0, 0); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_REG_SM1, __pyx_t_33) < (0)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":50
 * ECT_REG_SM1 = ECT_REG_SM0 + 0x08
 * 
 * ECT_COEDET_SDO       = 0x01             # <<<<<<<<<<<<<<
 * ECT_COEDET_SDOINFO   = 0x02
 * ECT_COEDET_PDOASSIGN = 0x04
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_COEDET_SDO, __pyx_mstate_global->__pyx_int_1) < (0)) __PYX_ERR(0, 50, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":51
 * 
 * ECT_COEDET_SDO       = 0x01
 * ECT_COEDET_SDOINFO   = 0x02             # <<<<<<<<<<<<<<
 * ECT_COEDET_PDOASSIGN = 0x04
 * ECT_COEDET_PDOCONFIG = 0x08
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_COEDET_SDOINFO, __pyx_mstate_global->__pyx_int_2) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":52
 * ECT_COEDET_SDO       = 0x01
 * ECT_COEDET_SDOINFO   = 0x02
 * ECT_COEDET_PDOASSIGN = 0x04             # <<<<<<<<<<<<<<
 * ECT_COEDET_PDOCONFIG = 0x08
 * ECT_COEDET_UPLOAD    = 0x10
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_COEDET_PDOASSIGN, __pyx_mstate_global->__pyx_int_4) < (0)) __PYX_ERR(0, 52, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":53
 * ECT_COEDET_SDOINFO   = 0x02
 * ECT_COEDET_PDOASSIGN = 0x04
 * ECT_COEDET_PDOCONFIG = 0x08             # <<<<<<<<<<<<<<
 * ECT_COEDET_UPLOAD    = 0x10
 * ECT_COEDET_SDOCA     = 0x20
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_COEDET_PDOCONFIG, __pyx_mstate_global->__pyx_int_8) < (0)) __PYX_ERR(0, 53, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":54
 * ECT_COEDET_PDOASSIGN = 0x04
 * ECT_COEDET_PDOCONFIG = 0x08
 * ECT_COEDET_UPLOAD    = 0x10             # <<<<<<<<<<<<<<
 * ECT_COEDET_SDOCA     = 0x20
 * 
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_COEDET_UPLOAD, __pyx_mstate_global->__pyx_int_16) < (0)) __PYX_ERR(0, 54, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":55
 * ECT_COEDET_PDOCONFIG = 0x08
 * ECT_COEDET_UPLOAD    = 0x10
 * ECT_COEDET_SDOCA     = 0x20             # <<<<<<<<<<<<<<
 * 
 * cdef class CdefTimeouts:
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ECT_COEDET_SDOCA, __pyx_mstate_global->__pyx_int_32) < (0)) __PYX_ERR(0, 55, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_33 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_12CdefTimeouts_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefTimeouts___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_33) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_33 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_12CdefTimeouts_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefTimeouts___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_33) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_33 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_12CdefSettings_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSettings___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSettings, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_33) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CdefSettings, (type(self), 0xe6b07d6, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CdefSettings__set_state(self, __pyx_state)
*/
  __pyx_t_33 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_12CdefSettings_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSettings___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSettings, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_33) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":121
 *         self.always_release_gil = False
 * 
 * settings = CdefSettings()             # <<<<<<<<<<<<<<
 * 
 * cpdef enum ec_datatype:
*/
  __pyx_t_5 = NULL;
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_33 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSettings, __pyx_callargs+__pyx_t_34, (1-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_33);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_settings, ((PyObject *)__pyx_t_33)) < (0)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_DECREF((PyObject *)__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":156
 *     int* sdo_write_timeout
 * 
 * def find_adapters():             # <<<<<<<<<<<<<<
 *     """Create a list of available network adapters.
 * 
*/
  __pyx_t_33 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_1find_adapters, 0, __pyx_mstate_global->__pyx_n_u_find_adapters, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_find_adapters, __pyx_t_33) < (0)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":172
 * 
 * 
 * @contextlib.contextmanager             # <<<<<<<<<<<<<<
 * def open(ifname):
 *     """Context manager function to create a Master object.
*/
  __pyx_t_5 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_contextlib); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_contextmanager); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_3open, 0, __pyx_mstate_global->__pyx_n_u_open, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_32};
    __pyx_t_33 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_33);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_open, __pyx_t_33) < (0)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":184
 * 
 * 
 * def al_status_code_to_string(code):             # <<<<<<<<<<<<<<
 *     """Look up text string that belongs to AL status code.
 * 
*/
  __pyx_t_33 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_6al_status_code_to_string, 0, __pyx_mstate_global->__pyx_n_u_al_status_code_to_string, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_al_status_code_to_string, __pyx_t_33) < (0)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":197
 * 
 * 
 * class Master(CdefMaster):             # <<<<<<<<<<<<<<
 *     """Representing a logical EtherCAT master device.
 * 
*/
  __pyx_t_33 = PyTuple_Pack(1, ((PyObject *)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_33); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_32, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_Master, __pyx_mstate_global->__pyx_n_u_Master, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_Representing_a_logical_EtherCAT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_2 != __pyx_t_33) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_33) < 0))) __PYX_ERR(0, 197, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __Pyx_Py3ClassCreate(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_Master, __pyx_t_2, __pyx_t_5, NULL, 0, 1); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Master, __pyx_t_33) < (0)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":252
 *     cdef readonly cpysoem.boolean context_initialized
 * 
 *     state = property(_get_state, _set_state)             # <<<<<<<<<<<<<<
 *     expected_wkc  = property(_get_expected_wkc)
 *     dc_time = property(_get_dc_time)
*/
  __pyx_t_32 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_get_state); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_set_state); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_32, __pyx_t_5, __pyx_t_33};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (3-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_state, __pyx_t_2) < (0)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":253
 * 
 *     state = property(_get_state, _set_state)
 *     expected_wkc  = property(_get_expected_wkc)             # <<<<<<<<<<<<<<
 *     dc_time = property(_get_dc_time)
 *     manual_state_change = property(_get_manual_state_change, _set_manual_state_change)
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_get_expected_wkc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_33, __pyx_t_5};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_expected_wkc, __pyx_t_2) < (0)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":254
 *     state = property(_get_state, _set_state)
 *     expected_wkc  = property(_get_expected_wkc)
 *     dc_time = property(_get_dc_time)             # <<<<<<<<<<<<<<
 *     manual_state_change = property(_get_manual_state_change, _set_manual_state_change)
 * 
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_get_dc_time); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_33};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_dc_time, __pyx_t_2) < (0)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":255
 *     expected_wkc  = property(_get_expected_wkc)
 *     dc_time = property(_get_dc_time)
 *     manual_state_change = property(_get_manual_state_change, _set_manual_state_change)             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_get_manual_state_change); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetNameInClass(__pyx_t_32, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_set_manual_state_change); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_33, __pyx_t_5, __pyx_t_32};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (3-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_manual_state_change, __pyx_t_2) < (0)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":288
 *         self.context_initialized = False
 * 
 *     def open(self, ifname, ifname_red=None):             # <<<<<<<<<<<<<<
 *         """Initialize and open network interface.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_3open, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_open, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_open, __pyx_t_2) < (0)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":316
 *         self.context_initialized = True
 * 
 *     def check_context_is_initialized(self):             # <<<<<<<<<<<<<<
 *         if not self.context_initialized:
 *             raise NetworkInterfaceNotOpenError("SOEM Network interface is not initialized or has been closed. Call Master.open() first")
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_5check_context_is_initialized, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_check_context_is_init, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_check_context_is_initialized, __pyx_t_2) < (0)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":335
 *         return ret_val
 * 
 *     cpdef cpysoem.boolean check_release_gil(self, release_gil):             # <<<<<<<<<<<<<<
 *         """Checks if the GIL should be released.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_7check_release_gil, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_check_release_gil, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_check_release_gil, __pyx_t_2) < (0)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":345
 *         return self.always_release_gil
 * 
 *     def config_init(self, usetable=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Enumerate and init all slaves.
 * 
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_release_gil, Py_None) < (0)) __PYX_ERR(0, 345, __pyx_L1_error)
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_9config_init, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_config_init, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_32, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_config_init, __pyx_t_32) < (0)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":370
 *         return ret_val
 * 
 *     def config_map(self):             # <<<<<<<<<<<<<<
 *         """Map all slaves PDOs in IO map.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_11config_map, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_config_map, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_config_map, __pyx_t_32) < (0)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":392
 *         return ret_val
 * 
 *     def config_overlap_map(self):             # <<<<<<<<<<<<<<
 *         """Map all slaves PDOs to overlapping IO map.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_13config_overlap_map, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_config_overlap_map, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_config_overlap_map, __pyx_t_32) < (0)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":415
 *         return ret_val
 * 
 *     def _collect_mailbox_errors(self):             # <<<<<<<<<<<<<<
 *         # collect SDO or mailbox errors that occurred during PDO configuration read in ecx_config_map_group
 *         error_list = []
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_15_collect_mailbox_errors, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster__collect_mailbox_erro, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_collect_mailbox_errors, __pyx_t_32) < (0)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":436
 *         return error_list
 * 
 *     def config_dc(self):             # <<<<<<<<<<<<<<
 *         """Locate DC slaves, measure propagation delays.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_17config_dc, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_config_dc, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_config_dc, __pyx_t_32) < (0)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":445
 *         return cpysoem.ecx_configdc(&self._ecx_contextt)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the network interface.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_19close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_close, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_32) < (0)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":453
 *         cpysoem.ecx_close(&self._ecx_contextt)
 * 
 *     def read_state(self):             # <<<<<<<<<<<<<<
 *         """Read all slaves states.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_21read_state, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_read_state, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_read_state, __pyx_t_32) < (0)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":462
 *         return cpysoem.ecx_readstate(&self._ecx_contextt)
 * 
 *     def write_state(self):             # <<<<<<<<<<<<<<
 *         """Write all slaves state.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_23write_state, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_write_state, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_write_state, __pyx_t_32) < (0)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":473
 *         return cpysoem.ecx_writestate(&self._ecx_contextt, 0)
 * 
 *     def state_check(self, int expected_state, timeout=50000):             # <<<<<<<<<<<<<<
 *         """Check actual slave state.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_25state_check, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_state_check, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[3]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_state_check, __pyx_t_32) < (0)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":500
 *         return result
 * 
 *     def send_processdata(self, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Transmit processdata to slaves.
 * 
*/
  __pyx_t_32 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyDict_SetItem(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_release_gil, Py_None) < (0)) __PYX_ERR(0, 500, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_27send_processdata, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_send_processdata, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_2, __pyx_t_32);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_send_processdata, __pyx_t_2) < (0)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":533
 *         return result
 * 
 *     def send_overlap_processdata(self, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Transmit overlap processdata to slaves.
 * 
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_release_gil, Py_None) < (0)) __PYX_ERR(0, 533, __pyx_L1_error)
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_29send_overlap_processdata, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_send_overlap_processd, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_32, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_send_overlap_processdata, __pyx_t_32) < (0)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":559
 *         return result
 * 
 *     def receive_processdata(self, timeout=2000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Receive processdata from slaves.
 * 
*/
  __pyx_t_32 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyDict_SetItem(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_release_gil, Py_None) < (0)) __PYX_ERR(0, 559, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_31receive_processdata, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster_receive_processdata, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[4]);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_2, __pyx_t_32);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_receive_processdata, __pyx_t_2) < (0)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":578
 *         return cpysoem.ecx_receive_processdata(&self._ecx_contextt, timeout)
 * 
 *     def _get_slave(self, int pos):             # <<<<<<<<<<<<<<
 *         if pos < 0:
 *             raise IndexError('requested slave device is not available')
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_33_get_slave, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster__get_slave, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_get_slave, __pyx_t_2) < (0)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":590
 *         return ethercat_slave
 * 
 *     def _get_state(self):             # <<<<<<<<<<<<<<
 *         """Can be used to check if all slaves are in Operational state, or to request a new state for all slaves.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_35_get_state, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster__get_state, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[24])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_get_state, __pyx_t_2) < (0)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":597
 *         return self._ec_slave[0].state
 * 
 *     def _set_state(self, value):             # <<<<<<<<<<<<<<
 *         self._ec_slave[0].state = value
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_37_set_state, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster__set_state, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_set_state, __pyx_t_2) < (0)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":600
 *         self._ec_slave[0].state = value
 * 
 *     def _get_expected_wkc(self):             # <<<<<<<<<<<<<<
 *         """Calculates the expected Working Counter"""
 *         return (self._ec_group[0].outputsWKC * 2) + self._ec_group[0].inputsWKC
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_39_get_expected_wkc, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster__get_expected_wkc, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[26])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_get_expected_wkc, __pyx_t_2) < (0)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":604
 *         return (self._ec_group[0].outputsWKC * 2) + self._ec_group[0].inputsWKC
 * 
 *     def _get_dc_time(self):             # <<<<<<<<<<<<<<
 *         """DC time in ns required to synchronize the EtherCAT cycle with SYNC0 cycles.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_41_get_dc_time, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster__get_dc_time, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[27])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_get_dc_time, __pyx_t_2) < (0)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":610
 *         return self._ec_DCtime
 * 
 *     def _set_manual_state_change(self, int manual_state_change):             # <<<<<<<<<<<<<<
 *         """Set manualstatechange variable in context.
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_43_set_manual_state_change, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster__set_manual_state_cha, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[28])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_set_manual_state_change, __pyx_t_2) < (0)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":624
 *         self._ecx_contextt.manualstatechange = manual_state_change
 * 
 *     def _get_manual_state_change(self):             # <<<<<<<<<<<<<<
 *         return self._ecx_contextt.manualstatechange
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_45_get_manual_state_change, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster__get_manual_state_cha, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[29])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefMaster, __pyx_mstate_global->__pyx_n_u_get_manual_state_change, __pyx_t_2) < (0)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_47__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10CdefMaster_49__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefMaster___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":629
 * 
 * 
 * class SdoError(Exception):             # <<<<<<<<<<<<<<
 *     """Sdo read or write abort
 * 
*/
  __pyx_t_2 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_33 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_32, __pyx_mstate_global->__pyx_n_u_SdoError, __pyx_mstate_global->__pyx_n_u_SdoError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_Sdo_read_or_write_abort_Attribut); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (__pyx_t_32 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_33, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 629, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":638
 *     """
 * 
 *     def __init__(self, slave_pos, index, subindex, abort_code, desc):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.index = index
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_8SdoError_1__init__, 0, __pyx_mstate_global->__pyx_n_u_SdoError___init, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_33, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_2) < (0)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":629
 * 
 * 
 * class SdoError(Exception):             # <<<<<<<<<<<<<<
 *     """Sdo read or write abort
 * 
*/
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_SdoError, __pyx_t_32, __pyx_t_33, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_SdoError, __pyx_t_2) < (0)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":645
 *         self.desc = desc
 * 
 * class Emergency(Exception):             # <<<<<<<<<<<<<<
 *     """Emergency message.
 * 
*/
  __pyx_t_32 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_5 = __Pyx_PEP560_update_bases(__pyx_t_32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_33 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_33, __pyx_t_5, __pyx_mstate_global->__pyx_n_u_Emergency, __pyx_mstate_global->__pyx_n_u_Emergency, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_Emergency_message_Attributes_sla); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_5 != __pyx_t_32) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_32) < 0))) __PYX_ERR(0, 645, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":657
 *     """
 * 
 *     def __init__(self, slave_pos, error_code, error_reg, b1, w1, w2):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9Emergency_1__init__, 0, __pyx_mstate_global->__pyx_n_u_Emergency___init, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_32) < (0)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":665
 *         self.w2 = w2
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         b1w1w2_bytes = bytes([self.b1]) + self.w1.to_bytes(length=2, byteorder='little') + self.w2.to_bytes(length=2, byteorder='little')
 *         b1w1w2_str = ','.join(format(x, '02x') for x in b1w1w2_bytes)
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9Emergency_3__str__, 0, __pyx_mstate_global->__pyx_n_u_Emergency___str, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_str, __pyx_t_32) < (0)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":645
 *         self.desc = desc
 * 
 * class Emergency(Exception):             # <<<<<<<<<<<<<<
 *     """Emergency message.
 * 
*/
  __pyx_t_32 = __Pyx_Py3ClassCreate(__pyx_t_33, __pyx_mstate_global->__pyx_n_u_Emergency, __pyx_t_5, __pyx_t_2, NULL, 0, 1); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Emergency, __pyx_t_32) < (0)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":671
 * 
 * 
 * class SdoInfoError(Exception):             # <<<<<<<<<<<<<<
 *     """Errors during Object directory info read
 * 
*/
  __pyx_t_5 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_33 = __Pyx_PEP560_update_bases(__pyx_t_5); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_33); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_33, __pyx_mstate_global->__pyx_n_u_SdoInfoError, __pyx_mstate_global->__pyx_n_u_SdoInfoError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_Errors_during_Object_directory_i); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (__pyx_t_33 != __pyx_t_5) {
    if (unlikely((PyDict_SetItemString(__pyx_t_32, "__orig_bases__", __pyx_t_5) < 0))) __PYX_ERR(0, 671, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":678
 *     """
 * 
 *     def __init__(self, message):             # <<<<<<<<<<<<<<
 *         self.message = message
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_12SdoInfoError_1__init__, 0, __pyx_mstate_global->__pyx_n_u_SdoInfoError___init, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_5) < (0)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":671
 * 
 * 
 * class SdoInfoError(Exception):             # <<<<<<<<<<<<<<
 *     """Errors during Object directory info read
 * 
*/
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_SdoInfoError, __pyx_t_33, __pyx_t_32, NULL, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_SdoInfoError, __pyx_t_5) < (0)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":682
 * 
 * 
 * class MailboxError(Exception):             # <<<<<<<<<<<<<<
 *     """Errors in mailbox communication
 * 
*/
  __pyx_t_33 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_33); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_32, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_MailboxError, __pyx_mstate_global->__pyx_n_u_MailboxError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_Errors_in_mailbox_communication); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_2 != __pyx_t_33) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_33) < 0))) __PYX_ERR(0, 682, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":691
 *     """
 * 
 *     def __init__(self, slave_pos, error_code, desc):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
*/
  __pyx_t_33 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_12MailboxError_1__init__, 0, __pyx_mstate_global->__pyx_n_u_MailboxError___init, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36])); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_33) < (0)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":682
 * 
 * 
 * class MailboxError(Exception):             # <<<<<<<<<<<<<<
 *     """Errors in mailbox communication
 * 
*/
  __pyx_t_33 = __Pyx_Py3ClassCreate(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_MailboxError, __pyx_t_2, __pyx_t_5, NULL, 0, 1); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MailboxError, __pyx_t_33) < (0)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":697
 * 
 * 
 * class PacketError(Exception):             # <<<<<<<<<<<<<<
 *     """Errors related to mailbox communication
 * 
*/
  __pyx_t_2 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_33 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_32, __pyx_mstate_global->__pyx_n_u_PacketError, __pyx_mstate_global->__pyx_n_u_PacketError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_Errors_related_to_mailbox_commun); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (__pyx_t_32 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_33, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 697, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":709
 *     # based on the comments in the soem code
 *     _code_desc = {
 *       1: 'Unexpected frame returned',             # <<<<<<<<<<<<<<
 *       3: 'Data container too small for type',
 *     }
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_int_1, __pyx_mstate_global->__pyx_kp_u_Unexpected_frame_returned) < (0)) __PYX_ERR(0, 709, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_int_3, __pyx_mstate_global->__pyx_kp_u_Data_container_too_small_for_typ) < (0)) __PYX_ERR(0, 709, __pyx_L1_error)
  if (__Pyx_SetNameInClass(__pyx_t_33, __pyx_mstate_global->__pyx_n_u_code_desc, __pyx_t_2) < (0)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":713
 *     }
 * 
 *     def __init__(self, slave_pos, error_code):             # <<<<<<<<<<<<<<
 *         self.slave_pos = slave_pos
 *         self.error_code = error_code
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_11PacketError_1__init__, 0, __pyx_mstate_global->__pyx_n_u_PacketError___init, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_33, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_2) < (0)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":717
 *         self.error_code = error_code
 * 
 *     def _get_desc(self):             # <<<<<<<<<<<<<<
 *         return self._code_desc[self.error_code]
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_11PacketError_3_get_desc, 0, __pyx_mstate_global->__pyx_n_u_PacketError__get_desc, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[38])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_33, __pyx_mstate_global->__pyx_n_u_get_desc, __pyx_t_2) < (0)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":720
 *         return self._code_desc[self.error_code]
 * 
 *     desc = property(_get_desc)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_6 = NULL;
  __pyx_t_4 = PyObject_GetItem(__pyx_t_33, __pyx_mstate_global->__pyx_n_u_get_desc);
  if (unlikely(!__pyx_t_4)) {
    PyErr_Clear();
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get_desc);
  }
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (__Pyx_SetNameInClass(__pyx_t_33, __pyx_mstate_global->__pyx_n_u_desc, __pyx_t_2) < (0)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":697
 * 
 * 
 * class PacketError(Exception):             # <<<<<<<<<<<<<<
 *     """Errors related to mailbox communication
 * 
*/
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PacketError, __pyx_t_32, __pyx_t_33, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_PacketError, __pyx_t_2) < (0)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":723
 * 
 * 
 * class ConfigMapError(Exception):             # <<<<<<<<<<<<<<
 *     """Errors during Object directory info read
 * 
*/
  __pyx_t_32 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_5 = __Pyx_PEP560_update_bases(__pyx_t_32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_33 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_33, __pyx_t_5, __pyx_mstate_global->__pyx_n_u_ConfigMapError, __pyx_mstate_global->__pyx_n_u_ConfigMapError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_Errors_during_Object_directory_i_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_5 != __pyx_t_32) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_32) < 0))) __PYX_ERR(0, 723, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":730
 *     """
 * 
 *     def __init__(self, error_list):             # <<<<<<<<<<<<<<
 *         self.error_list = error_list
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_14ConfigMapError_1__init__, 0, __pyx_mstate_global->__pyx_n_u_ConfigMapError___init, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_32) < (0)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":723
 * 
 * 
 * class ConfigMapError(Exception):             # <<<<<<<<<<<<<<
 *     """Errors during Object directory info read
 * 
*/
  __pyx_t_32 = __Pyx_Py3ClassCreate(__pyx_t_33, __pyx_mstate_global->__pyx_n_u_ConfigMapError, __pyx_t_5, __pyx_t_2, NULL, 0, 1); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ConfigMapError, __pyx_t_32) < (0)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":734
 * 
 * 
 * class EepromError(Exception):             # <<<<<<<<<<<<<<
 *     """EEPROM access error
 * 
*/
  __pyx_t_5 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_33 = __Pyx_PEP560_update_bases(__pyx_t_5); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_33); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_33, __pyx_mstate_global->__pyx_n_u_EepromError, __pyx_mstate_global->__pyx_n_u_EepromError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_EEPROM_access_error_Attributes_m); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (__pyx_t_33 != __pyx_t_5) {
    if (unlikely((PyDict_SetItemString(__pyx_t_32, "__orig_bases__", __pyx_t_5) < 0))) __PYX_ERR(0, 734, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":741
 *     """
 * 
 *     def __init__(self, message):             # <<<<<<<<<<<<<<
 *         self.message = message
 * 
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_11EepromError_1__init__, 0, __pyx_mstate_global->__pyx_n_u_EepromError___init, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[40])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (__Pyx_SetNameInClass(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_5) < (0)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":734
 * 
 * 
 * class EepromError(Exception):             # <<<<<<<<<<<<<<
 *     """EEPROM access error
 * 
*/
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_EepromError, __pyx_t_33, __pyx_t_32, NULL, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_EepromError, __pyx_t_5) < (0)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":745
 * 
 * 
 * class WkcError(Exception):             # <<<<<<<<<<<<<<
 *     """Working counter error.
 * 
*/
  __pyx_t_33 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_33); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_32, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_WkcError, __pyx_mstate_global->__pyx_n_u_WkcError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_Working_counter_error_Attributes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_2 != __pyx_t_33) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_33) < 0))) __PYX_ERR(0, 745, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":753
 *     """
 * 
 *     def __init__(self, message=None, wkc=None):             # <<<<<<<<<<<<<<
 *         self.message = message
 *         self.wkc = wkc
*/
  __pyx_t_33 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_8WkcError_1__init__, 0, __pyx_mstate_global->__pyx_n_u_WkcError___init, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[41])); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_33, __pyx_mstate_global->__pyx_tuple[5]);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_33) < (0)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pysoem/pysoem.pyx":745
 * 
 * 
 * class WkcError(Exception):             # <<<<<<<<<<<<<<
 *     """Working counter error.
 * 
*/
  __pyx_t_33 = __Pyx_Py3ClassCreate(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_WkcError, __pyx_t_2, __pyx_t_5, NULL, 0, 1); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_33);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_WkcError, __pyx_t_33) < (0)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pysoem/pysoem.pyx":757
 *         self.wkc = wkc
 * 
 * class NetworkInterfaceNotOpenError(Exception):             # <<<<<<<<<<<<<<
 *     """Error when a master or slave method is used and the context has not been initialized."""
 *     pass
*/
  __pyx_t_2 = PyTuple_Pack(1, ((PyObject *)(((PyTypeObject*)PyExc_Exception)))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_33 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_32, __pyx_mstate_global->__pyx_n_u_NetworkInterfaceNotOpenError, __pyx_mstate_global->__pyx_n_u_NetworkInterfaceNotOpenError, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_Error_when_a_master_or_slave_met); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  if (__pyx_t_32 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_33, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 757, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_NetworkInterfaceNotOpenError, __pyx_t_32, __pyx_t_33, NULL, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_NetworkInterfaceNotOpenError, __pyx_t_2) < (0)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13_CallbackData_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CallbackData___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[42])); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem__CallbackData, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_32) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__CallbackData, (type(self), 0x0955bee, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__CallbackData__set_state(self, __pyx_state)
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13_CallbackData_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CallbackData___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[43])); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem__CallbackData, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_32) < (0)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":770
 * 
 * 
 * class SiiOffset:             # <<<<<<<<<<<<<<
 *     """Item offsets in SII general section."""
 *     # Took it from ethercattype.h but no type was given.
*/
  __pyx_t_32 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_mstate_global->__pyx_empty_tuple, __pyx_mstate_global->__pyx_n_u_SiiOffset, __pyx_mstate_global->__pyx_n_u_SiiOffset, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_kp_u_Item_offsets_in_SII_general_sect); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);

  /* "pysoem/pysoem.pyx":773
 *     """Item offsets in SII general section."""
 *     # Took it from ethercattype.h but no type was given.
 *     MAN = 0x0008             # <<<<<<<<<<<<<<
 *     ID = 0x000A
 *     REV = 0x000B
*/
  if (__Pyx_SetNameInClass(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_MAN, __pyx_mstate_global->__pyx_int_8) < (0)) __PYX_ERR(0, 773, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":774
 *     # Took it from ethercattype.h but no type was given.
 *     MAN = 0x0008
 *     ID = 0x000A             # <<<<<<<<<<<<<<
 *     REV = 0x000B
 *     BOOT_RX_MBX = 0x0014
*/
  if (__Pyx_SetNameInClass(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_ID, __pyx_mstate_global->__pyx_int_10) < (0)) __PYX_ERR(0, 774, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":775
 *     MAN = 0x0008
 *     ID = 0x000A
 *     REV = 0x000B             # <<<<<<<<<<<<<<
 *     BOOT_RX_MBX = 0x0014
 *     BOOT_TX_MBX = 0x0016
*/
  if (__Pyx_SetNameInClass(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_REV, __pyx_mstate_global->__pyx_int_11) < (0)) __PYX_ERR(0, 775, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":776
 *     ID = 0x000A
 *     REV = 0x000B
 *     BOOT_RX_MBX = 0x0014             # <<<<<<<<<<<<<<
 *     BOOT_TX_MBX = 0x0016
 *     STD_RX_MBX = 0x0018
*/
  if (__Pyx_SetNameInClass(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_BOOT_RX_MBX, __pyx_mstate_global->__pyx_int_20) < (0)) __PYX_ERR(0, 776, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":777
 *     REV = 0x000B
 *     BOOT_RX_MBX = 0x0014
 *     BOOT_TX_MBX = 0x0016             # <<<<<<<<<<<<<<
 *     STD_RX_MBX = 0x0018
 *     STD_TX_MBX = 0x001A
*/
  if (__Pyx_SetNameInClass(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_BOOT_TX_MBX, __pyx_mstate_global->__pyx_int_22) < (0)) __PYX_ERR(0, 777, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":778
 *     BOOT_RX_MBX = 0x0014
 *     BOOT_TX_MBX = 0x0016
 *     STD_RX_MBX = 0x0018             # <<<<<<<<<<<<<<
 *     STD_TX_MBX = 0x001A
 *     MBX_PROTO = 0x001C
*/
  if (__Pyx_SetNameInClass(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_STD_RX_MBX, __pyx_mstate_global->__pyx_int_24) < (0)) __PYX_ERR(0, 778, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":779
 *     BOOT_TX_MBX = 0x0016
 *     STD_RX_MBX = 0x0018
 *     STD_TX_MBX = 0x001A             # <<<<<<<<<<<<<<
 *     MBX_PROTO = 0x001C
 * 
*/
  if (__Pyx_SetNameInClass(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_STD_TX_MBX, __pyx_mstate_global->__pyx_int_26) < (0)) __PYX_ERR(0, 779, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":780
 *     STD_RX_MBX = 0x0018
 *     STD_TX_MBX = 0x001A
 *     MBX_PROTO = 0x001C             # <<<<<<<<<<<<<<
 * 
 * 
*/
  if (__Pyx_SetNameInClass(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_MBX_PROTO, __pyx_mstate_global->__pyx_int_28) < (0)) __PYX_ERR(0, 780, __pyx_L1_error)

  /* "pysoem/pysoem.pyx":770
 * 
 * 
 * class SiiOffset:             # <<<<<<<<<<<<<<
 *     """Item offsets in SII general section."""
 *     # Took it from ethercattype.h but no type was given.
*/
  __pyx_t_5 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_mstate_global->__pyx_n_u_SiiOffset, __pyx_mstate_global->__pyx_empty_tuple, __pyx_t_32, NULL, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_SiiOffset, __pyx_t_5) < (0)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":804
 *     cdef public _emcy_callbacks
 * 
 *     name = property(_get_name)             # <<<<<<<<<<<<<<
 *     man = property(_get_eep_man)
 *     id = property(_get_eep_id)
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_name); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_33};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_name, __pyx_t_32) < (0)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":805
 * 
 *     name = property(_get_name)
 *     man = property(_get_eep_man)             # <<<<<<<<<<<<<<
 *     id = property(_get_eep_id)
 *     rev = property(_get_eep_rev)
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_eep_man); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_33, __pyx_t_5};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 805, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_man, __pyx_t_32) < (0)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":806
 *     name = property(_get_name)
 *     man = property(_get_eep_man)
 *     id = property(_get_eep_id)             # <<<<<<<<<<<<<<
 *     rev = property(_get_eep_rev)
 *     config_func = property(_get_PO2SOconfig, _set_PO2SOconfig)
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_eep_id); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_33};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_id, __pyx_t_32) < (0)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":807
 *     man = property(_get_eep_man)
 *     id = property(_get_eep_id)
 *     rev = property(_get_eep_rev)             # <<<<<<<<<<<<<<
 *     config_func = property(_get_PO2SOconfig, _set_PO2SOconfig)
 *     setup_func = property(_get_PO2SOconfigEx, _set_PO2SOconfigEx)
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_eep_rev); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_33, __pyx_t_5};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_rev, __pyx_t_32) < (0)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":808
 *     id = property(_get_eep_id)
 *     rev = property(_get_eep_rev)
 *     config_func = property(_get_PO2SOconfig, _set_PO2SOconfig)             # <<<<<<<<<<<<<<
 *     setup_func = property(_get_PO2SOconfigEx, _set_PO2SOconfigEx)
 *     state = property(_get_state, _set_state)
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_PO2SOconfig); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_PO2SOconfig); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_33, __pyx_t_2};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (3-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_config_func, __pyx_t_32) < (0)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":809
 *     rev = property(_get_eep_rev)
 *     config_func = property(_get_PO2SOconfig, _set_PO2SOconfig)
 *     setup_func = property(_get_PO2SOconfigEx, _set_PO2SOconfigEx)             # <<<<<<<<<<<<<<
 *     state = property(_get_state, _set_state)
 *     input = property(_get_input)
*/
  __pyx_t_2 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_PO2SOconfigEx); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_PO2SOconfigEx); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_33, __pyx_t_5};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (3-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_setup_func, __pyx_t_32) < (0)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":810
 *     config_func = property(_get_PO2SOconfig, _set_PO2SOconfig)
 *     setup_func = property(_get_PO2SOconfigEx, _set_PO2SOconfigEx)
 *     state = property(_get_state, _set_state)             # <<<<<<<<<<<<<<
 *     input = property(_get_input)
 *     output = property(_get_output, _set_output)
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_state); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_33, __pyx_t_2};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (3-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_state, __pyx_t_32) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":811
 *     setup_func = property(_get_PO2SOconfigEx, _set_PO2SOconfigEx)
 *     state = property(_get_state, _set_state)
 *     input = property(_get_input)             # <<<<<<<<<<<<<<
 *     output = property(_get_output, _set_output)
 *     al_status = property(_get_al_status)
*/
  __pyx_t_2 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_input); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_33};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_input, __pyx_t_32) < (0)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":812
 *     state = property(_get_state, _set_state)
 *     input = property(_get_input)
 *     output = property(_get_output, _set_output)             # <<<<<<<<<<<<<<
 *     al_status = property(_get_al_status)
 *     is_lost = property(_get_is_lost, _set_is_lost)
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_output); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_output); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_33, __pyx_t_2, __pyx_t_5};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (3-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_output, __pyx_t_32) < (0)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":813
 *     input = property(_get_input)
 *     output = property(_get_output, _set_output)
 *     al_status = property(_get_al_status)             # <<<<<<<<<<<<<<
 *     is_lost = property(_get_is_lost, _set_is_lost)
 *     od = property(_get_od)
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_al_status); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_2};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_al_status, __pyx_t_32) < (0)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":814
 *     output = property(_get_output, _set_output)
 *     al_status = property(_get_al_status)
 *     is_lost = property(_get_is_lost, _set_is_lost)             # <<<<<<<<<<<<<<
 *     od = property(_get_od)
 * 
*/
  __pyx_t_2 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_is_lost); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_is_lost); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_5, __pyx_t_33};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (3-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_is_lost, __pyx_t_32) < (0)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":815
 *     al_status = property(_get_al_status)
 *     is_lost = property(_get_is_lost, _set_is_lost)
 *     od = property(_get_od)             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, pos):
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_od); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_33, __pyx_t_5};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_od, __pyx_t_32) < (0)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":823
 *         self._emcy_callbacks = []
 * 
 *     def dc_sync(self, act, sync0_cycle_time, sync0_shift_time=0, sync1_cycle_time=None):             # <<<<<<<<<<<<<<
 *         """Activate or deactivate SYNC pulses at the slave.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_3dc_sync, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_dc_sync, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[44])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[6]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_dc_sync, __pyx_t_32) < (0)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":859
 *         return result
 * 
 *     def sdo_read(self, index, uint8_t subindex, int size=0, ca=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Read a CoE object.
 * 
*/
  __pyx_t_32 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyDict_SetItem(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_release_gil, Py_None) < (0)) __PYX_ERR(0, 859, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_5sdo_read, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_sdo_read, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[45])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[7]);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_5, __pyx_t_32);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_sdo_read, __pyx_t_5) < (0)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":949
 *         return result
 * 
 *     def sdo_write(self, index, uint8_t subindex, bytes data, ca=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Write to a CoE object.
 * 
*/
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_release_gil, Py_None) < (0)) __PYX_ERR(0, 949, __pyx_L1_error)
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_7sdo_write, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_sdo_write, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[46])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_32, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_sdo_write, __pyx_t_32) < (0)) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":986
 *             raise WkcError(wkc=result)
 * 
 *     def mbx_receive(self):             # <<<<<<<<<<<<<<
 *         """Read out the slaves out mailbox - to check for emergency messages.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_9mbx_receive, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_mbx_receive, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[47])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_mbx_receive, __pyx_t_32) < (0)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1010
 *         return wkt
 * 
 *     def write_state(self):             # <<<<<<<<<<<<<<
 *         """Write slave state.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_11write_state, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_write_state, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[48])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_write_state, __pyx_t_32) < (0)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1018
 *         return cpysoem.ecx_writestate(self._ecx_contextt, self._pos)
 * 
 *     def state_check(self, int expected_state, timeout=2000):             # <<<<<<<<<<<<<<
 *         """Wait for the slave to reach the state that was requested."""
 *         self._master.check_context_is_initialized()
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_13state_check, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_state_check, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[49])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[4]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_state_check, __pyx_t_32) < (0)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1023
 *         return cpysoem.ecx_statecheck(self._ecx_contextt, self._pos, expected_state, timeout)
 * 
 *     def reconfig(self, timeout=500):             # <<<<<<<<<<<<<<
 *         """Reconfigure slave.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_15reconfig, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_reconfig, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[50])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[8]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_reconfig, __pyx_t_32) < (0)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1033
 *         return cpysoem.ecx_reconfig_slave(self._ecx_contextt, self._pos, timeout)
 * 
 *     def recover(self, timeout=500):             # <<<<<<<<<<<<<<
 *         """Recover slave.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_17recover, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_recover, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[51])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[8]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_recover, __pyx_t_32) < (0)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1043
 *         return cpysoem.ecx_recover_slave(self._ecx_contextt, self._pos, timeout)
 * 
 *     def eeprom_read(self, int word_address, timeout=20000):             # <<<<<<<<<<<<<<
 *         """Read 4 byte from EEPROM
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_19eeprom_read, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_eeprom_read, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[52])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[9]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_eeprom_read, __pyx_t_32) < (0)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1059
 *         return PyBytes_FromStringAndSize(<char*>&tmp, 4)
 * 
 *     def eeprom_write(self, int word_address, bytes data, timeout=20000):             # <<<<<<<<<<<<<<
 *         """Write 2 byte (1 word) to EEPROM
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_21eeprom_write, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_eeprom_write, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[53])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[9]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_eeprom_write, __pyx_t_32) < (0)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1104
 *         return result
 * 
 *     def foe_write(self, filename, password, bytes data, timeout = 200000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """ Write given data to device using FoE
 * 
*/
  __pyx_t_32 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyDict_SetItem(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_release_gil, Py_None) < (0)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_23foe_write, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_foe_write, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[54])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[10]);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_5, __pyx_t_32);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_foe_write, __pyx_t_5) < (0)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pysoem/pysoem.pyx":1158
 *         return result
 * 
 *     def foe_read(self, filename, password, size, timeout = 200000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Read given filename from device using FoE
 * 
*/
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_release_gil, Py_None) < (0)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_25foe_read, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_foe_read, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[55])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[10]);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_32, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_foe_read, __pyx_t_32) < (0)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1199
 *             PyMem_Free(pbuf)
 * 
 *     def amend_mbx(self, mailbox, start_address, size):             # <<<<<<<<<<<<<<
 *         """Change the start address and size of a mailbox.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_27amend_mbx, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_amend_mbx, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[56])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_amend_mbx, __pyx_t_32) < (0)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1238
 *             raise AttributeError()
 * 
 *     def set_watchdog(self, wd_type, wd_time_ms):             # <<<<<<<<<<<<<<
 *         """Change the watchdog time of the PDI or Process Data watchdog.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_29set_watchdog, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_set_watchdog, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[57])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_watchdog, __pyx_t_32) < (0)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1275
 *             warnings.warn(f'The actual set watchdog time ({actual_wd_time_ms} ms) differs from the requested watchdog time ({wd_time_ms} ms) due to resolution limits of the hardware!', UserWarning)
 * 
 *     def get_watchdog(self, wd_type):             # <<<<<<<<<<<<<<
 *         """Get the watchdog time of the PDI or Process Data watchdog.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_31get_watchdog, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_get_watchdog, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[58])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_watchdog, __pyx_t_32) < (0)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1297
 *         return wd_time_ms
 * 
 *     def _get_watchdog_divider_ns(self):             # <<<<<<<<<<<<<<
 *         fprd_fpwr_timeout_us = 4000
 *         wd_div_reg = int.from_bytes(self._fprd(ECT_REG_WD_DIV, 2, fprd_fpwr_timeout_us),
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_33_get_watchdog_divider_ns, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_watchdog_divider, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[59])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_watchdog_divider_ns, __pyx_t_32) < (0)) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1305
 *         return wd_div_ns
 * 
 *     def get_max_watchdog_time(self):             # <<<<<<<<<<<<<<
 *         """Get the maximum watchdog timeout.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_35get_max_watchdog_time, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_get_max_watchdog_time, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[60])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_max_watchdog_time, __pyx_t_32) < (0)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1316
 *         return wd_time_ms_limit
 * 
 *     def add_emergency_callback(self, callback):             # <<<<<<<<<<<<<<
 *         """Get notified on EMCY messages from this slave.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_37add_emergency_callback, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave_add_emergency_callback, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[61])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_add_emergency_callback, __pyx_t_32) < (0)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1341
 *             callback(emergency_msg)
 * 
 *     def _disable_complete_access(self):             # <<<<<<<<<<<<<<
 *         """Helper function that stops config_map() from using "complete access" for SDO requests for this device.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_39_disable_complete_access, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__disable_complete_acce, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[62])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_disable_complete_access, __pyx_t_32) < (0)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1354
 *         self._ec_slave.CoEdetails &= ~ECT_COEDET_SDOCA
 * 
 *     def _fprd(self, int address, int size, timeout_us=2000):             # <<<<<<<<<<<<<<
 *         """Send and receive of the FPRD cmd primitive (Configured Address Physical Read)."""
 *         cdef unsigned char* data
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_41_fprd, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__fprd, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[63])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[4]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_fprd, __pyx_t_32) < (0)) __PYX_ERR(0, 1354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1367
 *             PyMem_Free(data)
 * 
 *     def _fpwr(self, int address, bytes data, timeout_us=2000):             # <<<<<<<<<<<<<<
 *         """Send and receive of the FPWR cmd primitive (Configured Address Physical Write)."""
 *         cdef int wkc = cpysoem.ecx_FPWR(self._ecx_contextt.port, self._ec_slave.configadr, address, <int>len(data), <unsigned char*>data, timeout_us)
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_43_fpwr, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__fpwr, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[64])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[4]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_fpwr, __pyx_t_32) < (0)) __PYX_ERR(0, 1367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1398
 *             raise Exception('unexpected error, Etype: {}'.format(err.Etype))
 * 
 *     def _get_name(self):             # <<<<<<<<<<<<<<
 *         """Name of the slave, read out from the slaves SII during config_init."""
 *         return (<bytes>self._ec_slave.name).decode('utf8')
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_45_get_name, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_name, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[65])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_name, __pyx_t_32) < (0)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1402
 *         return (<bytes>self._ec_slave.name).decode('utf8')
 * 
 *     def _get_eep_man(self):             # <<<<<<<<<<<<<<
 *         """Vendor ID of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_man
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_47_get_eep_man, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_eep_man, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[66])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_eep_man, __pyx_t_32) < (0)) __PYX_ERR(0, 1402, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1406
 *         return self._ec_slave.eep_man
 * 
 *     def _get_eep_id(self):             # <<<<<<<<<<<<<<
 *         """Product Code of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_id
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_49_get_eep_id, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_eep_id, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[67])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_eep_id, __pyx_t_32) < (0)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1410
 *         return self._ec_slave.eep_id
 * 
 *     def _get_eep_rev(self):             # <<<<<<<<<<<<<<
 *         """Revision Number of the slave, read out from the slaves SII during config_init."""
 *         return self._ec_slave.eep_rev
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_51_get_eep_rev, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_eep_rev, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[68])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_eep_rev, __pyx_t_32) < (0)) __PYX_ERR(0, 1410, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1414
 *         return self._ec_slave.eep_rev
 * 
 *     def _get_PO2SOconfig(self):             # <<<<<<<<<<<<<<
 *         """Slaves callback function that is called during config_map.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_53_get_PO2SOconfig, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_PO2SOconfig, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[69])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_PO2SOconfig, __pyx_t_32) < (0)) __PYX_ERR(0, 1414, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1423
 *         return <object>self._ec_slave.user
 * 
 *     def _get_PO2SOconfigEx(self):             # <<<<<<<<<<<<<<
 *         """Alternative callback function that is called during config_map.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_55_get_PO2SOconfigEx, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_PO2SOconfigEx, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[70])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_PO2SOconfigEx, __pyx_t_32) < (0)) __PYX_ERR(0, 1423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1436
 *         return <object>self._ec_slave.user
 * 
 *     def _set_PO2SOconfig(self, value):             # <<<<<<<<<<<<<<
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_57_set_PO2SOconfig, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__set_PO2SOconfig, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[71])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_PO2SOconfig, __pyx_t_32) < (0)) __PYX_ERR(0, 1436, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1444
 *             self._ec_slave.PO2SOconfig = _xPO2SOconfig
 * 
 *     def _set_PO2SOconfigEx(self, value):             # <<<<<<<<<<<<<<
 *         self._cd.func = value
 *         self._ec_slave.user = <void*>self._cd
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_59_set_PO2SOconfigEx, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__set_PO2SOconfigEx, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[72])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_PO2SOconfigEx, __pyx_t_32) < (0)) __PYX_ERR(0, 1444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1452
 *             self._ec_slave.PO2SOconfig = _xPO2SOconfigEx
 * 
 *     def _get_state(self):             # <<<<<<<<<<<<<<
 *         """Request a new state.
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_61_get_state, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_state, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[73])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_state, __pyx_t_32) < (0)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1459
 *         return self._ec_slave.state
 * 
 *     def _set_state(self, value):             # <<<<<<<<<<<<<<
 *         self._ec_slave.state = value
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_63_set_state, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__set_state, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[74])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_state, __pyx_t_32) < (0)) __PYX_ERR(0, 1459, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1462
 *         self._ec_slave.state = value
 * 
 *     def _get_input(self):             # <<<<<<<<<<<<<<
 *         num_bytes = self._ec_slave.Ibytes
 *         if (self._ec_slave.Ibytes == 0 and self._ec_slave.Ibits > 0):
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_65_get_input, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_input, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[75])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_input, __pyx_t_32) < (0)) __PYX_ERR(0, 1462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1468
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.inputs, num_bytes)
 * 
 *     def _get_output(self):             # <<<<<<<<<<<<<<
 *         num_bytes = self._ec_slave.Obytes
 *         if (self._ec_slave.Obytes == 0 and self._ec_slave.Obits > 0):
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_67_get_output, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_output, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[76])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_output, __pyx_t_32) < (0)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1474
 *         return PyBytes_FromStringAndSize(<char*>self._ec_slave.outputs, num_bytes)
 * 
 *     def _set_output(self, bytes value):             # <<<<<<<<<<<<<<
 *         memcpy(<char*>self._ec_slave.outputs, <char*>value, len(value))
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_69_set_output, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__set_output, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[77])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_output, __pyx_t_32) < (0)) __PYX_ERR(0, 1474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1477
 *         memcpy(<char*>self._ec_slave.outputs, <char*>value, len(value))
 * 
 *     def _get_al_status(self):             # <<<<<<<<<<<<<<
 *         return self._ec_slave.ALstatuscode
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_71_get_al_status, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_al_status, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[78])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_al_status, __pyx_t_32) < (0)) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1480
 *         return self._ec_slave.ALstatuscode
 * 
 *     def _get_is_lost(self):             # <<<<<<<<<<<<<<
 *         return self._ec_slave.islost
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_73_get_is_lost, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_is_lost, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[79])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_is_lost, __pyx_t_32) < (0)) __PYX_ERR(0, 1480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1483
 *         return self._ec_slave.islost
 * 
 *     def _set_is_lost(self, value):             # <<<<<<<<<<<<<<
 *         self._ec_slave.islost = value
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_75_set_is_lost, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__set_is_lost, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[80])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_set_is_lost, __pyx_t_32) < (0)) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1486
 *         self._ec_slave.islost = value
 * 
 *     def _get_od(self):             # <<<<<<<<<<<<<<
 *         logger.debug('ecx_readODlist()')
 *         cdef int result = cpysoem.ecx_readODlist(self._ecx_contextt, self._pos, &self._ex_odlist)
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_77_get_od, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave__get_od, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[81])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefSlave, __pyx_mstate_global->__pyx_n_u_get_od, __pyx_t_32) < (0)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_79__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[82])); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_32) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for pickling"
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_9CdefSlave_81__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefSlave___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[83])); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_32) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1515
 *     cdef cpysoem.ec_OElistt _ex_oelist
 * 
 *     index = property(_get_index)             # <<<<<<<<<<<<<<
 *     data_type = property(_get_data_type)
 *     name = property(_get_name)
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_index); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_33};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_index, __pyx_t_32) < (0)) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1516
 * 
 *     index = property(_get_index)
 *     data_type = property(_get_data_type)             # <<<<<<<<<<<<<<
 *     name = property(_get_name)
 *     object_code = property(_get_object_code)
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_data_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_33, __pyx_t_5};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_data_type, __pyx_t_32) < (0)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1517
 *     index = property(_get_index)
 *     data_type = property(_get_data_type)
 *     name = property(_get_name)             # <<<<<<<<<<<<<<
 *     object_code = property(_get_object_code)
 *     entries = property(_get_entries)
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_name); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_33};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_name, __pyx_t_32) < (0)) __PYX_ERR(0, 1517, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1518
 *     data_type = property(_get_data_type)
 *     name = property(_get_name)
 *     object_code = property(_get_object_code)             # <<<<<<<<<<<<<<
 *     entries = property(_get_entries)
 *     bit_length = property(_get_bit_length)
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_object_code); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_33, __pyx_t_5};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_object_code, __pyx_t_32) < (0)) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1519
 *     name = property(_get_name)
 *     object_code = property(_get_object_code)
 *     entries = property(_get_entries)             # <<<<<<<<<<<<<<
 *     bit_length = property(_get_bit_length)
 *     obj_access = property(_get_obj_access)
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_entries); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_33};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1519, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_entries, __pyx_t_32) < (0)) __PYX_ERR(0, 1519, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1520
 *     object_code = property(_get_object_code)
 *     entries = property(_get_entries)
 *     bit_length = property(_get_bit_length)             # <<<<<<<<<<<<<<
 *     obj_access = property(_get_obj_access)
 * 
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_bit_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_33, __pyx_t_5};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_bit_length, __pyx_t_32) < (0)) __PYX_ERR(0, 1520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1521
 *     entries = property(_get_entries)
 *     bit_length = property(_get_bit_length)
 *     obj_access = property(_get_obj_access)             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int item):
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_obj_access); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_33};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_obj_access, __pyx_t_32) < (0)) __PYX_ERR(0, 1521, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1528
 *         self._are_entries_read = False
 * 
 *     def _read_description(self):             # <<<<<<<<<<<<<<
 *         cdef int result
 *         if not self._is_description_read:
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_3_read_description, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject__read_description, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[84])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_read_description, __pyx_t_32) < (0)) __PYX_ERR(0, 1528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1537
 *           self._is_description_read = True
 * 
 *     def _read_entries(self):             # <<<<<<<<<<<<<<
 *         cdef int result
 *         if not self._are_entries_read:
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_5_read_entries, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject__read_entries, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[85])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_read_entries, __pyx_t_32) < (0)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1546
 *             self._are_entries_read = True
 * 
 *     def _get_index(self):             # <<<<<<<<<<<<<<
 *         return self._ex_odlist.Index[self._item]
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_7_get_index, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject__get_index, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[86])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_index, __pyx_t_32) < (0)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1549
 *         return self._ex_odlist.Index[self._item]
 * 
 *     def _get_data_type(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         return self._ex_odlist.DataType[self._item]
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_9_get_data_type, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject__get_data_type, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[87])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_data_type, __pyx_t_32) < (0)) __PYX_ERR(0, 1549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1553
 *         return self._ex_odlist.DataType[self._item]
 * 
 *     def _get_object_code(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         return self._ex_odlist.ObjectCode[self._item]
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_11_get_object_code, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject__get_object_code, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[88])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_object_code, __pyx_t_32) < (0)) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1557
 *         return self._ex_odlist.ObjectCode[self._item]
 * 
 *     def _get_name(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         return self._ex_odlist.Name[self._item]
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_13_get_name, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject__get_name, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[89])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_name, __pyx_t_32) < (0)) __PYX_ERR(0, 1557, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1561
 *         return self._ex_odlist.Name[self._item]
 * 
 *     def _get_entries(self):             # <<<<<<<<<<<<<<
 *         self._read_description()
 *         self._read_entries()
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_15_get_entries, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject__get_entries, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[90])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_entries, __pyx_t_32) < (0)) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1575
 *             return entries
 * 
 *     def _get_bit_length(self):             # <<<<<<<<<<<<<<
 *         cdef int sum = 0
 *         self._read_description()
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_17_get_bit_length, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject__get_bit_length, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[91])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_bit_length, __pyx_t_32) < (0)) __PYX_ERR(0, 1575, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1586
 *             return sum
 * 
 *     def _get_obj_access(self):             # <<<<<<<<<<<<<<
 *         # read functions to return correct value if no previous access to other variables is done
 *         self._read_description()
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_19_get_obj_access, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject__get_obj_access, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[92])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObject, __pyx_mstate_global->__pyx_n_u_get_obj_access, __pyx_t_32) < (0)) __PYX_ERR(0, 1586, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_21__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[93])); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_32) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for pickling"
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_13CdefCoeObject_23__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObject___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[94])); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_32) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1600
 *     cdef int _item
 * 
 *     name = property(_get_name)             # <<<<<<<<<<<<<<
 *     data_type = property(_get_data_type)
 *     bit_length = property(_get_bit_length)
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_get_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_33, __pyx_t_5};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_name, __pyx_t_32) < (0)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1601
 * 
 *     name = property(_get_name)
 *     data_type = property(_get_data_type)             # <<<<<<<<<<<<<<
 *     bit_length = property(_get_bit_length)
 *     obj_access = property(_get_obj_access)
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_get_data_type); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_33};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_data_type, __pyx_t_32) < (0)) __PYX_ERR(0, 1601, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1602
 *     name = property(_get_name)
 *     data_type = property(_get_data_type)
 *     bit_length = property(_get_bit_length)             # <<<<<<<<<<<<<<
 *     obj_access = property(_get_obj_access)
 * 
*/
  __pyx_t_33 = NULL;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_get_bit_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_33, __pyx_t_5};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_bit_length, __pyx_t_32) < (0)) __PYX_ERR(0, 1602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1603
 *     data_type = property(_get_data_type)
 *     bit_length = property(_get_bit_length)
 *     obj_access = property(_get_obj_access)             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int item):
*/
  __pyx_t_5 = NULL;
  __Pyx_GetNameInClass(__pyx_t_33, (PyObject*)__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_get_obj_access); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_33};
    __pyx_t_32 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_property, __pyx_callargs+__pyx_t_34, (2-__pyx_t_34) | (__pyx_t_34*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_obj_access, __pyx_t_32) < (0)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1608
 *         self._item = item
 * 
 *     def _get_name(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.Name[self._item]
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_3_get_name, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObjectEntry__get_name, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[95])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_get_name, __pyx_t_32) < (0)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1611
 *         return self._ex_oelist.Name[self._item]
 * 
 *     def _get_data_type(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.DataType[self._item]
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_5_get_data_type, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObjectEntry__get_data_typ, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[96])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_get_data_type, __pyx_t_32) < (0)) __PYX_ERR(0, 1611, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1614
 *         return self._ex_oelist.DataType[self._item]
 * 
 *     def _get_bit_length(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.BitLength[self._item]
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_7_get_bit_length, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObjectEntry__get_bit_leng, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[97])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_get_bit_length, __pyx_t_32) < (0)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1617
 *         return self._ex_oelist.BitLength[self._item]
 * 
 *     def _get_obj_access(self):             # <<<<<<<<<<<<<<
 *         return self._ex_oelist.ObjAccess[self._item]
 * 
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_9_get_obj_access, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObjectEntry__get_obj_acce, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[98])); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_6pysoem_6pysoem_CdefCoeObjectEntry, __pyx_mstate_global->__pyx_n_u_get_obj_access, __pyx_t_32) < (0)) __PYX_ERR(0, 1617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObjectEntry___reduce_cyth, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[99])); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_32) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ex_oelist cannot be converted to a Python object for pickling"
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_18CdefCoeObjectEntry_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CdefCoeObjectEntry___setstate_cy, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[100])); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_32) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "(tree fragment)":4
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
 * def __pyx_unpickle_CdefSettings(__pyx_type, long __pyx_checksum, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_result
 *     __Pyx_CheckUnpickleChecksum(__pyx_checksum, 0xe6b07d6, 0xdac0a2c, 0x9cc8369, b'always_release_gil, timeouts')
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_8__pyx_unpickle_CdefSettings, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_CdefSettings, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[101])); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_CdefSettings, __pyx_t_32) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "(tree fragment)":1
 * cdef extern from *:             # <<<<<<<<<<<<<<
 *     int __Pyx_CheckUnpickleChecksum(long, long, long, long, const char*) except -1
 *     int __Pyx_UpdateUnpickledDict(object, object, Py_ssize_t) except -1
*/
  __pyx_t_32 = __Pyx_CyFunction_New(&__pyx_mdef_6pysoem_6pysoem_10__pyx_unpickle__CallbackData, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__CallbackData, NULL, __pyx_mstate_global->__pyx_n_u_pysoem_pysoem, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[102])); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_32);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle__CallbackData, __pyx_t_32) < (0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pysoem/pysoem.pyx":1
 * # Copyright 2021 Benjamin Partzsch             # <<<<<<<<<<<<<<
 * #
 * # This file is part of the PySOEM project and licenced under the MIT license.
*/
  __pyx_t_32 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_32) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pysoem.pysoem", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pysoem.pysoem");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_property = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_property); if (!__pyx_builtin_property) __PYX_ERR(0, 252, __pyx_L1_error)

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_keys.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_keys.method_name = &__pyx_mstate->__pyx_n_u_keys;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "EnumType":52
 *     ('ECT_BIT8', __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(ECT_BIT8)),
 * 
 * ], module=globals().get("__module__", 'pysoem.pysoem'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_module_2, __pyx_mstate_global->__pyx_kp_u_pysoem_pysoem); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "pysoem/pysoem.pyx":288
 *         self.context_initialized = False
 * 
 *     def open(self, ifname, ifname_red=None):             # <<<<<<<<<<<<<<
 *         """Initialize and open network interface.
 * 
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);

  /* "pysoem/pysoem.pyx":345
 *         return self.always_release_gil
 * 
 *     def config_init(self, usetable=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Enumerate and init all slaves.
 * 
*/
  __pyx_mstate_global->__pyx_tuple[2] = PyTuple_Pack(1, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[2])) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);

  /* "pysoem/pysoem.pyx":473
 *         return cpysoem.ecx_writestate(&self._ecx_contextt, 0)
 * 
 *     def state_check(self, int expected_state, timeout=50000):             # <<<<<<<<<<<<<<
 *         """Check actual slave state.
 * 
*/
  __pyx_mstate_global->__pyx_tuple[3] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_50000); if (unlikely(!__pyx_mstate_global->__pyx_tuple[3])) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[3]);

  /* "pysoem/pysoem.pyx":559
 *         return result
 * 
 *     def receive_processdata(self, timeout=2000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Receive processdata from slaves.
 * 
*/
  __pyx_mstate_global->__pyx_tuple[4] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_2000); if (unlikely(!__pyx_mstate_global->__pyx_tuple[4])) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[4]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[4]);

  /* "pysoem/pysoem.pyx":753
 *     """
 * 
 *     def __init__(self, message=None, wkc=None):             # <<<<<<<<<<<<<<
 *         self.message = message
 *         self.wkc = wkc
*/
  __pyx_mstate_global->__pyx_tuple[5] = PyTuple_Pack(2, Py_None, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[5])) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[5]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[5]);

  /* "pysoem/pysoem.pyx":823
 *         self._emcy_callbacks = []
 * 
 *     def dc_sync(self, act, sync0_cycle_time, sync0_shift_time=0, sync1_cycle_time=None):             # <<<<<<<<<<<<<<
 *         """Activate or deactivate SYNC pulses at the slave.
 * 
*/
  __pyx_mstate_global->__pyx_tuple[6] = PyTuple_Pack(2, __pyx_mstate_global->__pyx_int_0, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[6])) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[6]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[6]);

  /* "pysoem/pysoem.pyx":859
 *         return result
 * 
 *     def sdo_read(self, index, uint8_t subindex, int size=0, ca=False, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """Read a CoE object.
 * 
*/
  __pyx_mstate_global->__pyx_tuple[7] = PyTuple_Pack(2, __pyx_mstate_global->__pyx_int_0, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[7])) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[7]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[7]);

  /* "pysoem/pysoem.pyx":1023
 *         return cpysoem.ecx_statecheck(self._ecx_contextt, self._pos, expected_state, timeout)
 * 
 *     def reconfig(self, timeout=500):             # <<<<<<<<<<<<<<
 *         """Reconfigure slave.
 * 
*/
  __pyx_mstate_global->__pyx_tuple[8] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_500); if (unlikely(!__pyx_mstate_global->__pyx_tuple[8])) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[8]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[8]);

  /* "pysoem/pysoem.pyx":1043
 *         return cpysoem.ecx_recover_slave(self._ecx_contextt, self._pos, timeout)
 * 
 *     def eeprom_read(self, int word_address, timeout=20000):             # <<<<<<<<<<<<<<
 *         """Read 4 byte from EEPROM
 * 
*/
  __pyx_mstate_global->__pyx_tuple[9] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_20000); if (unlikely(!__pyx_mstate_global->__pyx_tuple[9])) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[9]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[9]);

  /* "pysoem/pysoem.pyx":1104
 *         return result
 * 
 *     def foe_write(self, filename, password, bytes data, timeout = 200000, *, release_gil=None):             # <<<<<<<<<<<<<<
 *         """ Write given data to device using FoE
 * 
*/
  __pyx_mstate_global->__pyx_tuple[10] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_200000); if (unlikely(!__pyx_mstate_global->__pyx_tuple[10])) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[10]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[10]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<11; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 9; } index[] = {{1},{3},{3},{33},{82},{18},{266},{85},{103},{146},{179},{225},{36},{4},{179},{472},{86},{25},{32},{27},{185},{6},{30},{97},{25},{116},{3},{2},{3},{1},{1},{1},{8},{27},{7},{23},{16},{12},{6},{2},{15},{9},{47},{46},{50},{18},{17},{14},{13},{10},{11},{15},{39},{128},{101},{67},{21},{14},{16},{27},{30},{7},{11},{10},{11},{13},{31},{33},{13},{18},{36},{38},{34},{33},{28},{34},{31},{33},{29},{28},{26},{24},{23},{29},{30},{31},{27},{10},{28},{30},{34},{23},{28},{35},{21},{21},{35},{21},{39},{28},{16},{20},{22},{21},{29},{15},{21},{30},{35},{27},{22},{22},{12},{30},{32},{9},{27},{29},{34},{15},{15},{26},{28},{24},{21},{22},{22},{20},{22},{19},{17},{21},{20},{34},{26},{28},{22},{21},{20},{32},{19},{17},{21},{22},{18},{19},{31},{22},{21},{18},{17},{18},{19},{22},{21},{21},{12},{30},{32},{14},{23},{8},{8},{8},{8},{8},{8},{8},{8},{11},{20},{20},{14},{16},{18},{17},{10},{13},{13},{13},{13},{12},{16},{10},{10},{11},{11},{14},{19},{27},{19},{15},{18},{14},{14},{14},{14},{13},{18},{11},{20},{9},{18},{17},{34},{2},{10},{7},{7},{3},{9},{12},{21},{6},{10},{28},{8},{11},{11},{20},{21},{20},{3},{12},{9},{11},{10},{10},{8},{17},{12},{21},{9},{8},{17},{10},{3},{17},{8},{22},{7},{9},{24},{18},{9},{6},{18},{2},{12},{10},{10},{3},{9},{2},{8},{3},{2},{28},{17},{17},{18},{5},{4},{10},{10},{11},{23},{11},{9},{11},{11},{10},{18},{10},{14},{4},{9},{7},{7},{5},{4},{8},{5},{24},{7},{11},{11},{11},{12},{6},{7},{5},{4},{3},{10},{10},{9},{14},{8},{14},{12},{8},{13},{8},{9},{6},{5},{20},{5},{15},{10},{8},{7},{3},{9},{16},{18},{14},{15},{14},{12},{9},{11},{12},{12},{12},{17},{10},{10},{12},{24},{21},{9},{15},{16},{7},{11},{10},{10},{12},{24},{12},{1},{2},{6},{10},{2},{5},{8},{5},{13},{7},{4},{5},{4},{6},{6},{6},{7},{7},{8},{3},{19},{6},{11},{14},{11},{7},{13},{6},{10},{15},{4},{8},{10},{7},{4},{9},{10},{11},{2},{4},{3},{6},{8},{4},{3},{3},{3},{11},{11},{8},{13},{14},{12},{11},{10},{27},{28},{14},{12},{17},{13},{10},{19},{8},{7},{10},{17},{13},{11},{6},{7},{3},{8},{9},{4},{4},{24},{16},{16},{18},{12},{24},{12},{11},{10},{12},{10},{12},{19},{8},{10},{6},{4},{10},{5},{9},{6},{13},{5},{11},{10},{7},{8},{3},{16},{16},{16},{3},{8},{5},{4},{7},{10},{3},{8},{6},{12},{8},{4},{5},{6},{2},{2},{4},{8},{9},{10},{10},{16},{11},{7},{18},{3},{3},{12},{11},{1},{11},{82},{55},{11},{97},{44},{37},{36},{76},{353},{117},{2},{23},{108},{76},{21},{70},{31},{31},{142},{32},{59},{59},{51},{123},{103},{284},{26},{30},{204},{15},{31},{11},{11},{15},{9},{37},{16},{23},{44},{9},{14},{182},{106},{54},{25},{25},{25},{141},{25},{120},{25},{25},{28},{108},{57},{97},{11},{11},{11},{13},{11},{17},{19},{15},{17},{17},{17},{17},{113},{25},{192},{9},{21},{78},{223},{149},{102},{207},{2},{36},{128},{56},{56},{55},{36},{64},{20}};
    #if (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (5931 bytes) */
const char* const cstring = "BZh91AY&SYd\"\242l\000\004\266\177\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\300@@@@@@@@@@@@\000@\000`\030\367u\240|w\007\241\272\315\367}\271\367\325}I\321(\027gd,I\254\264\311\230\r\364\001\240tj\205n\305nc*hQ\356\360\233R\024\002\206\007\274%\020\204L\023$\361\000j~\243L\206\210\364\324\315\032\023\323@\324\322\237\222\023\311\261'\223\n1\222i\241\251\345<\322\214\323\324\321\265\036\244\320I\020\023!2\004F\251\372S\312x$\314\231M\032df\246\236\231# 4\000h\365\032\000\000\006\200\000\000\tM\001\"\024\322\232yL\324g\252h1\2426\243G\250z\2316\240\006\232\007\250\365\006\232\001\352\000\000\000i\240=C@\022jQ\021\243\324\220z\232z\236\220\321\352\014@h<\221\352i\243i\224\375S\324\320\r2\007\251\345\001\246\215\000\320\323\322z\236\246\201\261OS\322\020\250A\223\364Q\345\032\033H\017S&\214@\323\0014\321\352d2\006A\2024\3041\014F\214\206\230\232\000\000a\"A\000\204\302h\304\t\202jy\t6\232\236J4\375D\375P\365\007\244\r\r\000\000\000\000\000\000h\007\376B\300\341\245\203ls\354\353\322\244X\346\303\242\374\256\362\242\267;\227\211\2133\226s\267~\253w\370*\364\306\274J{&\002\t\025O\210)a\257\263\323\327\352\356\355\356\273\223p`\340zC\037\355\252\377\322\273\206\302\261\312|xt\t\356\307\211\200\355|\276\332\363\271\336\377[k\005\306\3705q=\353\261\313\255\272\353\265wa\22127\240\033\307\\e+\t\301\200\247>\n\202\342\270\300!\002\034\3412|<\220PE!\345\n\004\253^\215E\221\233\0022\"\210\242\212\242(\301`\254Q\222\n\013\024\024\210\252\310#\025U`\253\021F\020ETDRb]x\272@\230\013\260\034@\306'\250\030\320\327\332v\377\314\233n\236*+\373~\215\267\355s\033p^\242k(\267[P\305-\3127\3412L\234\236G\000\2777\360\002\376VYa\224L\250e\252*\010\013$Y\005\200(\010\213!\212\320)D\014\000^\003\000\200\030\2465$Y@[(\240\017]\350\241w\324--j\376\022\255PQ\"\226\250m!\214E\253\032\203\032\006\316[\026\n\305c.a\003\021!\252\224\315\2433(h\222D\222\232Y\216\013\212U\255\\\341MT\212C\206\245F\244\014\204\n\244\357$\226ARI\027\017X\355r\343\211\262\372;\261\310\341\327@""\030_\010\006KZ\243b\210\203\273\270\326\330\031\255\364\r\210\304\270\203b#\3374\311]i\267-\246d\324\244\245\037G\023nG}\214\301\202\333V\001\251P*P(\324\034v\035\223\260^\360\202Yj\025UJ(\2534\367c\245\326J\353\020\340f\262\013\030\323\031\370I\201\237\202\245[#\n\330\252\013Q\202x\034\270\0226H\314\316\326\267\203\000\301\216\024\231(\340\307\2300\267q\301Q\340@m\245\261\353\010\027UA\001%\210\tC\010\260\277\367p\335j~y\300\314\207\321\201\005\304\303\342\354x\032\377\356\364\207\372\250J_\302\277\2029\215\266K\270;R\327}\013^\030)J\330JW~\343\027#\364\343[\0037\362\315z\021\211\033\341\006?\331\224\312\342\0203YH\242\223}\017Z\307\245\t\233'hY\"1\301\363eL\n\014\266\274Z>u+M&p!\010\313\373\220&@\301\341\252mO\323\n\205\212\327\305\3778\377kb\247\212|3x\003\332\037\346\030\321\232\256\267\203\236\303\004#B\262\243\245\272\364\244%d\217{q\330\272\340\215z\216I\213\357M\213\266I$\265\026F\\\366\326\370\241\237\232\373\263F\007\257\307\250\315Og\253\345\361\366\312L_)\361\320E\005Q{=\201\355\315m\252\311\r\024;c\232\201\312\222\024\263\333k\310\344\304\355q\014u\316\333i\225\025\355\345\373\242\311\177\355(\032\031\255\013\020P)\257\232\263\226\333y\371\367\276O\254l\353dy\230\034\251S\020\320\332\266\207\n\364{\256\375oO\240g!#0\314\317\274\265\200v\016\352\277e\204\322b\205;::D[\212\300\250~F\357\316*rrUSJ\270\352\276\231V\363\365p\334`]\002u\377\250N3\024\340\003M\010\202\260Q\006\361\235\311\204\364\243\304x\032X\375[\341\031z\010\024(\"\262-6\375\2046\316\373\312\354\322\252\252 \210\366\314\311\213\224\244\206\207N;\304'\034\260J\031J\332\2253\320x\326C\246\2266\225\022\201\025\251Up\"\265Z\245,\325\316\005\212A\031\301wj\360\263\300\205\225\216\361\316\231\001\317\301@\250\232\343\214\037r,\374\267uHK\025\213\345\331\354\364\302\3663\221I#\351\213\367z\217\201\363\014\033F\346\027\230]t^\257\244\036\362\220\250\003QGZ2\262\n\036\370\207\233^\275\333=\313\347\003\236\305\374w\344g\351\374e\337\312\336\271\271.z\304t&G\304\241:\377\331\352\214F\360z\305[\271""\362\201\tD\364%&Re/(l\351\024T\336|\375\322\323S\"\246_\213\362s\327\233\233\232\376\221\211\355\262\313.\317\020\337\225\260\240\267\324\335.\210\217\0217\242\313\271\320\224\206'^\347\324\266P8WY\0044\316\326\220\346K\022p\024'\231\277\224\345\355\363\033\214\000\300\363\316cl\375\277\273\255s\361\227\217\016\332\351\357\261\302WX\305\355\267\261\343\332~Eo\2756\373^\\^\215/\241\025\347\251\276\004vM\214\201\2034\353\223Fa\003\004j\005\000\231\243\000t\326\216\246O\203_\200`\235\235\233t\\\314\016\245\305\244p\t\013M\003\244b >\247\341<\0165\267\373\336^!:)\177bl\343\000\277\215\320\317D\251B4\222\"\006D\007~\216\347\205\0363\272\355W\311E\311\354\272\022d\371\211\217\225\365\3363\270p\236\333}\035\215\345$>\374>\2274\316\337\367\211\264\273\326d\223\274r\311H\264\267\210\213\341\363\367z\362\331\203QuI)c\355\340\342\336\241V.\013D\300\335v\022\2765UFb\374\262\324[\331\341\026\025\260\227E\326J\030\3517P\235V\004~|\374h\372\205\271\223m\345H\223\317\265\357\357U$~\260\263E\320E\214\261\"5\271\222>X\221\350\375\253\260\272\367\206(\314/\253\343\273\215\034\210t\210\177\355\341,\343\314\211\362\337>47\347\271\222\260\252\222^\277\251\031iFJ\310\0349\027Y\333\355y\3435\217<\366s\237\037\202\322\225\212\341W%2KI\331J\205,\032w\2441\345\311\220\022\010\345T\221L\215\030\303Ac4F\240\310\251\024\313\270\312\313\036@$e\334fG\241\206u\234\211\r\020\237>\222A\206\207c\330\004wc$\360\204\274\243\2316rp\333\002\341\324p\350,\257\215\207D\271\271\016M\236Z\274\327g\275\177\236\273\216q\366\360\r\371\303\036\252\324\227\271\000\330\026d\214M\201\216\236\254Z0 >\323\236RD\221\215\230y\3763 .\222M\351\310\350K\260;,(\254l\r$m6\211|\363\247@2\235\374\030Aw4\200\210\223\226\300M\245\210\035\377\371\325\307v\205n\370\002\302p\307\010\275c)\272\247D\276.\327\272\366Hi4\210G\221\363\363o\257\232\311\330\246\210'\"A\003\231JP\254b\214\022t\353\345P\270\334\370\263\253\007\253\247 \310\351\377_\371^\235\333L\027d\272\0219\237\027\314\007\027|\305\037\331\263\357\246\265\232\373\326""\0355u\334\017(\020\215\353\2309\270d\264k\001\353\020\365\241\243\265\230\2075\026\242E\004\215\203J\220\235\261\"LP\364\237gB\370\216#\235\034\341\324y\203\254\255E\013f\235\232\264\245\250>\213Lt\364\377\307i\266\352\\\023%b\343\364\260\201\037L(oa\236P\216LK\3313\347\313\242d$\244XI\303\032\220]\353\315\277xK\236\275X\202\357\350t\315V8\310\342\250\203y\323\347R\334o\377GO\321\360\371\027DU}u\313\017\330\036d\260\336\222\372\362\344h=m\037-{\354\234\232\204T#\266S\031\000D\347\360t\002\255{FWz\210(AD\320QI\027\371\342>\010XM\207\301\224D\323%\023\346\301\034\377\037\336\336\261'Y\344\201\321\331o\227b\023}\207{\330NGJx\345t\257IV)]\211\230G\243?\372Tc>\354\331\220\331\363\267\207\017MzV\003LVj\344[6\203l\204I\320hu>\014\200(\264\3608\314CS\022\361\201\"9\2655\rL\2273\277\023T\022$\247\216m\361\036\210\346}\221h\355-\322\\\226\353\246\273K\025-.I&\322\022\244\353JQ\013\205q\2269\016\366md\260+*\211L\304r\023\016F\t\302\341%pl=0\202\001\304\261\022\347\tb\204p\004\030Y,b\324\021\020\340t\243\233M\312\227Bno1\346\203\030\2661\313p\323\245\313\231r\274[d\335\230w\003\n\276\375d\215g\003\267r'6\347 \261m\342\213G\203\214\233\271\020\355\342\035\355j\316d\352\334\244P\033C\036SZ\254\310\210\217\217\253\273\323O\350\334\276\236\257.\220\316\014\032\027)\245\276)L\373\026q\311@\275\341\2677\355 \360WJ\310\200E\230\271O|\375\3755,Y\250#cu:'\305L\264M\227\r\276\246\357`Y\335\231\342\036\0347\025 \014K'\3569.N\256:\202\343\271\3736\311m,E\020\212\365\334\016\005A\316\265z\301\n\341\332\366\263t\321:\"\211\334\207e\355\244&]dc%\352M1\210\274D\256\020\256\21045FHb\336eT^\301%ypw|\307\334\373\376\343\340\036\320RJ\200Q+\016\311\331\216\265\021\2307\315v\036w'Q\240\330F\016b\211\347\365\375u\321\003\325u\235\273\021\256\034\273\226\377\244\037z\232`\206\217\017X\r\206\000]\276\3376YQ\006@\374\354L8\245\305\033\370q\227{\265\305p\354\250u]\310\361hj<{\020\232)\266\307\026\266\241&\244\311\240\265n`\362\222\213\326\035\004\206\250^}`1I\262\235\321\330!\236|\233\202W\307\205\374-""\243m\014\0323\242\017\222V|g\264\214r\326\210cH8#\233\240\347KXA\340\232A5\221s\013\027=e\206\232k\214K9RC.\025\360!g\032l\314\304D\240\t\031B$\202\321\007\005\301oG}\245\203v2\023\332\372\234\247\003i\033\361\255\0050\2455L\3607'#Z\320\223\004\252\324C0\353\321\202\216\325\2645Q\274\02486\374\034\344f\036\211d$\354\355\335BH\266\215\006\320\307\260\255\316\304\227i\3074\016\2265\221,2\322Jg\3650\265d}\342\324\304\305?8\211\341\265e\233$,J\215W\2028w.c\311x\320&O7\273\352\276C\327\305\016\023#O\240\220H9&\351\301\262K\265JR\217V.\3626CW|W\214\330do\005@\014z<%o\305\334\313\335\326\221\022\340\235\312\316\247|\314\036\024&U\202\332`\\gT\343\034(\310\022I\010L\303A\345E\013]\027\3078\3142\035&\034\235\202\345\2552\235\010f\222&,\241\272,\310\256 ^\005\330\357!,\006\330Q)\304\222\221S\233\3270(\312\232\365 \211\262\224I+\304/\035\270\311\335\"p\227\002\2332\027_\026\201\272N\t\036s$\013\030bc\306\020\017hK>\001\0205A\207!l\007\020\315l\016M\024\244\334\232\233\235\225\370\336WArA6\202\222!\013\004\331\373Y_\004\211\2138\355\357\221\rg2\211ma\013F\336\315\345k\311\221GzF\202\301\005B\020\207\004\227\214\256\3314\266\222\313\030\253-\316l\202\307q) \217D\316\016j\372\241p}d\373\255\211\317\020.H\224\000\350\350p \210\242\304\364,\320\333m\266\322\022\252$X\n\261U\277T\304\265\307,,\036M\245$8\200\312\375\032\036\005\3230\320Ex\271\346q\266\034\233j\271\346NE01h\2234\373\274\017wa\362]\212G\317\250\013\362\022H\322\032\024\032\271\307\303\013\343\210H\341e\263&\227o#\317*\001W\177\035\032\311\314\267>c\n*kZ\374q\311-\003\264\206\330z\030@\306\3216\240\234\263\321\004\206\326d\027\350\307\205[\327!\001;\240A\014\213\034\027\206\344\341\026\223\277\026\332Tb$y,\376G\227\264\323\351\260\371\267i\006\215j\033\201\014\373\246\003d\001\354\205\276S\326LHo>e\251j\357\240p[PSaC\261\036\276\202\273\212\027\364LT\321\262<\250\204j}\367\352\316\272\033H\244t\314\324\220q\246-k\220\213n\271\354~6V\256\302\232I>\363F\331g(\213\211<$\364\334\221\255\34013\350BjER\nZ&\263\320X\332\226""\311/q\217\252\311f\331\202\332\r\210\216\333\031\216\360\016P\204\332\304\356\024\2012`\371\273\021\327\346\371\3544\231\204\006\017yu\335W\253\372;9q\320@\r\213\254\3162\033Z&B2L\237\2371\2336\006~\217fx\326\333\031\004\010\177Y\346\014\221\257@v0`!\266&\322\264\n\201\331\326<_\\\027\027\n\300\225\211\233\243\244/\235\243\261\005[\352+v\n(L\356\334O\233H\007\200o#\361n\327\343\206J\364[\246\350c\326\244\344`\324\357\341\303\323i\257R\255\314C/\201\303a\221\331\222\332\3166\216\312B\031ljd\017T\243#Jm\260c`\330\323\210pn\275\240M\312t\035pX\261\235\334\264t\016\225\016r\257\2218\310\223\276\316\363a\314^\214\261\270\206H\223J\"!\203\223\204\224\234\2349\022\271Q\207\304\347\021\355\014\007Q\322C@\344[$\230;cQ\247fy\225q}\307\327\241+\204\212\236\2641\030!\211Q\026\021x\320\023\250\303\360\2571L\232\330S\001\225b0\343\262\025\210D\202\204\000\222\\@k!\370B+\035'\216\272\365&\233\314\361f\264N.s\2216\331+\241@\326u\273\211A\014\207\031\245\324\211\323@1\246g\236\220D\304$X\023\334\025\320y8\203\320+\210\020\031\221\223i\213C \001\236\331\245\220\322RL\026\214z\014/\356fr\214\201y\345\010\222q\021\r!z5\260\364\224\300\240\273A\342^\215I\241\243J\\M\276\242\333\231\257$\257+\306\030MI\324\230S\0207#\323\314\216\213B\001\255Xl\231L\331\320g\372\311A\314\275\273 \221\250\305E\334\346\375aY\203\345\207_\034\326\263h:N)\243\001=\262\324ff%P\302\272o\323\257\353^\230\313\235Z\227\253\207\311\341a\226\357G\010\034/\244\317sl\2214G\255\004\325\301\340V\226Q\325\2109\350\325\371\345<mR\222\210\210I\224\033\273\250\"\243~~I&\202\252\301`\214\355\300`P(\023\206\324\271\246\026\2240\020\306\2312R\336\343l\035VX\345\020\371\203\251.A\034\266\300Yd\023\223\253\341\322\250#\224j\3464\321m\322\354\224v*p\3179\022\3108\232\013*\001\2413\264TB\216$U\r\323\211t\332\226\2466\214\310\005a\315i\342\304*\211\005\"\002\014\260X\305\311\243\002X\030\364\242\007\005o\345\335\247m\201ik\273\222\014\014\307\"\224U\306\232\0269S\262\275r\266\225-\"\3526\343\025\306\236\026\236\343\247\276\252T""\\G\034\341\255pr\360\203t\305\233;C\313\"m4\020\263\212\tt9\273\236k\277>\266\327X\"\017W{k\342\006g\322\366l9\244f\\\207HJr\306\362\220\034\031Wl\341\322\330\027\266[\340\224\326\320\223\301C\215\370\316\262\220(\317>9\312m\255#\006\333H\031\030\311Y8\324\202f\211Ai\\R\300b\324V@H'\020\217\222e\327\244\223wi\367\001`g~\310\"\322\357\030\363\276-\032\364\016e\246h\342\350GG9\331\332trc{\341=\250\221I\233F\265U\013\240?\246a\247\251\007S\031\215\t\336\022P\005\343%\255\\R\003\031\2356\344\357L*\022vd\2068oJ\001hp\320X\3431T\345X\271\344\241)\000t\301Yca\003m\000b\302\001N\351W\204\225j3\262\t\020\003\204[\027\305\252\241\344!\322=\243\266\266\261\"\323\002\355c\262a%\"\206(\031eclC,\001\323\036U\300m\"\307\3319\273A\024\270*@\2060\327M\322]\334x\031X,\263'<\003D\333&\334\276\217N*\r\230\211\344<\313%4\210\2665@\370\211\217\025`\202\346V\2757$\214b\005?\202\n\022\354y\302\235RW\2334\020\302\261\006\336\204#\241.\377\003\2413\253\315\337\322r\355C\376M\313bFfH\336z\221l\315V7sc\354\n]\366r5\2232\261\305\352\024\331\263kr(\035\t5\231\004\340\316\346\304m6\235\3257\232\257\227\346\311\004\332B\033m\014e\242\326\240R\240\307_Q\257'k\324 \356\247\277\261\177W\201p\033z\261\3504|~[\2752\222\344\353R\254\240\023RY\024\200\327\3509E\312\272w\037\003F6\232\274\247\334s\375\212\252R\034j$Vh\222\251I\264s\222\240\221\234y\327~oF\356\305\211|\242C\032?\016\005\034b\3449RVI\230\353\005,\340}K\325\340\024\206\227\207\247\257\363\373\222\340p?/\2106\256s\220\257\225\373\251\220\017\217\337F7\276\001\030\350\215\030n\014\310\351\033\240\224_\275S\235\333\273\031\255m\345\333\033\025^}U\365\223\017r>\252)\2513\227q\231\206\036uMM\313\346\256\355OtoD\376D\330\021H# \2622O\342((\251\005\000O\334\320\023\2400\350\365\002y(\203\325j4\215\025!\221@\324\251C:%\260\234\344\021'\221>\2367h{\24775\233\005DP6b\003<\277\2024\365\221\242a\210\034w\271\361\013\367\345\362\211/\016\351\313\226z\246\304?\227P\341\024i\240\375\035\277\037\314\323\020!\277\200\217@H\356\241\035\330\354A0/\023\305.""\303\313\241#\367\204&zjgfH<_\325\330\366\261\030\355\346\210q=\025$\241#\317\353\r\033\362\354~\276N\203\347C\232\006\376pYy_3F\275[\375!X\263\337xZ\357\216\365\347\276\307\370\333\3476y\374\207\353\022\322\326l\034\272\2550\324\346K\341a\362\202\r\351\243\244Z\177|\223\224\355\232\243Pa\250ll\014B\327\300\265\253\360a\036*\242j\363\246\252\376\201\254.\270\303\\p\337\333\226a\256n\346\222\375\372Xm%\255\350u\343l\351[B#\240s\372Sk\nt\023\014\364\261(\2614\354,4>\337\315\323t\361\252i,\302\037\363\212\036\217\345\305?\323FM}\034+\350Gy-\304\267\252\237\246\267psK&\245\261\3445\224\330\024\263\247\356\220\2363\373\362\271\332\n\275\325Ft([i\253\327\367\3615\232\347\026\311\260\212\362\007\274\331oV\301\357b(\273,\033\367)\376.Q\343eG'u\232\333\377\267\036\323\220\332;\214\177\221\214\024Q\210\326\361l\245u8\355\374U\005\024]#\223b\256\300\251E\351Nh\r\264\274&\236\235N!xM6#\352\202\276\301\240\325L\374Y6X\2153\365\220\343\211\\<mJ\014\352|\321\317v\335W4\341M\361\233S:\255~\342\322\333)\034\231^\n\244\256\325T\373\305+p2\001\334\rL\277\234$\345\263\225sTh\274\037R6\007\014v]!E\363Vcw\225\253\333\336G\233\323\361\257\360n\376\373\007\253\2639\302\234\31537\264\312\364\234s0\032\240g\214\361\003P4Z\345M\350\203\010\213\007\233\303\315\273\323d{\033\316\253e\022&v0\277J\207\212\024\233]Yc\231\235\366h#\343\203\006\260-\214\343C\321\315\014\306\377\r\014\324cN\327bx\274\252~\025\005\250+\031\267\224\306\024\3758\337\332\213\374\016\250y'\323E;\33710\335:~ln\361\211F\236\230b\022<\351D\207T\363\243!>\377N\375s\025}\335\205\304\304]b\222/U\254?\020V\264\215*\323r\243\371\033\305\232\332-4\261\334\354\233)\214_)8\260\236F|\231\271\271\253>h+~\312e\231J\275\372faK\267\364b\232\374\216\305\244\223\226\217\261p\350\260XJ\005\010\370F\373,p'\027,:\034\326<z\024p\355s\350\3273\350K8\036\260\366\324&~Vf^e\376\302\375r\216(\032\310\251;\035s\251#;\316d\316\\\205*\247:\"I*\001v\233u\022\340=*.2h\013\202\333\336\177AqxK\300\346\010!\332\r\244\233\024\227(vH\373\351]>\2749DH.\"\0217""\"\216\317\226\376\346\335\367w\003^\363\263\317+\t$\350\373f}\306S\270M\274\272W\026|>S\037\327%\363\204+\233'm:\267\027\354S\361\340UR\311\272\360w\005O\366\207\201:\223\263\306y*\204\324\242\236\204\357\244']=\302{N\222\277\202\370\360\223}\245^N\261\251j\341\215z|\227\037\252QeK\024\276\254\370\363\336\370xcc\335\271\272+N9+gk%\302U\225\003\253\220\266E\333\323\321\021\332s\020\220\220\250\243\032\250\227\325\030S\273\240)\341\337B\310\250\204\246\021\325\210\326-\210\375\333\027(\243feb\254\331NI\202\203%\333\025\304w\014)VAM]\3050\033c\276c\217\032D\236E\334\360l{8keU\242:\210\222{\361\310\334\267o\023y\360\326\346p\004E\033\262\323\244\331\326\314\325>)\206M\013\r\351\244\367\030\366\244\352\365\234\366M\350R\034m\203I\324\374\236\t}\357\007\356\207\333\375y|\254\365<\277\307\342\3147\232\030\365N@\335\254(\343T\"[NZ\315\025\254\265\310\351\377\361w$S\205\t\006B*&\300";
    PyObject *data = __Pyx_DecompressString(cstring, 5931, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (5882 bytes) */
const char* const cstring = "x\332\275Y\313c\323\306\326\307\251)!\244\020'\201\360j\253@\n\001B \017Bo\240P\023;\\S\022\347a\010\205\257W\225\245\261#b[\216$'N\241-\313,\265\324\322K/\275\364\322K/g\351e\376\204\374\t\33793\222%?\002\245\267\337\307C\232\307\321\314\231\363\370\235s\306\023w\247\213wg\213\021\311\224\004Y\313\231\222\232#\272`j\232`d\245LFHi\320\333\313\223hte-\276$H\262L\014C \272\256\351}\002\374a\217\260i\352j\262`\022c\236\365\361O\026\350\2444\021\306\rS\2771\317\277p\007\031\221\263\342\256\256\232\204OG\263DO\223\234\274\347\322M\366\035\271\274\221\221v\210\230\327\014a\\\315\231\260\0014US\325r\202\226\022\314M\302\t\232\344l}Q\326\024\342\322s\206p\244\215H'\351V\032\030P\r\223\350M\272\344\224K\240\240\330\222{p\2007w\177i\316\357v\316\033\302\233\251\211i\037\311t7\222\231\211YN\022e\033\357n\222\234 \tY\tw\027`\200\235\t\204cnj\212\240\032B\301 \212 \345\024v`\324\036)\232\302\246d\0109\315\024\222\004\276Vs \024)\243\376F\224I\266\250!(\005]\315\245\205x\362-\221MAQuxi\372\036\220\24648\253\244\374\267z\375G\266\341\272\310\200\340\335\235$\201\365@\275\244(\223<\352\332`\312\006\353\024\226$5\223\324\212\\n\360o]\321\242M\033u8Rs JF\006\262\312f\0139U\226p\225\2173\362\177ah\n1\344V\001\342\210\256\346\233\3548,\353$#\231\240cS\353\316\272\360\377\317\373'\365\377\327\016\0303I\026\270H\031\304d\232Y\217\305\004\360}\242K\031\301\000S\001\272\311e-G\2265p.sS2\205\205=0\373\034\232\275B2j\022(M\222\331\023`\023UF\377\000\242\234\260\022]\271=\373\375,s\n\235\240\351\031\202QH\312\031\3110\0103\227dA\315\230\260!\232\2151)\304R\302\236V\020r\204K9\017t\376\017LtA\340\221I\353\272\224\003\307b\202\027\341s0\357\353\216a\253\340\226\360\365\242\2241\310\344\032\311\353\304 9\023\355\037\254VK\203\2662B\024\226\320\027\302\t\327\237\025\262\243\312.\304-\202\210\210$o\002#\346\256\246o\201H\200&%\311\204\261'\303\3216\321\367%\260t\366\265\2323L)'\177\022\"\215y\341)\212\330\363\236\246\336\r\000\367\002\310\311\361T\300\217\224\232\026\0211&\205\204K\324\334\010""\226\320I\323\337\346\231\200\346\177]PHj\035\351~\235\364\366U4\021\035\\4\325,\321\n\346\274\3404X0Y\217\304\231\373\273\241\004\3070\3208,\001\0279\020(QZ\226cA\342\350\365x\014\371\253\013J\231]i\317\000\0263D2\210\230V3\002,\251\027\230\n\371\244\340L2a=\215=g\337\256\307\243K\302r\207~T\016\267>\244E\000B\024f\010,g4@\351Ia\001y\342\312\233\324\362$7~CH\251\272a\002R\t\317Q51\027\030\205\024\270:Qp\202\243g4g\002p\036=\337u\212u1\210p\331$5\335\374G\301\202\255\330\002\026F\236\310jJ\205\363\203\312\370\374\347\302\0363%f|\222l\026\030\024\230\302\256d\312\233\212\226fj\027\306\023\233 qP\022\262%\343\024\032/iO\035\014\016\024\340\2112\"\350\004\306Jv\nIQ\304&\265\010\216\231IJ\362\026j\243\220c\2603\372\"G\212yf2BJ\227`K\235\230\005=G\224\rP<\367\224\002j\237\237\341#\016\370\227\240rwKv\005\330\266>#\231\027\204\t\370;~c\3621r\016\226F`:\2430\233C;\362Y\342\273\337\025\325\220\222\031B\344\"\363\300x\304'\340\361\033\3360b\201\257\037\205v\016?L\313*F\232\274`\200\035\317\303\202\252\301'\024!k\334\000\260K\245\010\204\245\224\256e\2310u\262] \006J\252MI\234\234\273\024\200\241\226)0K\312\250Y\3254\\\203\332\224te\027Pe4\247\201\252RR!c\n\"p\244\024d\"\212\356\3479-w\033d\273\003\316\005\2632\372\231(\346\367\014\215d'e-\223\341\301\302h\216\260,\010\002\2043\000\360\233\006\241:=\376r:\306\236\373\0252\3554\201\241\034\320\033\336\3218\016r\270v\275]\332\001GC\301\030$\223\232\024\211,2\252\t\267[\024\035FLw\250(j\n\212\335\351\303\371E\036\010\014\021\214\034c\205\2010\31737D\255\035\242;\241_\022Vx\354\323\270\273#\302\345Uy+\203\037\265ox\304~\330',\004\374\375M\376\253%t\371\016\227\360\235\246*\212\0171x\303\234V\320e\362\250\340\371\035s\225\366\376\004\204P\210=h\226\273<\270\210Y\346\346\314\2528\013\357~\007\313\013+R\036\004\373$\036O\210k\257\304\245'\257Xs=\021NDY+\301\006\305\005\307\371\261\370i\351L6\255Pf\007\022\305\366iL[ \023\360\0100\020.h\204\003rK\207\241w\347H\347\036]i>\276\221C\225&\246\230\004\267\310\220\\\332\334<\212\002+\r""\314[\310Q\0049\000\273\243\346@\245\"\017\260-\024\2378\305'\017\360Q\336?\3066\237\203$KW\211\321eF\315)\244\330e\274\343\214\2378\2363\tm\026\353\332fY\232\343\303\330n\323>\036y\370\367Z\335\245\327\234\353*:w\326\201>\321\251\013D\346\"\206\237\202\tOf\216\322>\354:\226\010\321\247}.+\345 \362\212\316\316\220T\247;>gX\3271\210\037\370\007\215O/ft\371N\336$\362\226\213f\242j\210\276\314\252\203\314\227\304\371\3470\347\362\367yj\253\310\235c\270x\347(\204\300\316A\r\220.#\345\333&1\006\373\272L\345\355G\202$\204\250\230\\\351\032\232\030Z\264o\026B\254\322\\\374#$GL9\242\005i\370Fy\302\334\344c\335\2110\376vw\343\363\315v5\277uW\365\254q\304\032\316\324\321\013L\212N\246\002z\316\3463\004h<\357s(Ry]i\351\356\352\276.Z\334J|z=\316\225s\364L\264\3306\347\230c\301h\033'$/\252J\227A\260\341.\243:\331i\033Us\371\202\331>f\210`\213\355\243.\010\371\206\264\366\235\241\276\351\\\316S\2507\346f^ \322\035U!\272\230\363\237\3148RN\306G\344dte\335\350\306\226\321\311V\367\024\3337\237Es\316&};\002N\031{9\331\033\000\021C\222\311\020\264c\220\031\2677\232\322H\033\035\216\264\021qd+z\322ra\321\233v\247\274Q`Qt|\327\033\204\2016Q\342\310\016\361Y\247[\001\267\216\264qdt\335\262\315\235\371`\2337'x\371k\370\333\335=\3217\333\351\214\354\024KR\236\3355\265\366\200\236\347\331\321\205\204\370$\226\230r\336\323\316{\306y\317:\357{\316{\316y\337w\336\337\263w<\376<\032^\306\346B<\032\211&\304\225H<\274\276\036{\3326\266\020_^\214=\365\215Am\337\332[\010\267\366c\313\213~\212\027+\317\343\341\010\016D\342K\341\030[>\266\234\210>\215\256M\315\371:\323\263\276\316\314\264\2573\347\237a\334\307\027\022\270Wb-\266\314X[\213\206\237\363O\260\305\351\327\242O\305\365\245\273^s\312mnD\304H\354\245\257\227\210-E\341\254\261\216\241\265\370Bt}=\022N\260#\262\261Hlq1\272\026]^\2106\207\342\213b$\3743v_,\307\026\342\221\250\217\263\027\313L\244\021~T\267\307\317\352\3668\347no\256e\216\035\367el=\366\344ysY\346p\314$|M\317:\\'o6:\247\320\364L\275\313\300\344\303\214\006\330`<\232\304{\277b^""\217Eb\3131'M\217\345 \371,d\341\265\230\221\322K\341e\314\327AF\211\270\377\246\327\337n\356\314#\340r|9\312\227rnlbn\231\274\254\231q\210\331\354\243\370\n\247YY\2136\233\000U\304d\263\276fs\365\2261\226aA\366\007\303{E\370\037Q!![\206\264e\215\244\326\242/\327\303\213\315e\331C\014/\374\304\033\321\265\265\370\332z\"\342\324'\330\342E\211{q\355\276\233\033\303\000^\363\270\223\315\266G\240\252qv\237\272\261%\263\031\367\335\244\360\256l$\331\344\027,\242WII\274j2\272\2437\214\352xs\341\006\317f\203\255'\232\232\310k\271\316\313\265&\332Ky\020\273\"!\316\253\032$U: \223\232#Fr*9\265;\265;-\262_A\2346\254\346U\005\311B\n\3474\035\302\233,\271,\211\262\002\177?\222-v\244\210\242\310\256-Qo\000\250Y\354\002\003\360\211h\352`\031\270(K\035\331\221\330\003\265+C4\341%\200\3272\216\310\277}\027\022\315\244\323i\340\375\222/\351\3642\315\316\364\322\273\304pZ\220\202Hi\242c\352\327,\207\234\220\351d\370\nI\026\322\334\026\025\264B\347uD\232\005\263\032P\342\275\205\243vh\341\312\270\260/\364\372\003.\230\003H\304\251f\360\265G\300C\341\330\336\317\006\336\2176\315\237\322\010\336z\313\222S3\220\242,\342\017@\315\372\203\205$\1775\222RA\345`2)(\340\\\336\0147\3027\343zJ\323\263\260&\246\210\370\237%\207\356\365\260\0106\211\375\2661\274\263\342\026&2U\210\242\003;\300\"\374{\256\245A\300\355\311cg2\331\232B\266U\257\255\365\252\277\000\363\240\302K5\375\t\246?\255\364W\266\035\325\232W\334z9\247?\323<\252\204\353\232\3774\223\321\266\032\270\275\352u2T_^\352\225\201^^\332\222\214\036\221\231\242\324\235<DTUEM\341\356\374\211\371\0138\006;\232\003X\374t*:\233\003\026\316)\321{\361\277\261E\366\014.\373\214j\232\031\222ajt\356\374\\\327D\037\315\3413\327E0\374\022\316\227\351\211Y\222M\"\263\300\023\360\353t\rl\260\233[l\000z2 \021\263\232R\310\020\357\r-]k*\217\313\326\367T\314B\036\311rd\027\037E\023<\210\033\245'|\237\3345\005\253K89\027:\3768\005\241L\311\003\036\346\0255\257\345\363\032l\222\327I^\322\211\350/\020\241\231':X\241\377\332\023\257M\213<\2674\nN\017@\275""\2201y\333\321\0136\231\375\262V!\307\256\360HKu\3306\325z\223\306\346vL\206;\360g\033$\316%\320q]\322rA\342U\316]\312e7\351vRm\357\226\270\363\246\316\035@3\362a\077\077\250\016f\271#Ag\307\315\321\233\2319^rb\235}T9\336^\203\267\027P\235\005\225\277\214:\3526\204%\346\216\205\370\n,\257\254\362\227\t\320v\356\311}\371\274\330%\267wo\224\341]\3103\2703\324t\216(x\255\217\377\301\263\260,`\277\220\271?\373\360\036\254\0039\202\023\3559\003^Eb\230\212\010\266\227B\005\260\034\316($\231\303\2029a,\272\013\014\310\031\376C\035\357\033\233j\312l\366\247Z\346\301\212puxn\352\332.\2169h\355\201\266\231\315Cf\301\034\244\220\007\241\223\002(\323=-\264\231\221\025\314\324\367\240\323\002a\017\003\323\007\274\304w/\362!\303\001\374\001\354q\032\020\224\274\244\307k\211\354\026\331\355;D\340\005\316\0133\034\030\304\350\014(\274\273\2059\245\342\312\311W\237\025?\004\032\301[\345\251\362be\364C\340\360\364\261\343\263t&R\0137\202\247\367\267\354K\245p\351M%P9_\r5\202g\366\267\033\301\276\375\261\375\214}\325^m\364\016X'\254m\373\204\275]:U\276W9Q\331\256\236\254]\252\207\033\275\347\254\255\022\320\237\332\237\242\301\331J\370\360\374\261\223_\355\377\233\016OUB\215\336\253\245\307\225\321\312,\275\377\264f\326\357\323\265_\351\257R\243\367\014=3^\016\224G`\257\263\025\ty\272Y\036-\317UB48W\221\016\277<vr\200\236\371\326~\017\214\302\006\337\330ji\273\034,/W\257\327\206k\257\353\357\351\033\205*o\351\333<\315o7zO\355\317\356\033\326h\243\377\234%Y\333\215\376Ak\334\036m\364\236l\364\017\321\241\211r\270\274QY\254\216\036\360i\276A?\260H\373o\224C\370\365\024\020Z\263t\344\373j\000\016C\317\315T\246*\341J\242\312\2319\350\355o\036\346\234\245\225VK\n\2759_\035\253\246i\364\005}\361\222/x\321[\220\023I%\223\336zP\235\241\217\226\353\022\r\302\332\007G\234C\245j\206f\2624k\036\036;V\010D{\340\025\355\371\t_?\365<\357\361\235\217\035\214\006\377U=G\037\343\326\207\023\307N\216X\212=N\257\314V\226j!N\372\247-\203:\372\007\351\340\325\322\224\307\375a\337\261\223\375\373\367\254/,v""^x\\\260C\007(\242\253\245\207e\035\365P\331\346\203\260\214C\367\025\252\242\037%t\326J\320\013\250\244\373\225\233\325H-P\273\\_=\350\277l?\0075\016\323\333?T\023\265\221\372\211\372o\364\347\327\215\311;\345\367\325\271\3320\215\276\241o\370\331r\270\3209\353\227\322hi\226^\277\017\306\023@.f\254\264\275^\n\226\302\007\375\003\260\343}\373&\035{X\215\326\206j\341Z\202\306\340\363_\350/\377i\014\014[\313\360\345\024\320\r\014Y\367\354/\354\251\306\3607\266do\303\024\035\376\216\311{\373\200\035\177\327J\332\260\366\231\375\202\265\000\326\340\220q\331\331\2636Pu\030\206K;p\021H\300\210\257\322\253\377\252\206\016\207\231t\351E\244\004\373\357w\364\333\277\377\314\222\232r9\207\024\223e\251\274}\320\3775\375\372vy\025\332&\235Y\250M!?\337[\323\326j\243\357\364>\001\266\003\36608\322\351A\353\272=R\n\224\206\231\257\301\212\340i\245\360\2070j\010\214\010\264\3268\003j\340\354~\010s\255\352\026\352\324\272i\207\321\027\271I\214Y\222o\354\202}\001d4\203\222p}\022\255\371:39\346\246Y\220`\244\334\007\3768\303\224\215\323w\312\333\225`%\314\230\370\020\346\"4Q\365\203\3265`\0306\275l\377\304\226\210Wg\253Ep\342\037\350\253\237\351\317L/ 9\023\224v\305\0167\006\316Y\031TQ\243\177\200\016\200B\\\276\371Zt\220\211\310\277A\337\200\365%\363\327\276\213\364\342]\020\304pE\254E\352\275t%A\023/\032}_\355'\254\021\273\007<\371\364\260\365\034\304\025\0021\322\323\327J\022H\252\027\245\217\353\314\303\311\321\232\017\200\377\242e\332\363\270\363\341\300\307\247\233\336\217Z\274H/\314\2414\252A\372\360y=\\O\320\325\327\364\365\033t\234\230\225@\001\204\254\263(\3503L\007p\316\244}\262\024*]-\255\225\366\200\355\241\312\223\212Z\335\256\001v\234\336OYaz\366Z)\215Vp\320\313|\206\001\307\251\375\231}i\377\275=\215\374\014Z\223x\234\003.r\300\340\223g\255\010=\317\010/\331\367KWJ\317\312\311J\300QI\243w\330\212\330}\245Q\264\274\210\325g\217\263\255\315\362l\371\267\352H\255\247\006\300p\336\016\37005XyVUk\177\373\323\267\365\000~\212_\364\342$\332\373\246\235.%8Z\316YCV\030\264\014""\313\335)\207\016\334\006\034d\320\303}\020\354\300e{\212\033\3467\266\006\362P\350\364\223\332X-]_\255\023\272\236@\321\017Y\267!\232\004\351\030D\215\312j\343/C\200\337\345AF4\030\256\205\016/\373\243\216\003\231\326\234=\004\266\311%~\000\372\221\000\036$\253@/O`\000l\364\216\322Q\240\257\214\321\271\305\232R\277NWE*\002$\223\206\037J\006\255[\366\252\215fw\316\333\343\240\367\"\006\031\364\360\031KB1\2405\377^\212\226C\345\353\000\246oQ\274\265).(\210\237v\262\364Uy\325\337\355/;b\274Yr\0327\000\257\373]\267\030A\237\030\251\006\252\303\325\3275\243>Z\237\255\357\322\227 \016BI\232\2467\351\246\332\0309\337\370\352\254\265n\007\376\231\375or\034\376\234\375\017\270l\271\031{\362\007\254\274l?Em9\260\326=\204\242L\317\373e\372-@5\303gD\215K\266\003\241\367\021\203\243\225\220\243H\364\222Y\372\355\275\n\003\302\007\256\222`\206\216\206!\n\005\352\260\322p\323\245\316Y\262=\210\236_\332\252\014VXD\\\264F\351\0203\241\306\020\017\312_\224\347+\257\253\2735\030\030>\350\305\260\326S\276\t\236\r^|\212\236\372\032\202\024\3541lE\355\020\2758_\035\245\017\336\323\367\277C\204\376#\360\033\006\352\337z\336\365p\247\201\330`\351\210\304\210u\003V\320z\006\361G/].\257\225\215\312\225J\244\332\007j\231\256\255\272\324*\207xl\276erd6\366\214\347\020|\346'\313!Y\262Y\370\211Y\030\215\366\237s\347\\\266\231{\377\333\n\267\020\301p{\377\0102_\017\344\322\372\342\264l-\256\250ez\361\016\307\361\263\326*>~aN\031\241\343\220\341@\326\010(i@x\273\010\201\000\242\351u\300\311q\310\031f\312R\343\326\244\367\240\223a\032F\230\247/^\321W2\225!\227\333\242[\231Fh\330z\212\252r?\236\200\250a\200\243\336\231\206\307\\\254\036\252_\255\277\241\257\300KY\036\331N~\013pF\006\235O\24219c7x\016\351`)\030&$\004\316\241x0z\307\310\306J\233\354S\026-0UD\2278n\275`Q\370\035\270\024\240D%]]\255J\215\201\363\364<fBp\304\263\326\177XL\r4\006F\254]\236i\300\002(\237\226\235\032\335w\232\265\336\225nbvq\300\277\362\370\303\300\372\300&|\361\316\t\236G\206;'\266\300\375\307\312R\227 \307\002p%Hg#\010""\254\365\013tm\375\2231n\254\224,\237d'OB\262\037\202x\361\2311\256)w\016\275\n@\314\263r\272\262Q]\300\210\311'`=4\350\321\322\267\230\004\320{Q\340n\204\256\254\322U\314q\333\220x\002e|\320\213\372\001\346\243\3105\204\365\007\340\310\220\271~\315r\237~\310j\256\225\2242+\010rP\321`\340\352w\324\343\006\335~\177\276\216y\021#k\237\300\257Wq)\236St\315\360\247\370\234\360\327\220\253\025\226.\261m\317\225\267\351\335'\000\013\363\365u8wc\034\215}\034\334\254\001\322\375\237\322`i\036\300c\273)J\206\221\210\372\247\230u|Y\226\350\235\307U\003\2243\206\3365\006\325\035\356\263`\031\366-\210\366\336\207#\200\352\237k\226G\372\300\010\023\373[\226\300\272\306\213&\370\250\024\342\371\331\000fy\350\010gx\036\306\003\301\200\333it\035\207\260\024\340\300y\202ek/\241\334\030\2554\027\177\000*4K\276\256\317A\370@\312Ud\027O\341\003o\261`)\257r\201\0240C\261\225\3225\200\37318\376\205\352h\365~\355*\324fP\310@\254\003\240I\322$B\023\315\030\324\330\001\250\337\r\204\021\361\303=Oz\270\032\216\333\033%\266z\021\312\343\240\275\014\3100[~\007%^\010\322\235\267t\345\025\313\227_s\333\352\337\177\210\206\310\355\204q\004\365\306\023\372\004)\234\202\243Y\316\035|\264\234\003\257\276\200\205\220\315rj\267B\303\322\376\031\004\377/k\251z\030\353\263%\004\004:\371\250\252\324\256\327\317\323U\316\r\034\nN\244\3039\214\300\002\036g\241'\206\257X\317\263\036T^{f\346\253\324X\202?\n\0063Z\232\342\336\024\300k\202[\326*\r\216X\177\224V\235\220\205\257\247x\272G\364\321\032\200\r\236\316\013\364\020T/\333\013`\311C\337\272\033E\252'\252\005\032yI_n\034\014\tX\037\321\357\356A\031>R;N\243\033t\003\352\344\2674\370\200>pW\373\254\322\327\271O\230.\277@\203B\315\255v\326\266\230\003\274\257\317\321\325\265\216\322\026\352\022\340\001\244'Q\211\333C\221\026\377\240\177\374\331M\\\316E\304\021R\303}\373\272^N|\010\300j\307OC\345s\311\216@\276\226\200\004,X\r7\202L\224\326\242=\001@\027<\361ag\177\003R\021,&{\0071V\203g\037\364\2160\217\331\000\270\031\002Y\366\327\236\326\357\321\225u\272""\016lC\006\335\274h\310\323\374.\335E\316q\245m\226\374O\226\025\220\312\375\352\255\332F\035=\326?\262\332d\212\2068\254~\0267ao\227\333p\236P\345z\365\022\2606u\320:\022\246\301\037\351\217\250\373C\241]\263\375M\2539\337\335\340\237V\346AG\347j\333\035\006\017q\004\213\271\306\235{\030W^\326gi\034\322W8?(o\257\333\225\304\337)AZ\002\224s\313\360!@\203\200\263\255Y1\004\246q\277M\001\r\224\263\207\327\320\014\037@8\000\323\375\332\376\031\027-\007\351\304\303\352\"\234 T\037e\027\013\230\215;\205\212\004E\370c\372\0322 \236\207o\323\355w\340\273\357]O\216\340+\322\023\355i\226'\310\330\235R\210^y\214).\310\t\303\000\270\010\rB\3048\274\322\375\236\303\273\030\273[\341\027c(\3530\252\354C\340 \010~K\205\273\225S\325\007\265-p\230Fp\310\372\241\324\013\307\013\241\306\337\203\024\301\024\356`h\373\261v\272.\261\373\301#\276{h\027\341\310\001\337w\230\254\374Y\315\325\303\316\265\"|\324!\312\033~E\207h\360&\275\371\270\272\2157q\037?\r\273\336c+\370S\"\260\362\341c\307\207\350 .\373]y\027\257\004\376\027(\311t\030";
    PyObject *data = __Pyx_DecompressString(cstring, 5882, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (14461 bytes) */
const char* const bytes = ",02x04xData container too small for typeEEPROM access error\n    \n    Attributes:\n        message (str): error message\n    EEPROM write errorEmergency message.\n\n    Attributes:\n        slave_pos (int): position of the slave\n        error_code (int): error code\n        error_reg (int): error register\n        b1 (int): data byte [0]\n        w1 (int): data bytes [1,2]\n        w2 (int): data bytes [3,4]\n    Error when a master or slave method is used and the context has not been initialized.Errors during Object directory info read\n    \n    Attributes:\n        message (str): error message\n    Errors during Object directory info read\n    \n    Attributes:\n        error_list (str): a list of exceptions of type MailboxError or SdoError\n    Errors in mailbox communication\n    \n    Attributes:\n        slave_pos (int): position of the slave\n        error_code (int): error code\n        desc (str): error description\n    Errors related to mailbox communication \n    \n    Attributes:\n        slave_pos (int): position of the slave\n        error_code (int): error code\n        message (str): error message\n        desc (str): error description\n    Item offsets in SII general section.NoneNote that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.Representing a logical EtherCAT master device.\n\n    For each network interface you can have a Master instance.\n\n    Attributes:\n        slaves: Gets a list of the slaves found during config_init. The slave instances are of type :class:`CdefSlave`.\n        sdo_read_timeout: timeout for SDO read access for all slaves connected\n        sdo_write_timeout: timeout for SDO write access for all slaves connected\n        always_release_gil : true to always release the GIL\n    SOEM Network interface is not initialized or has been closed. Call Master.open() firstSdo List Info read failedSdo ObjectE""ntry Info read failedSdo Object Info read failedSdo read or write abort\n    \n    Attributes:\n        slave_pos (int): position of the slave\n        abort_code (int): specified sdo abort code\n        desc (str): error description\n    Slave The actual set watchdog time (This way of catching emergency messages is deprecated, use the add_emergency_callback() function!Unexpected frame returnedWorking counter error.\n\n    Attributes:\n        message (str): error message\n        wkc (int): Working counter\n    :  , , ().?add_notecould not open interface {}disableecx_readODdescription()ecx_readODlist()ecx_readOE()enablegcio map size: {}isenabled ms) differs from the requested watchdog time ( ms) due to resolution limits of the hardware!no default __reduce__ due to non-trivial __cinit__pysoem.collectionspysoem.contextlibpysoem.loggingpysoem.pysoempysoem.syspysoem.timepysoem.warningsrequested slave device is not availableself._ec_slave,self._ecx_contextt,self._ex_odlist,self._the_masters_settings cannot be converted to a Python object for picklingself._ecx_context,self._ex_odlist,self._ex_oelist cannot be converted to a Python object for picklingself._ex_oelist cannot be converted to a Python object for picklingsrc/pysoem/pysoem.pyx<stringsource>unexpected errorunexpected error, Etype: {}wd_time_ms is limited to {} msAdapterBOOT_RX_MBXBOOT_STATEBOOT_TX_MBX_CallbackData_CallbackData.__reduce_cython___CallbackData.__setstate_cython__CdefCoeObjectCdefCoeObjectEntryCdefCoeObjectEntry.__reduce_cython__CdefCoeObjectEntry.__setstate_cython__CdefCoeObjectEntry._get_bit_lengthCdefCoeObjectEntry._get_data_typeCdefCoeObjectEntry._get_nameCdefCoeObjectEntry._get_obj_accessCdefCoeObject.__reduce_cython__CdefCoeObject.__setstate_cython__CdefCoeObject._get_bit_lengthCdefCoeObject._get_data_typeCdefCoeObject._get_entriesCdefCoeObject._get_indexCdefCoeObject._get_nameCdefCoeObject._get_obj_accessCdefCoeObject._get_object_codeCdefCoeObject._read_descriptionCdefCoeObject._read_entriesC""defMasterCdefMaster.__reduce_cython__CdefMaster.__setstate_cython__CdefMaster._collect_mailbox_errorsCdefMaster._get_dc_timeCdefMaster._get_expected_wkcCdefMaster._get_manual_state_changeCdefMaster._get_slaveCdefMaster._get_stateCdefMaster._set_manual_state_changeCdefMaster._set_stateCdefMaster.check_context_is_initializedCdefMaster.check_release_gilCdefMaster.closeCdefMaster.config_dcCdefMaster.config_initCdefMaster.config_mapCdefMaster.config_overlap_mapCdefMaster.openCdefMaster.read_stateCdefMaster.receive_processdataCdefMaster.send_overlap_processdataCdefMaster.send_processdataCdefMaster.state_checkCdefMaster.write_stateCdefSettingsCdefSettings.__reduce_cython__CdefSettings.__setstate_cython__CdefSlaveCdefSlave.__reduce_cython__CdefSlave.__setstate_cython__CdefSlave._disable_complete_accessCdefSlave._fprdCdefSlave._fpwrCdefSlave._get_PO2SOconfigCdefSlave._get_PO2SOconfigExCdefSlave._get_al_statusCdefSlave._get_eep_idCdefSlave._get_eep_manCdefSlave._get_eep_revCdefSlave._get_inputCdefSlave._get_is_lostCdefSlave._get_nameCdefSlave._get_odCdefSlave._get_outputCdefSlave._get_stateCdefSlave._get_watchdog_divider_nsCdefSlave._set_PO2SOconfigCdefSlave._set_PO2SOconfigExCdefSlave._set_is_lostCdefSlave._set_outputCdefSlave._set_stateCdefSlave.add_emergency_callbackCdefSlave.amend_mbxCdefSlave.dc_syncCdefSlave.eeprom_readCdefSlave.eeprom_writeCdefSlave.foe_readCdefSlave.foe_writeCdefSlave.get_max_watchdog_timeCdefSlave.get_watchdogCdefSlave.mbx_receiveCdefSlave.reconfigCdefSlave.recoverCdefSlave.sdo_readCdefSlave.sdo_writeCdefSlave.set_watchdogCdefSlave.state_checkCdefSlave.write_stateCdefTimeoutsCdefTimeouts.__reduce_cython__CdefTimeouts.__setstate_cython__ConfigMapErrorConfigMapError.__init__ECT_BIT1ECT_BIT2ECT_BIT3ECT_BIT4ECT_BIT5ECT_BIT6ECT_BIT7ECT_BIT8ECT_BOOLEANECT_COEDET_PDOASSIGNECT_COEDET_PDOCONFIGECT_COEDET_SDOECT_COEDET_SDOCAECT_COEDET_SDOINFOECT_COEDET_UPLOADECT_DOMAINECT_INTEGER16ECT_INTEGER24ECT_INTEGER32ECT_INTEGER64ECT_INTEGER8ECT_OCTET_STRINGECT_REAL32ECT""_REAL64ECT_REG_SM0ECT_REG_SM1ECT_REG_WD_DIVECT_REG_WD_TIME_PDIECT_REG_WD_TIME_PROCESSDATAECT_TIME_DIFFERENCEECT_TIME_OF_DAYECT_UNICODE_STRINGECT_UNSIGNED16ECT_UNSIGNED24ECT_UNSIGNED32ECT_UNSIGNED64ECT_UNSIGNED8ECT_VISIBLE_STRINGEepromErrorEepromError.__init__EmergencyEmergency.__init__Emergency.__str__Emergency.__str__.<locals>.genexprIDINIT_STATEIntEnumIntFlagMANMBX_PROTOMailboxErrorMailboxError.__init__MasterNONE_STATENetworkInterfaceNotOpenErrorOP_STATEPREOP_STATEPacketErrorPacketError.__init__PacketError._get_desc__Pyx_PyDict_NextRefREVSAFEOP_STATESTATE_ACKSTATE_ERRORSTD_RX_MBXSTD_TX_MBXSdoErrorSdoError.__init__SdoInfoErrorSdoInfoError.__init__SiiOffsetWkcErrorWkcError.__init__abort_codeactactual_wd_time_msadaptersadd_emergency_callbackaddressal_statusal_status_code_to_stringalways_release_gilamend_mbxappendasyncio.coroutinesb1b1w1w2_bytesb1w1w2_strbit_lengthbufbyteordercacallback_cdcdcheck_context_is_initializedcheck_release_gil__class_getitem__cline_in_tracebackclosecode_code_desccoe_objectcoe_objects_collect_mailbox_errorscollectionsconfig_dcconfig_funcconfig_initconfig_mapconfig_overlap_mapcontextlibcontextmanagerdatadata_typedc_syncdc_timedebugdesc__dict___dict_disable_complete_access__doc___ec_adapterec_datatypeeeprom_readeeprom_writeencodeentriesentryenumerrerror_codeerror_listerror_regethercat_slaveexc_infoexpected_stateexpected_wkcfilenamefind_adaptersfoe_readfoe_writeformat_fprdfprd_fpwr_timeout_us_fpwrfpwr_timeout_usfrom_bytes__func__genexprgetgetLogger_get_PO2SOconfig_get_PO2SOconfigEx_get_al_status_get_bit_length_get_data_type_get_dc_time_get_desc_get_eep_id_get_eep_man_get_eep_rev_get_entries_get_expected_wkc_get_index_get_input_get_is_lost_get_manual_state_changeget_max_watchdog_time_get_name_get_obj_access_get_object_code_get_od_get_output_get_slave_get_stateget_watchdog_get_watchdog_divider_ns__getstate__iidifnameifname_redinindex__init__input_is_coroutineis_lostitemitemskeyslengthlittleloggerloggingmailbox__main__manmanual_state_changemastermbx""_receive_member_names___members__message__metaclass__module__module____mro_entries__name__name__namedtuple__new__nextnum_bytesobj_accessobject_codeodopenoutoutputpasswordpbufpdipoppos__prepare__processdatapropertypysoem.pysoem__pyx_checksum__pyx_result__pyx_state__pyx_type__pyx_unpickle_CdefSettings__pyx_unpickle__CallbackData__pyx_vtable____qualname___read_description_read_entriesread_statereceive_processdatareconfigrecover__reduce____reduce_cython____reduce_ex__release_gilresultret_valrevsdo_readsdo_writeselfsendsend_overlap_processdatasend_processdata_set_PO2SOconfig_set_PO2SOconfigEx_set_is_lost_set_manual_state_change__set_name___set_output_set_stateset_watchdogsetdefault__setstate____setstate_cython__settingssetup_funcsignedsizesize_inoutslaveslave_posslavesstart_addressstatestate_checkstd_buffer__str__subindexsumsync0_cycle_timesync0_shift_timesync1_cycle_timesys__test__throwtimetimeouttimeout_ustmpto_bytesupdateuse_setstateusetableutf8valuevaluesw1w2warnwarningswd_div_nswd_div_regwd_time_mswd_time_ms_limitwd_time_regwd_typewd_type_to_reg_mapwkcwktword_addresswrite_statex\200\001\330\004+\2501\250F\260!\200\001\360\016\000\0054\3203D\300A\330\004\016\210k\230\033\240A\240[\260\001\260\030\270\021\330\004\017\210q\330\004\n\210$\210l\230#\230Q\330\010\020\220\007\220q\230\007\230q\240\013\2505\260\007\260q\270\t\300\033\310A\330\010\026\220k\240\021\330\004\013\2101\320\0044\260A\360\030\000\t\r\210H\320\0241\260\021\330\010#\240?\260!\2604\3207G\300t\3107\320R`\320`a\330\010\017\320\017(\250\001\250\027\260\001\260\025\260a\200\001\330\004*\250!\2506\260\021\320\0046\260a\360\006\000\t\020\320\017\037\230|\2501\250A\330\010\036\230i\240q\250\004\250N\270'\300\024\300Z\310|\320[d\320dj\320jp\320pq\330\010\013\2104\210s\220!\330\014\026\220a\220q\330\014\022\220(\230!\330\010\t\330\014\023\320\023,\250A\250W\260F\270!\340\014\026\220a\220q\320\0046\260a\360\014\000\t\r\320\014)\250\021\330\010\013\2101\330\014\023\2204\320\0278\270\001\330\010\026\320\0263\2601""\260A\260T\270\021\320\0046\260a\340\010\014\210H\320\0241\260\021\330\010\026\220o\240Q\240d\320*:\270$\270g\320EU\320UV\320\0046\260a\360\032\000\t\r\320\014)\250\021\330\010\026\220o\240Q\240a\240t\320+;\2703\320>N\310a\320\0048\270\001\340\010\036\230i\240q\250\004\250N\270'\300\024\300Z\310|\320[d\320di\320il\320lm\320mt\360\000\000u\001E\002\360\000\000E\002K\002\360\000\000K\002L\002\330\010\013\2104\210s\220!\330\014\022\220(\230!\320\0049\270\026\320?U\320UV\360,\000\t\027\220d\230(\320\"4\260M\300\021\330\010\013\2104\210\177\230c\240\021\330\014\022\320\022#\2401\340\010\014\210H\320\0241\260\021\360\n\000\t\014\2105\220\003\2201\330\014\023\2201\330\014\031\230\021\340\014\023\320\023#\240<\250r\260\025\260a\260q\330\014\031\230\021\340\010\013\2105\220\003\2201\330\014\r\360\006\000\t\014\2101\330\014\025\220T\320\031*\250!\2507\260*\270D\300\001\300\034\310Q\340\014\034\230L\250\001\250\024\320-=\270T\300\027\310\007\310z\320YZ\330./\250|\2706\300\024\320E[\320[l\320lm\320mn\360\006\000\t\026\220]\240!\2404\320'7\260q\270\001\330\014\023\2203\220g\230S\240\004\240A\340\014\020\220\003\2207\230*\320$<\270E\300\023\300A\300T\320I[\320[]\320]^\330\020\024\220N\240!\2401\240A\340\020\023\2205\230\003\2301\330\024\036\230a\230q\330\020\024\320\024%\240Q\240a\240q\340\010\013\2104\210w\220b\230\001\330\014\017\210u\220C\220q\330\024\036\230a\230q\330\014\022\220(\230!\2304\230q\340\010\t\330\014\023\320\023,\250A\250W\260F\270!\340\014\017\210u\220C\220q\330\020\032\230!\2301\320\004#\320#9\270\021\360\024\000\t\027\220d\320\032,\250A\250Q\330\010\014\320\014)\250\021\330\010\014\210J\220a\360\006\000\t\014\2101\330\014\026\220d\320\032.\250a\250q\340\014\035\320\035-\250Q\250a\250t\3203C\3001\340\010\013\2108\2202\220Q\330\n\016\210e\2205\230\001\230\024\230Q\330\016\022\220'\230\027\240\001\240\024\240[\260\001\260\021\330\010\017\210q\240A\200A\360\n\000\t\r\320\014#\2401\330\017\031\230\021\230!\2304\230q\200A\330\010\013\2104\210r\220\021\330\014\022""\220*\230A\230Q\330\010\013\2104\210s\220$\220a\330\014\022\220*\230A\230Q\330\010\031\230\031\240!\2403\240a\240q\330\010\026\220k\240\021\330\010\026\320\026'\240q\250\004\250A\330\010\026\220m\2401\240D\250\n\260!\2603\260a\260q\330\010\026\320\026/\250q\260\004\260A\330\010\017\210q\200A\340\010\013\2104\210t\2201\330\014\022\220&\230\001\230\021\330\014\034\230K\240q\250\004\250O\2704\270x\300t\310=\320XY\320Y]\320]^\330\014\017\210t\2207\230\"\230A\330\020\026\220l\240!\2401\330\014\020\320\020%\240Q\200A\330\010\013\2104\210t\2201\330\014\022\320\022.\250a\250q\200A\340\010\013\2104\210t\2201\330\n\020\220\006\220a\220q\330\n\032\320\0320\260\001\260\024\260_\300D\310\010\320PT\320TU\330\n\r\210T\220\027\230\002\230!\330\016\024\220L\240\001\240\021\330\n\016\320\016&\240a\200A\360\010\000\t\014\2104\210t\220:\230Q\330\014\023\2201\340\010\017\210x\220t\230:\240Q\200A\360\020\000\t\014\2104\210t\220:\230Q\330\014\023\2201\340\010\017\210x\220t\230:\240Q\200A\360\014\000\t\r\320\014)\250\021\360\006\000\t\032\320\0316\260a\260q\270\004\320<L\310A\310T\320QZ\320Z[\340\010\014\210I\220T\230\021\330\014\021\220\025\220a\330\014\017\210r\220\021\330\020\026\220b\230\t\240\021\240#\240R\240y\260\001\260\023\260B\260i\270q\300\001\330\010\016\210f\220A\320\025&\240g\250Q\250a\340\010\025\220T\320\0311\260\021\330\010\013\2103\210a\210|\2302\230Q\330\014\022\220.\240\001\240\021\340\010\017\210q\200A\360\016\000\t\025\220D\320\0301\260\021\330\010\033\2307\240\"\240J\250b\260\001\330\010\017\210q\200A\330\010\024\220D\230\n\240!\330\010\014\210D\220\n\230(\240#\240R\240t\2504\250z\270\027\300\002\300!\330\014\030\230\001\330\010\017\320\017(\250\001\250\027\260\004\260J\270i\300q\200A\330\010\024\220D\230\n\240!\330\010\014\210D\220\n\230(\240#\240R\240t\2504\250z\270\027\300\002\300!\330\014\030\230\001\330\010\017\320\017(\250\001\250\027\260\004\260J\270j\310\001\200A\330\010\014\210D\220\010\230\001\330\010\014\210J\220h\230g\240T\250\021\330\010\013\2106\220""\023\220A\330\014\020\220\n\230/\250\021\340\014\020\220\n\230/\250\021\200A\360\022\000\t\r\210H\320\0241\260\021\360\006\000\020\034\2301\230A\230Q\330\010\036\230o\250Q\250d\3202B\300$\300g\310Q\310e\320ST\360\006\000\t\023\220-\230q\240\004\320$4\260A\260Q\330\014\020\220\003\2207\230*\320$<\270E\300\023\300A\300T\320I[\320[]\320]^\330\020\024\220N\240!\2401\240A\340\020\024\320\024%\240Q\240a\240q\340\010\017\210q\320\004A\300\021\360\034\000\t\r\210H\320\0241\260\021\330\010\013\2104\210s\220!\2206\230\023\230A\330\014\022\220.\240\001\340\010\016\210a\210w\220a\220u\320\034,\250F\260!\330\010!\320!1\260\021\260$\3206F\300d\310'\320Q_\320_d\320de\330\010\013\2104\210w\220b\230\001\330\014\022\220+\230Q\230a\200A\360\026\000\t\r\210H\320\0241\260\021\340\010\032\230!\330\010\013\2108\2203\220a\340\014\020\220\006\220a\220}\240E\250\021\250'\260\025\260j\300\002\300!\3001\330\014\020\220\n\230#\230Q\230b\240\r\250Q\330\014\020\220\n\230#\230Q\230b\240\014\250A\330\014\020\220\n\230*\240A\330\014\020\220\n\230)\2401\340\014\020\220\006\220a\220q\330\0270\260\001\260\027\270\001\270\024\270Z\300s\310!\3104\310w\320V[\320[e\320eg\320gh\320hi\330\027\030\330\r\025\220S\230\001\340\014\020\220\006\220a\220}\240E\250\021\250'\260\025\260j\300\002\300!\3001\330\014\020\220\n\230#\230Q\230b\240\r\250Q\330\014\020\220\n\230#\230Q\230b\240\014\250A\330\014\020\220\n\230*\240A\330\014\020\220\n\230*\240A\340\014\020\220\006\220a\220q\330\0270\260\001\260\027\270\001\270\024\270Z\300s\310!\3104\310w\320V[\320[e\320eg\320gh\320hi\330\027\030\340\014\022\220.\240\001\200A\360\016\000\t\r\210H\320\0241\260\021\330\010\014\320\014\034\230G\2401\240A\200A\360\n\000\t\r\210H\320\0241\260\021\330\010\026\220o\240Q\240d\320*:\270$\270a\200A\360\030\000\t\r\210H\320\0241\260\021\340\010\037\230q\330\010\t\330\014\023\2201\330\014\033\2301\340\010\013\2108\2207\320\032,\250E\260\021\330\014\022\220.\240\001\330\010\033\2304\320\0375\260Q\330\010\013\210;\220b\230\001\330\014\022\220.""\240\001\320!A\300\027\310\001\310\021\340\010\024\220D\320\0301\260\021\330\010\026\220c\230\022\230:\240Q\240k\260\022\2601\340\010\014\210F\220!\320\023%\240Q\240a\330\023\036\230i\240q\250\003\250:\260Z\270w\300a\330\023\024\340\010\034\230L\250\002\250*\260B\260a\340\010\013\320\013\035\230S\240\001\330\014\024\220E\230\021\320\032:\270!\320;|\320|}\360\000\000~\001z\002\360\000\000z\002{\002\200A\330\010\014\210J\220a\220r\230\031\240!\200A\340\010\020\220\004\220J\230a\230r\240\034\250R\250s\260\"\260D\270\n\300!\3002\300Q\200A\330\010\014\210J\220i\230q\200A\330\010\014\210J\220j\240\001\200A\360\026\000\t\r\210J\220o\240Q\240a\200A\330\010\014\210K\220q\200A\330\010\014\210M\230\021\330\010\014\210I\220Q\330\010\014\210L\230\001\330\010\014\210N\230!\330\010\014\210H\220A\200A\330\010\014\210M\230\021\330\010\014\210N\230!\200A\330\010\014\210M\230\021\330\010\014\210N\230!\330\010\014\210H\220A\200A\330\010\014\210M\230\021\330\010\014\210N\230!\330\010\014\210M\230\021\330\010\014\210F\220!\330\010\014\210F\220!\330\010\014\210F\220!\200A\330\010\014\210N\230!\200A\360\030\000\t\r\210N\320\032/\250q\200A\340\010\025\220Q\340\010\025\220]\240!\2401\240D\320(8\270\001\270\021\330\014\017\210s\220'\230\032\2401\330\020\032\230'\240\021\240(\250!\2503\250a\330+.\250a\330+.\250a\330+.\320.A\320AT\320TU\320UX\320Xc\320cj\320jk\320kl\330\021\024\220G\230:\240Q\330\020\032\230'\240\021\240,\250a\250s\260!\330/2\260!\3306I\310\021\310#\310[\320X_\320_`\320`a\330\021\024\220G\230:\240Q\330\020\032\230'\240\021\240+\250Q\250c\260\021\330.1\260\021\340\020\032\230'\240\021\240)\2501\250A\330\010\017\210q\200A\330\010\014\320\014\036\230a\330\010\014\210N\230!\340\010\013\2104\210{\230'\240\021\240$\240h\250c\260\021\330\014\023\2201\340\014\026\220a\330\014\020\220\005\220U\230!\2304\230{\250'\260\021\260$\260g\270Q\270a\330\020\030\320\030*\250!\2501\330\020\025\220^\2401\240D\250\001\330\020\027\220w\230a\230q\330\014\023\2201\200A\340\010\014\320\014\036\230a\330""\010\014\210N\230!\330\010\013\2104\210{\230'\240\021\240$\240h\250c\260\021\330\014\023\2204\220{\240*\250A\250Q\340\014\023\2201\200A\330\010\014\320\014\036\230a\330\010\017\210t\220;\230e\2401\240D\250\001\200A\330\010\014\320\014\036\230a\330\010\017\210t\220;\230i\240q\250\004\250A\200A\330\010\014\320\014\036\230a\330\010\017\210t\220;\230k\250\021\250$\250a\200A\360\014\000\t\r\320\014)\250\021\360\006\000\t\032\320\031.\250a\250q\260\004\3204D\300A\300T\310\031\320RS\340\010\014\210I\220T\230\021\330\014\021\220\025\220a\330\014\017\210r\220\021\330\020\026\220b\230\t\240\021\240$\240b\250\t\260\021\260$\260b\270\t\300\021\300!\330\010\016\210f\220A\320\025&\240g\250Q\250a\340\010\025\220T\320\0311\260\021\330\010\013\2103\210a\210|\2302\230Q\330\014\022\220.\240\001\240\021\330\010\017\210q\200A\330\010\016\210a\210w\220d\230*\240J\250g\260W\270C\270q\300\001\200A\330\010\016\210f\220A\220Q\330\010!\240\037\260\001\260\024\3205E\300T\310\027\320PQ\320QU\320UV\330\010\013\2104\210w\220b\230\001\330\014\022\220,\230a\230q\340\010\026\220a\330\010\014\210E\220\025\220a\220t\230;\240a\330\014\031\230\035\240a\240q\330\014\026\320\026&\240d\250!\330\014\026\220n\240A\240T\250\021\330\014\027\220w\230a\230q\340\010\017\210q\200A\360\014\000\t\r\320\014)\250\021\330\010\026\220m\2401\240A\240T\250\021\200A\360\014\000\t\r\320\014)\250\021\330\010\026\220n\240A\240Q\240d\250!\200A\360\020\000\t\r\320\014)\250\021\330\010\026\220o\240Q\240a\240t\320+;\2701\200A\360\020\000\t \230q\330\010\t\330\014\023\2201\330\014\033\2301\340\010\013\2108\2207\320\032,\250E\260\021\330\014\022\220.\240\001\330\010\024\220D\320\0301\260\021\330\010\026\220c\230\033\240A\240T\250\026\250q\3200B\300!\300:\310S\320PQ\330(2\260!\330(/\250q\330\010\025\220\\\240\022\240:\250R\250q\330\010\017\210q\200A\330\010\037\230q\330\010\025\220S\230\013\2401\240D\250\006\250a\320/?\270s\300!\330$.\250a\330$+\2501\330\010\024\220C\220s\230+\240R\240q\330\010\017\210q\200A\330\010\027\220q\330""\010\014\320\014\036\230a\330\010\014\210N\230!\330\010\013\2104\210{\230'\240\021\240$\240h\250c\260\021\330\014\023\2204\220{\240*\250A\250Q\340\014\020\220\005\220U\230!\2304\230{\250'\260\021\260$\260g\270Q\270a\330\020\027\220t\230;\240j\260\001\260\021\330\014\023\2201\200A\330\010\017\210t\220>\240\021\200A\360\010\000\t\020\210t\2201\200A\340\010\017\210t\220:\230Q\200A\360\n\000\t\020\210t\220:\230Q\200A\330\010\017\210t\220:\230Q\200A\360\n\000\t\020\210t\220:\230Q\230b\240\001\200A\340\010\020\220\007\220t\230:\240V\2507\260!\2601\200A\330\010\017\210t\220;\230a\230t\2401\200A\330\010\017\210t\220;\230e\2401\240D\250\001\200A\330\010\017\210t\220;\230f\240A\240T\250\021\200A\330\010\017\210t\220;\230i\240q\250\004\250A\200A\330\010\017\210t\220;\230j\250\001\250\024\250Q\200A\330\010\027\220u\230A\230Q\230d\240&\250\002\250$\250c\260\031\270!\2707\300#\300Z\310z\320Y[\320[_\320_b\320bk\320kl\320ls\320sv\360\000\000w\001A\002\360\000\000A\002B\002\330\010\025\220S\230\005\230W\240A\330\010\017\210x\220q\230\004\230N\250!\2504\250{\270'\300\021\300$\300j\320PX\320XY\320YZ\200A\360\014\000\t\014\210<\220w\230a\330\014\023\2201\330\010\017\210t\2201\320\004B\320BY\320YZ\360\024\000\t\027\220d\230(\320\"4\260M\300\021\340\010\013\2104\210\177\230c\240\021\330\014\022\320\022#\2401\340\010\014\210H\320\0241\260\021\360\006\000\t\031\230\003\2301\230A\340\010\013\2101\330\014\025\220T\320\031+\2501\250J\260j\300\006\300f\310A\340\014\034\230M\250\021\250$\320.>\270d\300'\310\030\320QX\320XY\320Yb\320bl\320lr\360\000\000s\001C\002\360\000\000C\002I\002\360\000\000I\002J\002\360\010\000\t\023\220-\230q\240\004\320$4\260A\260Q\330\014\023\2203\220g\230S\240\004\240A\330\014\020\320\020!\240\021\240!\2401\340\010\017\210q\200\001\330\004\n\210+\220Q\320\004\027\220~\240Q\330\010\014\210K\220q\330\010\014\210G\2201\320\004>\320>R\320RS\360\024\000\t\r\210H\320\0241\260\021\340\010\013\320\013\034\230C\230q\330\023\037\230q\240\004\320$4\260D\270\007\270u\320DV\320VW""\340\023 \240\001\240\024\320%5\260T\270\027\300\005\320EW\320Wi\320ij\320\004;\320;R\320RS\360\024\000\t\027\220d\230(\320\"4\260M\300\021\330\010\013\2104\210\177\230c\240\021\330\014\022\320\022#\2401\340\010\014\210H\320\0241\260\021\360\n\000\t\020\320\017\037\230|\2502\250U\260!\2601\330\010\025\220Q\360\006\000\t\014\2101\330\014\025\220T\320\031*\250!\250:\260Z\270|\3106\320QR\340\014\034\230L\250\001\250\024\320-=\270T\300\027\310\010\320PW\320WX\320Xa\320ak\320kl\320lx\320x~\320~\177\360\010\000\t\023\220-\230q\240\004\320$4\260A\260Q\330\014\026\220a\220q\330\014\023\2203\220g\230S\240\004\240A\330\014\020\320\020!\240\021\240!\2401\360\006\000\t\n\330\014\023\320\023,\250A\250W\260F\270!\340\014\026\220a\220q\200\001\360\010\000\005\016\210T\220\033\230D\240\r\250T\260\027\270\004\270A\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\230:\240W\250E\260\023\260D\270\014\300G\3105\320PS\320SW\320W]\320]d\320di\320il\320lp\320pw\320w~\320~\177\330\004\007\200q\330\010\017\320\017.\250d\260!\2607\270+\300W\310A\340\010\017\320\017.\250d\260!\2607\270+\300Q\200\001\360\010\000\005\016\210T\320\021&\240d\250!\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220E\230\024\230Q\330\010\022\220!\330\010\027\220q\340\010\027\220t\230:\240W\250A\330\004\007\200q\330\010\017\320\017-\250T\260\021\260'\270\033\300G\3101\340\010\017\320\017-\250T\260\021\260'\270\033\300A\320\004@\320@V\320VW\360 \000\t\027\220d\230(\320\"4\260M\300\021\330\010\014\210H\320\0241\260\021\340\010\030\230\003\2301\230A\340\010\013\2101\330\014\025\220T\320\031+\2501\250G\260:\270T\300\026\300q\340\014\034\230M\250\021\250$\320.>\270d\300'\310\027\320PZ\320Z[\330/5\3205E\300V\3104\320Oe\320ew\320wx\320xy\360\006\000\t\026\220]\240!\2404\320'7\260q\270\001\330\014\020\220\003\2207\230*\320$<\270E\300\023\300A\300T\320I[\320[]\320]^\330\020\024\220N\240!\2401\240A\340\020\024\320\024%\240Q\240a\240q\340""\010\013\2104\210w\220b\230\001\330\014\022\220(\230!\2304\230q\200\001\320\004\036\230a\360\016\000\t\r\210H\320\0241\260\021\330\010\026\320\026(\250\001\250\024\320-=\270T\300\027\310\001\320\004&\240a\360&\000\t\014\210;\220c\230\021\330\014\035\230Y\240a\240q\250\004\320,<\270F\300'\310\021\310!\340\014\035\320\0350\260\001\260\021\260$\3206F\300a\300t\310?\320Z`\320`g\320gh\320hq\320q{\360\000\000|\001C\002\360\000\000C\002D\002\360\000\000D\002E\002\330\010\013\2108\2203\220a\330\014\022\220/\240\021\320\"?\270w\300a\300q\340\010\014\320\014#\2401\320\004.\250a\360\"\000\t\027\220d\320\032,\250A\250Q\330\010\014\320\014)\250\021\330\010\013\2101\330\014\023\2204\320\0270\260\001\330\010\026\320\026+\2501\250A\250T\260\021\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220=\240\010\250\001\250\021\330\004\007\200|\2207\230!\330\010/\250q\3200@\300\016\310a\330\004\013\2101\200\001\340\004\037\230q\320 0\260\013\270;\300k\320QR\330\004\023\220<\230x\240q\250\001\330\004\007\200|\2207\230!\330\010.\250a\250\177\270n\310A\330\004\013\2101\320\004\037\230q\360\016\000\t\r\210H\320\0241\260\021\330\010\026\320\026)\250\021\250$\320.>\270d\300'\310\021\320\004*\320*?\270q\360\032\000\t\027\220d\320\032,\250A\250Q\330\010\014\320\014)\250\021\330\010\013\2101\330\014\023\2204\320\0273\2601\260A\330\010\026\320\026.\250a\250q\260\004\3204D\300A\200\001\360\024\000\005\023\320\022)\250\021\250%\250w\260a\260q";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 479; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 61) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 479; i < 567; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 567; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 479;
      for (Py_ssize_t i=0; i<88; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab;
    double const c_constants[] = {1000000.0};
    for (int i = 0; i < 1; i++) {
      numbertab[i] = PyFloat_FromDouble(c_constants[i]);
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 1;
    int8_t const cint_constants_1[] = {0,1,2,3,4,8,10,11,16,20,22,24,26,28,32,40};
    int16_t const cint_constants_2[] = {500,1024,1040,1056,2000,2048,20000};
    int32_t const cint_constants_4[] = {50000L,65535L,200000L,9788398L,241895382L};
    for (int i = 0; i < 28; i++) {
      numbertab[i] = PyLong_FromLong((i < 16 ? cint_constants_1[i - 0] : (i < 23 ? cint_constants_2[i - 16] : cint_constants_4[i - 23])));
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<29; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 3;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 4;
    unsigned int flags : 10;
    unsigned int first_line : 11;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR), 172};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_ifname, __pyx_mstate->__pyx_n_u_master};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_open, __pyx_mstate->__pyx_kp_b_iso88591__8, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR), 667};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_x};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_genexpr, __pyx_mstate->__pyx_kp_b_iso88591_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_d_G1F_a_vWE_Q_q_t_WA_q_T_G1_T, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_6, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 156};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_ec_adapter, __pyx_mstate->__pyx_n_u_Adapter, __pyx_mstate->__pyx_n_u_adapters};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_find_adapters, __pyx_mstate->__pyx_kp_b_iso88591_43DA_k_A_q_l_Q_q_q_5_q_A_k_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 184};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_code};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_al_status_code_to_string, __pyx_mstate->__pyx_kp_b_iso88591_waq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 288};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ifname, __pyx_mstate->__pyx_n_u_ifname_red, __pyx_mstate->__pyx_n_u_ret_val};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_open, __pyx_mstate->__pyx_kp_b_iso88591_a_c_Yaq_F_0_6Fat_Z_gghhqq_C_C_D, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 316};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_check_context_is_initialized, __pyx_mstate->__pyx_kp_b_iso88591_A_4t1_aq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 335};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_release_gil};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_check_release_gil, __pyx_mstate->__pyx_kp_b_iso88591_A_wa_1_t1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 1, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 345};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_usetable, __pyx_mstate->__pyx_n_u_release_gil, __pyx_mstate->__pyx_n_u_ret_val, __pyx_mstate->__pyx_n_u_i};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_config_init, __pyx_mstate->__pyx_kp_b_iso88591_9_d_AQ_Ja_1_d_aq_Qat3C1_82Q_e5, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 370};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_cd_2, __pyx_mstate->__pyx_n_u_ret_val, __pyx_mstate->__pyx_n_u_slave, __pyx_mstate->__pyx_n_u_error_list};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_config_map, __pyx_mstate->__pyx_kp_b_iso88591_A_aq_4DAT_RS_IT_a_r_b_b_b_fA_gQa, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 392};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_cd_2, __pyx_mstate->__pyx_n_u_ret_val, __pyx_mstate->__pyx_n_u_slave, __pyx_mstate->__pyx_n_u_error_list};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_config_overlap_map, __pyx_mstate->__pyx_kp_b_iso88591_A_6aq_LATQZZ_IT_a_r_b_Ry_Biq_fA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 415};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_error_list, __pyx_mstate->__pyx_n_u_err};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_collect_mailbox_errors, __pyx_mstate->__pyx_kp_b_iso88591_A_Q_1D_8_s_1_3a_a_a_AATTUUXXccjj, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 436};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_config_dc, __pyx_mstate->__pyx_kp_b_iso88591_A_m1AT, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 445};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_A_1_4q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 453};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_read_state, __pyx_mstate->__pyx_kp_b_iso88591_A_nAQd, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 462};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_write_state, __pyx_mstate->__pyx_kp_b_iso88591_A_oQat_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 473};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_expected_state, __pyx_mstate->__pyx_n_u_timeout};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_state_check, __pyx_mstate->__pyx_kp_b_iso88591_6a_oQat_3_Na, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 1, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 500};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_release_gil};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_send_processdata, __pyx_mstate->__pyx_kp_b_iso88591_a_d_AQ_1_4_0_1AT, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 1, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 533};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_release_gil};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_send_overlap_processdata, __pyx_mstate->__pyx_kp_b_iso88591_6a_1_4_8_31AT, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 1, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 559};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_timeout, __pyx_mstate->__pyx_n_u_release_gil};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_receive_processdata, __pyx_mstate->__pyx_kp_b_iso88591_q_d_AQ_1_4_31A_aq_4DA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 578};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pos, __pyx_mstate->__pyx_n_u_ethercat_slave};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_slave, __pyx_mstate->__pyx_kp_b_iso88591_A_4r_AQ_4s_a_AQ_3aq_k_q_A_m1D_3a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 590};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[24] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_state, __pyx_mstate->__pyx_kp_b_iso88591_A_t_Qb, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[24])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 597};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_value};
    __pyx_mstate_global->__pyx_codeobj_tab[25] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_set_state, __pyx_mstate->__pyx_kp_b_iso88591_A_Jar, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[25])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 600};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[26] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_expected_wkc, __pyx_mstate->__pyx_kp_b_iso88591_A_Jar_Rs_D_2Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[26])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 604};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[27] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_dc_time, __pyx_mstate->__pyx_kp_b_iso88591_A_t1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[27])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 610};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_manual_state_change};
    __pyx_mstate_global->__pyx_codeobj_tab[28] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_set_manual_state_change, __pyx_mstate->__pyx_kp_b_iso88591_A_N_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[28])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 624};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[29] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_manual_state_change, __pyx_mstate->__pyx_kp_b_iso88591_A_t, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[29])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[30] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[30])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[31] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[31])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 638};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_slave_pos, __pyx_mstate->__pyx_n_u_index, __pyx_mstate->__pyx_n_u_subindex, __pyx_mstate->__pyx_n_u_abort_code, __pyx_mstate->__pyx_n_u_desc};
    __pyx_mstate_global->__pyx_codeobj_tab[32] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_A_M_IQ_L_N_HA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[32])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {7, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 657};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_slave_pos, __pyx_mstate->__pyx_n_u_error_code, __pyx_mstate->__pyx_n_u_error_reg, __pyx_mstate->__pyx_n_u_b1, __pyx_mstate->__pyx_n_u_w1, __pyx_mstate->__pyx_n_u_w2};
    __pyx_mstate_global->__pyx_codeobj_tab[33] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_A_M_N_M_F_F_F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[33])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 665};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_b1w1w2_bytes, __pyx_mstate->__pyx_n_u_b1w1w2_str, __pyx_mstate->__pyx_n_u_genexpr, __pyx_mstate->__pyx_n_u_genexpr};
    __pyx_mstate_global->__pyx_codeobj_tab[34] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_str, __pyx_mstate->__pyx_kp_b_iso88591_A_uAQd_c_7_ZzY___bbkkllssv_w_A_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[34])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 678};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_message};
    __pyx_mstate_global->__pyx_codeobj_tab[35] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_A_Kq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[35])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 691};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_slave_pos, __pyx_mstate->__pyx_n_u_error_code, __pyx_mstate->__pyx_n_u_desc};
    __pyx_mstate_global->__pyx_codeobj_tab[36] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_A_M_N_HA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[36])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 713};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_slave_pos, __pyx_mstate->__pyx_n_u_error_code};
    __pyx_mstate_global->__pyx_codeobj_tab[37] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_A_M_N, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[37])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 717};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[38] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_desc, __pyx_mstate->__pyx_kp_b_iso88591_A_t_at1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[38])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 730};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_error_list};
    __pyx_mstate_global->__pyx_codeobj_tab[39] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_A_N, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[39])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 741};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_message};
    __pyx_mstate_global->__pyx_codeobj_tab[40] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_A_Kq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[40])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 753};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_message, __pyx_mstate->__pyx_n_u_wkc};
    __pyx_mstate_global->__pyx_codeobj_tab[41] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_init, __pyx_mstate->__pyx_kp_b_iso88591_Q_Kq_G1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[41])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[42] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_T_D_T_A_G1F_a_vWE_Q_q_t_WE_D_G5, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[42])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[43] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_1F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[43])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 823};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_act, __pyx_mstate->__pyx_n_u_sync0_cycle_time, __pyx_mstate->__pyx_n_u_sync0_shift_time, __pyx_mstate->__pyx_n_u_sync1_cycle_time};
    __pyx_mstate_global->__pyx_codeobj_tab[44] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_dc_sync, __pyx_mstate->__pyx_kp_b_iso88591_RRS_H_1_Cq_q_4D_uDVVW_5T_EWWiij, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[44])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 1, 11, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 859};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_index, __pyx_mstate->__pyx_n_u_subindex, __pyx_mstate->__pyx_n_u_size, __pyx_mstate->__pyx_n_u_ca, __pyx_mstate->__pyx_n_u_release_gil, __pyx_mstate->__pyx_n_u_pbuf, __pyx_mstate->__pyx_n_u_std_buffer, __pyx_mstate->__pyx_n_u_size_inout, __pyx_mstate->__pyx_n_u_result, __pyx_mstate->__pyx_n_u_err};
    __pyx_mstate_global->__pyx_codeobj_tab[45] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_sdo_read, __pyx_mstate->__pyx_kp_b_iso88591_9_UUV_d_4M_4_c_1_H_1_5_1_1_r_aq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[45])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 1, 9, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 949};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_index, __pyx_mstate->__pyx_n_u_subindex, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_ca, __pyx_mstate->__pyx_n_u_release_gil, __pyx_mstate->__pyx_n_u_size, __pyx_mstate->__pyx_n_u_result, __pyx_mstate->__pyx_n_u_err};
    __pyx_mstate_global->__pyx_codeobj_tab[46] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_sdo_write, __pyx_mstate->__pyx_kp_b_iso88591_VVW_d_4M_H_1_1A_1_T_1G_T_q_M_d, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[46])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 986};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_buf, __pyx_mstate->__pyx_n_u_wkt, __pyx_mstate->__pyx_n_u_err};
    __pyx_mstate_global->__pyx_codeobj_tab[47] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_mbx_receive, __pyx_mstate->__pyx_kp_b_iso88591_A_H_1_1AQ_oQd2B_gQeST_q_4AQ_7_E, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[47])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1010};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[48] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_write_state, __pyx_mstate->__pyx_kp_b_iso88591_A_H_1_oQd_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[48])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1018};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_expected_state, __pyx_mstate->__pyx_n_u_timeout};
    __pyx_mstate_global->__pyx_codeobj_tab[49] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_state_check, __pyx_mstate->__pyx_kp_b_iso88591_6a_H_1_oQd_gEUUV, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[49])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1023};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_timeout};
    __pyx_mstate_global->__pyx_codeobj_tab[50] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_reconfig, __pyx_mstate->__pyx_kp_b_iso88591_q_H_1_d, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[50])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1033};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_timeout};
    __pyx_mstate_global->__pyx_codeobj_tab[51] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_recover, __pyx_mstate->__pyx_kp_b_iso88591_a_H_1_T, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[51])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1043};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_word_address, __pyx_mstate->__pyx_n_u_timeout, __pyx_mstate->__pyx_n_u_tmp};
    __pyx_mstate_global->__pyx_codeobj_tab[52] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_eeprom_read, __pyx_mstate->__pyx_kp_b_iso88591_4A_H_1_47Gt7R_a_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[52])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1059};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_word_address, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_timeout, __pyx_mstate->__pyx_n_u_tmp, __pyx_mstate->__pyx_n_u_result};
    __pyx_mstate_global->__pyx_codeobj_tab[53] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_eeprom_write, __pyx_mstate->__pyx_kp_b_iso88591_A_H_1_4s_6_A_awau_F_1_6Fd_Q__dd, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[53])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 1, 9, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1104};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_filename, __pyx_mstate->__pyx_n_u_password, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_timeout, __pyx_mstate->__pyx_n_u_release_gil, __pyx_mstate->__pyx_n_u_result, __pyx_mstate->__pyx_n_u_size, __pyx_mstate->__pyx_n_u_err};
    __pyx_mstate_global->__pyx_codeobj_tab[54] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_foe_write, __pyx_mstate->__pyx_kp_b_iso88591_BBYYZ_d_4M_4_c_1_H_1_1A_1_T_1Jj, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[54])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 1, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1158};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_filename, __pyx_mstate->__pyx_n_u_password, __pyx_mstate->__pyx_n_u_size, __pyx_mstate->__pyx_n_u_timeout, __pyx_mstate->__pyx_n_u_release_gil, __pyx_mstate->__pyx_n_u_pbuf, __pyx_mstate->__pyx_n_u_size_inout, __pyx_mstate->__pyx_n_u_result, __pyx_mstate->__pyx_n_u_err};
    __pyx_mstate_global->__pyx_codeobj_tab[55] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_foe_read, __pyx_mstate->__pyx_kp_b_iso88591_RRS_d_4M_4_c_1_H_1_2U_1_Q_1_T_Z, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[55])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1199};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_mailbox, __pyx_mstate->__pyx_n_u_start_address, __pyx_mstate->__pyx_n_u_size, __pyx_mstate->__pyx_n_u_fpwr_timeout_us};
    __pyx_mstate_global->__pyx_codeobj_tab[56] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_amend_mbx, __pyx_mstate->__pyx_kp_b_iso88591_A_H_1_83a_a_E_j_1_Qb_Q_Qb_A_A_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[56])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 9, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1238};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_wd_type, __pyx_mstate->__pyx_n_u_wd_time_ms, __pyx_mstate->__pyx_n_u_fprd_fpwr_timeout_us, __pyx_mstate->__pyx_n_u_wd_type_to_reg_map, __pyx_mstate->__pyx_n_u_wd_time_ms_limit, __pyx_mstate->__pyx_n_u_wd_div_ns, __pyx_mstate->__pyx_n_u_wd_time_reg, __pyx_mstate->__pyx_n_u_actual_wd_time_ms};
    __pyx_mstate_global->__pyx_codeobj_tab[57] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_set_watchdog, __pyx_mstate->__pyx_kp_b_iso88591_A_H_1_q_1_1_87_E_4_5Q_b_A_D_1_c, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[57])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1275};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_wd_type, __pyx_mstate->__pyx_n_u_fprd_fpwr_timeout_us, __pyx_mstate->__pyx_n_u_wd_type_to_reg_map, __pyx_mstate->__pyx_n_u_wd_div_ns, __pyx_mstate->__pyx_n_u_wd_time_reg, __pyx_mstate->__pyx_n_u_wd_time_ms};
    __pyx_mstate_global->__pyx_codeobj_tab[58] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_watchdog, __pyx_mstate->__pyx_kp_b_iso88591_A_q_1_1_87_E_D_1_c_AT_q0B_SPQ_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[58])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1297};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_fprd_fpwr_timeout_us, __pyx_mstate->__pyx_n_u_wd_div_reg, __pyx_mstate->__pyx_n_u_wd_div_ns};
    __pyx_mstate_global->__pyx_codeobj_tab[59] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_watchdog_divider_ns, __pyx_mstate->__pyx_kp_b_iso88591_A_q_S_1D_a_s_a_1_Cs_Rq_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[59])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1305};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_wd_div_ns, __pyx_mstate->__pyx_n_u_wd_time_ms_limit};
    __pyx_mstate_global->__pyx_codeobj_tab[60] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_max_watchdog_time, __pyx_mstate->__pyx_kp_b_iso88591_A_D_1_7_Jb_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[60])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1316};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_callback};
    __pyx_mstate_global->__pyx_codeobj_tab[61] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_add_emergency_callback, __pyx_mstate->__pyx_kp_b_iso88591_A_H_1_G1A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[61])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1341};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[62] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_disable_complete_access, __pyx_mstate->__pyx_kp_b_iso88591_A_JoQa, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[62])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1354};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_address, __pyx_mstate->__pyx_n_u_size, __pyx_mstate->__pyx_n_u_timeout_us, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_wkc};
    __pyx_mstate_global->__pyx_codeobj_tab[63] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_fprd, __pyx_mstate->__pyx_kp_b_iso88591_6a_1A_iq_N_Z_ddjjppq_4s_aq_AWF, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[63])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1367};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_address, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_timeout_us, __pyx_mstate->__pyx_n_u_wkc};
    __pyx_mstate_global->__pyx_codeobj_tab[64] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_fpwr, __pyx_mstate->__pyx_kp_b_iso88591_8_iq_N_Z_ddiillmmt_u_E_E_K_K_L, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[64])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1398};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[65] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_name, __pyx_mstate->__pyx_kp_b_iso88591_A_t_V7_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[65])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1402};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[66] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_eep_man, __pyx_mstate->__pyx_kp_b_iso88591_A_t_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[66])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1406};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[67] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_eep_id, __pyx_mstate->__pyx_kp_b_iso88591_A_t_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[67])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1410};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[68] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_eep_rev, __pyx_mstate->__pyx_kp_b_iso88591_A_t_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[68])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1414};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[69] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_PO2SOconfig, __pyx_mstate->__pyx_kp_b_iso88591_A_4t_Q_1_xt_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[69])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1423};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[70] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_PO2SOconfigEx, __pyx_mstate->__pyx_kp_b_iso88591_A_4t_Q_1_xt_Q_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[70])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1436};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_value};
    __pyx_mstate_global->__pyx_codeobj_tab[71] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_set_PO2SOconfig, __pyx_mstate->__pyx_kp_b_iso88591_A_D_JhgT_6_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[71])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1444};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_value};
    __pyx_mstate_global->__pyx_codeobj_tab[72] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_set_PO2SOconfigEx, __pyx_mstate->__pyx_kp_b_iso88591_A_D_JhgT_6_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[72])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1452};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[73] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_state, __pyx_mstate->__pyx_kp_b_iso88591_A_t_Q_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[73])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1459};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_value};
    __pyx_mstate_global->__pyx_codeobj_tab[74] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_set_state, __pyx_mstate->__pyx_kp_b_iso88591_A_Jiq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[74])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1462};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_num_bytes};
    __pyx_mstate_global->__pyx_codeobj_tab[75] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_input, __pyx_mstate->__pyx_kp_b_iso88591_A_D_D_Rt4z_Jiq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[75])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1468};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_num_bytes};
    __pyx_mstate_global->__pyx_codeobj_tab[76] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_output, __pyx_mstate->__pyx_kp_b_iso88591_A_D_D_Rt4z_Jj, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[76])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1474};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_value};
    __pyx_mstate_global->__pyx_codeobj_tab[77] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_set_output, __pyx_mstate->__pyx_kp_b_iso88591_A_awd_JgWCq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[77])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1477};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[78] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_al_status, __pyx_mstate->__pyx_kp_b_iso88591_A_t_Q_3, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[78])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1480};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[79] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_is_lost, __pyx_mstate->__pyx_kp_b_iso88591_A_t_Q_3, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[79])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1483};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_value};
    __pyx_mstate_global->__pyx_codeobj_tab[80] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_set_is_lost, __pyx_mstate->__pyx_kp_b_iso88591_A_Jj, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[80])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1486};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_result, __pyx_mstate->__pyx_n_u_coe_objects, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_coe_object};
    __pyx_mstate_global->__pyx_codeobj_tab[81] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_od, __pyx_mstate->__pyx_kp_b_iso88591_A_fAQ_5ET_PQQUUV_4wb_aq_a_E_at_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[81])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[82] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[82])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[83] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[83])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1528};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_result};
    __pyx_mstate_global->__pyx_codeobj_tab[84] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_read_description, __pyx_mstate->__pyx_kp_b_iso88591_A_4t1_aq_0__D_PTTU_T_L_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[84])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1537};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_result};
    __pyx_mstate_global->__pyx_codeobj_tab[85] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_read_entries, __pyx_mstate->__pyx_kp_b_iso88591_A_4t1_Kq_O4xt_XYY_t7_A_l_1_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[85])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1546};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[86] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_index, __pyx_mstate->__pyx_kp_b_iso88591_A_t_fAT, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[86])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1549};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[87] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_data_type, __pyx_mstate->__pyx_kp_b_iso88591_A_a_t_iq_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[87])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1553};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[88] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_object_code, __pyx_mstate->__pyx_kp_b_iso88591_A_a_t_k_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[88])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1557};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[89] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_name, __pyx_mstate->__pyx_kp_b_iso88591_A_a_t_e1D, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[89])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1561};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_entries, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_entry};
    __pyx_mstate_global->__pyx_codeobj_tab[90] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_entries, __pyx_mstate->__pyx_kp_b_iso88591_A_a_N_4_hc_1_a_U_4_gQa_1_1D_waq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[90])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1575};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_sum, __pyx_mstate->__pyx_n_u_i};
    __pyx_mstate_global->__pyx_codeobj_tab[91] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_bit_length, __pyx_mstate->__pyx_kp_b_iso88591_A_q_a_N_4_hc_4_AQ_U_4_gQa_t_j_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[91])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1586};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[92] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_obj_access, __pyx_mstate->__pyx_kp_b_iso88591_A_a_N_4_hc_4_AQ_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[92])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[93] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[93])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[94] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[94])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1608};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[95] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_name, __pyx_mstate->__pyx_kp_b_iso88591_A_t_e1D, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[95])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1611};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[96] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_data_type, __pyx_mstate->__pyx_kp_b_iso88591_A_t_iq_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[96])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1614};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[97] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_bit_length, __pyx_mstate->__pyx_kp_b_iso88591_A_t_j_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[97])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1617};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[98] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_src_pysoem_pysoem_pyx, __pyx_mstate->__pyx_n_u_get_obj_access, __pyx_mstate->__pyx_kp_b_iso88591_A_t_j_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[98])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[99] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[99])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[100] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[100])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[101] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_CdefSettings, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_xq_7_a_nA_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[101])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 4};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[102] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle__CallbackData, __pyx_mstate->__pyx_kp_b_iso88591_q_0_kQR_7_q0_a_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[102])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* Generator.init */
  if (likely(__pyx_Generator_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall (used by PyObjectCallOneArg) */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by RejectKeywords) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyObjectCall2Args (used by PyObjectCallMethod1) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod1) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod1 (used by append) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_mstate_global->__pyx_n_u_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* RaiseUnboundLocalError */
static void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* PyObjectVectorCallKwBuilder (used by PyObjectVectorCallMethodKwBuilder) */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectVectorCallMethodKwBuilder */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames) {
    PyObject *result;
    PyObject *obj = PyObject_GetAttr(args[0], name);
    if (unlikely(!obj))
        return NULL;
    result = __Pyx_Object_Vectorcall_CallFromBuilder(obj, args+1, nargsf-1, kwnames);
    Py_DECREF(obj);
    return result;
}
#endif

/* PyObjectFormat */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* obj, PyObject* format_spec) {
    int ret;
    _PyUnicodeWriter writer;
    if (likely(PyFloat_CheckExact(obj))) {
        _PyUnicodeWriter_Init(&writer);
        ret = _PyFloat_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else if (likely(PyLong_CheckExact(obj))) {
        _PyUnicodeWriter_Init(&writer);
        ret = _PyLong_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else {
        return PyObject_Format(obj, format_spec);
    }
    if (unlikely(ret == -1)) {
        _PyUnicodeWriter_Dealloc(&writer);
        return NULL;
    }
    return _PyUnicodeWriter_Finish(&writer);
}
#endif

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_mstate_global->__pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    if (max_char > 1114111) max_char = 1114111;
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - result_ulength < 0))
        goto overflow;
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = values[i];
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (__Pyx_PyUnicode_READY(uval) == (-1))
            goto bad;
        #endif
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(ulength < 0)) goto bad;
        #endif
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    Py_ssize_t i;
    PyObject *result = NULL;
    PyObject *value_tuple = PyTuple_New(value_count);
    if (unlikely(!value_tuple)) return NULL;
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    for (i=0; i<value_count; i++) {
        if (__Pyx_PyTuple_SET_ITEM(value_tuple, i, values[i]) != (0)) goto bad;
        Py_INCREF(values[i]);
    }
    result = PyUnicode_Join(__pyx_mstate_global->__pyx_empty_unicode, value_tuple);
bad:
    Py_DECREF(value_tuple);
    return result;
#endif
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_mstate_global->__pyx_n_u_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* GetTopmostException (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* py_dict_keys */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_keys, d);
}

/* DictGetItem */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    if (unlikely(__Pyx_PyDict_GetItemRef(d, key, &value) == 0)) { // no value, no error
        if (unlikely(PyTuple_Check(key))) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args)) {
                PyErr_SetObject(PyExc_KeyError, args);
                Py_DECREF(args);
            }
        } else {
            PyErr_SetObject(PyExc_KeyError, key);
        }
    }
    return value;
}
#endif

/* PyFloatBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_DivideObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a, result;
    CYTHON_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    if (likely(PyFloat_CheckExact(op1))) {
        a = __Pyx_PyFloat_AS_DOUBLE(op1);
        
    } else
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsZero(op1)) {
            a = 0.0;
            
        } else if (__Pyx_PyLong_IsCompact(op1)) {
            a = (double) __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                case 2:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                        a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -2)
                                a = -a;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                case 3:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                        a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -3)
                                a = -a;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                case 4:
                    if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                        a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -4)
                                a = -a;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                default:
        #endif
                    a = PyLong_AsDouble(op1);
                    if (unlikely(a == -1.0 && PyErr_Occurred())) return NULL;
        #if CYTHON_USE_PYLONG_INTERNALS
            }
        }
        #endif
    } else {
        return (inplace ? __Pyx_PyNumber_InPlaceDivide(op1, op2) : __Pyx_PyNumber_Divide(op1, op2));
    }
        result = a / b;
        return PyFloat_FromDouble(result);
}
#endif

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_MultiplyCObj(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceMultiply : PyNumber_Multiply)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_MultiplyCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long a = intval;
    long b;
    const PY_LONG_LONG lla = intval;
    PY_LONG_LONG llb;
    if (unlikely(__Pyx_PyLong_IsZero(op2))) {
        return __Pyx_NewRef(op2);
    }
    const int is_positive = __Pyx_PyLong_IsPos(op2);
    const digit* digits = __Pyx_PyLong_Digits(op2);
    const Py_ssize_t size = __Pyx_PyLong_DigitCount(op2);
    if (likely(size == 1)) {
        b = (long) digits[0];
        if (!is_positive) b *= -1;
    } else {
        switch (size) {
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                    b = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) b *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                    llb = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) llb *= -1;
                    goto calculate_long_long;
                }
                break;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                    b = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) b *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                    llb = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) llb *= -1;
                    goto calculate_long_long;
                }
                break;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                    b = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) b *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                    llb = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) llb *= -1;
                    goto calculate_long_long;
                }
                break;
        }
        return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
    }
    calculate_long:
        CYTHON_UNUSED_VAR(a);
        CYTHON_UNUSED_VAR(b);
        llb = b;
        goto calculate_long_long;
    calculate_long_long:
        {
            PY_LONG_LONG llx;
            llx = lla * llb;
            return PyLong_FromLongLong(llx);
        }
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_MultiplyCObj(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long a = intval;
    double b = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) * (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_MultiplyCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op2))) {
        return __Pyx_Unpacked___Pyx_PyLong_MultiplyCObj(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op2)) {
        return __Pyx_Float___Pyx_PyLong_MultiplyCObj(op2, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_MultiplyCObj(op1, op2, inplace);
}
#endif

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a;
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla;
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return __Pyx_NewRef(op2);
    }
    const int is_positive = __Pyx_PyLong_IsPos(op1);
    const digit* digits = __Pyx_PyLong_Digits(op1);
    const Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
    if (likely(size == 1)) {
        a = (long) digits[0];
        if (!is_positive) a *= -1;
    } else {
        switch (size) {
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
        }
        return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    calculate_long:
        {
            long x;
            x = a + b;
            return PyLong_FromLong(x);
        }
    calculate_long_long:
        {
            PY_LONG_LONG llx;
            llx = lla + llb;
            return PyLong_FromLongLong(llx);
        }
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_AddObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) + (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_AddObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_AddObjC(op1, op2, inplace);
}
#endif

/* decode_c_bytes (used by decode_bytes) */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    ctx = PyUnicode_FromString(name);
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
    if (nogil)
        PyGILState_Release(state);
}

/* AllocateExtensionType */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final) {
    if (is_final || likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
        allocfunc alloc_func = __Pyx_PyType_GetSlot(t, tp_alloc, allocfunc);
        return alloc_func(t, 0);
    } else {
        newfunc tp_new = __Pyx_PyType_TryGetSlot(&PyBaseObject_Type, tp_new, newfunc);
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (!tp_new) {
            PyObject *new_str = PyUnicode_FromString("__new__");
            if (likely(new_str)) {
                PyObject *o = PyObject_CallMethodObjArgs((PyObject *)&PyBaseObject_Type, new_str, t, NULL);
                Py_DECREF(new_str);
                return o;
            } else
                return NULL;
        } else
    #endif
        return tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
    }
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod0 (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u__6);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* ListPack */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...) {
    va_list va;
    PyObject *l = PyList_New(n);
    va_start(va, n);
    if (unlikely(!l)) goto end;
    for (Py_ssize_t i=0; i<n; ++i) {
        PyObject *arg = va_arg(va, PyObject*);
        Py_INCREF(arg);
        if (__Pyx_PyList_SET_ITEM(l, i, arg) != (0)) {
            Py_CLEAR(l);
            goto end;
        }
    }
    end:
    va_end(va);
    return l;
}

/* GetAttr (used by Globals) */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n)))
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* Globals */
static PyObject* __Pyx_Globals(void) {
    return __Pyx_NewRef(__pyx_mstate_global->__pyx_d);
}

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* Py3UpdateBases */
static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base = NULL, *meth, *new_base, *result, *new_bases = NULL;
#if CYTHON_ASSUME_SAFE_SIZE
    size_bases = PyTuple_GET_SIZE(bases);
#else
    size_bases = PyTuple_Size(bases);
    if (size_bases < 0) return NULL;
#endif
    for (i = 0; i < size_bases; i++) {
#if CYTHON_AVOID_BORROWED_REFS
        Py_CLEAR(base);
#endif
#if CYTHON_ASSUME_SAFE_MACROS
        base = PyTuple_GET_ITEM(bases, i);
#else
        base = PyTuple_GetItem(bases, i);
        if (!base) goto error;
#endif
#if CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(base);
#endif
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_mstate_global->__pyx_n_u_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                PyObject *base_from_list;
#if CYTHON_ASSUME_SAFE_MACROS
                base_from_list = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base_from_list);
                Py_INCREF(base_from_list);
#else
                base_from_list = PyTuple_GetItem(bases, j);
                if (!base_from_list) goto error;
                Py_INCREF(base_from_list);
                if (PyList_SetItem(new_bases, j, base_from_list) < 0) goto error;
#endif
            }
        }
#if CYTHON_ASSUME_SAFE_SIZE
        j = PyList_GET_SIZE(new_bases);
#else
        j = PyList_Size(new_bases);
        if (j < 0) goto error;
#endif
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
#if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(base);
#endif
    return result;
error:
    Py_XDECREF(new_bases);
#if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(base);
#endif
    return NULL;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases;
#if CYTHON_ASSUME_SAFE_SIZE
    nbases = PyTuple_GET_SIZE(bases);
#else
    nbases = PyTuple_Size(bases);
    if (nbases < 0) return NULL;
#endif
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
#if CYTHON_ASSUME_SAFE_MACROS
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *tmp = PyTuple_GetItem(bases, i);
        if (!tmp) return NULL;
#endif
        tmptype = Py_TYPE(tmp);
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
        metaclass = &PyType_Type;
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectLookupSpecial (used by Py3ClassCreate) */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_mstate_global->__pyx_n_u_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_module_2, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_mstate_global->__pyx_n_u_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
    Py_XDECREF(owned_metaclass);
    return result;
}

/* GetNameInClass */
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    PyObject *dict;
    assert(PyType_Check(nmspace));
#if CYTHON_USE_TYPE_SLOTS
    dict = ((PyTypeObject*)nmspace)->tp_dict;
    Py_XINCREF(dict);
#else
    dict = PyObject_GetAttr(nmspace, __pyx_mstate_global->__pyx_n_u_dict);
#endif
    if (likely(dict)) {
        result = PyObject_GetItem(dict, name);
        Py_DECREF(dict);
        if (result) {
            return result;
        }
    }
    PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CheckUnpickleChecksum */
static void __Pyx_RaiseUnpickleChecksumError(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    PyObject *pickle_module = PyImport_ImportModule("pickle");
    if (unlikely(!pickle_module)) return;
    PyObject *pickle_error = PyObject_GetAttrString(pickle_module, "PickleError");
    Py_DECREF(pickle_module);
    if (unlikely(!pickle_error)) return;
    if (checksum2 == checksum1) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x) = (%s))",
            checksum, checksum1, members);
    } else if (checksum3 == checksum2) {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, members);
    } else {
        PyErr_Format(pickle_error, "Incompatible checksums (0x%x vs (0x%x, 0x%x, 0x%x) = (%s))",
            checksum, checksum1, checksum2, checksum3, members);
    }
    Py_DECREF(pickle_error);
}
static int __Pyx_CheckUnpickleChecksum(long checksum, long checksum1, long checksum2, long checksum3, const char *members) {
    int found = 0;
    found |= checksum1 == checksum;
    found |= checksum2 == checksum;
    found |= checksum3 == checksum;
    if (likely(found))
        return 0;
    __Pyx_RaiseUnpickleChecksumError(checksum, checksum1, checksum2, checksum3, members);
    return -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int8_t(int8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int8_t neg_one = (int8_t) -1, const_zero = (int8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int8_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int8_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int8_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int8_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int8_t __Pyx_PyLong_As_int8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int8_t neg_one = (int8_t) -1, const_zero = (int8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int8_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int8_t) -1;
        val = __Pyx_PyLong_As_int8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int8_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int8_t) >= 2 * PyLong_SHIFT)) {
                            return (int8_t) (((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int8_t) >= 3 * PyLong_SHIFT)) {
                            return (int8_t) (((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int8_t) >= 4 * PyLong_SHIFT)) {
                            return (int8_t) (((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int8_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int8_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int8_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int8_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int8_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int8_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int8_t) (((int8_t)-1)*(((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int8_t) ((((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int8_t) (((int8_t)-1)*(((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int8_t) ((((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int8_t) (((int8_t)-1)*(((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int8_t) ((((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int8_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int8_t, long, PyLong_AsLong(x))
        } else if ((sizeof(int8_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int8_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int8_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int8_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int8_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int8_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int8_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int8_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int8_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int8_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int8_t) 1) << (sizeof(int8_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int8_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int8_t");
    return (int8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int8_t");
    return (int8_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_ec_state(ec_state value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const ec_state neg_one = (ec_state) -1, const_zero = (ec_state) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(ec_state) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(ec_state) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(ec_state) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(ec_state) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(ec_state) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(ec_state),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(ec_state));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyLong_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyLong_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT)) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT)) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT)) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint8_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint8_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint8_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint8_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint8_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint8_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint8_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint8_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint8_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint8_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint8_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint8_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint8_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint8_t) 1) << (sizeof(uint8_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint8_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint16_t __Pyx_PyLong_As_uint16_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint16_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint16_t) -1;
        val = __Pyx_PyLong_As_uint16_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint16_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint16_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) >= 2 * PyLong_SHIFT)) {
                            return (uint16_t) (((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint16_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) >= 3 * PyLong_SHIFT)) {
                            return (uint16_t) (((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint16_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) >= 4 * PyLong_SHIFT)) {
                            return (uint16_t) (((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint16_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint16_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint16_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint16_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint16_t) (((uint16_t)-1)*(((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint16_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint16_t) ((((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint16_t) (((uint16_t)-1)*(((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint16_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint16_t) ((((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint16_t) (((uint16_t)-1)*(((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint16_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint16_t) ((((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint16_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint16_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint16_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint16_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint16_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint16_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint16_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint16_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint16_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint16_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint16_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint16_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint16_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint16_t) 1) << (sizeof(uint16_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint16_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint16_t");
    return (uint16_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint16_t");
    return (uint16_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint16_t(uint16_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint16_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint16_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint16_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint16_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint16_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint16_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint8_t(uint8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint8_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int32_t(int32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyLong_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint32_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint32_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint32_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint32_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint32_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint32_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint32_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint32_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint32_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint32_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint32_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint32_t) 1) << (sizeof(uint32_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint32_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int32_t __Pyx_PyLong_As_int32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int32_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int32_t) -1;
        val = __Pyx_PyLong_As_int32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) >= 2 * PyLong_SHIFT)) {
                            return (int32_t) (((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) >= 3 * PyLong_SHIFT)) {
                            return (int32_t) (((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) >= 4 * PyLong_SHIFT)) {
                            return (int32_t) (((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int32_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int32_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int32_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int32_t) (((int32_t)-1)*(((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int32_t) ((((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int32_t) (((int32_t)-1)*(((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int32_t) ((((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int32_t) (((int32_t)-1)*(((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int32_t) ((((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int32_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, long, PyLong_AsLong(x))
        } else if ((sizeof(int32_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int32_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int32_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int32_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int32_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int32_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int32_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int32_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int32_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int32_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int32_t) 1) << (sizeof(int32_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int32_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int32_t");
    return (int32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int32_t");
    return (int32_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyLong_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_ec_err_type(ec_err_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const ec_err_type neg_one = (ec_err_type) -1, const_zero = (ec_err_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(ec_err_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(ec_err_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(ec_err_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(ec_err_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(ec_err_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(ec_err_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(ec_err_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum____pyx_t_6pysoem_6pysoem_ec_datatype(enum __pyx_t_6pysoem_6pysoem_ec_datatype value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_6pysoem_6pysoem_ec_datatype neg_one = (enum __pyx_t_6pysoem_6pysoem_ec_datatype) -1, const_zero = (enum __pyx_t_6pysoem_6pysoem_ec_datatype) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_6pysoem_6pysoem_ec_datatype) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_6pysoem_6pysoem_ec_datatype) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum __pyx_t_6pysoem_6pysoem_ec_datatype) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_6pysoem_6pysoem_ec_datatype) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_6pysoem_6pysoem_ec_datatype) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_6pysoem_6pysoem_ec_datatype),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum __pyx_t_6pysoem_6pysoem_ec_datatype));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* UpdateUnpickledDict */
static int __Pyx__UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    PyObject *state_dict = __Pyx_PySequence_ITEM(state, index);
    if (unlikely(!state_dict)) {
        return -1;
    }
    int non_empty = PyObject_IsTrue(state_dict);
    if (non_empty == 0) {
        Py_DECREF(state_dict);
        return 0;
    } else if (unlikely(non_empty == -1)) {
        return -1;
    }
    PyObject *dict;
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    dict = PyObject_GetAttrString(obj, "__dict__");
    #else
    dict = PyObject_GenericGetDict(obj, NULL);
    #endif
    if (unlikely(!dict)) {
        Py_DECREF(state_dict);
        return -1;
    }
    int result;
    if (likely(PyDict_CheckExact(dict))) {
        result = PyDict_Update(dict, state_dict);
    } else {
        PyObject *obj_result = __Pyx_PyObject_CallMethod1(dict, __pyx_mstate_global->__pyx_n_u_update, state_dict);
        if (likely(obj_result)) {
            Py_DECREF(obj_result);
            result = 0;
        } else {
            result = -1;
        }
    }
    Py_DECREF(state_dict);
    Py_DECREF(dict);
    return result;
}
static int __Pyx_UpdateUnpickledDict(PyObject *obj, PyObject *state, Py_ssize_t index) {
    Py_ssize_t state_size = __Pyx_PyTuple_GET_SIZE(state);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(state_size == -1)) return -1;
    #endif
    if (state_size <= index) {
        return 0;
    }
    return __Pyx__UpdateUnpickledDict(obj, state, index);
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module_2);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__7);
    }
    goto done;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* IterNextPlain (used by CoroutineBase) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *__Pyx_GetBuiltinNext_LimitedAPI(void) {
    if (unlikely(!__pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache))
        __pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache = __Pyx_GetBuiltinName(__pyx_mstate_global->__pyx_n_u_next);
    return __pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next_Plain(PyObject *iterator) {
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    PyObject *result;
    PyObject *next = __Pyx_GetBuiltinNext_LimitedAPI();
    if (unlikely(!next)) return NULL;
    result = PyObject_CallFunctionObjArgs(next, iterator, NULL);
    return result;
#else
    (void)__Pyx_GetBuiltinName; // only for early limited API
    iternextfunc iternext = __Pyx_PyObject_GetIterNextFunc(iterator);
    assert(iternext);
    return iternext(iterator);
#endif
}

/* ReturnWithStopIteration (used by CoroutineBase) */
static void __Pyx__ReturnWithStopIteration(PyObject* value, int async);
static CYTHON_INLINE void __Pyx_ReturnWithStopIteration(PyObject* value, int async, int iternext) {
    if (value == Py_None) {
        if (async || !iternext)
            PyErr_SetNone(async ? PyExc_StopAsyncIteration : PyExc_StopIteration);
        return;
    }
    __Pyx__ReturnWithStopIteration(value, async);
}
static void __Pyx__ReturnWithStopIteration(PyObject* value, int async) {
#if CYTHON_COMPILING_IN_CPYTHON
    __Pyx_PyThreadState_declare
#endif
    PyObject *exc;
    PyObject *exc_type = async ? PyExc_StopAsyncIteration : PyExc_StopIteration;
#if CYTHON_COMPILING_IN_CPYTHON
    if ((PY_VERSION_HEX >= (0x030C00A6)) || unlikely(PyTuple_Check(value) || PyExceptionInstance_Check(value))) {
        if (PY_VERSION_HEX >= (0x030e00A1)) {
            exc = __Pyx_PyObject_CallOneArg(exc_type, value);
        } else {
            PyObject *args_tuple = PyTuple_New(1);
            if (unlikely(!args_tuple)) return;
            Py_INCREF(value);
            PyTuple_SET_ITEM(args_tuple, 0, value);
            exc = PyObject_Call(exc_type, args_tuple, NULL);
            Py_DECREF(args_tuple);
        }
        if (unlikely(!exc)) return;
    } else {
        Py_INCREF(value);
        exc = value;
    }
    #if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    #if CYTHON_USE_EXC_INFO_STACK
    if (!__pyx_tstate->exc_info->exc_value)
    #else
    if (!__pyx_tstate->exc_type)
    #endif
    {
        Py_INCREF(exc_type);
        __Pyx_ErrRestore(exc_type, exc, NULL);
        return;
    }
    #endif
#else
    exc = __Pyx_PyObject_CallOneArg(exc_type, value);
    if (unlikely(!exc)) return;
#endif
    PyErr_SetObject(exc_type, exc);
    Py_DECREF(exc);
}

/* CoroutineBase (used by Generator) */
#if !CYTHON_COMPILING_IN_LIMITED_API
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6 && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#endif // CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void
__Pyx_Coroutine_Undelegate(__pyx_CoroutineObject *gen) {
#if CYTHON_USE_AM_SEND
    gen->yieldfrom_am_send = NULL;
#endif
    Py_CLEAR(gen->yieldfrom);
}
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            #if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
            value = PyObject_GetAttr(ev, __pyx_mstate_global->__pyx_n_u_value);
            if (unlikely(!value)) goto limited_api_failure;
            #else
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            #endif
            Py_DECREF(ev);
        }
        else if (unlikely(PyTuple_Check(ev))) {
            Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(ev);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely(tuple_size < 0)) {
                Py_XDECREF(tb);
                Py_DECREF(ev);
                Py_DECREF(et);
                return -1;
            }
            #endif
            if (tuple_size >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#elif CYTHON_ASSUME_SAFE_MACROS
                value = PySequence_ITEM(ev, 0);
#else
                value = PySequence_GetItem(ev, 0);
                if (!value) goto limited_api_failure;
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_GetAttr(ev, __pyx_mstate_global->__pyx_n_u_value);
#else
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
#endif
    Py_DECREF(ev);
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!value)) return -1;
#endif
    *pvalue = value;
    return 0;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL || !CYTHON_ASSUME_SAFE_MACROS
  limited_api_failure:
    Py_XDECREF(et);
    Py_XDECREF(tb);
    Py_XDECREF(ev);
    return -1;
#endif
}
static CYTHON_INLINE
__Pyx_PySendResult __Pyx_Coroutine_status_from_result(PyObject **retval) {
    if (*retval) {
        return PYGEN_NEXT;
    } else if (likely(__Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, retval) == 0)) {
        return PYGEN_RETURN;
    } else {
        return PYGEN_ERROR;
    }
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
static void __Pyx_Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
__Pyx_PySendResult __Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, PyObject **result, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(__Pyx_Coroutine_get_is_running(self));  // Callers should ensure is_running
    if (unlikely(self->resume_label == -1)) {
        __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
        return PYGEN_ERROR;
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    retval = self->body(self, tstate, value);
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    *result = retval;
    if (self->resume_label == -1) {
        return likely(retval) ? PYGEN_RETURN : PYGEN_ERROR;
    }
    return PYGEN_NEXT;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
#define __Pyx_Coroutine_MethodReturnFromResult(gen, result, retval, iternext)\
    ((result) == PYGEN_NEXT ? (retval) : __Pyx__Coroutine_MethodReturnFromResult(gen, result, retval, iternext))
static PyObject *
__Pyx__Coroutine_MethodReturnFromResult(PyObject* gen, __Pyx_PySendResult result, PyObject *retval, int iternext) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (likely(result == PYGEN_RETURN)) {
        int is_async = 0;
        #ifdef __Pyx_AsyncGen_USED
        is_async = __Pyx_AsyncGen_CheckExact(gen);
        #endif
        __Pyx_ReturnWithStopIteration(retval, is_async, iternext);
        Py_XDECREF(retval);
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE __Pyx_PySendResult
__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen, PyObject** retval) {
    __Pyx_PySendResult result;
    PyObject *val = NULL;
    assert(__Pyx_Coroutine_get_is_running(gen));
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    result = __Pyx_Coroutine_SendEx(gen, val, retval, 0);
    Py_XDECREF(val);
    return result;
}
#if CYTHON_USE_AM_SEND
static __Pyx_PySendResult
__Pyx_Coroutine_SendToDelegate(__pyx_CoroutineObject *gen, __Pyx_pyiter_sendfunc gen_am_send, PyObject *value, PyObject **retval) {
    PyObject *ret = NULL;
    __Pyx_PySendResult delegate_result, result;
    assert(__Pyx_Coroutine_get_is_running(gen));
    delegate_result = gen_am_send(gen->yieldfrom, value, &ret);
    if (delegate_result == PYGEN_NEXT) {
        assert (ret != NULL);
        *retval = ret;
        return PYGEN_NEXT;
    }
    assert (delegate_result != PYGEN_ERROR || ret == NULL);
    __Pyx_Coroutine_Undelegate(gen);
    result = __Pyx_Coroutine_SendEx(gen, ret, retval, 0);
    Py_XDECREF(ret);
    return result;
}
#endif
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval = NULL;
    __Pyx_PySendResult result = __Pyx_Coroutine_AmSend(self, value, &retval);
    return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 0);
}
static __Pyx_PySendResult
__Pyx_Coroutine_AmSend(PyObject *self, PyObject *value, PyObject **retval) {
    __Pyx_PySendResult result;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        *retval = __Pyx_Coroutine_AlreadyRunningError(gen);
        return PYGEN_ERROR;
    }
    #if CYTHON_USE_AM_SEND
    if (gen->yieldfrom_am_send) {
        result = __Pyx_Coroutine_SendToDelegate(gen, gen->yieldfrom_am_send, value, retval);
    } else
    #endif
    if (gen->yieldfrom) {
        PyObject *yf = gen->yieldfrom;
        PyObject *ret;
      #if !CYTHON_USE_AM_SEND
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
      #endif
        {
            #if !CYTHON_COMPILING_IN_LIMITED_API || __PYX_LIMITED_VERSION_HEX >= 0x03080000
            if (value == Py_None && PyIter_Check(yf))
                ret = __Pyx_PyIter_Next_Plain(yf);
            else
            #endif
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_mstate_global->__pyx_n_u_send, value);
        }
        if (likely(ret)) {
            __Pyx_Coroutine_unset_is_running(gen);
            *retval = ret;
            return PYGEN_NEXT;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, retval);
    } else {
        result = __Pyx_Coroutine_SendEx(gen, value, retval, 0);
    }
    __Pyx_Coroutine_unset_is_running(gen);
    return result;
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    __Pyx_PySendResult result;
    PyObject *retval = NULL;
    CYTHON_UNUSED_VAR(gen);
    assert(__Pyx_Coroutine_get_is_running(gen));
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        result = __Pyx_Coroutine_Close(yf, &retval);
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        result = __Pyx_Coroutine_Close(yf, &retval);
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        result = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf);
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
        result = PYGEN_RETURN;
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
        result = PYGEN_RETURN;
    } else
    #endif
    {
        PyObject *meth;
        result = PYGEN_RETURN;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_mstate_global->__pyx_n_u_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval)) {
                result = PYGEN_ERROR;
            }
        }
    }
    Py_XDECREF(retval);
    return result == PYGEN_ERROR ? -1 : 0;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __Pyx_PySendResult result;
    PyObject *retval = NULL;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    }
    #if CYTHON_USE_AM_SEND
    if (gen->yieldfrom_am_send) {
        result = __Pyx_Coroutine_SendToDelegate(gen, gen->yieldfrom_am_send, Py_None, &retval);
    } else
    #endif
    if (gen->yieldfrom) {
        PyObject *yf = gen->yieldfrom;
        PyObject *ret;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && (PY_VERSION_HEX < 0x030A00A3 || !CYTHON_USE_AM_SEND)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
            ret = __Pyx_PyIter_Next_Plain(yf);
        if (likely(ret)) {
            __Pyx_Coroutine_unset_is_running(gen);
            return ret;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, &retval);
    } else {
        result = __Pyx_Coroutine_SendEx(gen, Py_None, &retval, 0);
    }
    __Pyx_Coroutine_unset_is_running(gen);
    return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 1);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    PyObject *retval = NULL;
    __Pyx_PySendResult result;
    CYTHON_UNUSED_VAR(arg);
    result = __Pyx_Coroutine_Close(self, &retval);
    if (unlikely(result == PYGEN_ERROR))
        return NULL;
    Py_XDECREF(retval);
    Py_RETURN_NONE;
}
static __Pyx_PySendResult
__Pyx_Coroutine_Close(PyObject *self, PyObject **retval) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PySendResult result;
    PyObject *yf;
    int err = 0;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        *retval = __Pyx_Coroutine_AlreadyRunningError(gen);
        return PYGEN_ERROR;
    }
    yf = gen->yieldfrom;
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    result = __Pyx_Coroutine_SendEx(gen, NULL, retval, 1);
    if (result == PYGEN_ERROR) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_Coroutine_unset_is_running(gen);
        if (!__Pyx_PyErr_Occurred()) {
            return PYGEN_RETURN;
        } else if (likely(__Pyx_PyErr_ExceptionMatches2(PyExc_GeneratorExit, PyExc_StopIteration))) {
            __Pyx_PyErr_Clear();
            return PYGEN_RETURN;
        }
        return PYGEN_ERROR;
    } else if (likely(result == PYGEN_RETURN && *retval == Py_None)) {
        __Pyx_Coroutine_unset_is_running(gen);
        return PYGEN_RETURN;
    } else {
        const char *msg;
        Py_DECREF(*retval);
        *retval = NULL;
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
            msg = "async generator ignored GeneratorExit";
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        __Pyx_Coroutine_unset_is_running(gen);
        return PYGEN_ERROR;
    }
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen)))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    yf = gen->yieldfrom;
    if (yf) {
        __Pyx_PySendResult result;
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                goto propagate_exception;
            goto throw_here;
        }
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_mstate_global->__pyx_n_u_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    __Pyx_Coroutine_unset_is_running(gen);
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        Py_DECREF(yf);
        if (ret) {
            __Pyx_Coroutine_unset_is_running(gen);
            return ret;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, &ret);
        __Pyx_Coroutine_unset_is_running(gen);
        return __Pyx_Coroutine_MethodReturnFromResult(self, result, ret, 0);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
propagate_exception:
    {
        PyObject *retval = NULL;
        __Pyx_PySendResult result = __Pyx_Coroutine_SendEx(gen, NULL, &retval, 0);
        __Pyx_Coroutine_unset_is_running(gen);
        return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 0);
    }
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, "throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    {
        int e = __Pyx_call_type_traverse((PyObject*)gen, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    #if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    if (gen->gi_weakreflist != NULL)
    #endif
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        {
            destructor del = __Pyx_PyObject_GetSlot(gen, tp_del, destructor);
            if (del) del(self);
        }
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
#if CYTHON_USE_TP_FINALIZE
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *retval = NULL;
        __Pyx_PySendResult result = __Pyx_Coroutine_Close(self, &retval);
        if (result == PYGEN_ERROR) {
            PyErr_WriteUnraisable(self);
        } else {
            Py_XDECREF(retval);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
}
#endif
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx__Coroutine_get_frame(__pyx_CoroutineObject *self)
{
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyObject *frame;
    #if PY_VERSION_HEX >= 0x030d0000
    Py_BEGIN_CRITICAL_SECTION(self);
    #endif
    frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        PyObject *globals = PyDict_New();
        if (unlikely(!globals)) return NULL;
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            globals,                        /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        Py_DECREF(globals);
        if (unlikely(!frame))
            return NULL;
        if (unlikely(self->gi_frame)) {
            Py_DECREF(frame);
            frame = self->gi_frame;
        } else {
            self->gi_frame = frame;
        }
    }
    Py_INCREF(frame);
    #if PY_VERSION_HEX >= 0x030d0000
    Py_END_CRITICAL_SECTION();
    #endif
    return frame;
#else
    CYTHON_UNUSED_VAR(self);
    Py_RETURN_NONE;
#endif
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context) {
    CYTHON_UNUSED_VAR(context);
    PyObject *frame = self->gi_frame;
    if (frame)
        return __Pyx_NewRef(frame);
    return __Pyx__Coroutine_get_frame(self);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->yieldfrom_am_send = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_LIMITED_API
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    gen->gi_weakreflist = NULL;
#endif
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}
static char __Pyx_Coroutine_test_and_set_is_running(__pyx_CoroutineObject *gen) {
    char result;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    result = gen->is_running;
    gen->is_running = 1;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
    return result;
}
static void __Pyx_Coroutine_unset_is_running(__pyx_CoroutineObject *gen) {
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    assert(gen->is_running);
    gen->is_running = 0;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
}
static char __Pyx_Coroutine_get_is_running(__pyx_CoroutineObject *gen) {
    char result;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    result = gen->is_running;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
    return result;
}
static PyObject *__Pyx_Coroutine_get_is_running_getter(PyObject *gen, void *closure) {
    CYTHON_UNUSED_VAR(closure);
    char result = __Pyx_Coroutine_get_is_running((__pyx_CoroutineObject*)gen);
    if (result) Py_RETURN_TRUE;
    else Py_RETURN_FALSE;
}
#if __PYX_HAS_PY_AM_SEND == 2
static void __Pyx_SetBackportTypeAmSend(PyTypeObject *type, __Pyx_PyAsyncMethodsStruct *static_amsend_methods, __Pyx_pyiter_sendfunc am_send) {
    Py_ssize_t ptr_offset = (char*)(type->tp_as_async) - (char*)type;
    if (ptr_offset < 0 || ptr_offset > type->tp_basicsize) {
        return;
    }
    memcpy((void*)static_amsend_methods, (void*)(type->tp_as_async), sizeof(*type->tp_as_async));
    static_amsend_methods->am_send = am_send;
    type->tp_as_async = __Pyx_SlotTpAsAsync(static_amsend_methods);
}
#endif
static PyObject *__Pyx_Coroutine_fail_reduce_ex(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    __Pyx_TypeName self_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE((PyObject*)self));
    PyErr_Format(PyExc_TypeError, "cannot pickle '" __Pyx_FMT_TYPENAME "' object",
                         self_type_name);
    __Pyx_DECREF_TypeName(self_type_name);
    return NULL;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {"__reduce_ex__", (PyCFunction) __Pyx_Coroutine_fail_reduce_ex, METH_O, 0},
    {"__reduce__", (PyCFunction) __Pyx_Coroutine_fail_reduce_ex, METH_NOARGS, 0},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {"gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     PyDoc_STR("object being iterated by 'yield from', or None")},
    {"gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {"__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {"__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     PyDoc_STR("name of the generator"), 0},
    {"__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     PyDoc_STR("qualified name of the generator"), 0},
    {"gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     PyDoc_STR("Frame of the generator"), 0},
    {"gi_running", __Pyx_Coroutine_get_is_running_getter, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) PyObject_GenericGetAttr},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    {Py_am_send, (void *)__Pyx_Coroutine_AmSend},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_WEAKREF |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | __Pyx_TPFLAGS_HAVE_AM_SEND,
    __pyx_GeneratorType_slots
};
#if __PYX_HAS_PY_AM_SEND == 2
static __Pyx_PyAsyncMethodsStruct __pyx_Generator_as_async;
#endif
static int __pyx_Generator_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_GeneratorType_spec, NULL);
    if (unlikely(!mstate->__pyx_GeneratorType)) {
        return -1;
    }
#if __PYX_HAS_PY_AM_SEND == 2
    __Pyx_SetBackportTypeAmSend(mstate->__pyx_GeneratorType, &__pyx_Generator_as_async, &__Pyx_Coroutine_AmSend);
#endif
    return 0;
}
static PyObject *__Pyx_Generator_GetInlinedResult(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *retval = NULL;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    }
    __Pyx_PySendResult result = __Pyx_Coroutine_SendEx(gen, Py_None, &retval, 0);
    __Pyx_Coroutine_unset_is_running(gen);
    (void) result;
    assert (result == PYGEN_RETURN || result == PYGEN_ERROR);
    assert ((result == PYGEN_RETURN && retval != NULL) || (result == PYGEN_ERROR && retval == NULL));
    return retval;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
